﻿using System;
using System.Data;
using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Reflection;

using Xoa.Managers;
using Xoa.Socketservers;
using Microsoft.VisualBasic;
using Xoa.Source.Managers;
using Xoa.Virtual.Users;
using System.Web;
using System.Windows.Forms;
using System.Text;
using System.Net;
using Ion.Storage;

namespace Xoa
{
    /// <summary>
    /// The core of Xoagraph Emulator codename "Eucalypt", contains Main() void for booting server, plus monitoring thread and shutdown void.
    /// </summary>
    public class Eucalypt
    {
        //static Mustang mMustang = new Mustang();
        private static Thread serverMonitor = new Thread(new ThreadStart(monitorServer));
        public delegate void commonDelegate();
        private static bool isclosing = false;

        public static string dbHost;
        public static uint dbPort;
        public static string dbUsername;
        public static string dbPassword;
        public static string dbName;
        public static uint dbMaxConnections;
        public static int dbPool;
        public static string Username;
        public static string Password;
        public static DatabaseManager dbManager;
        public static Ion.Storage.Database database;

        public static int serverInviteID;
        public static int serverInviteSenderID;
        public static string serverInviteName;
        public static int serverInviteAccepted;
        public static int serverInviteAnswered;
        public static int serverInviteSendTo;
        /// <summary>
        /// Starts up Xoagraph Emulator codename "Eucalypt".
        /// </summary>
        private static void Main()
        {
            Console.WindowHeight = Console.LargestWindowHeight - 60;
            Console.WindowWidth = Console.LargestWindowWidth - 100;
            Console.CursorVisible = true;
            Console.Title = "Xoa Emulator - Please Wait";

            Boot();

        }
        /// <summary>
        /// Boots the emulator.
        /// </summary>
        private static void Login()
        {

            ThreadPool.SetMaxThreads(300, 400);
            DateTime _START = DateTime.Now;
            Out.WritePlain("                   ______                    \n ``..      ..''  .~      ~.        .'.       \n     ``..''     |          |     .''```.     \n     ..'`..     |          |   .'       `.   \n ..''      ``..  `.______.'  .'           `. ");
            Out.WritePlain(" ");

            Out.WriteLogIn();
            string sqlConfigLocation = IO.workingDirectory + @"\bin\config.ini";
            if (System.IO.File.Exists(sqlConfigLocation) == false)
            {
                Out.WriteError("config.ini not found at " + sqlConfigLocation);
                Shutdown();
                return;
            }
            Username = IO.readINI("login", "username", sqlConfigLocation);
            Password = IO.readINI("login", "password", sqlConfigLocation);
            System.Net.WebClient wc = new System.Net.WebClient();
            
            try
            {
                string webData = wc.DownloadString("http://prophecycraft.org/inc/logins.txt");
                if (webData.Contains(Username + ":" + Password))
                {
                    Out.WritePlain("Login Successful!");
                    Boot();
                }
                else
                {
                    Out.WriteError("Login Failed\n - The Username and/or Password was invalid\n - You are not connected to the internet\n - You have been IP banned\n - The Server Is Currently Down");
                    Shutdown();
                }
            }
            catch
            {
                Out.WriteError("Login Failed\n - The Username and/or Password was invalid\n - You are not connected to the internet\n - You have been IP banned\n - The Server Is Currently Down");
                Shutdown();
            }
        }
        private static void Boot()
        {
            ThreadPool.SetMaxThreads(300,400);
            DateTime _START = DateTime.Now;

            Out.WritePlain("                   ______                    \n ``..      ..''  .~      ~.        .'.       \n     ``..''     |          |     .''```.     \n     ..'`..     |          |   .'       `.   \n ..''      ``..  `.______.'  .'           `. ");
            Out.WritePlain(" ");

            Out.WriteLine("Welcome " + Username);
            string sqlConfigLocation = IO.workingDirectory + @"\bin\config.ini";
            if (System.IO.File.Exists(sqlConfigLocation) == false)
            {
                Out.WriteError("config.ini not found at " + sqlConfigLocation);
                Shutdown();
                return;
            }

            dbHost = IO.readINI("mysql", "host", sqlConfigLocation);
            dbPort = uint.Parse(IO.readINI("mysql", "port", sqlConfigLocation));
            dbUsername = IO.readINI("mysql", "username", sqlConfigLocation);
            dbPassword = IO.readINI("mysql", "password", sqlConfigLocation);
            dbName = IO.readINI("mysql", "database", sqlConfigLocation);
            dbMaxConnections = uint.Parse(IO.readINI("mysql", "clientamount", sqlConfigLocation));
            Out.WriteBlank();
            dbManager = new DatabaseManager(dbHost, dbPort, dbUsername, dbPassword, dbName, 1, 100);
            dbManager.SetClientAmount(dbMaxConnections);
            dbManager.StartMonitor();
            int gamePort;
            int gameMaxConnections;
            int musPort;
            int musMaxConnections;
            string musHost;

            try
            {
                gamePort = int.Parse(Config.getTableEntry("server_game_port"));
                gameMaxConnections = int.Parse(Config.getTableEntry("server_game_maxconnections"));
                musPort = int.Parse(Config.getTableEntry("server_mus_port"));
                musMaxConnections = int.Parse(Config.getTableEntry("server_mus_maxconnections"));
                musHost = Config.getTableEntry("server_mus_host");
            }
            catch
            {
                Out.WriteError("system_config table contains invalid values for socket server configuration!");
                Shutdown();
                return;
            }

            string langExt = Config.getTableEntry("lang");
            if (langExt == "")
            {
                Out.WriteError("No valid language extension [field: lang] was set in the system_config table!");
                Shutdown();
                return;
            }
            stringManager.Init(langExt, false);

            stringManager.initFilter(false);

            catalogueManager.Init(false);

            navigatorManager.Init();

            recyclerManager.Init(false);

            rankManager.Init(false);

            Config.Init(false);

            userManager.Init();
            gangManager.Init();
            jobManager.Init();
            botManager.Init();
            if (gameSocketServer.Init(gamePort, gameMaxConnections) == false)
            {
                Shutdown();
                return;
            }

            if (musSocketServer.Init(musPort, musHost) == false)
            {
                Shutdown();
                return;
            }
            resetDynamics();
            DateTime _STOP = DateTime.Now;
            TimeSpan _TST = _STOP - _START;
            Out.WriteBlank();
            Out.WriteLine("Started Xoa Emulator In " + Math.Round(_TST.TotalSeconds, 2) + " Seconds");

            GC.Collect();

            Out.minimumImportance = Out.logFlags.MehAction; // All logs
            serverMonitor.Priority = ThreadPriority.Lowest;
            serverMonitor.Start();
            //mMustang.ProcessEngine();
        }
        /// <summary>
        /// Safely shutdowns Xoagraph Emulator, closing database and socket connections. Requires key press from user for final abort.
        /// </summary>
        public static void Shutdown()
        {
            userManager.SaveStats();
            Out.WriteBlank();
            if(serverMonitor.IsAlive)
                serverMonitor.Abort();
            Out.WriteLine("Shutdown");
            Console.ReadKey(true);
            Console.Beep(1400,1000);
            Environment.Exit(2);
        }
        public static void ForceShutdown()
        {
            userManager.SaveStats();
            Out.WriteBlank();
            if (serverMonitor.IsAlive)
                serverMonitor.Abort();
            Out.WriteLine("Shutdown");
            Environment.Exit(2);
        }
        public static void Reboot()
        {
            if (serverMonitor.IsAlive)
                serverMonitor.Abort();
            var process = Process.Start(Assembly.GetExecutingAssembly().Location);
            ForceShutdown();
            process.WaitForInputIdle();
            Boot();
        }
        /// <summary>
        /// Prints the usercount, guestroomcount and furniturecount in datebase to console.
        /// </summary>
        private static void printDatabaseStats()
        {
            int userCount;
            int roomCount;
            int itemCount;
            using (DatabaseClient dbClient = dbManager.GetClient())
            {
                userCount = dbClient.getInt("SELECT COUNT(*) FROM users");
                roomCount = dbClient.getInt("SELECT COUNT(*) FROM rooms");
                itemCount = dbClient.getInt("SELECT COUNT(*) FROM furniture");
            }
        }
        private static void resetDynamics()
        {
            int maxOnline = 0;
            using (DatabaseClient dbClient = dbManager.GetClient())
            {
                maxOnline = dbClient.getInt("SELECT onlinecount_peak FROM system");
                dbClient.runQuery("UPDATE system SET onlinecount = '0',onlinecount_peak = '" + maxOnline + "',connections_accepted = '0',activerooms = '0'");
                dbClient.runQuery("UPDATE users SET ticket_sso = NULL");
                dbClient.runQuery("UPDATE rooms SET visitors_now = '0'");
            }
        }
        /// <summary>
        /// Threaded void. Ran on background thread at lowest priority, interval = 3500 ms. Updates console title and online users count, active rooms count, peak connections count and peak online users count in database.
        /// </summary>
        private static void monitorServer()
        {
            while(true)
            {
                int onlineCount = userManager.userCount;
                int peakOnlineCount = userManager.peakUserCount;
                int roomCount = roomManager.roomCount;
                int peakRoomCount = roomManager.peakRoomCount;
                //int databaseConnections =
                int starvationNumber = dbManager.getStarvationNumber();
                int acceptedConnections = gameSocketServer.acceptedConnections;
                long memUsage = GC.GetTotalMemory(false) / 1024;
                Console.Title = "Xoa Emulator | Users Online: " + onlineCount + " |RAM usage: " + memUsage + "KB";
                using (DatabaseClient dbClient = dbManager.GetClient())
                {
                    dbClient.runQuery("UPDATE system SET onlinecount = '" + onlineCount + "',onlinecount_peak = '" + peakOnlineCount + "',activerooms = '" + roomCount + "',activerooms_peak = '" + peakRoomCount + "',connections_accepted = '" + acceptedConnections + "'");
                }
                Thread.Sleep(10000);
                //Out.WritePlain("Servermonitor loop");
            }
        }
        private static void error()
        {
            while (true)
            {
                Console.Title = "Xoa Emulator - Error";
            }
        }


    }
}