﻿using System;
using System.Data;
using System.Text;
using System.Threading;
using System.Collections;

using Xoa.Virtual.Users;
using Xoa.Managers;
using Xoa.Virtual.Rooms.Pathfinding;
using Ion.Storage;
using Xoa;
namespace Xoa.Virtual.Rooms.Bots
{
    /// <summary>
    /// Represents a computer controlled virtual user with an artifical intelligence (AI). The bot walks around in virtual rooms on specified coordinates, interacts with other virtual users and serves drinks and food.
    /// </summary>
    public class virtualBot
    {
        #region Declares
        /// <summary>
        /// The virtual room the bot roams in.
        /// </summary>
        internal virtualRoom Room;
        internal virtualRoomBot roomBot;
        /// <summary>
        /// The status manager that keeps status strings for the user in room.
        /// </summary>
        internal virtualRoomBotStatusManager statusManager;
        /// <summary>
        /// The name of the bot.
        /// </summary>
        internal string Name;
        internal bool Working;
        internal bool coolDown;
        internal int CDTimer;
        /// <summary>
        /// The mission/motto that the bot has.
        /// </summary>
        internal string Mission;
        /// <summary>
        /// The bot's figure string.
        /// </summary>
        internal string Figure;
        /// <summary>
        /// The bot's uniform string.
        /// </summary>
        internal string Uniform;
        /// <summary>
        /// The ID of the bot in the virtual room.
        /// </summary>
        internal int ID;
        /// <summary>
        /// The RoomID of the bot in the virtual room.
        /// </summary>
        internal int RoomID;
        /// <summary>
        /// If A Bot Is A Cop Or Not
        /// </summary>
        internal int Cop;
        /// <summary>
        /// If A Bot Is A Doctor Or Not
        /// </summary>
        internal int Doctor;
        /// <summary>
        /// If A Bot Is A Fighter Or Not
        /// </summary>
        internal int Fighter;
        /// <summary>
        /// If A Bot Rings The Police Or Not
        /// </summary>
        internal int RingCops;
        /// <summary>
        /// The Bots Health
        /// </summary>
        internal int Health;
        internal string Colour;
        internal string Animal;
        internal int Owner;
        /// <summary>
        /// The Bots str
        /// </summary>
        internal int Str;
        /// <summary>
        /// If A Bot Is Dead
        /// </summary>
        internal int Dead;
        /// <summary>
        //The Bots Home Room
        /// </summary>
        internal int Home;
        /// <summary>
        /// Indicates if the bot uses 'freeroam', which allows it to walk everywhere where it can go to. Astar pathfinding is used.
        /// </summary>
        internal int freeRoam;
        /// <summary>
        /// The message that the bot will use (on random) to shouting people near the bot.
        /// </summary>
        internal string noShoutingMessage;
        internal string JobMotto;
        /// <summary>
        /// Specifies if the bot can walk.
        /// </summary>
        internal bool walkLock;
        internal int WorkTime;
        internal bool calledCops;
        internal int BBJ;
        internal int Xspot;
        internal int Yspot;
        internal byte Zz;
        internal string Type;
        internal int Banker;
        private delegate void statusVoid(string Key, string Value, int Length);
        /// <summary>
        /// Handles the random determining of actions.
        /// </summary>
        private Thread aiHandler;
        /// <summary>
        /// Contains the texts that the bot can 'say' on random.
        /// </summary>
        /// <summary>
        /// Contains the coordinate's where the bot can walk to. Ignored if freeroam is enabled.
        /// </summary>
        private Coord[] Coords;
        /// <summary>
        /// The virtualRoomUser object the bot is currently serving an item to.
        /// </summary>
        private virtualRoomUser Customer;
        /// <summary>
        /// The chatTrigger object that was invoked by the current customer.
        /// </summary>
        public Thread deadLooper;
        public Thread WorkStart;
        #endregion

        #region Constructors/destructors
        /// <summary>
        /// Contains the bot's animation statuses.
        /// </summary>
        private Hashtable Statuses;
        /// <summary>
        /// Initializes a new virtualBot object, loading the walk squares, chat texts etc.
        /// </summary>
        /// <param name="botID">The database ID of the bot.</param>
        /// <param name="roomUID">The ID that identifies this bot in room.</param>
        /// <param name="Room">The virtualRoom object where the bot is in.</param>
        internal virtualBot(int botID)
        {

            DataRow dRow;
            using (DatabaseClient dbClient = Xoa.Eucalypt.dbManager.GetClient())
            {
                dRow = dbClient.getRow("SELECT * FROM roombots WHERE id = '" + botID + "'");
            }
            this.Name = Convert.ToString(dRow["name"]);
            this.Mission = Convert.ToString(dRow["mission"]);
            this.Figure = Convert.ToString(dRow["figure"]);
            this.freeRoam = Convert.ToInt32(dRow["freeroam"]);
            this.noShoutingMessage = Convert.ToString(dRow["message_noshouting"]);
            this.RoomID = Convert.ToInt32(dRow["roomid"]);
            this.Cop = Convert.ToInt32(dRow["cop"]);
            this.Fighter = Convert.ToInt32(dRow["fighter_bot"]);
            this.RingCops = Convert.ToInt32(dRow["ring_police"]);
            this.Health = Convert.ToInt32(dRow["health"]);
            this.Str = Convert.ToInt32(dRow["str"]);
            this.Owner = Convert.ToInt32(dRow["bot_owner"]);
            this.Doctor = Convert.ToInt32(dRow["doctor"]);
            this.BBJ = Convert.ToInt32(dRow["bbj"]);
            this.Uniform = Convert.ToString(dRow["uniform"]);
            this.JobMotto = Convert.ToString(dRow["jobmotto"]);
            this.Animal = Convert.ToString(dRow["animal"]);
            this.Colour = Convert.ToString(dRow["hex"]);
            this.Xspot = Convert.ToInt32(dRow["x"]);
            this.Yspot = Convert.ToInt32(dRow["y"]);
            this.Zz = Convert.ToByte(dRow["z"]);
            this.Banker = Convert.ToInt32(dRow["banker"]);
            this.coolDown = false;
            this.CDTimer = 0;
            this.Type = Animal;
            this.calledCops = false;
        }
        #endregion
    }
}
