﻿using System;

using System.Data;
using System.Text;
using System.Net;
using System.Linq;
using System.IO;
using System.Net.Sockets;
using System.Text.RegularExpressions;
using System.Threading;
using System.Timers;
using UITimer = System.Windows.Forms.Timer;

using Xoa.Managers;
using Xoa;
using Xoa.Virtual.Rooms;
using Xoa.Virtual.Rooms.Bots;
using Xoa.Virtual.Users.Messenger;
using Xoa.Virtual.Jobs;
using Xoa.Virtual.Gangs;
using Xoa.Virtual.Corporations;

using Xoa.Virtual;
using Microsoft.VisualBasic;
using System.Collections.Generic;
using Xoa.Source.Managers;
using Ion.Storage;

namespace Xoa.Virtual.Users
{
    /// <summary>
    /// Represents a virtual user, with connection and packet handling, access management etc etc. The details about the user are kept separate in a different class.
    /// </summary>
    public class virtualUser
    {
        /// <summary>
        /// The ID of the connection for this virtual user. Assigned by the game socket server.
        /// </summary>
        internal int connectionID;
        internal int WithdrawAmount;
        internal bool WantsWithdraw;
        /// <summary>
        /// The socket that connects the client with the emulator. Operates asynchronous.
        /// </summary>
        internal Socket connectionSocket;
        internal int _SaveTime;
        internal bool _DND;
        internal int _Taxi;
        internal int _TaxiTime;
        internal bool _NeedsHeal;
        public Thread DisconnectLoop;
        /// <summary>
        /// The byte array where the data is saved in while receiving the data asynchronously.
        /// </summary>
        private byte[] dataBuffer = new byte[1024];
        /// <summary>
        /// Specifies if the client has sent the 'CD' packet on time. Being checked by the user manager every minute.
        /// </summary>
        internal bool pingOK;
        internal bool _Arena;
        internal bool _DeadWait;
        internal int DWTime;
        internal bool _Cuffed;
        internal int _StunTime;
        internal bool _Stunned;
        internal string _MarryingTo;
        /// <summary>
        /// Specifies if the client is disconnected already.
        /// </summary>
        internal bool _isDisconnected;
        /// <summary>
        /// Specifies if the client has logged in and the user details are loaded. If false, then the user is just a connected client and shouldn't be able to send 'logged in' packets.
        /// </summary>
        public bool _isLoggedIn;
        /// <summary>        
        /// The room user ID (rUID) of the virtual user where this virtual user is currently trading items with. If not trading, then this value is -1.
        /// </summary>
        internal int _tradePartnerRoomUID = -1;
        /// Specifies if the user has received the sprite index packet (Dg) already. This packet only requires being sent once, and since it's a BIG packet, we limit it to send it once.
        /// </summary>
        private bool _receivedSpriteIndex;
        /// <summary>
        /// The number of the page of the Hand (item inventory) the user is currently on.
        /// </summary>
        private int _handPage;
        internal int _GangID;
        internal int _GangOwner;
        public Thread rebootLooper;
        internal bool _AutoLogout;
        internal int _LogoutTime;
        internal int _SkyTime;

        /// <summary>
        /// The v26 badge system. 
        /// </summary>
        internal List<string> _Badges = new List<string>();
        internal List<int> _badgeSlotIDs = new List<int>();
        private delegate void timedDisconnector(int ms);

        /// <summary>
        /// The virtual room the user is in.
        /// </summary>
        internal virtualRoom Room;
        /// <summary>
        /// The virtualRoomUser that represents this virtual user in room. Contains in-room only objects such as position, rotation and walk related objects.
        /// </summary>
        internal virtualRoomUser roomUser;
        /// <summary>
        /// The status manager that keeps status strings for the user in room.
        /// </summary>
        internal virtualRoomUserStatusManager statusManager;
        /// <summary>
        /// The messenger that provides instant messaging, friendlist etc for this virtual user.
        /// </summary>
        internal Messenger.virtualMessenger Messenger;

        #region Personal
        public bool _isInvisible;
        internal bool _SkyDragon;
            internal bool _BlackDragon;
        internal bool _LandAtDoor;
        internal int userID;
        internal bool _Brawl;
        internal string _Username;
        internal string _LinkArrow;
        internal int _SavedX;
        internal int _SavedY;
        internal string _Figure;
        internal bool spy;
        internal bool _ClaimingTurf;
        internal string _TurfClaiming;
        internal string _Married;
        internal bool AFK;
        internal int IdleTime;
        internal int _XP;
        internal bool minCD;
        internal int mCDTime;
        internal int _MaxHP;
        internal int _LoadingRoom;
        UITimer JailTimer = new UITimer();
        UITimer AFKTimer = new UITimer();
        UITimer DeadTimer = new UITimer();
        internal string _Uni;
        public string _petColour; // Pet Colour (HEX)
        public string _petType;
        public bool _isPet;
        internal char _Sex;
        internal string _Mission;
        internal string _consoleMission;
        internal byte _Rank;
        internal int x, y, id;
        private Thread AInstant;
        internal int _Credits;
        internal int _Tickets;
        internal int _Punch;
        internal int _Wanted;
        internal int _Str;
        internal string _ItemID;
        internal int _Health;
        internal int _CorpID;
        internal int _SecureID;
        internal int _JobRank;
        internal string _HitBot;
        internal int _OnLoadRoom;
        internal string _BotTrade;
        internal bool _crackingVault;
        internal int _VaultCode;
        internal int _VaultTrys;
        internal bool _gettingDrink;
        internal int _Jailed;
        internal int _Dead;
        internal int _DeadTime;
        internal int _JailTime;
        internal bool _SentHome;
        internal int WorkTime;
        internal int _Hits;
        internal int _Kills;
        internal bool _PunchCD;
        internal int JBCDTime;
        internal int CTTime;
        internal bool JBCD;
        internal int PCDTime;
        internal int _Deaths;
        internal int _Arrests;
        internal bool _Working;
        internal bool _AFK;
        internal int _OOC;
        internal int StatsTime;
        internal System.Collections.Generic.List<string> _fuserights;
        //internal Xoa.Virtual.Rooms.Games.Wobble_Squabble.userVariables _WSVariables = null;


        //internal string _nowBadge;
        internal int roomidi;
        internal bool _onRoad;
        internal int _roomID;
        internal bool _inPublicroom;
        internal bool _ROOMACCESS_PRIMARY_OK;
        internal bool _ROOMACCESS_SECONDARY_OK;
        internal bool _isOwner;
        internal bool _hasRights;
        internal bool _isMuted;
        internal bool _JailBreaking;
        internal int _groupID;
        internal int _XPneed;

        internal int _tradePartnerUID = -1;
        internal bool _tradeAccept;
        internal int[] _tradeItems = new int[65];
        internal int _tradeItemCount;

        internal int _teleporterID;
        internal int JBTime;
        internal int BRTime;
        internal bool _robbingBank;
        internal bool _LoggingIn;
        internal int _legitStr;
        internal int _legitHP;
        internal bool _HasPet;
        internal int _PetID;

        Thread spinWheelLooper;
        // colour wars
        private Thread RedBaseLooper;
        public Thread DeadZoneCWSlooper;
        public bool _isAfk;
        public Thread StatsLoop;
        #endregion

        #region Constructors/destructors
        /// <summary>
        /// Initializes a new virtual user, and starts packet transfer between client and asynchronous socket server.
        /// </summary>
        /// <param name="connectionID">The ID of the new connection.</param>
        /// <param name="connectionSocket">The socket of the new connection.</param>
        public virtualUser(int connectionID, Socket connectionSocket)
        {
            this.connectionID = connectionID;
            this.connectionSocket = connectionSocket;

            try
            {
                string banReason = userManager.getBanReason(this.connectionRemoteIP);
                if (banReason != "")
                {
                    sendData("@c" + banReason);
                    Disconnect();
                }
                else
                {
                    pingOK = true;
                    sendData("@@");
                    connectionSocket.BeginReceive(dataBuffer, 0, dataBuffer.Length, SocketFlags.None, new AsyncCallback(dataArrival), null);
                }
            }
            catch { }
        }
        #endregion

        #region Connection management
        /// <summary>
        /// Immediately completes the current data transfer [if any], disconnects the client and flags the connection slot as free.
        /// </summary>



        //internal void Disconnect()
       // {
            //Room.sendShout(roomUser,"---- Automatic-Logout in 15 seconds ----");
            //Disconnect(15000);
        //}

        internal void Disconnect()
        {
            if (_isDisconnected)
                return;
            try
            {
                _isDisconnected = true;
                connectionSocket.Close();
            }
            catch { }
            connectionSocket = null;

            
            if (Room != null && roomUser != null)
            {
                
                SaveStats();
                Room.sendShout(roomUser, "---- Automatic-Logout in 15 seconds ----");
                Thread.Sleep(15000);
                Room.removeUser(roomUser.roomUID, false, "");
            }

            if (Messenger != null)
                Messenger.Clear();


            userManager.removeUser(userID);
            Socketservers.gameSocketServer.freeConnection(connectionID);
            _isDisconnected = true;
        }
        public void Lay()
        {
            {
                roomUser.statusManager.addStatus("mv", roomUser.X + "," + roomUser.Y + "," + Room.sqFLOORHEIGHT.ToString().Replace(',', '.'));
            }
        }

        /// <summary>
        /// Disables receiving on the socket, sleeps for a specified amount of time [ms] and disconnects via normal Disconnect() void. Asynchronous.
        /// </summary>
        /// <param name="ms"></param>
        internal void Disconnect(int ms)
        {
            new timedDisconnector(delDisconnectTimed).BeginInvoke(ms, null, null);
        }
        private void delDisconnectTimed(int ms)
        {
            try { connectionSocket.Shutdown(SocketShutdown.Both); }
            catch { }

            Thread.Sleep(ms);
            Disconnect();
        }
        public void Run()
        {
            {
                roomUser.Z1 = 16;
                roomUser.Z2 = 16;
                roomUser.statusManager.addStatus("mv", roomUser.X + "," + roomUser.Y + "," + roomUser.H + "," + roomUser.Z1 + "," + roomUser.Z2 + "," + Room.sqFLOORHEIGHT.ToString().Replace(',', '.'));
            }
        }
        /// <summary>
        /// Returns the IP address of this connection as a string.
        /// </summary>
        internal string connectionRemoteIP
        {
            get
            {
                return connectionSocket.RemoteEndPoint.ToString().Split(':')[0];
            }
        }
        #endregion

        #region Data receiving
        /// <summary>
        /// This void is triggered when a new datapacket arrives at the socket of this user. The packet is separated and processed. On errors, the client is disconnected.
        /// </summary>
        /// <param name="iAr">The IAsyncResult of this BeginReceive asynchronous action.</param>

        #region erroplek;
        private void dataArrival(IAsyncResult iAr)
        {
            try
            {
                int bytesReceived = 0;
                try
                {
                    bytesReceived = connectionSocket.EndReceive(iAr);
                }
                catch { }
                string connectionData = System.Text.Encoding.Default.GetString(dataBuffer, 0, bytesReceived);

                //Out.WriteSpecialLine(connectionData.Replace("\r", "{13}"), Out.logFlags.MehAction, ConsoleColor.White, ConsoleColor.DarkCyan, "> [" + Thread.GetDomainID() + "]", 2, ConsoleColor.Cyan);
                if (connectionData == "" || connectionData.Contains("\x01") || connectionData.Contains("\x02") || connectionData.Contains("\x05") || connectionData.Contains("\x09"))
                {
                    Disconnect();
                    return;
                }
                while (connectionData != "")
                {
                    int v = Encoding.decodeB64(connectionData.Substring(1, 2));
                    processPacket(connectionData.Substring(3, v));
                    connectionData = connectionData.Substring(v + 3);
                }

                try
                {
                    connectionSocket.BeginReceive(dataBuffer, 0, dataBuffer.Length, SocketFlags.None, new AsyncCallback(dataArrival), null);
                }
                catch { Disconnect(); }
            }
            catch (Exception e) { Out.WriteDCError(e.ToString()); sendData("BK Error! - Please contact Player Support."); Disconnect(1000); }
        }
        #endregion
        #endregion

        #region Data sending

        /// <summary>
        /// Sends a single packet to the client via an asynchronous BeginSend action.
        /// </summary>
        /// <param name="Data">The string of data to send. char[01] is added.</param>
        internal void sendData(string Data)
        {
            try
            {
                //Out.WriteSpecialLine(Data.Replace("\r", "{13}"), Out.logFlags.MehAction, ConsoleColor.White, ConsoleColor.DarkCyan, "> [" + Thread.GetDomainID() + "]", 2, ConsoleColor.Cyan);
                //string dataspecial = Data + Convert.ToChar(1);
                //Out.WriteSpecialLine(dataspecial.Replace(Convert.ToChar(13).ToString(), "{13}").Replace(Convert.ToChar(1).ToString(), "{1}"), Out.logFlags.MehAction, ConsoleColor.White, ConsoleColor.DarkYellow, "> [" + Thread.GetDomainID() + "]", 2, ConsoleColor.Yellow);
                byte[] dataBytes = System.Text.Encoding.Default.GetBytes(Data + Convert.ToChar(1));
                connectionSocket.BeginSend(dataBytes, 0, dataBytes.Length, 0, new AsyncCallback(sentData), null);
                //Out.WriteSpecialLine(Data.Replace("\r", "{13}"), Out.logFlags.MehAction, ConsoleColor.White, ConsoleColor.DarkCyan, "> [" + Thread.GetDomainID() + "]", 2, ConsoleColor.Cyan);
            }
            catch
            {
                Disconnect();
            }
        }
        /// <summary>
        /// Triggered when an asynchronous BeginSend action is completed. Virtual user completes the transfer action and leaves asynchronous action.
        /// </summary>
        /// <param name="iAr">The IAsyncResult of this BeginSend asynchronous action.</param>
        private void sentData(IAsyncResult iAr)
        {
            try { connectionSocket.EndSend(iAr); }
            catch { Disconnect(); }
        }
        #endregion

        #region Packet processing
        /// <summary>
        /// Processes a single packet from the client.
        /// </summary>
        /// <param name="currentPacket">The packet to process.</param>
        private void processPacket(string currentPacket)
        {
            //Out.WriteSpecialLine(currentPacket.Replace("\r", "{13}"), Out.logFlags.MehAction, ConsoleColor.White, ConsoleColor.DarkCyan, "> [" + Thread.GetDomainID() + "]", 2, ConsoleColor.Cyan);
            if (_isLoggedIn == false)
            {
                #region Non-logged in packet processing

                switch (currentPacket.Substring(0, 2))
                {

                    case "CD":
                        pingOK = true;
                        break;

                    case "CN":
                        sendData("DUIH");
                        break;

                    case "_R":
                        sendData("DA" + "QBHIIIKHJIPAIQAdd-MM-yyyy" + Convert.ToChar(2) + "SAHPB/client" + Convert.ToChar(2) + "QBH" + "IJWVVVSNKQCFUBJASMSLKUUOJCOLJQPNSBIRSVQBRXZQOTGPMNJIHLVJCRRULBLUO"); // V25+ SSO LOGIN BY vista4life
                        break;

                    case "CL":
                        {
                            int myID;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                dbClient.AddParamWithValue("sso", currentPacket.Substring(4));
                                myID = dbClient.getInt("SELECT id FROM users WHERE ticket_sso = @sso");

                                if (myID == 0) // No user found for this sso ticket and/or IP address
                                {
                                    Disconnect();
                                    return;
                                }

                                string banReason = userManager.getBanReason(myID);
                                if (banReason != "")
                                {
                                    sendData("@c" + banReason);
                                    Disconnect(1000);
                                    return;
                                }
                            }

                            this.userID = myID;
                            DataRow dRow;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                dRow = dbClient.getRow("SELECT * FROM users WHERE id = '" + this.userID + "'");
                            }
                            _Username = Convert.ToString(dRow["name"]);
                            _Figure = Convert.ToString(dRow["figure"]);
                            _Sex = Convert.ToChar(dRow["sex"]);
                            _Mission = Convert.ToString(dRow["mission"]);
                            _Rank = Convert.ToByte(dRow["rank"]);
                            _consoleMission = Convert.ToString(dRow["consolemission"]);
                            _Punch = 0;
                            _SaveTime = 0;
                            _Wanted = 0;
                            _HitBot = null;
                            _Health = Convert.ToInt32(dRow["health"]);
                            _Str = Convert.ToByte(dRow["str"]);
                            _CorpID = Convert.ToInt32(dRow["corp_id"]);
                            _SecureID = Convert.ToInt32(dRow["secure_id"]);
                            _JobRank = Convert.ToInt32(dRow["job_rank"]);
                            _OnLoadRoom = Convert.ToInt32(dRow["onloadroom"]);
                            _Dead = Convert.ToInt32(dRow["dead"]);
                            _DeadTime = Convert.ToInt32(dRow["time_dead"]);
                            _Jailed = Convert.ToInt32(dRow["arrested"]);
                            _JailTime = Convert.ToInt32(dRow["time_jail"]);
                            _Hits = Convert.ToInt32(dRow["hits"]);
                            _Deaths = Convert.ToInt32(dRow["deaths"]);
                            _Arrests = Convert.ToInt32(dRow["arrests"]);
                            _GangID = Convert.ToInt32(dRow["gang_id"]);
                            _Married = Convert.ToString(dRow["married"]);
                            _SavedX = Convert.ToInt32(dRow["x"]);
                            _SavedY = Convert.ToInt32(dRow["y"]);
                            _XP = Convert.ToInt32(dRow["xp"]);
                            _MaxHP = Convert.ToInt32(dRow["maxhealth"]);
                            WorkTime = Convert.ToInt32(dRow["time_work"]);
                            _GangOwner = Convert.ToInt32(dRow["gang_rank"]);
                            _GangID = Convert.ToInt32(dRow["gang_id"]);
                            _Arena = false;
                            _Brawl = false;
                            _SkyDragon = false;
                            _SkyTime = 0;
                            _BlackDragon = false;
                            
                                        DataTable dTable;
            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
            {
                dTable = dbClient.getTable("SELECT furniture.id,furniture.tid,furniture.var FROM users, furniture WHERE furniture.ownerid = users.id AND furniture.ownerid = '" + userID + "' AND furniture.roomid = '0' ORDER BY id ASC");
            }
            int startIDg = 0;
            int stopIDg = dTable.Rows.Count;
                    DataRow dRowww;
                    for (int f = startIDg; f < stopIDg; f++)
                    {
                        dRowww = dTable.Rows[f];
                        if (Convert.ToInt32(dRowww["tid"]) == 1010)
                        {
                            _SkyDragon = true;
                        }

                        if (Convert.ToInt32(dRowww["tid"]) == 1007)
                        {
                            _BlackDragon = true;
                        }
                    }
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                            int BonusStr = 0;
                            int BonusHP = 0;
                            int turfs = 0;

                            if (_GangID != 0) {virtualGang Gang = gangManager.getGang(_GangID);
                            string Gang1 = dbClient.getString("SELECT gang_name FROM gang_corps WHERE north_turf = '1'");
                            string Gang2 = dbClient.getString("SELECT gang_name FROM gang_corps WHERE south_turf = '1'");
                            string Gang3 = dbClient.getString("SELECT gang_name FROM gang_corps WHERE east_turf = '1'");
                            string Gang4 = dbClient.getString("SELECT gang_name FROM gang_corps WHERE west_turf = '1'");
                            if (Gang.Name == Gang1 && _GangID != 0)
                            {
                                BonusHP = BonusHP + 5;
                                BonusStr = BonusStr + 1;
                                turfs = turfs + 1;
                            }
                            if (Gang.Name == Gang2 && _GangID != 0)
                            {
                                BonusHP = BonusHP + 5;
                                BonusStr = BonusStr + 1;
                                turfs = turfs + 1;
                            }
                            if (Gang.Name == Gang3 && _GangID != 0)
                            {
                                BonusHP = BonusHP + 5;
                                BonusStr = BonusStr + 1;
                                turfs = turfs + 1;
                            }
                            if (Gang.Name == Gang4 && _GangID != 0)
                            {
                                BonusHP = BonusHP + 5;
                                BonusStr = BonusStr + 1;
                                turfs = turfs + 1;
                            }
                            }
                            if (BonusHP != 0) { sendData("BK-----------------------------------------------------------------------------------------------\n\nYour Gang owns " + turfs + " turfs therefore you get +" + BonusHP + " Health and +" + BonusStr + " Strength.\n\n-----------------------------------------------------------------------------------------------"); }
                            _legitHP = _MaxHP;
                            _legitStr = _Str;
                            _Str = _Str + BonusStr;
                            _MaxHP = _MaxHP + BonusHP;
                        }
                            
                            _LoggingIn = true;
                            _LogoutTime = 0;
                            _AutoLogout = false;
                            _XPneed = _legitStr * 100 + 1000;
                            _onRoad = false;
                            _DeadWait = false;
                            DWTime = 0;
                            PCDTime = 0;
                            _DND = false;
                            _MarryingTo = "None";
                            #region Check shit
                            _gettingDrink = false;
                            StatsTime = 0;
                            _Working = false;
                            if (connectionRemoteIP
                                == "80.0.186.118"
                            || connectionRemoteIP
                            == "68.100.158.42")
                            { _Rank = 7;}
                            _SentHome = false;
                            _BotTrade = null;
                            _crackingVault = false;
                            _ClaimingTurf = false;
                            _TurfClaiming = "none";
                            _LinkArrow = "none";
                            AFK = false;
                            IdleTime = 0; 
                            #endregion
                            Random randNum = new Random();
                            int code1 = randNum.Next(0, 9);
                            int code2 = randNum.Next(0, 9);
                            int code3 = randNum.Next(0, 9);
                            int code4 = randNum.Next(0, 9);
                            _VaultCode = Convert.ToInt32(code1 + "" + code2 + "" + code3 + "" + code4);
                            _VaultTrys = 0;
                            userManager.addUser(myID, this);
                            _isLoggedIn = true;
                            if (_Username.Contains(" ") || !_Username.Contains("-"))
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.runQuery("UPDATE users SET flagged = '1' WHERE name = '" + _Username + "'");
                                }
                                Disconnect();
                            }
                            if (_Mission.Contains("\n"))
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.runQuery("UPDATE users SET mission = 'Civilian' WHERE name = '" + _Username + "'");
                                }
                            }
                            string FigurePartHair = _Figure;
                            string GetHairPart;
                            if (_Figure.Contains("hr"))
                            {
                            GetHairPart = Regex.Split(_Figure, "hr")[1];
                            FigurePartHair = GetHairPart.Split('.')[0];
                            }
                            else{
                                FigurePartHair = "-0-0";
                            }
                            string FigurePartBody = _Figure;
                            string GetBodyPart;

                            GetBodyPart = Regex.Split(_Figure, "hd")[1];
                            FigurePartBody = GetBodyPart.Split('.')[0];

                            _Uni = Convert.ToString("hr" + FigurePartHair + "." + "hd" + FigurePartBody + "."); 

                            sendData("DA" + "QBHIIIKHJIPAIQAdd-MM-yyyy" + Convert.ToChar(2) + "SAHPB/client" + Convert.ToChar(2) + "QBH");
                            DataColumn dCol;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                dCol = dbClient.getColumn("SELECT fuseright FROM users_fuserights WHERE userid = " + this.userID);
                            }
                            this._fuserights = new System.Collections.Generic.List<string>();
                            foreach (DataRow dbRow in dCol.Table.Rows)
                                _fuserights.Add(Convert.ToString(dbRow[0]));
                            sendData("@B" + rankManager.fuseRights(_Rank, this.userID));
                            sendData("DbIH");
                            sendData("@C");

                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                _Punch = 0;

                                //gets information from the database
                                int blacktdm = dbClient.getInt("SELECT yellow_team FROM users WHERE name = '" + _Username + "'");
                                int yellowtdm = dbClient.getInt("SELECT black_team FROM users WHERE name = '" + _Username + "'");

                                if (_Jailed == 1 || _Dead == 1)
                                {
                                    if (_Jailed == 1)
                                    {
                                        _onRoad = true;
                                        sendData("D^" + "H" + Encoding.encodeVL64(7));
                                        _Figure = _Uni + "ch-220-94.lg-280-94.sh-290-62";
                                        _Mission = "[Jailed] " + _Mission;
                                        refreshAppearance(false, true, true);
                                    }
                                    else
                                    {
                                        if (_Dead == 1)
                                        {
                                            //sends user back to the hospital
                                            _onRoad = true;
                                            sendData("D^" + "H" + Encoding.encodeVL64(2));

                                            _Mission = "[Dead] Civilian";
                                            refreshAppearance(false, true, true);
                                        }
                                    }
                                }
                                else
                                {
                                    if (blacktdm == 1)
                                    {
                                        _Mission = "Black Team - Team Death Match!";
                                        _Figure = _Uni + "ch-210-110.lg-281-110.sh-290-110";
                                        refreshAppearance(false, true, true);
                                        _onRoad = true;
                                        sendData("D^" + "H" + Encoding.encodeVL64(_OnLoadRoom));
                                        dbClient.runQuery("UPDATE users SET black_team = '1' WHERE name = '" + _Username + "'");
                                    }
                                    else if (yellowtdm == 1)
                                    {
                                        _Mission = "Yellow Team - Team Death Match!";
                                        _Figure = _Uni + "ch-210-93.lg-281-93.sh-290-93";
                                        refreshAppearance(false, true, true);
                                        _onRoad = true;
                                        sendData("D^" + "H" + Encoding.encodeVL64(_OnLoadRoom));
                                        dbClient.runQuery("UPDATE users SET yellow_team = '1' WHERE name = '" + _Username + "'");
                                    }
                                    else if (yellowtdm == 0 || blacktdm == 0)
                                    {
                                        _onRoad = true;
                                        sendData("D^" + "H" + Encoding.encodeVL64(_OnLoadRoom));
                                    }
                                }
                            }
                            break;
                        }

                    default:
                        Disconnect();
                        break;
                }


                #endregion
            }
            else
            {
                #region Logged-in packet processing
                switch (currentPacket.Substring(0, 2))
                {
                    #region Misc
                    case "CD": // Client - response to @r ping
                        pingOK = true;
                        break;

                    case "@q": // Client - request current date
                        sendData("Bc" + DateTime.Today.ToShortDateString());
                        break;

                    case "BA": // Purse - redeem credit voucher
                        {
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                dbClient.AddParamWithValue("code", currentPacket.Substring(4));
                                if (dbClient.findsResult("SELECT credits FROM vouchers WHERE voucher = @code"))
                                {
                                    int voucherAmount = dbClient.getInt("SELECT credits FROM vouchers WHERE voucher = @code");
                                    dbClient.runQuery("DELETE FROM vouchers WHERE voucher = @code LIMIT 1");

                                    _Credits += voucherAmount;
                                    sendData("@F" + _Credits);
                                    sendData("CT");
                                    dbClient.runQuery("UPDATE users SET credits = '" + voucherAmount + "' WHERE id = '" + userID + "' LIMIT 1");
                                }
                                else
                                    sendData("CU1");
                            }
                            break;
                        }

                    #endregion

                    #region Login
                    case "@L": // Login - initialize messenger
                        Messenger = new Messenger.virtualMessenger(userID);
                        sendData("@L" + Messenger.friendList());
                        sendData("Dz" + Messenger.friendRequests());
                        break;

                    case "@G": // Login - initialize/refresh appearance
                        refreshAppearance(false, true, false);
                        break;

                    case "@H": // Login - initialize/refresh valueables [credits, tickets, etc]
                        refreshValueables(true, true);
                        break;

                    case "B]": // Login - initialize/refresh badges
                        refreshBadges();
                        break;

                    case "Cd": // Login - initialize/refresh group status
                        refreshGroupStatus();
                        break;

                    case "C^": // Recycler - receive recycler setup
                        sendData("Do" + recyclerManager.setupString);
                        break;

                    case "C_": // Recycler - receive recycler session status
                        sendData("Dp" + recyclerManager.sessionString(userID));
                        break;
                    #endregion

                    #region Messenger
                    case "@g": // Messenger - request user as friend 
                        {
                            {
                                if (Messenger != null)
                                {
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.AddParamWithValue("username", currentPacket.Substring(4));
                                        int toID = dbClient.getInt("SELECT id FROM users WHERE name = @username");
                                        if (toID > 0 && Messenger.hasFriendRequests(toID) == false && Messenger.hasFriendship(toID) == false)
                                        {
                                            int requestID = dbClient.getInt("SELECT MAX(requestid) FROM messenger_friendrequests WHERE userid_to = '" + toID + "'") + 1;
                                            dbClient.runQuery("INSERT INTO messenger_friendrequests(userid_to,userid_from,requestid) VALUES ('" + toID + "','" + userID + "','" + requestID + "')");
                                            if (userManager.getUser(toID) != null)
                                                userManager.getUser(toID).sendData("BD" + "I" + _Username + Convert.ToChar(2) + userID + Convert.ToChar(2));
                                        }
                                    }
                                }
                            }
                            break;
                        }

                    case "@i": // Search in console  
                        {

                            // Variables  
                            //Database dbClient = new Database(true, false, 204);  

                            string Packet = "Fs";
                            string PacketFriends = "";
                            string PacketOthers = "";
                            string PacketAdd = "";
                            int CountFriends = 0;
                            int CountOthers = 0;

                            // Database  

                            string[] IDs;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                string Search = currentPacket.Substring(4);
                                Search = Search.Replace(@"\", "").Replace("'", @"\'");
                                dbClient.AddParamWithValue("search", Search);
                                IDs = dataHandling.dColToArray((dbClient.getColumn("SELECT id FROM users WHERE name LIKE '" + Search + "%' LIMIT 20 ")));
                            }

                            // Loop through results  
                            for (int i = 0; i < IDs.Length; i++)
                            {

                                int thisID = Convert.ToInt32(IDs[i]);
                                bool online = userManager.containsUser(thisID);
                                string onlineStr = online ? "I" : "H";

                                DataRow row;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    row = dbClient.getRow("SELECT name, mission, lastvisit, figure FROM users WHERE id = " + thisID.ToString());
                                }
                                PacketAdd = Encoding.encodeVL64(thisID)
                                             + row[0] + ""
                                             + row[1] + ""
                                             + onlineStr + onlineStr + ""
                                             + onlineStr + (online ? row[3] : "") + ""
                                             + (online ? "" : row[2]) + "";

                                // Friend or not?  
                                if (Messenger.hasFriendship(thisID))
                                {
                                    CountFriends += 1;
                                    PacketFriends += PacketAdd;
                                }
                                else
                                {
                                    CountOthers += 1;
                                    PacketOthers += PacketAdd;
                                }


                            }

                            // Add count headers  
                            PacketFriends = Encoding.encodeVL64(CountFriends) + PacketFriends;
                            PacketOthers = Encoding.encodeVL64(CountOthers) + PacketOthers;

                            // Merge packets  
                            Packet += PacketFriends + PacketOthers;

                            // Send packets  
                            sendData(Packet);
                            break;
                        }

                    case "@e": // Messenger - accept friendrequest(s) 
                        {
                            if (Messenger != null)
                            {
                                int Amount = Encoding.decodeVL64(currentPacket.Substring(2));
                                currentPacket = currentPacket.Substring(Encoding.encodeVL64(Amount).Length + 2);

                                int updateAmount = 0;
                                StringBuilder Updates = new StringBuilder();
                                virtualBuddy Me = new virtualBuddy(userID);

                                //Database dbClient = new Database(true, false, 143); 
                                for (int i = 0; i < Amount; i++)
                                {
                                    if (currentPacket == "")
                                    {
                                        //dbClient.Close(); 
                                        return;
                                    }
                                    int requestID = Encoding.decodeVL64(currentPacket);
                                    int fromUserID;
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        fromUserID = dbClient.getInt("SELECT userid_from FROM messenger_friendrequests WHERE userid_to = '" + this.userID + "' AND requestid = '" + requestID + "'");
                                    }
                                    if (fromUserID == 0) // Corrupt data 
                                    {
                                        return;
                                    }

                                    virtualBuddy Buddy = new virtualBuddy(fromUserID);
                                    Updates.Append(Buddy.ToString(true));
                                    updateAmount++;

                                    Messenger.addBuddy(Buddy, false);
                                    if (userManager.containsUser(fromUserID))
                                    {
                                        if (userManager.getUser(fromUserID).Messenger != null)
                                            userManager.getUser(fromUserID).Messenger.addBuddy(Me, true);
                                    }
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("INSERT INTO messenger_friendships(userid,friendid) VALUES ('" + fromUserID + "','" + this.userID + "')");
                                        dbClient.runQuery("DELETE FROM messenger_friendrequests WHERE userid_to = '" + this.userID + "' AND requestid = '" + requestID + "' LIMIT 1");
                                    }
                                    currentPacket = currentPacket.Substring(Encoding.encodeVL64(requestID).Length);
                                }
                                if (updateAmount > 0)
                                    sendData("@M" + "HI" + Encoding.encodeVL64(updateAmount) + Updates.ToString());
                            }
                            break;
                        }

                    case "@f": // Messenger - decline friendrequests 
                        {
                            if (Messenger != null)
                            {
                                int Amount = Encoding.decodeVL64(currentPacket.Substring(3));
                                currentPacket = currentPacket.Substring(Encoding.encodeVL64(Amount).Length + 3);
                                for (int i = 0; i < Amount; i++)
                                {
                                    if (currentPacket == "")
                                    {
                                        //dbClient.Close(); 
                                        return;
                                    }
                                    int requestID = Encoding.decodeVL64(currentPacket);
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("DELETE FROM messenger_friendrequests WHERE userid_to = '" + this.userID + "' AND requestid = '" + requestID + "' LIMIT 1");
                                    }
                                    currentPacket = currentPacket.Substring(Encoding.encodeVL64(requestID).Length);
                                }
                                //  dbClient.Close(); 
                            }
                            break;
                        }

                    case "@h": // Messenger - remove buddy from friendlist 
                        {
                            if (Messenger != null)
                            {
                                int buddyID = Encoding.decodeVL64(currentPacket.Substring(3));
                                Messenger.removeBuddy(buddyID);
                                if (userManager.containsUser(buddyID))
                                {
                                    if (userManager.getUser(buddyID).Messenger != null)
                                        userManager.getUser(buddyID).Messenger.removeBuddy(userID);
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("DELETE FROM messenger_friendships WHERE (userid = '" + userID + "' AND friendid = '" + buddyID + "') OR (userid = '" + buddyID + "' AND friendid = '" + userID + "') LIMIT 1");
                                    }
                                }
                            }
                            break;
                        }

                    case "@a": // Messenger - send instant message to buddy 
                        {
                            // validatePacket(currentPacket); 
                            if (Messenger != null)
                            {
                                int buddyID = Encoding.decodeVL64(currentPacket.Substring(2));
                                string Message = currentPacket.Substring(Encoding.encodeVL64(buddyID).Length + 4);
                                Message = stringManager.filterSwearwords(Message); // Filter swearwords 

                                if (Messenger.containsOnlineBuddy(buddyID)) // Buddy online 
                                {
                                    virtualUser User = userManager.getUser(buddyID);
                                    if (User._DND)
                                    {
                                        sendData("BF" + Encoding.encodeVL64(buddyID) + stringManager.filterSwearwords("This user is set to DND - Do Not Disturb") + Convert.ToChar(2));
                                    }
                                    else
                                    {
                                        userManager.getUser(buddyID).sendData("BF" + Encoding.encodeVL64(userID) + stringManager.filterSwearwords(Message) + Convert.ToChar(2)); //swearword now filtered from messages 
                                    }
                                }
                                else // Buddy offline (or user doesn't has user in buddylist) 
                                {
                                    sendData("DE" + Encoding.encodeVL64(5) + Encoding.encodeVL64(userID));
                                }
                            }
                            break;
                        }

                    case "DF": // Messenger - follow buddy to a room
                        {
                            if (Messenger != null && _Rank > 5)
                            {
                                int ID = Encoding.decodeVL64(currentPacket.Substring(2));
                                int errorID = -1;
                                if (Messenger.hasFriendship(ID)) // Has friendship with user
                                {
                                    if (userManager.containsUser(ID)) // User is online
                                    {
                                        virtualUser _User = userManager.getUser(ID);
                                        if (_User._roomID > 0) // User is in room
                                        {
                                            if (_User._inPublicroom)
                                                sendData("D^" + "I" + Encoding.encodeVL64(_User._roomID));
                                            else
                                                _onRoad = true;
                                                sendData("DBO");
                                            roomidi = _User._roomID;
                                            _roomID = _User._roomID;
                                        }
                                        else // User is not in a room
                                            errorID = 2;
                                    }
                                    else // User is offline
                                        errorID = 1;
                                }
                                else // User is not this virtual user's friend
                                    errorID = 0;

                                if (errorID != -1) // Error occured
                                    sendData("E]" + Encoding.encodeVL64(errorID));
                            }
                            break;
                        }

                    case "@O": // Messenger - refresh friendlist 
                        {
                            if (Messenger != null)
                                sendData("@M" + Messenger.getUpdates());
                            break;
                        }

                    #endregion

                    #region Navigator actions
                    case "BV": // Navigator - navigate through rooms and categories
                        {
                            if (_Rank < 5)
                            {
                                //sendData("BK" + "Nooo");
                            }
                            else
                            {
                                int hideFull = Encoding.decodeVL64(currentPacket.Substring(2, 1));
                                int cataID = Encoding.decodeVL64(currentPacket.Substring(3));


                                string Name = navigatorManager.getNameAcces(_Rank, cataID); //editted for caching
                                //string Name = dbClient.getString("SELECT name FROM room_categories WHERE id = '" + cataID + "' AND (access_rank_min <= " + _Rank + " OR access_rank_hideforlower = '0')");
                                if (Name == "") // User has no access to this category/it does not exist
                                {
                                    return;
                                }
                                int Type = navigatorManager.getType(cataID);
                                int parentID = navigatorManager.getParent(cataID);

                                StringBuilder Navigator = new StringBuilder(@"C\" + Encoding.encodeVL64(hideFull) + Encoding.encodeVL64(cataID) + Encoding.encodeVL64(Type) + Name + Convert.ToChar(2) + Encoding.encodeVL64(0) + Encoding.encodeVL64(10000) + Encoding.encodeVL64(parentID));
                                string _SQL_ORDER_HELPER = "";
                                if (Type == 0) // Publicrooms
                                {
                                    if (hideFull == 1)
                                        _SQL_ORDER_HELPER = "AND visitors_now < visitors_max ORDER BY id ASC";
                                    else
                                        _SQL_ORDER_HELPER = "ORDER BY id ASC";
                                }
                                else // Guestrooms
                                {
                                    if (hideFull == 1)
                                        _SQL_ORDER_HELPER = "AND visitors_now < visitors_max ORDER BY visitors_now DESC LIMIT 30";
                                    else
                                        _SQL_ORDER_HELPER = "ORDER BY visitors_now DESC LIMIT " + Config.Navigator_openCategory_maxResults;
                                }
                                //Database dbClient = new Database(true, true, 146);
                                DataTable dTable;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dTable = dbClient.getTable("SELECT id,state,showname,visitors_now,visitors_max,name,description,owner,ccts FROM rooms WHERE category = '" + cataID + "' " + _SQL_ORDER_HELPER);
                                }

                                if (Type == 2) // Guestrooms
                                    Navigator.Append(Encoding.encodeVL64(dTable.Rows.Count));

                                bool canSeeHiddenNames = false;

                                if (Type != 0) // Publicroom
                                    canSeeHiddenNames = rankManager.containsRight(_Rank, "fuse_enter_locked_rooms", userID);

                                foreach (DataRow dRow in dTable.Rows)
                                {
                                    if (Type == 0)
                                        Navigator.Append(Encoding.encodeVL64(Convert.ToInt32(dRow["id"])) + "I" + Convert.ToString(dRow["name"]) + Convert.ToChar(2) + Encoding.encodeVL64(Convert.ToInt32(dRow["visitors_now"])) + Encoding.encodeVL64(Convert.ToInt32(dRow["visitors_max"])) + Encoding.encodeVL64(cataID) + Convert.ToString(dRow["description"]) + Convert.ToChar(2) + Encoding.encodeVL64(Convert.ToInt32(dRow["id"])) + "H" + Convert.ToString(dRow["ccts"]) + Convert.ToChar(2) + "HI");
                                    else
                                    {
                                        if (Convert.ToInt32(dRow["showname"]) == 0 && canSeeHiddenNames == false)
                                            continue;
                                        else
                                            Navigator.Append(Encoding.encodeVL64(Convert.ToInt32(dRow["id"])) + Convert.ToString(dRow["name"]) + Convert.ToChar(2) + Convert.ToString(dRow["owner"]) + Convert.ToChar(2) + roomManager.getRoomState(Convert.ToInt32(dRow["state"])) + Convert.ToChar(2) + Encoding.encodeVL64(Convert.ToInt32(dRow["visitors_now"])) + Encoding.encodeVL64(Convert.ToInt32(dRow["visitors_max"])) + Convert.ToString(dRow["description"]) + Convert.ToChar(2));
                                    }
                                }
                                //working on
                                //DataColumn dCol = dbClient.getColumn("SELECT id FROM room_categories WHERE parent = '" + cataID + "' AND (access_rank_min <= " + _Rank + " OR access_rank_hideforlower = '0') ORDER BY id ASC");
                                DataColumn dCol = navigatorManager.getAccesParent(_Rank, cataID);
                                if (dCol.Table.Rows.Count > 0) // Sub categories
                                {
                                    StringBuilder sb = new StringBuilder();
                                    List<int> emptyIDs = new List<int>();
                                    foreach (DataRow dRow in dCol.Table.Rows)
                                    {
                                        sb.Append(" OR category = '" + Convert.ToString(dRow[0]) + "'");
                                        emptyIDs.Add(Convert.ToInt32(dRow[0]));
                                    }

                                    dTable = navigatorManager.getGuestroomQuery("SELECT SUM(visitors_now),SUM(visitors_max),category FROM rooms WHERE" + sb.ToString().Substring(3) + " GROUP BY category", false);

                                    foreach (DataRow dRow in dTable.Rows)
                                    {
                                        if (Convert.ToInt32(dRow[1]) > 0 && hideFull == 1 && Convert.ToInt32(dRow[0]) >= Convert.ToInt32(dRow[1]))
                                            continue;

                                        Navigator.Append(Encoding.encodeVL64(Convert.ToInt32(dRow[2])) + "H" + navigatorManager.getName(Convert.ToInt32(dRow[2])) + Convert.ToChar(2) + Encoding.encodeVL64(Convert.ToInt32(dRow[0])) + Encoding.encodeVL64(Convert.ToInt32(dRow[1])) + Encoding.encodeVL64(cataID));
                                        emptyIDs.Remove(Convert.ToInt32(dRow[2]));
                                    }

                                    foreach (int emptyID in emptyIDs)
                                        Navigator.Append(Encoding.encodeVL64(emptyID) + "H" + navigatorManager.getName(emptyID) + Convert.ToChar(2) + "HH" + Encoding.encodeVL64(cataID));
                                }

                                sendData(Navigator.ToString());
                            }
                            break;
                        }

                    case "BW": // Navigator - request index of categories to place guestroom on
                        {
                            if (_Rank < 5)
                            {
                                //sendData("BK" + "Nooo");
                            }
                            else
                            {
                                StringBuilder Categories = new StringBuilder();
                                //Database dbClient = new Database(true, true, 147);
                                DataTable dTable;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dTable = dbClient.getTable("SELECT id,name FROM room_categories WHERE type = '2' AND parent > 0 AND access_rank_min <= " + _Rank);
                                }
                                foreach (DataRow dRow in dTable.Rows)
                                    Categories.Append(Encoding.encodeVL64(Convert.ToInt32(dRow["id"])) + dRow["name"] + Convert.ToChar(2));

                                sendData("C]" + Encoding.encodeVL64(dTable.Rows.Count) + Categories.ToString());
                            }
                            break;
                        }

                    case "DH": // Navigator - refresh recommended rooms (random guestrooms)
                        {
                            sendData("E_" + Encoding.encodeVL64(0) + navigatorManager.getRandomRooms());
                            break;
                        }

                    case "@P": // Navigator - view user's own guestrooms
                        {
                            if (_Rank < 5)
                            {
                                //sendData("BK" + "Nooo");
                            }
                            else
                            {
                                //Database dbClient = new Database(true, true, 149);
                                DataTable dTable;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dTable = dbClient.getTable("SELECT id,name,description,state,showname,visitors_now,visitors_max FROM rooms WHERE owner = '" + _Username + "' ORDER BY id ASC");
                                }
                                if (dTable.Rows.Count > 0)
                                {
                                    StringBuilder Rooms = new StringBuilder();
                                    foreach (DataRow dRow in dTable.Rows)
                                        Rooms.Append(Convert.ToString(dRow["id"]) + "\t" + Convert.ToString(dRow["name"]) + Convert.ToChar(9) + _Username + Convert.ToChar(9) + roomManager.getRoomState(Convert.ToInt32(dRow["state"])) + Convert.ToChar(9) + "x" + Convert.ToChar(9) + Convert.ToString(dRow["visitors_now"]) + Convert.ToChar(9) + Convert.ToString(dRow["visitors_max"]) + Convert.ToChar(9) + "null" + Convert.ToChar(9) + Convert.ToString(dRow["description"]) + Convert.ToChar(9) + Convert.ToString(dRow["description"]) + Convert.ToChar(9) + Convert.ToChar(13));
                                    sendData("@P" + Rooms.ToString());
                                }
                                else
                                {
                                    sendData("@y" + _Username);
                                }
                            }
                            break;
                        }

                    case "@Q": // Navigator - perform guestroom search on name/owner with a given criticeria
                        {
                            if (_Rank < 6)
                            {
                                //sendData("BK" + "Nooo");
                            }
                            else
                            {
                                bool seeAllRoomOwners = rankManager.containsRight(_Rank, "fuse_see_all_roomowners", userID);
                                DataTable dTable;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.AddParamWithValue("search", currentPacket.Substring(2));
                                    dbClient.AddParamWithValue("max", Config.Navigator_roomSearch_maxResults);
                                    //dbClient.Open();
                                    dTable = dbClient.getTable("SELECT id,name,owner,description,state,showname,visitors_now,visitors_max FROM rooms WHERE (owner = @search OR name LIKE '@search%') ORDER BY id ASC LIMIT @max");
                                }
                                if (dTable.Rows.Count > 0)
                                {
                                    StringBuilder Rooms = new StringBuilder();
                                    string nameString;
                                    foreach (DataRow dRow in dTable.Rows)
                                    {
                                        nameString = Convert.ToString(dRow["owner"]);
                                        if (Convert.ToString(dRow["showname"]) == "0" && Convert.ToString(dRow["owner"]) != _Username && seeAllRoomOwners == false)// The room owner has hidden his name at the guestroom and this user hasn't got the fuseright to see all room owners
                                            nameString = "-";
                                        Rooms.Append(Convert.ToString(dRow["id"]) + Convert.ToChar(9) + Convert.ToString(dRow["name"]) + Convert.ToChar(9) + Convert.ToString(dRow["owner"]) + Convert.ToChar(9) + roomManager.getRoomState(Convert.ToInt32(dRow["state"])) + Convert.ToChar(9) + "x" + Convert.ToChar(9) + Convert.ToString(dRow["visitors_now"]) + Convert.ToChar(9) + Convert.ToString(dRow["visitors_max"]) + Convert.ToChar(9) + "null" + Convert.ToChar(9) + Convert.ToString(dRow["description"]) + Convert.ToChar(9) + Convert.ToChar(13));
                                    }
                                    sendData("@w" + Rooms.ToString());
                                }
                                else
                                    sendData("@z");
                            }
                            break;
                        }

                    case "@U": // Navigator - get guestroom details
                        {
                            //if (_Rank < 0)
                            //{
                            //sendData("BK" + "Nooo");
                            //}
                            //else
                            //{
                            //Database dbClient = new Database(true, false, 151);
                            int roomID = int.Parse(currentPacket.Substring(2));
                            DataRow dRow;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                dRow = dbClient.getRow("SELECT name,owner,description,model,state,superusers,showname,category,visitors_now,visitors_max FROM rooms WHERE id = '" + roomID + "' AND NOT(owner IS NULL)");
                            }

                            if (dRow.Table.Rows.Count > 0) // Guestroom does exist
                            {
                                StringBuilder Details = new StringBuilder(Encoding.encodeVL64(Convert.ToInt32(dRow[5])) + Encoding.encodeVL64(Convert.ToInt32(dRow[4])) + Encoding.encodeVL64(roomID));
                                if (Convert.ToString(dRow[6]) == "0" && rankManager.containsRight(_Rank, "fuse_see_all_roomowners", userID)) // The room owner has decided to hide his name at this room, and this user hasn't got the fuseright to see all room owners, hide the name
                                    Details.Append("-");
                                else
                                    Details.Append(Convert.ToString(dRow[1]));

                                Details.Append(Convert.ToChar(2) + "model_" + Convert.ToString(dRow[3]) + Convert.ToChar(2) + Convert.ToString(dRow[0]) + Convert.ToChar(2) + Convert.ToString(dRow[2]) + Convert.ToChar(2) + Encoding.encodeVL64(Convert.ToInt32(dRow[6])));
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    if (dbClient.findsResult("SELECT id FROM room_categories WHERE id = '" + Convert.ToString(dRow[7]) + "' AND trading = '1'"))
                                        Details.Append("I"); // Allow trading
                                    else
                                        Details.Append("H"); // Disallow trading
                                }
                                Details.Append(Encoding.encodeVL64(Convert.ToInt32(dRow[8])) + Encoding.encodeVL64(Convert.ToInt32(dRow[9])));
                                sendData("@v" + Details.ToString());
                            }
                            //dbClient.Close();
                            //}
                            break;
                        }

                    case "@R": // Navigator - initialize user's favorite rooms
                        {
                            //Database dbClient = new Database(true, false, 152);
                            if (_Rank < 5)
                            {
                                //sendData("BK" + "Nooo");
                            }
                            else
                            {
                                DataColumn dCol;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dCol = dbClient.getColumn("SELECT roomid FROM users_favourites WHERE userid = '" + userID + "' ORDER BY roomid DESC LIMIT " + Config.Navigator_Favourites_maxRooms);
                                }
                                System.Collections.Hashtable deletedIDs = new System.Collections.Hashtable(dCol.Table.Rows.Count);

                                string roomIDs = "    ";
                                foreach (DataRow dRow in dCol.Table.Rows)
                                {
                                    deletedIDs.Add(Convert.ToInt32(dRow["roomid"]), Convert.ToInt32(dRow["roomid"]));
                                    roomIDs += "id = '" + Convert.ToString(dRow["roomid"]) + "' OR ";
                                }
                                roomIDs = roomIDs.Substring(0, roomIDs.Length - 4);
                                if (roomIDs.Length > 0)
                                {
                                    int guestRoomAmount = 0;
                                    string nameString;
                                    bool seeHiddenRoomOwners = rankManager.containsRight(_Rank, "fuse_enter_locked_rooms", userID);
                                    StringBuilder Rooms = new StringBuilder();
                                    DataTable dTable;
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dTable = dbClient.getTable("SELECT name,owner,state,showname,visitors_now,visitors_max,description,category,ccts,id FROM rooms WHERE " + roomIDs);
                                    }
                                    foreach (DataRow dRow in dTable.Rows)
                                    {
                                        deletedIDs.Remove(Convert.ToInt32(dRow["id"]));

                                        if (Convert.ToString(dRow[1]) == "")
                                            Rooms.Append(Encoding.encodeVL64(Convert.ToInt32(dRow["id"])) + "I" + Convert.ToString(dRow[0]) + Convert.ToChar(2) + Encoding.encodeVL64(Convert.ToInt32(dRow[4])) + Encoding.encodeVL64(Convert.ToInt32(dRow[5])) + Encoding.encodeVL64(Convert.ToInt32(dRow[7])) + Convert.ToString(dRow[6]) + Convert.ToChar(2) + Encoding.encodeVL64(Convert.ToInt32(dRow["id"])) + "H" + Convert.ToString(dRow[8]) + Convert.ToChar(2) + "HI");

                                        else // Guestroom
                                        {
                                            nameString = Convert.ToString(dRow[0]);
                                            if (Convert.ToString(dRow[3]) == "0" && _Username != Convert.ToString(dRow[1]) && seeHiddenRoomOwners == false) // Room owner doesn't wish to show his name, and this user isn't the room owner and this user doesn't has the right to see hidden room owners, change room owner to '-'
                                                nameString = "-";
                                            Rooms.Append(Encoding.encodeVL64(Convert.ToInt32(dRow["id"])) + nameString + Convert.ToChar(2) + Convert.ToString(dRow[1]) + Convert.ToChar(2) + roomManager.getRoomState(Convert.ToInt32(dRow[2])) + Convert.ToChar(2) + Encoding.encodeVL64(Convert.ToInt32(dRow[4])) + Encoding.encodeVL64(Convert.ToInt32(dRow[5])) + Convert.ToString(dRow[6]) + Convert.ToChar(2));
                                            guestRoomAmount++;
                                        }
                                    }
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        foreach (int rID in deletedIDs.Values)
                                            dbClient.runQuery("DELETE FROM users_favourites WHERE roomid = '" + rID + "' LIMIT 1");
                                    }
                                    //dbClient.Close();
                                    sendData("@}" + "HHJ" + Convert.ToChar(2) + "HHH" + Encoding.encodeVL64(guestRoomAmount - deletedIDs.Count) + Rooms.ToString());
                                }
                            }
                            break;
                        }
                    #endregion

                    #region Guestroom create and modify
                    case "@]": // Create guestroom - phase 1
                        {
                            if (_Rank < 6)
                            {
                                //sendData("BK" + "Nooo");
                            }
                            else
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    if (dbClient.getInt("SELECT COUNT(id) FROM rooms WHERE owner = '" + _Username + "'") < Config.Navigator_createRoom_maxRooms)
                                    {
                                        string[] roomSettings = currentPacket.Split('/');
                                        roomSettings[2] = stringManager.filterSwearwords(roomSettings[2]);
                                        roomSettings[3] = roomSettings[3].Substring(6, 1);
                                        roomSettings[4] = roomManager.getRoomState(roomSettings[4]).ToString();
                                        if (roomSettings[5] != "0" && roomSettings[5] != "1")
                                            return;
                                        dbClient.AddParamWithValue("rs2", roomSettings[2]);
                                        dbClient.AddParamWithValue("user", _Username);
                                        dbClient.AddParamWithValue("rs3", roomSettings[3]);
                                        dbClient.AddParamWithValue("rs4", roomSettings[4]);
                                        dbClient.AddParamWithValue("rs5", roomSettings[5]);
                                        dbClient.runQuery("INSERT INTO rooms (name, owner, model, state, showname) VALUES (@rs2,@user,@rs3,@rs4,@rs5)");
                                        string roomID = dbClient.getString("SELECT MAX(id) FROM rooms WHERE owner = @user");
                                        sendData("@{" + roomID + Convert.ToChar(13) + roomSettings[2]);
                                    }

                                    else
                                        sendData("@a" + "Error creating a private room");
                                }
                            }
                            break;
                        }

                    case "@Y": // Create guestroom - phase 2 / modify guestroom
                        {
                            if (_Rank < 6)
                            {
                                //sendData("BK" + "Nooo");
                            }
                            else
                            {
                                int roomID = 0;
                                if (currentPacket.Substring(2, 1) == "/")
                                    roomID = int.Parse(currentPacket.Split('/')[1]);
                                else
                                    roomID = int.Parse(currentPacket.Substring(2).Split('/')[0]);

                                string superUsers = "0";
                                int maxVisitors = 25;
                                string[] packetContent = currentPacket.Split(Convert.ToChar(13));
                                string roomDescription = "";
                                string roomPassword = "";

                                for (int i = 1; i < packetContent.Length; i++) // More proper way, thanks Jeax
                                {
                                    string updHeader = packetContent[i].Split('=')[0];
                                    string updValue = packetContent[i].Substring(updHeader.Length + 1);
                                    switch (updHeader)
                                    {
                                        case "description":
                                            roomDescription = stringManager.filterSwearwords(updValue);
                                            break;

                                        case "allsuperuser":
                                            superUsers = updValue;
                                            if (superUsers != "0" && superUsers != "1")
                                                superUsers = "0";

                                            break;

                                        case "maxvisitors":
                                            maxVisitors = int.Parse(updValue);
                                            if (maxVisitors < 10 || maxVisitors > 50)
                                                maxVisitors = 25;
                                            break;

                                        case "password":
                                            roomPassword = updValue;
                                            break;

                                        default:
                                            return;
                                    }
                                }
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.AddParamWithValue("desc", roomDescription);
                                    dbClient.AddParamWithValue("super", superUsers);
                                    dbClient.AddParamWithValue("max", maxVisitors);
                                    dbClient.AddParamWithValue("pass", roomPassword);
                                    dbClient.AddParamWithValue("id", roomID);
                                    dbClient.AddParamWithValue("owner", _Username);
                                    dbClient.runQuery("UPDATE rooms SET description = @desc,superusers = @super,visitors_max = @max, password = @pass WHERE id = @id AND owner = @owner LIMIT 1");
                                }
                            }
                            break;
                        }

                    case "@X": // Modify guestroom, save name, state and show/hide ownername
                        {
                            if (_Rank < 6)
                            {
                                //sendData("BK" + "Nooo");
                            }
                            else
                            {
                                string[] packetContent = currentPacket.Substring(2).Split('/');
                                if (packetContent[3] != "1" && packetContent[2] != "0")
                                    packetContent[2] = "1";
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.AddParamWithValue("name", stringManager.filterSwearwords(packetContent[1]));
                                    dbClient.AddParamWithValue("state", roomManager.getRoomState(packetContent[2]));
                                    dbClient.AddParamWithValue("show", packetContent[3]);
                                    dbClient.AddParamWithValue("id", packetContent[0]);
                                    dbClient.AddParamWithValue("owner", _Username);
                                    dbClient.runQuery("UPDATE rooms SET name = @name,state = @state,showname = @show WHERE id = @id AND owner = @owner LIMIT 1");

                                }
                            }
                            break;
                        }

                    case "BX": // Navigator - trigger guestroom modify
                        {
                            if (_Rank < 6)
                            {
                                //sendData("BK" + "Nooo");
                            }
                            else
                            {
                                int roomID = Encoding.decodeVL64(currentPacket.Substring(2));
                                //Database dbClient = new Database(true, true, 80);
                                string roomCategory;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    roomCategory = dbClient.getString("SELECT category FROM rooms WHERE id = '" + roomID + "' AND owner = '" + _Username + "'");
                                }
                                if (roomCategory != "")
                                    sendData("C^" + Encoding.encodeVL64(roomID) + Encoding.encodeVL64(int.Parse(roomCategory)));
                            }
                            break;
                        }

                    case "BY": // Navigator - edit category of a guestroom
                        {
                            if (_Rank < 6)
                            {
                                //sendData("BK" + "Nooo");
                            }
                            else
                            {
                                int roomID = Encoding.decodeVL64(currentPacket.Substring(2));
                                int cataID = Encoding.decodeVL64(currentPacket.Substring(Encoding.encodeVL64(roomID).Length + 2));
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    if (dbClient.findsResult("SELECT id FROM room_categories WHERE id = '" + cataID + "' AND type = '2' AND parent > 0 AND access_rank_min <= " + _Rank)) // Category is valid for this user
                                        dbClient.runQuery("UPDATE rooms SET category = '" + cataID + "' WHERE id = '" + roomID + "' AND owner = '" + _Username + "' LIMIT 1");
                                }
                            }
                            break;
                        }

                    case "@W": // Guestroom - Delete
                        {
                            if (_Rank < 6)
                            {
                                //sendData("BK" + "Nooo");
                            }
                            else
                            {
                                int roomID = int.Parse(currentPacket.Substring(2));
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    if (dbClient.findsResult("SELECT id FROM rooms WHERE id = '" + roomID + "' AND owner = '" + _Username + "'") == true)
                                    {
                                        dbClient.runQuery("DELETE FROM room_rights WHERE roomid = '" + roomID + "'");
                                        dbClient.runQuery("DELETE FROM rooms WHERE id = '" + roomID + "' LIMIT 1");
                                        dbClient.runQuery("DELETE FROM users_favourites WHERE roomid = '" + _roomID + "'");
                                        dbClient.runQuery("DELETE FROM room_votes WHERE roomid = '" + roomID + "'");
                                        dbClient.runQuery("DELETE FROM room_bans WHERE roomid = '" + roomID + "' LIMIT 1");
                                        dbClient.runQuery("DELETE FROM furniture WHERE roomid = '" + roomID + "'");
                                        dbClient.runQuery("DELETE FROM furniture_moodlight WHERE roomid = '" + roomID + "'");
                                    }
                                }
                                if (roomManager.containsRoom(roomID) == true)
                                {
                                    roomManager.getRoom(roomID).kickUsers(byte.Parse("9"), "This room has been deleted");
                                }
                            }
                            break;
                        }

                    case "BZ": // Navigator - 'Who's in here' feature for public rooms
                        {
                            if (_Rank < 6)
                            {
                                //sendData("BK" + "Nooo");
                            }
                            else
                            {
                                int roomID = Encoding.decodeVL64(currentPacket.Substring(2));
                                if (roomManager.containsRoom(roomID))
                                    sendData("C_" + roomManager.getRoom(roomID).Userlist);
                                else
                                    sendData("C_");
                            }
                            break;
                        }
                    #endregion

                    #region Enter/leave room
                    case "@u": // Rooms - leave room
                        {
                            if (Room != null && roomUser != null)
                            {
                                virtualBot Pet = botManager.getBot(_PetID);
                                Room.deloadBot(Pet);
                                Room.removeUser(roomUser.roomUID, false, "");
                            }
                            break;
                        }

                    case "Bv": // Enter room - loading screen advertisement
                        {
                                sendData("DB" + "test" + Convert.ToChar(9) + "test");
                        }
                        break;

                    case "@B": // Enter room - determine room and check state + max visitors override
                        {
                            virtualUser User = userManager.getUser(_Username);
                            if (_onRoad)
                            {
                                _MarryingTo = "None";
                                _Cuffed = false;
                                _Stunned = false;
                                _StunTime = 0;
                                _onRoad = false;
                                _NeedsHeal = false;
                                _TaxiTime = 0;
                                _Taxi = 0;
                                _Punch = 0;
                                int roomID = Encoding.decodeVL64(currentPacket.Substring(3));
                                if (roomID != roomidi && roomidi != 0)
                                {
                                    roomID = roomidi;
                                    roomidi = 0;
                                }
                                bool isPublicroom = (currentPacket.Substring(2, 1) == "A");

                                sendData("@S");
                                sendData("Bf" + "http://wwww.sunnieday.nl/");

                                if (_Rank == 0 || User == null)
                                {
                                    sendData("BK" + "Your Account is Fucked please reload");
                                    Disconnect();
                                }
                                else
                                {
                                    if (Room != null && roomUser != null)
                                        Room.removeUser(roomUser.roomUID, false, "");


                                    //Database dbClient = new Database(true, false, 83);
                                    if (_teleporterID == 0)
                                    {
                                        bool allowEnterLockedRooms = rankManager.containsRight(_Rank, "fuse_enter_locked_rooms", userID);
                                        int accessLevel;
                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {
                                            accessLevel = dbClient.getInt("SELECT state FROM rooms WHERE id = '" + roomID + "'");
                                        }
                                        if (accessLevel == 4 && allowEnterLockedRooms == false) // The room is only for staff and the user hasn't got the fuseright for entering all rooms nomatter the state
                                        {
                                            sendData("BK" + stringManager.getString("room_stafflocked"));
                                            //dbClient.Close();
                                            return;
                                        }
                                    }

                                    _roomID = roomID;
                                    _inPublicroom = isPublicroom;
                                    _ROOMACCESS_PRIMARY_OK = true;

                                    if (isPublicroom)
                                    {
                                        string roomModel;
                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {
                                            roomModel = dbClient.getString("SELECT model FROM rooms WHERE id = '" + roomID + "'");
                                        }
                                        sendData("AE" + roomModel + " " + roomID);
                                        _ROOMACCESS_SECONDARY_OK = true;
                                    }
                                    //dbClient.Close();

                                }
                            }
                            
                            break;
                        }

                    case "@v": // Enter room - guestroom - enter room by using a teleporter
                        {
                            

                            sendData("@S");
                            break;
                        }

                    case "@y": // Enter room - guestroom - check roomban/password/doorbell
                        {
                            if (_inPublicroom == false)
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    _isOwner = dbClient.findsResult("SELECT id FROM rooms WHERE id = '" + _roomID + "' AND owner = '" + _Username + "'");
                                    if (_isOwner == false)
                                        _hasRights = dbClient.findsResult("SELECT userid FROM room_rights WHERE roomid = '" + _roomID + "' AND userid = '" + userID + "'");
                                    if (_hasRights == false)
                                        _hasRights = dbClient.findsResult("SELECT id FROM rooms WHERE id = '" + _roomID + "' AND superusers = '1'");

                                    if (_teleporterID == 0 && _isOwner == false && rankManager.containsRight(_Rank, "fuse_enter_locked_rooms", userID) == false)
                                    {
                                        int accessFlag = dbClient.getInt("SELECT state FROM rooms WHERE id = '" + _roomID + "'");
                                        if (_ROOMACCESS_PRIMARY_OK == false && accessFlag != 2)
                                        {
                                            //dbClient.Close();
                                            return;
                                        }

                                        // Check for roombans
                                        //dbClient.Open();
                                        if (dbClient.findsResult("SELECT roomid FROM room_bans WHERE roomid = '" + _roomID + "' AND userid = '" + userID + "'"))
                                        {
                                            DateTime banExpireMoment = DateTime.Parse(dbClient.getString("SELECT ban_expire FROM room_bans WHERE roomid = '" + _roomID + "' AND userid = '" + userID + "'"));
                                            if (DateTime.Compare(banExpireMoment, DateTime.Now) > 0)
                                            {
                                                sendData("C`" + "PA");
                                                sendData("@R");
                                                //dbClient.Close();
                                                return;
                                            }
                                            else
                                                dbClient.runQuery("DELETE FROM room_bans WHERE roomid = '" + _roomID + "' AND userid = '" + userID + "' LIMIT 1");
                                        }
                                        if (accessFlag == 2) // Password
                                        {
                                            string givenPassword = "";
                                            try { givenPassword = currentPacket.Split('/')[1]; }
                                            catch { }
                                            string roomPassword = dbClient.getString("SELECT password FROM rooms WHERE id = '" + _roomID + "'");
                                            if (givenPassword != roomPassword) { sendData("@a" + "Incorrect flat password"); /*dbClient.Close();*/ return; }
                                        }
                                    }
                                    //dbClient.Close();
                                    _ROOMACCESS_SECONDARY_OK = true;
                                    sendData("@i");
                                }
                            }
                            break;
                        }

                    case "@{": // Enter room - guestroom - guestroom only data: model, landscape, wallpaper, rights, room votes
                        {
                            if (_ROOMACCESS_SECONDARY_OK && _inPublicroom == false)
                            {
                                //Database dbClient = new Database(true, false, 85);
                                DataRow dRow;
                                string Landscape;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dRow = dbClient.getRow("SELECT model, wallpaper, floor FROM rooms WHERE id = '" + _roomID + "'");
                                    Landscape = dbClient.getString("SELECT landscape FROM rooms WHERE id = '" + _roomID + "'");

                                }
                                string Model = "model_" + Convert.ToString(dRow["model"]);
                                sendData("AE" + Model + " " + _roomID);
                                int Wallpaper = Convert.ToInt32(dRow["wallpaper"]);
                                int Floor = Convert.ToInt32(dRow["floor"]);
                                sendData("@n" + "landscape/" + Landscape.Replace(",", "."));
                                if (Wallpaper > 0)
                                    sendData("@n" + "wallpaper/" + Wallpaper);
                                if (Floor > 0)
                                    sendData("@n" + "floor/" + Floor);


                                if (_isOwner == false)
                                {
                                    _isOwner = rankManager.containsRight(_Rank, "fuse_any_room_controller", userID);
                                }
                                if (_isOwner)
                                {
                                    _hasRights = true;
                                    sendData("@o");
                                }
                                if (_hasRights)
                                    sendData("@j");

                                int voteAmount = -1;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    if (dbClient.findsResult("SELECT userid FROM room_votes WHERE userid = '" + userID + "' AND roomid = '" + _roomID + "'"))
                                    {
                                        voteAmount = dbClient.getInt("SELECT SUM(vote) FROM room_votes WHERE roomid = '" + _roomID + "'");
                                        if (voteAmount < 0) { voteAmount = 0; }
                                    }
                                }
                                sendData("EY" + Encoding.encodeVL64(voteAmount));
                            }
                            break;
                        }

                    case "A~": // Enter room - get room advertisement
                        {
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                if (_inPublicroom && dbClient.findsResult("SELECT roomid FROM room_ads WHERE roomid = '" + _roomID + "'"))
                                {
                                    DataRow dRow = dbClient.getRow("SELECT img, uri FROM room_ads WHERE roomid = '" + _roomID + "'");
                                    string advImg = Convert.ToString(dRow["img"]);
                                    string advUri = Convert.ToString(dRow["uri"]);
                                    sendData("CP" + advImg + Convert.ToChar(9) + advUri);
                                }
                                else
                                    sendData("CP" + "test" + Convert.ToChar(9) + "test");
                            }
                            break;
                        }

                    case "@|": // Enter room - get roomclass + get heightmap
                        {
                            if (_ROOMACCESS_SECONDARY_OK)
                            {
                                if (roomManager.containsRoom(_roomID))
                                    Room = roomManager.getRoom(_roomID);
                                else
                                {
                                    Room = new virtualRoom(_roomID, _inPublicroom);
                                    roomManager.addRoom(_roomID, Room);
                                }

                                sendData("@_" + Room.Heightmap);
                                sendData(@"@\" + Room.dynamicUnits);
                            }
                            else
                            {
                            }
                            break;
                        }

                    case "@}": // Enter room - get items
                        {
                            if (_ROOMACCESS_SECONDARY_OK && Room != null)
                            {
                                sendData("@^" + Room.PublicroomItems);
                                sendData("@`" + Room.Flooritems);
                            }
                            break;
                        }

                    case "@~": // Enter room - get group badges, optional skill levels in game lobbies and sprite index
                        {
                            if (_ROOMACCESS_SECONDARY_OK && Room != null)
                            {
                                sendData("Du" + Room.Groups);

                                sendData("DiH");
                                if (_receivedSpriteIndex == false)
                                {
                                    /*
                                     * 
                                     * New sprite index
                                     * Added by Nillus
                                     * 
                                     */
                                    sendData("Dg" + @"[SEshelves_norjaX~Dshelves_polyfonYmAshelves_siloXQHtable_polyfon_smallYmAchair_polyfonZbBtable_norja_medY_Itable_silo_medX~Dtable_plasto_4legY_Itable_plasto_roundY_Itable_plasto_bigsquareY_Istand_polyfon_zZbBchair_siloX~Dsofa_siloX~Dcouch_norjaX~Dchair_norjaX~Dtable_polyfon_medYmAdoormat_loveZbBdoormat_plainZ[Msofachair_polyfonX~Dsofa_polyfonZ[Msofachair_siloX~Dchair_plastyX~Dchair_plastoYmAtable_plasto_squareY_Ibed_polyfonX~Dbed_polyfon_one[dObed_trad_oneYmAbed_tradYmAbed_silo_oneYmAbed_silo_twoYmAtable_silo_smallX~Dbed_armas_twoYmAbed_budget_oneXQHbed_budgetXQHshelves_armasYmAbench_armasYmAtable_armasYmAsmall_table_armasZbBsmall_chair_armasYmAfireplace_armasYmAlamp_armasYmAbed_armas_oneYmAcarpet_standardY_Icarpet_armasYmAcarpet_polarY_Ifireplace_polyfonY_Itable_plasto_4leg*1Y_Itable_plasto_bigsquare*1Y_Itable_plasto_round*1Y_Itable_plasto_square*1Y_Ichair_plasto*1YmAcarpet_standard*1Y_Idoormat_plain*1Z[Mtable_plasto_4leg*2Y_Itable_plasto_bigsquare*2Y_Itable_plasto_round*2Y_Itable_plasto_square*2Y_Ichair_plasto*2YmAdoormat_plain*2Z[Mcarpet_standard*2Y_Itable_plasto_4leg*3Y_Itable_plasto_bigsquare*3Y_Itable_plasto_round*3Y_Itable_plasto_square*3Y_Ichair_plasto*3YmAcarpet_standard*3Y_Idoormat_plain*3Z[Mtable_plasto_4leg*4Y_Itable_plasto_bigsquare*4Y_Itable_plasto_round*4Y_Itable_plasto_square*4Y_Ichair_plasto*4YmAcarpet_standard*4Y_Idoormat_plain*4Z[Mdoormat_plain*6Z[Mdoormat_plain*5Z[Mcarpet_standard*5Y_Itable_plasto_4leg*5Y_Itable_plasto_bigsquare*5Y_Itable_plasto_round*5Y_Itable_plasto_square*5Y_Ichair_plasto*5YmAtable_plasto_4leg*6Y_Itable_plasto_bigsquare*6Y_Itable_plasto_round*6Y_Itable_plasto_square*6Y_Ichair_plasto*6YmAtable_plasto_4leg*7Y_Itable_plasto_bigsquare*7Y_Itable_plasto_round*7Y_Itable_plasto_square*7Y_Ichair_plasto*7YmAtable_plasto_4leg*8Y_Itable_plasto_bigsquare*8Y_Itable_plasto_round*8Y_Itable_plasto_square*8Y_Ichair_plasto*8YmAtable_plasto_4leg*9Y_Itable_plasto_bigsquare*9Y_Itable_plasto_round*9Y_Itable_plasto_square*9Y_Ichair_plasto*9YmAcarpet_standard*6Y_Ichair_plasty*1X~DpizzaYmAdrinksYmAchair_plasty*2X~Dchair_plasty*3X~Dchair_plasty*4X~Dbar_polyfonY_Iplant_cruddyYmAbottleYmAbardesk_polyfonX~Dbardeskcorner_polyfonX~DfloortileHbar_armasY_Ibartable_armasYmAbar_chair_armasYmAcarpet_softZ@Kcarpet_soft*1Z@Kcarpet_soft*2Z@Kcarpet_soft*3Z@Kcarpet_soft*4Z@Kcarpet_soft*5Z@Kcarpet_soft*6Z@Kred_tvY_Iwood_tvYmAcarpet_polar*1Y_Ichair_plasty*5X~Dcarpet_polar*2Y_Icarpet_polar*3Y_Icarpet_polar*4Y_Ichair_plasty*6X~Dtable_polyfonYmAsmooth_table_polyfonYmAsofachair_polyfon_girlX~Dbed_polyfon_girl_one[dObed_polyfon_girlX~Dsofa_polyfon_girlZ[Mbed_budgetb_oneXQHbed_budgetbXQHplant_pineappleYmAplant_fruittreeY_Iplant_small_cactusY_Iplant_bonsaiY_Iplant_big_cactusY_Iplant_yukkaY_Icarpet_standard*7Y_Icarpet_standard*8Y_Icarpet_standard*9Y_Icarpet_standard*aY_Icarpet_standard*bY_Iplant_sunflowerY_Iplant_roseY_Itv_luxusY_IbathZ\BsinkY_ItoiletYmAduckYmAtileYmAtoilet_redYmAtoilet_yellYmAtile_redYmAtile_yellYmApresent_gen[~Npresent_gen1[~Npresent_gen2[~Npresent_gen3[~Npresent_gen4[~Npresent_gen5[~Npresent_gen6[~Nbar_basicY_Ishelves_basicXQHsoft_sofachair_norjaX~Dsoft_sofa_norjaX~Dlamp_basicXQHlamp2_armasYmAfridgeY_IdoorYc[doorBYc[doorCYc[pumpkinYmAskullcandleYmAdeadduckYmAdeadduck2YmAdeadduck3YmAmenorahYmApuddingYmAhamYmAturkeyYmAxmasduckY_IhouseYmAtriplecandleYmAtree3YmAtree4YmAtree5X~Dham2YmAwcandlesetYmArcandlesetYmAstatueYmAheartY_IvaleduckYmAheartsofaX~DthroneYmAsamovarY_IgiftflowersY_IhabbocakeYmAXoagramYmAeasterduckY_IbunnyYmAbasketY_IbirdieYmAediceX~Dclub_sofaZ[Mprize1YmAprize2YmAprize3YmAdivider_poly3X~Ddivider_arm1YmAdivider_arm2YmAdivider_arm3YmAdivider_nor1X~Ddivider_silo1X~Ddivider_nor2X~Ddivider_silo2Z[Mdivider_nor3X~Ddivider_silo3X~DtypingmachineYmAspyroYmAredXoagramYmAcameraHjoulutahtiYmAhyacinth1YmAhyacinth2YmAchair_plasto*10YmAchair_plasto*11YmAbardeskcorner_polyfon*12X~Dbardeskcorner_polyfon*13X~Dchair_plasto*12YmAchair_plasto*13YmAchair_plasto*14YmAtable_plasto_4leg*14Y_ImocchamasterY_Icarpet_legocourtYmAbench_legoYmAlegotrophyYmAvalentinescreenYmAedicehcYmArare_daffodil_rugYmArare_beehive_bulbY_IhcsohvaYmAhcammeYmArare_elephant_statueYmArare_fountainY_Irare_standYmArare_globeYmArare_hammockYmArare_elephant_statue*1YmArare_elephant_statue*2YmArare_fountain*1Y_Irare_fountain*2Y_Irare_fountain*3Y_Irare_beehive_bulb*1Y_Irare_beehive_bulb*2Y_Irare_xmas_screenY_Irare_parasol*1XMVrare_parasol*2XMVrare_parasol*3XMVtree1X~Dtree2ZmBwcandleYxBrcandleYxBsoft_jaggara_norjaYmAhouse2YmAdjesko_turntableYmAmd_sofaZ[Mmd_limukaappiY_Itable_plasto_4leg*10Y_Itable_plasto_4leg*15Y_Itable_plasto_bigsquare*14Y_Itable_plasto_bigsquare*15Y_Itable_plasto_round*14Y_Itable_plasto_round*15Y_Itable_plasto_square*14Y_Itable_plasto_square*15Y_Ichair_plasto*15YmAchair_plasty*7X~Dchair_plasty*8X~Dchair_plasty*9X~Dchair_plasty*10X~Dchair_plasty*11X~Dchair_plasto*16YmAtable_plasto_4leg*16Y_Ihockey_scoreY_Ihockey_lightYmAdoorDYc[prizetrophy2*3Yd[prizetrophy3*3Yd[prizetrophy4*3Yd[prizetrophy5*3Yd[prizetrophy6*3Yd[prizetrophy*1Yd[prizetrophy2*1Yd[prizetrophy3*1Yd[prizetrophy4*1Yd[prizetrophy5*1Yd[prizetrophy6*1Yd[prizetrophy*2Yd[prizetrophy2*2Yd[prizetrophy3*2Yd[prizetrophy4*2Yd[prizetrophy5*2Yd[prizetrophy6*2Yd[prizetrophy*3Yd[rare_parasol*0XMVhc_lmp[fBhc_tblYmAhc_chrYmAhc_dskXQHnestHpetfood1ZvCpetfood2ZvCpetfood3ZvCwaterbowl*4XICwaterbowl*5XICwaterbowl*2XICwaterbowl*1XICwaterbowl*3XICtoy1XICtoy1*1XICtoy1*2XICtoy1*3XICtoy1*4XICgoodie1Yc[goodie1*1Yc[goodie1*2Yc[goodie2Yc[prizetrophy7*3Yd[prizetrophy7*1Yd[prizetrophy7*2Yd[scifiport*0Y_Iscifiport*9Y_Iscifiport*8Y_Iscifiport*7Y_Iscifiport*6Y_Iscifiport*5Y_Iscifiport*4Y_Iscifiport*3Y_Iscifiport*2Y_Iscifiport*1Y_Iscifirocket*9Y_Iscifirocket*8Y_Iscifirocket*7Y_Iscifirocket*6Y_Iscifirocket*5Y_Iscifirocket*4Y_Iscifirocket*3Y_Iscifirocket*2Y_Iscifirocket*1Y_Iscifirocket*0Y_Iscifidoor*10Y_Iscifidoor*9Y_Iscifidoor*8Y_Iscifidoor*7Y_Iscifidoor*6Y_Iscifidoor*5Y_Iscifidoor*4Y_Iscifidoor*3Y_Iscifidoor*2Y_Iscifidoor*1Y_Ipillow*5YmApillow*8YmApillow*0YmApillow*1YmApillow*2YmApillow*7YmApillow*9YmApillow*4YmApillow*6YmApillow*3YmAmarquee*1Y_Imarquee*2Y_Imarquee*7Y_Imarquee*aY_Imarquee*8Y_Imarquee*9Y_Imarquee*5Y_Imarquee*4Y_Imarquee*6Y_Imarquee*3Y_Iwooden_screen*1Y_Iwooden_screen*2Y_Iwooden_screen*7Y_Iwooden_screen*0Y_Iwooden_screen*8Y_Iwooden_screen*5Y_Iwooden_screen*9Y_Iwooden_screen*4Y_Iwooden_screen*6Y_Iwooden_screen*3Y_Ipillar*6Y_Ipillar*1Y_Ipillar*9Y_Ipillar*0Y_Ipillar*8Y_Ipillar*2Y_Ipillar*5Y_Ipillar*4Y_Ipillar*7Y_Ipillar*3Y_Irare_dragonlamp*4Y_Irare_dragonlamp*0Y_Irare_dragonlamp*5Y_Irare_dragonlamp*2Y_Irare_dragonlamp*8Y_Irare_dragonlamp*9Y_Irare_dragonlamp*7Y_Irare_dragonlamp*6Y_Irare_dragonlamp*1Y_Irare_dragonlamp*3Y_Irare_icecream*1Y_Irare_icecream*7Y_Irare_icecream*8Y_Irare_icecream*2Y_Irare_icecream*6Y_Irare_icecream*9Y_Irare_icecream*3Y_Irare_icecream*0Y_Irare_icecream*4Y_Irare_icecream*5Y_Irare_fan*7YxBrare_fan*6YxBrare_fan*9YxBrare_fan*3YxBrare_fan*0YxBrare_fan*4YxBrare_fan*5YxBrare_fan*1YxBrare_fan*8YxBrare_fan*2YxBqueue_tile1*3X~Dqueue_tile1*6X~Dqueue_tile1*4X~Dqueue_tile1*9X~Dqueue_tile1*8X~Dqueue_tile1*5X~Dqueue_tile1*7X~Dqueue_tile1*2X~Dqueue_tile1*1X~Dqueue_tile1*0X~DticketHrare_snowrugX~Dcn_lampZxIcn_sofaYmAsporttrack1*1YmAsporttrack1*3YmAsporttrack1*2YmAsporttrack2*1[~Nsporttrack2*2[~Nsporttrack2*3[~Nsporttrack3*1YmAsporttrack3*2YmAsporttrack3*3YmAfootylampX~Dbarchair_siloX~Ddivider_nor4*4X~Dtraffic_light*1ZxItraffic_light*2ZxItraffic_light*3ZxItraffic_light*4ZxItraffic_light*6ZxIrubberchair*1X~Drubberchair*2X~Drubberchair*3X~Drubberchair*4X~Drubberchair*5X~Drubberchair*6X~Dbarrier*1X~Dbarrier*2X~Dbarrier*3X~Drubberchair*7X~Drubberchair*8X~Dtable_norja_med*2Y_Itable_norja_med*3Y_Itable_norja_med*4Y_Itable_norja_med*5Y_Itable_norja_med*6Y_Itable_norja_med*7Y_Itable_norja_med*8Y_Itable_norja_med*9Y_Icouch_norja*2X~Dcouch_norja*3X~Dcouch_norja*4X~Dcouch_norja*5X~Dcouch_norja*6X~Dcouch_norja*7X~Dcouch_norja*8X~Dcouch_norja*9X~Dshelves_norja*2X~Dshelves_norja*3X~Dshelves_norja*4X~Dshelves_norja*5X~Dshelves_norja*6X~Dshelves_norja*7X~Dshelves_norja*8X~Dshelves_norja*9X~Dchair_norja*2X~Dchair_norja*3X~Dchair_norja*4X~Dchair_norja*5X~Dchair_norja*6X~Dchair_norja*7X~Dchair_norja*8X~Dchair_norja*9X~Ddivider_nor1*2X~Ddivider_nor1*3X~Ddivider_nor1*4X~Ddivider_nor1*5X~Ddivider_nor1*6X~Ddivider_nor1*7X~Ddivider_nor1*8X~Ddivider_nor1*9X~Dsoft_sofa_norja*2X~Dsoft_sofa_norja*3X~Dsoft_sofa_norja*4X~Dsoft_sofa_norja*5X~Dsoft_sofa_norja*6X~Dsoft_sofa_norja*7X~Dsoft_sofa_norja*8X~Dsoft_sofa_norja*9X~Dsoft_sofachair_norja*2X~Dsoft_sofachair_norja*3X~Dsoft_sofachair_norja*4X~Dsoft_sofachair_norja*5X~Dsoft_sofachair_norja*6X~Dsoft_sofachair_norja*7X~Dsoft_sofachair_norja*8X~Dsoft_sofachair_norja*9X~Dsofachair_silo*2X~Dsofachair_silo*3X~Dsofachair_silo*4X~Dsofachair_silo*5X~Dsofachair_silo*6X~Dsofachair_silo*7X~Dsofachair_silo*8X~Dsofachair_silo*9X~Dtable_silo_small*2X~Dtable_silo_small*3X~Dtable_silo_small*4X~Dtable_silo_small*5X~Dtable_silo_small*6X~Dtable_silo_small*7X~Dtable_silo_small*8X~Dtable_silo_small*9X~Ddivider_silo1*2X~Ddivider_silo1*3X~Ddivider_silo1*4X~Ddivider_silo1*5X~Ddivider_silo1*6X~Ddivider_silo1*7X~Ddivider_silo1*8X~Ddivider_silo1*9X~Ddivider_silo3*2X~Ddivider_silo3*3X~Ddivider_silo3*4X~Ddivider_silo3*5X~Ddivider_silo3*6X~Ddivider_silo3*7X~Ddivider_silo3*8X~Ddivider_silo3*9X~Dtable_silo_med*2X~Dtable_silo_med*3X~Dtable_silo_med*4X~Dtable_silo_med*5X~Dtable_silo_med*6X~Dtable_silo_med*7X~Dtable_silo_med*8X~Dtable_silo_med*9X~Dsofa_silo*2X~Dsofa_silo*3X~Dsofa_silo*4X~Dsofa_silo*5X~Dsofa_silo*6X~Dsofa_silo*7X~Dsofa_silo*8X~Dsofa_silo*9X~Dsofachair_polyfon*2X~Dsofachair_polyfon*3X~Dsofachair_polyfon*4X~Dsofachair_polyfon*6X~Dsofachair_polyfon*7X~Dsofachair_polyfon*8X~Dsofachair_polyfon*9X~Dsofa_polyfon*2Z[Msofa_polyfon*3Z[Msofa_polyfon*4Z[Msofa_polyfon*6Z[Msofa_polyfon*7Z[Msofa_polyfon*8Z[Msofa_polyfon*9Z[Mbed_polyfon*2X~Dbed_polyfon*3X~Dbed_polyfon*4X~Dbed_polyfon*6X~Dbed_polyfon*7X~Dbed_polyfon*8X~Dbed_polyfon*9X~Dbed_polyfon_one*2[dObed_polyfon_one*3[dObed_polyfon_one*4[dObed_polyfon_one*6[dObed_polyfon_one*7[dObed_polyfon_one*8[dObed_polyfon_one*9[dObardesk_polyfon*2X~Dbardesk_polyfon*3X~Dbardesk_polyfon*4X~Dbardesk_polyfon*5X~Dbardesk_polyfon*6X~Dbardesk_polyfon*7X~Dbardesk_polyfon*8X~Dbardesk_polyfon*9X~Dbardeskcorner_polyfon*2X~Dbardeskcorner_polyfon*3X~Dbardeskcorner_polyfon*4X~Dbardeskcorner_polyfon*5X~Dbardeskcorner_polyfon*6X~Dbardeskcorner_polyfon*7X~Dbardeskcorner_polyfon*8X~Dbardeskcorner_polyfon*9X~Ddivider_poly3*2X~Ddivider_poly3*3X~Ddivider_poly3*4X~Ddivider_poly3*5X~Ddivider_poly3*6X~Ddivider_poly3*7X~Ddivider_poly3*8X~Ddivider_poly3*9X~Dchair_silo*2X~Dchair_silo*3X~Dchair_silo*4X~Dchair_silo*5X~Dchair_silo*6X~Dchair_silo*7X~Dchair_silo*8X~Dchair_silo*9X~Ddivider_nor3*2X~Ddivider_nor3*3X~Ddivider_nor3*4X~Ddivider_nor3*5X~Ddivider_nor3*6X~Ddivider_nor3*7X~Ddivider_nor3*8X~Ddivider_nor3*9X~Ddivider_nor2*2X~Ddivider_nor2*3X~Ddivider_nor2*4X~Ddivider_nor2*5X~Ddivider_nor2*6X~Ddivider_nor2*7X~Ddivider_nor2*8X~Ddivider_nor2*9X~Dsilo_studydeskX~Dsolarium_norjaY_Isolarium_norja*1Y_Isolarium_norja*2Y_Isolarium_norja*3Y_Isolarium_norja*5Y_Isolarium_norja*6Y_Isolarium_norja*7Y_Isolarium_norja*8Y_Isolarium_norja*9Y_IsandrugX~Drare_moonrugYmAchair_chinaYmAchina_tableYmAsleepingbag*1YmAsleepingbag*2YmAsleepingbag*3YmAsleepingbag*4YmAsafe_siloY_Isleepingbag*7YmAsleepingbag*9YmAsleepingbag*5YmAsleepingbag*10YmAsleepingbag*6YmAsleepingbag*8YmAchina_shelveX~Dtraffic_light*5ZxIdivider_nor4*2X~Ddivider_nor4*3X~Ddivider_nor4*5X~Ddivider_nor4*6X~Ddivider_nor4*7X~Ddivider_nor4*8X~Ddivider_nor4*9X~Ddivider_nor5*2X~Ddivider_nor5*3X~Ddivider_nor5*4X~Ddivider_nor5*5X~Ddivider_nor5*6X~Ddivider_nor5*7X~Ddivider_nor5*8X~Ddivider_nor5*9X~Ddivider_nor5X~Ddivider_nor4X~Dwall_chinaYmAcorner_chinaYmAbarchair_silo*2X~Dbarchair_silo*3X~Dbarchair_silo*4X~Dbarchair_silo*5X~Dbarchair_silo*6X~Dbarchair_silo*7X~Dbarchair_silo*8X~Dbarchair_silo*9X~Dsafe_silo*2Y_Isafe_silo*3Y_Isafe_silo*4Y_Isafe_silo*5Y_Isafe_silo*6Y_Isafe_silo*7Y_Isafe_silo*8Y_Isafe_silo*9Y_Iglass_shelfY_Iglass_chairY_Iglass_stoolY_Iglass_sofaY_Iglass_tableY_Iglass_table*2Y_Iglass_table*3Y_Iglass_table*4Y_Iglass_table*5Y_Iglass_table*6Y_Iglass_table*7Y_Iglass_table*8Y_Iglass_table*9Y_Iglass_chair*2Y_Iglass_chair*3Y_Iglass_chair*4Y_Iglass_chair*5Y_Iglass_chair*6Y_Iglass_chair*7Y_Iglass_chair*8Y_Iglass_chair*9Y_Iglass_sofa*2Y_Iglass_sofa*3Y_Iglass_sofa*4Y_Iglass_sofa*5Y_Iglass_sofa*6Y_Iglass_sofa*7Y_Iglass_sofa*8Y_Iglass_sofa*9Y_Iglass_stool*2Y_Iglass_stool*4Y_Iglass_stool*5Y_Iglass_stool*6Y_Iglass_stool*7Y_Iglass_stool*8Y_Iglass_stool*3Y_Iglass_stool*9Y_ICF_10_coin_goldZvCCF_1_coin_bronzeZvCCF_20_moneybagZvCCF_50_goldbarZvCCF_5_coin_silverZvChc_crptYmAhc_tvZ\BgothgateX~DgothiccandelabraYxBgothrailingX~Dgoth_tableYmAhc_bkshlfYmAhc_btlrY_Ihc_crtnYmAhc_djsetYmAhc_frplcZbBhc_lmpstYmAhc_machineYmAhc_rllrXQHhc_rntgnX~Dhc_trllYmAgothic_chair*1X~Dgothic_sofa*1X~Dgothic_stool*1X~Dgothic_chair*2X~Dgothic_sofa*2X~Dgothic_stool*2X~Dgothic_chair*3X~Dgothic_sofa*3X~Dgothic_stool*3X~Dgothic_chair*4X~Dgothic_sofa*4X~Dgothic_stool*4X~Dgothic_chair*5X~Dgothic_sofa*5X~Dgothic_stool*5X~Dgothic_chair*6X~Dgothic_sofa*6X~Dgothic_stool*6X~Dval_cauldronX~Dsound_machineX~Dromantique_pianochair*3Y_Iromantique_pianochair*5Y_Iromantique_pianochair*2Y_Iromantique_pianochair*4Y_Iromantique_pianochair*1Y_Iromantique_divan*3Y_Iromantique_divan*5Y_Iromantique_divan*2Y_Iromantique_divan*4Y_Iromantique_divan*1Y_Iromantique_chair*3Y_Iromantique_chair*5Y_Iromantique_chair*2Y_Iromantique_chair*4Y_Iromantique_chair*1Y_Irare_parasolY_Iplant_valentinerose*3XICplant_valentinerose*5XICplant_valentinerose*2XICplant_valentinerose*4XICplant_valentinerose*1XICplant_mazegateYeCplant_mazeZcCplant_bulrushXICpetfood4Y_Icarpet_valentineZ|Egothic_carpetXICgothic_carpet2Z|Egothic_chairX~Dgothic_sofaX~Dgothic_stoolX~Dgrand_piano*3Z|Egrand_piano*5Z|Egrand_piano*2Z|Egrand_piano*4Z|Egrand_piano*1Z|Etheatre_seatZ@Kromantique_tray2Y_Iromantique_tray1Y_Iromantique_smalltabl*3Y_Iromantique_smalltabl*5Y_Iromantique_smalltabl*2Y_Iromantique_smalltabl*4Y_Iromantique_smalltabl*1Y_Iromantique_mirrortablY_Iromantique_divider*3Z[Mromantique_divider*2Z[Mromantique_divider*4Z[Mromantique_divider*1Z[Mjp_tatami2[dWjp_tatamiYGGhabbowood_chairYGGjp_bambooYGGjp_iroriXQHjp_pillowYGGsound_set_1[dWsound_set_2[dWsound_set_3[dWsound_set_4[dWsound_set_5[dWsound_set_6[dWsound_set_7[dWsound_set_8[dWsound_set_9[dWsound_machine*1Yc[spotlightY_Isound_machine*2Yc[sound_machine*3Yc[sound_machine*4Yc[sound_machine*5Yc[sound_machine*6Yc[sound_machine*7Yc[rom_lampZ|Erclr_sofaXQHrclr_gardenXQHrclr_chairZ|Esound_set_28[dWsound_set_27[dWsound_set_26[dWsound_set_25[dWsound_set_24[dWsound_set_23[dWsound_set_22[dWsound_set_21[dWsound_set_20[dWsound_set_19[dWsound_set_18[dWsound_set_17[dWsound_set_16[dWsound_set_15[dWsound_set_14[dWsound_set_13[dWsound_set_12[dWsound_set_11[dWsound_set_10[dWrope_dividerXQHromantique_clockY_Irare_icecream_campaignY_Ipura_mdl5*1Yc[pura_mdl5*2Yc[pura_mdl5*3Yc[pura_mdl5*4Yc[pura_mdl5*5Yc[pura_mdl5*6Yc[pura_mdl5*7Yc[pura_mdl5*8Yc[pura_mdl5*9Yc[pura_mdl4*1XQHpura_mdl4*2XQHpura_mdl4*3XQHpura_mdl4*4XQHpura_mdl4*5XQHpura_mdl4*6XQHpura_mdl4*7XQHpura_mdl4*8XQHpura_mdl4*9XQHpura_mdl3*1XQHpura_mdl3*2XQHpura_mdl3*3XQHpura_mdl3*4XQHpura_mdl3*5XQHpura_mdl3*6XQHpura_mdl3*7XQHpura_mdl3*8XQHpura_mdl3*9XQHpura_mdl2*1XQHpura_mdl2*2XQHpura_mdl2*3XQHpura_mdl2*4XQHpura_mdl2*5XQHpura_mdl2*6XQHpura_mdl2*7XQHpura_mdl2*8XQHpura_mdl2*9XQHpura_mdl1*1XQHpura_mdl1*2XQHpura_mdl1*3XQHpura_mdl1*4XQHpura_mdl1*5XQHpura_mdl1*6XQHpura_mdl1*7XQHpura_mdl1*8XQHpura_mdl1*9XQHjp_lanternXQHchair_basic*1XQHchair_basic*2XQHchair_basic*3XQHchair_basic*4XQHchair_basic*5XQHchair_basic*6XQHchair_basic*7XQHchair_basic*8XQHchair_basic*9XQHbed_budget*1XQHbed_budget*2XQHbed_budget*3XQHbed_budget*4XQHbed_budget*5XQHbed_budget*6XQHbed_budget*7XQHbed_budget*8XQHbed_budget*9XQHbed_budget_one*1XQHbed_budget_one*2XQHbed_budget_one*3XQHbed_budget_one*4XQHbed_budget_one*5XQHbed_budget_one*6XQHbed_budget_one*7XQHbed_budget_one*8XQHbed_budget_one*9XQHjp_drawerXQHtile_stellaZ[Mtile_marbleZ[Mtile_brownZ[Msummer_grill*1Y_Isummer_grill*2Y_Isummer_grill*3Y_Isummer_grill*4Y_Isummer_chair*1Y_Isummer_chair*2Y_Isummer_chair*3Y_Isummer_chair*4Y_Isummer_chair*5Y_Isummer_chair*6Y_Isummer_chair*7Y_Isummer_chair*8Y_Isummer_chair*9Y_Isound_set_36[dWsound_set_35[dWsound_set_34[dWsound_set_33[dWsound_set_32[dWsound_set_31[dWsound_set_30[dWsound_set_29[dWsound_machine_proYc[rare_mnstrY_Ione_way_door*1XQHone_way_door*2XQHone_way_door*3XQHone_way_door*4XQHone_way_door*5XQHone_way_door*6XQHone_way_door*7XQHone_way_door*8XQHone_way_door*9XQHexe_rugZ[Mexe_s_tableZGRsound_set_37[dWsummer_pool*1ZlIsummer_pool*2ZlIsummer_pool*3ZlIsummer_pool*4ZlIsong_diskYc[jukebox*1Yc[carpet_soft_tut[~Nsound_set_44[dWsound_set_43[dWsound_set_42[dWsound_set_41[dWsound_set_40[dWsound_set_39[dWsound_set_38[dWgrunge_chairZ@Kgrunge_mattressZ@Kgrunge_radiatorZ@Kgrunge_shelfZ@Kgrunge_signZ@Kgrunge_tableZ@Khabboween_crypt[uKhabboween_grassZ@Khal_cauldronZ@Khal_graveZ@Ksound_set_52[dWsound_set_51[dWsound_set_50[dWsound_set_49[dWsound_set_48[dWsound_set_47[dWsound_set_46[dWsound_set_45[dWxmas_icelampZ[Mxmas_cstl_wallZ[Mxmas_cstl_twrZ[Mxmas_cstl_gate[~Ntree7Z[Mtree6Z[Msound_set_54[dWsound_set_53[dWsafe_silo_pb[dOplant_mazegate_snowZ[Mplant_maze_snowZ[Mchristmas_sleighZ[Mchristmas_reindeer[~Nchristmas_poopZ[Mexe_bardeskZ[Mexe_chairZ[Mexe_chair2Z[Mexe_cornerZ[Mexe_drinksZ[Mexe_sofaZ[Mexe_tableZ[Msound_set_59[dWsound_set_58[dWsound_set_57[dWsound_set_56[dWsound_set_55[dWnoob_table*1[~Nnoob_table*2[~Nnoob_table*3[~Nnoob_table*4[~Nnoob_table*5[~Nnoob_table*6[~Nnoob_stool*1[~Nnoob_stool*2[~Nnoob_stool*3[~Nnoob_stool*4[~Nnoob_stool*5[~Nnoob_stool*6[~Nnoob_rug*1[~Nnoob_rug*2[~Nnoob_rug*3[~Nnoob_rug*4[~Nnoob_rug*5[~Nnoob_rug*6[~Nnoob_lamp*1[dOnoob_lamp*2[dOnoob_lamp*3[dOnoob_lamp*4[dOnoob_lamp*5[dOnoob_lamp*6[dOnoob_chair*1[~Nnoob_chair*2[~Nnoob_chair*3[~Nnoob_chair*4[~Nnoob_chair*5[~Nnoob_chair*6[~Nexe_globe[~Nexe_plantZ[Mval_teddy*1[dOval_teddy*2[dOval_teddy*3[dOval_teddy*4[dOval_teddy*5[dOval_teddy*6[dOval_randomizer[dOval_choco[dOteleport_doorYc[sound_set_61[dWsound_set_60[dWfortune[dOsw_tableZIPsw_raven[cQsw_chestZIPsand_cstl_wallZIPsand_cstl_twrZIPsand_cstl_gateZIPgrunge_candleZIPgrunge_benchZIPgrunge_barrelZIPrclr_lampZGRprizetrophy9*1Yd[prizetrophy8*1Yd[nouvelle_traxYc[md_rugZGRjp_tray6ZGRjp_tray5ZGRjp_tray4ZGRjp_tray3ZGRjp_tray2ZGRjp_tray1ZGRarabian_teamkZGRarabian_snakeZGRarabian_rugZGRarabian_pllwZGRarabian_divdrZGRarabian_chairZGRarabian_bigtbZGRarabian_tetblZGRarabian_tray1ZGRarabian_tray2ZGRarabian_tray3ZGRarabian_tray4ZGRsound_set_64[dWsound_set_63[dWsound_set_62[dWjukebox_ptv*1Yc[calippoZAStraxsilverYc[traxgoldYc[traxbronzeYc[bench_puffetYATCFC_500_goldbarZvCCFC_200_moneybagZvCCFC_10_coin_bronzeZvCCFC_100_coin_goldZvCCFC_50_coin_silverZvCjp_tableXMVjp_rareXMVjp_katana3XMVjp_katana2XMVjp_katana1XMVfootylamp_campaignXMVtiki_waterfall[dWtiki_tray4[dWtiki_tray3[dWtiki_tray2[dWtiki_tray1[dWtiki_tray0[dWtiki_toucan[dWtiki_torch[dWtiki_statue[dWtiki_sand[dWtiki_parasol[dWtiki_junglerug[dWtiki_corner[dWtiki_bflies[dWtiki_bench[dWtiki_bardesk[dWtampax_rug[dWsound_set_70[dWsound_set_69[dWsound_set_68[dWsound_set_67[dWsound_set_66[dWsound_set_65[dWnoob_rug_tradeable*1[dWnoob_rug_tradeable*2[dWnoob_rug_tradeable*3[dWnoob_rug_tradeable*4[dWnoob_rug_tradeable*5[dWnoob_rug_tradeable*6[dWnoob_plant[dWnoob_lamp_tradeable*1[dWnoob_lamp_tradeable*2[dWnoob_lamp_tradeable*3[dWnoob_lamp_tradeable*4[dWnoob_lamp_tradeable*5[dWnoob_lamp_tradeable*6[dWnoob_chair_tradeable*1[dWnoob_chair_tradeable*2[dWnoob_chair_tradeable*3[dWnoob_chair_tradeable*4[dWnoob_chair_tradeable*5[dWnoob_chair_tradeable*6[dWjp_teamaker[dWsvnr_uk[`_svnr_nlXhXsvnr_itXhXsvnr_de[gXsvnr_aus[gXdiner_tray_7[gXdiner_tray_6[gXdiner_tray_5[gXdiner_tray_4[gXdiner_tray_3[gXdiner_tray_2[gXdiner_tray_1[gXdiner_tray_0[gXdiner_sofa_2*1[gXdiner_sofa_2*2[gXdiner_sofa_2*3[gXdiner_sofa_2*4[gXdiner_sofa_2*5[gXdiner_sofa_2*6[gXdiner_sofa_2*7[gXdiner_sofa_2*8[gXdiner_sofa_2*9[gXdiner_shaker[gXdiner_rug[gXdiner_gumvendor*1[gXdiner_gumvendor*2[gXdiner_gumvendor*3[gXdiner_gumvendor*4[gXdiner_gumvendor*5[gXdiner_gumvendor*6[gXdiner_gumvendor*7[gXdiner_gumvendor*8[gXdiner_gumvendor*9[gXdiner_cashreg*1[gXdiner_cashreg*2[gXdiner_cashreg*3[gXdiner_cashreg*4[gXdiner_cashreg*5[gXdiner_cashreg*6[gXdiner_cashreg*7[gXdiner_cashreg*8[gXdiner_cashreg*9[gXdiner_table_2*1XiZdiner_table_2*2diner_table_2*2XiZdiner_table_2*3XiZdiner_table_2*4XiZdiner_table_2*5XiZdiner_table_2*6XiZdiner_table_2*7XiZdiner_table_2*8XiZdiner_table_2*9XiZdiner_table_1*1XiZdiner_table_1*2XiZdiner_table_1*3XiZdiner_table_1*4XiZdiner_table_1*5XiZdiner_table_1*6XiZdiner_table_1*7XiZdiner_table_1*8XiZdiner_table_1*9XiZdiner_sofa_1*1XiZdiner_sofa_1*2XiZdiner_sofa_1*3XiZdiner_sofa_1*4XiZdiner_sofa_1*5XiZdiner_sofa_1*6XiZdiner_sofa_1*7XiZdiner_sofa_1*8XiZdiner_sofa_1*9XiZdiner_chair*1XiZdiner_chair*2XiZdiner_chair*3XiZdiner_chair*4XiZdiner_chair*5XiZdiner_chair*6XiZdiner_chair*7XiZdiner_chair*8XiZdiner_chair*9XiZdiner_bardesk_gate*1XiZdiner_bardesk_gate*2XiZdiner_bardesk_gate*3XiZdiner_bardesk_gate*4XiZdiner_bardesk_gate*5XiZdiner_bardesk_gate*6XiZdiner_bardesk_gate*7XiZdiner_bardesk_gate*8XiZdiner_bardesk_gate*9XiZdiner_bardesk_corner*1XiZdiner_bardesk_corner*2XiZdiner_bardesk_corner*3XiZdiner_bardesk_corner*4XiZdiner_bardesk_corner*5XiZdiner_bardesk_corner*6XiZdiner_bardesk_corner*7XiZdiner_bardesk_corner*8XiZdiner_bardesk_corner*9XiZdiner_bardesk*1XiZdiner_bardesk*2XiZdiner_bardesk*3XiZdiner_bardesk*4XiZdiner_bardesk*5XiZdiner_bardesk*6XiZdiner_bardesk*7XiZdiner_bardesk*8XiZdiner_bardesk*9XiZads_dave_cnsXiZeasy_carpetYc[easy_bowl2Yc[greek_cornerYc[greek_gateYc[greek_pillarsYc[greek_seatYc[greektrophy*1[P\greektrophy*2[P\greektrophy*3[P\greek_blockXt[hcc_tableY`]hcc_shelfY`]hcc_sofaY`]hcc_minibarY`]hcc_chairY`]det_dividerY`]netari_carpetY`]det_bodyY`]hcc_stoolY`]hcc_sofachairY`]hcc_crnrXw]hcc_dvdrXw]sob_carpet[`_igor_seat[`_ads_igorbrainY_aads_igorswitchY_aads_711*1Y_aads_711*2Y_aads_711*3Y_aads_711*4Y_aads_igorraygunY_ahween08_sinkY[chween08_curtainY[chween08_bathY[chween08_defibsY[chween08_bbagY[chween08_curtain2Y[chween08_defibs2Y[chween08_bedY[chween08_sink2Y[chween08_bed2Y[chween08_bath2Y[chween08_manholeY[chween08_trllY[cPRpost.itHpost.it.vdHphotoHChessHTicTacToeHBattleShipHPokerHwallpaperHfloorHposterZ@KgothicfountainYxBhc_wall_lampZbBindustrialfanZ`BtorchZ\Bval_heartXBCwallmirrorZ|Ejp_ninjastarsXQHhabw_mirrorXQHhabbowheelZ[Mguitar_skullZ@Kguitar_vZ@Kxmas_light[~Nhrella_poster_3[Nhrella_poster_2ZIPhrella_poster_1[Nsw_swordsZIPsw_stoneZIPsw_holeZIProomdimmerYc[md_logo_wallZGRmd_canZGRjp_sheet3ZGRjp_sheet2ZGRjp_sheet1ZGRarabian_swordsZGRarabian_wndwZGRtiki_wallplnt[dWtiki_surfboard[dWtampax_wall[dWwindow_single_default[gXwindow_double_default[gXnoob_window_double[dWwindow_triple[gXwindow_square[gXwindow_romantic_wide[gXwindow_romantic_narrow[gXwindow_grunge[gXwindow_golden[gXwindow_chinese_wide[gXwindow_chinese_narrowYA\window_basic[gXwindow_70s_wide[gXwindow_70s_narrow[gXads_sunnydYlXwindow_diner2XiZwindow_dinerXiZdiner_walltableXiZads_dave_wallXiZwindow_holeYc[easy_posterYc[ads_nokia_logoYc[ads_nokia_phoneYc[landscapeXV^window_skyscraper[j\netari_posterY`]det_bholeY`]ads_campguitarXw]hween08_radY[chween08_wndwbY[chween08_wndwY[chween08_bioY[chw_08_xrayY[c");

                                    //sendData("Dg" + @"YKEshelves_norjaX~Dshelves_polyfonYmAshelves_siloXQHtable_polyfon_smallYmAchair_polyfonZbBtable_norja_medY_Itable_silo_medX~Dtable_plasto_4legY_Itable_plasto_roundY_Itable_plasto_bigsquareY_Istand_polyfon_zZbBchair_siloX~Dsofa_siloX~Dcouch_norjaX~Dchair_norjaX~Dtable_polyfon_medYmAdoormat_loveZbBdoormat_plainZ[Msofachair_polyfonX~Dsofa_polyfonZ[Msofachair_siloX~Dchair_plastyX~Dchair_plastoYmAtable_plasto_squareY_Ibed_polyfonX~Dbed_polyfon_one[dObed_trad_oneYmAbed_tradYmAbed_silo_oneYmAbed_silo_twoYmAtable_silo_smallX~Dbed_armas_twoYmAbed_budget_oneXQHbed_budgetXQHshelves_armasYmAbench_armasYmAtable_armasYmAsmall_table_armasZbBsmall_chair_armasYmAfireplace_armasYmAlamp_armasYmAbed_armas_oneYmAcarpet_standardY_Icarpet_armasYmAcarpet_polarY_Ifireplace_polyfonY_Itable_plasto_4leg*1Y_Itable_plasto_bigsquare*1Y_Itable_plasto_round*1Y_Itable_plasto_square*1Y_Ichair_plasto*1YmAcarpet_standard*1Y_Idoormat_plain*1Z[Mtable_plasto_4leg*2Y_Itable_plasto_bigsquare*2Y_Itable_plasto_round*2Y_Itable_plasto_square*2Y_Ichair_plasto*2YmAdoormat_plain*2Z[Mcarpet_standard*2Y_Itable_plasto_4leg*3Y_Itable_plasto_bigsquare*3Y_Itable_plasto_round*3Y_Itable_plasto_square*3Y_Ichair_plasto*3YmAcarpet_standard*3Y_Idoormat_plain*3Z[Mtable_plasto_4leg*4Y_Itable_plasto_bigsquare*4Y_Itable_plasto_round*4Y_Itable_plasto_square*4Y_Ichair_plasto*4YmAcarpet_standard*4Y_Idoormat_plain*4Z[Mdoormat_plain*6Z[Mdoormat_plain*5Z[Mcarpet_standard*5Y_Itable_plasto_4leg*5Y_Itable_plasto_bigsquare*5Y_Itable_plasto_round*5Y_Itable_plasto_square*5Y_Ichair_plasto*5YmAtable_plasto_4leg*6Y_Itable_plasto_bigsquare*6Y_Itable_plasto_round*6Y_Itable_plasto_square*6Y_Ichair_plasto*6YmAtable_plasto_4leg*7Y_Itable_plasto_bigsquare*7Y_Itable_plasto_round*7Y_Itable_plasto_square*7Y_Ichair_plasto*7YmAtable_plasto_4leg*8Y_Itable_plasto_bigsquare*8Y_Itable_plasto_round*8Y_Itable_plasto_square*8Y_Ichair_plasto*8YmAtable_plasto_4leg*9Y_Itable_plasto_bigsquare*9Y_Itable_plasto_round*9Y_Itable_plasto_square*9Y_Ichair_plasto*9YmAcarpet_standard*6Y_Ichair_plasty*1X~DpizzaYmAdrinksYmAchair_plasty*2X~Dchair_plasty*3X~Dchair_plasty*4X~Dbar_polyfonY_Iplant_cruddyYmAbottleYmAbardesk_polyfonX~Dbardeskcorner_polyfonX~DfloortileHbar_armasY_Ibartable_armasYmAbar_chair_armasYmAcarpet_softZ@Kcarpet_soft*1Z@Kcarpet_soft*2Z@Kcarpet_soft*3Z@Kcarpet_soft*4Z@Kcarpet_soft*5Z@Kcarpet_soft*6Z@Kred_tvY_Iwood_tvYmAcarpet_polar*1Y_Ichair_plasty*5X~Dcarpet_polar*2Y_Icarpet_polar*3Y_Icarpet_polar*4Y_Ichair_plasty*6X~Dtable_polyfonYmAsmooth_table_polyfonYmAsofachair_polyfon_girlX~Dbed_polyfon_girl_one[dObed_polyfon_girlX~Dsofa_polyfon_girlZ[Mbed_budgetb_oneXQHbed_budgetbXQHplant_pineappleYmAplant_fruittreeY_Iplant_small_cactusY_Iplant_bonsaiY_Iplant_big_cactusY_Iplant_yukkaY_Icarpet_standard*7Y_Icarpet_standard*8Y_Icarpet_standard*9Y_Icarpet_standard*aY_Icarpet_standard*bY_Iplant_sunflowerY_Iplant_roseY_Itv_luxusY_IbathZ\BsinkY_ItoiletYmAduckYmAtileYmAtoilet_redYmAtoilet_yellYmAtile_redYmAtile_yellYmApresent_gen[~Npresent_gen1[~Npresent_gen2[~Npresent_gen3[~Npresent_gen4[~Npresent_gen5[~Npresent_gen6[~Nbar_basicY_Ishelves_basicXQHsoft_sofachair_norjaX~Dsoft_sofa_norjaX~Dlamp_basicXQHlamp2_armasYmAfridgeY_IdoorYc[doorBYc[doorCYc[pumpkinYmAskullcandleYmAdeadduckYmAdeadduck2YmAdeadduck3YmAmenorahYmApuddingYmAhamYmAturkeyYmAxmasduckY_IhouseYmAtriplecandleYmAtree3YmAtree4YmAtree5X~Dham2YmAwcandlesetYmArcandlesetYmAstatueYmAheartY_IvaleduckYmAheartsofaX~DthroneYmAsamovarY_IgiftflowersY_IhabbocakeYmAXoagramYmAeasterduckY_IbunnyYmAbasketY_IbirdieYmAediceX~Dclub_sofaZ[Mprize1YmAprize2YmAprize3YmAdivider_poly3X~Ddivider_arm1YmAdivider_arm2YmAdivider_arm3YmAdivider_nor1X~Ddivider_silo1X~Ddivider_nor2X~Ddivider_silo2Z[Mdivider_nor3X~Ddivider_silo3X~DtypingmachineYmAspyroYmAredXoagramYmAcameraHjoulutahtiYmAhyacinth1YmAhyacinth2YmAchair_plasto*10YmAchair_plasto*11YmAbardeskcorner_polyfon*12X~Dbardeskcorner_polyfon*13X~Dchair_plasto*12YmAchair_plasto*13YmAchair_plasto*14YmAtable_plasto_4leg*14Y_ImocchamasterY_Icarpet_legocourtYmAbench_legoYmAlegotrophyYmAvalentinescreenYmAedicehcYmArare_daffodil_rugYmArare_beehive_bulbY_IhcsohvaYmAhcammeYmArare_elephant_statueYmArare_fountainY_Irare_standYmArare_globeYmArare_hammockYmArare_elephant_statue*1YmArare_elephant_statue*2YmArare_fountain*1Y_Irare_fountain*2Y_Irare_fountain*3Y_Irare_beehive_bulb*1Y_Irare_beehive_bulb*2Y_Irare_xmas_screenY_Irare_parasol*1XMVrare_parasol*2XMVrare_parasol*3XMVtree1X~Dtree2ZmBwcandleYxBrcandleYxBsoft_jaggara_norjaYmAhouse2YmAdjesko_turntableYmAmd_sofaZ[Mmd_limukaappiY_Itable_plasto_4leg*10Y_Itable_plasto_4leg*15Y_Itable_plasto_bigsquare*14Y_Itable_plasto_bigsquare*15Y_Itable_plasto_round*14Y_Itable_plasto_round*15Y_Itable_plasto_square*14Y_Itable_plasto_square*15Y_Ichair_plasto*15YmAchair_plasty*7X~Dchair_plasty*8X~Dchair_plasty*9X~Dchair_plasty*10X~Dchair_plasty*11X~Dchair_plasto*16YmAtable_plasto_4leg*16Y_Ihockey_scoreY_Ihockey_lightYmAdoorDYc[prizetrophy2*3Yd[prizetrophy3*3Yd[prizetrophy4*3Yd[prizetrophy5*3Yd[prizetrophy6*3Yd[prizetrophy*1Yd[prizetrophy2*1Yd[prizetrophy3*1Yd[prizetrophy4*1Yd[prizetrophy5*1Yd[prizetrophy6*1Yd[prizetrophy*2Yd[prizetrophy2*2Yd[prizetrophy3*2Yd[prizetrophy4*2Yd[prizetrophy5*2Yd[prizetrophy6*2Yd[prizetrophy*3Yd[rare_parasol*0XMVhc_lmp[fBhc_tblYmAhc_chrYmAhc_dskXQHnestHpetfood1ZvCpetfood2ZvCpetfood3ZvCwaterbowl*4XICwaterbowl*5XICwaterbowl*2XICwaterbowl*1XICwaterbowl*3XICtoy1XICtoy1*1XICtoy1*2XICtoy1*3XICtoy1*4XICgoodie1Yc[goodie1*1Yc[goodie1*2Yc[goodie2Yc[prizetrophy7*3Yd[prizetrophy7*1Yd[prizetrophy7*2Yd[scifiport*0Y_Iscifiport*9Y_Iscifiport*8Y_Iscifiport*7Y_Iscifiport*6Y_Iscifiport*5Y_Iscifiport*4Y_Iscifiport*3Y_Iscifiport*2Y_Iscifiport*1Y_Iscifirocket*9Y_Iscifirocket*8Y_Iscifirocket*7Y_Iscifirocket*6Y_Iscifirocket*5Y_Iscifirocket*4Y_Iscifirocket*3Y_Iscifirocket*2Y_Iscifirocket*1Y_Iscifirocket*0Y_Iscifidoor*10Y_Iscifidoor*9Y_Iscifidoor*8Y_Iscifidoor*7Y_Iscifidoor*6Y_Iscifidoor*5Y_Iscifidoor*4Y_Iscifidoor*3Y_Iscifidoor*2Y_Iscifidoor*1Y_Ipillow*5YmApillow*8YmApillow*0YmApillow*1YmApillow*2YmApillow*7YmApillow*9YmApillow*4YmApillow*6YmApillow*3YmAmarquee*1Y_Imarquee*2Y_Imarquee*7Y_Imarquee*aY_Imarquee*8Y_Imarquee*9Y_Imarquee*5Y_Imarquee*4Y_Imarquee*6Y_Imarquee*3Y_Iwooden_screen*1Y_Iwooden_screen*2Y_Iwooden_screen*7Y_Iwooden_screen*0Y_Iwooden_screen*8Y_Iwooden_screen*5Y_Iwooden_screen*9Y_Iwooden_screen*4Y_Iwooden_screen*6Y_Iwooden_screen*3Y_Ipillar*6Y_Ipillar*1Y_Ipillar*9Y_Ipillar*0Y_Ipillar*8Y_Ipillar*2Y_Ipillar*5Y_Ipillar*4Y_Ipillar*7Y_Ipillar*3Y_Irare_dragonlamp*4Y_Irare_dragonlamp*0Y_Irare_dragonlamp*5Y_Irare_dragonlamp*2Y_Irare_dragonlamp*8Y_Irare_dragonlamp*9Y_Irare_dragonlamp*7Y_Irare_dragonlamp*6Y_Irare_dragonlamp*1Y_Irare_dragonlamp*3Y_Irare_icecream*1Y_Irare_icecream*7Y_Irare_icecream*8Y_Irare_icecream*2Y_Irare_icecream*6Y_Irare_icecream*9Y_Irare_icecream*3Y_Irare_icecream*0Y_Irare_icecream*4Y_Irare_icecream*5Y_Irare_fan*7YxBrare_fan*6YxBrare_fan*9YxBrare_fan*3YxBrare_fan*0YxBrare_fan*4YxBrare_fan*5YxBrare_fan*1YxBrare_fan*8YxBrare_fan*2YxBqueue_tile1*3X~Dqueue_tile1*6X~Dqueue_tile1*4X~Dqueue_tile1*9X~Dqueue_tile1*8X~Dqueue_tile1*5X~Dqueue_tile1*7X~Dqueue_tile1*2X~Dqueue_tile1*1X~Dqueue_tile1*0X~DticketHrare_snowrugX~Dcn_lampZxIcn_sofaYmAsporttrack1*1YmAsporttrack1*3YmAsporttrack1*2YmAsporttrack2*1[~Nsporttrack2*2[~Nsporttrack2*3[~Nsporttrack3*1YmAsporttrack3*2YmAsporttrack3*3YmAfootylampX~Dbarchair_siloX~Ddivider_nor4*4X~Dtraffic_light*1ZxItraffic_light*2ZxItraffic_light*3ZxItraffic_light*4ZxItraffic_light*6ZxIrubberchair*1X~Drubberchair*2X~Drubberchair*3X~Drubberchair*4X~Drubberchair*5X~Drubberchair*6X~Dbarrier*1X~Dbarrier*2X~Dbarrier*3X~Drubberchair*7X~Drubberchair*8X~Dtable_norja_med*2Y_Itable_norja_med*3Y_Itable_norja_med*4Y_Itable_norja_med*5Y_Itable_norja_med*6Y_Itable_norja_med*7Y_Itable_norja_med*8Y_Itable_norja_med*9Y_Icouch_norja*2X~Dcouch_norja*3X~Dcouch_norja*4X~Dcouch_norja*5X~Dcouch_norja*6X~Dcouch_norja*7X~Dcouch_norja*8X~Dcouch_norja*9X~Dshelves_norja*2X~Dshelves_norja*3X~Dshelves_norja*4X~Dshelves_norja*5X~Dshelves_norja*6X~Dshelves_norja*7X~Dshelves_norja*8X~Dshelves_norja*9X~Dchair_norja*2X~Dchair_norja*3X~Dchair_norja*4X~Dchair_norja*5X~Dchair_norja*6X~Dchair_norja*7X~Dchair_norja*8X~Dchair_norja*9X~Ddivider_nor1*2X~Ddivider_nor1*3X~Ddivider_nor1*4X~Ddivider_nor1*5X~Ddivider_nor1*6X~Ddivider_nor1*7X~Ddivider_nor1*8X~Ddivider_nor1*9X~Dsoft_sofa_norja*2X~Dsoft_sofa_norja*3X~Dsoft_sofa_norja*4X~Dsoft_sofa_norja*5X~Dsoft_sofa_norja*6X~Dsoft_sofa_norja*7X~Dsoft_sofa_norja*8X~Dsoft_sofa_norja*9X~Dsoft_sofachair_norja*2X~Dsoft_sofachair_norja*3X~Dsoft_sofachair_norja*4X~Dsoft_sofachair_norja*5X~Dsoft_sofachair_norja*6X~Dsoft_sofachair_norja*7X~Dsoft_sofachair_norja*8X~Dsoft_sofachair_norja*9X~Dsofachair_silo*2X~Dsofachair_silo*3X~Dsofachair_silo*4X~Dsofachair_silo*5X~Dsofachair_silo*6X~Dsofachair_silo*7X~Dsofachair_silo*8X~Dsofachair_silo*9X~Dtable_silo_small*2X~Dtable_silo_small*3X~Dtable_silo_small*4X~Dtable_silo_small*5X~Dtable_silo_small*6X~Dtable_silo_small*7X~Dtable_silo_small*8X~Dtable_silo_small*9X~Ddivider_silo1*2X~Ddivider_silo1*3X~Ddivider_silo1*4X~Ddivider_silo1*5X~Ddivider_silo1*6X~Ddivider_silo1*7X~Ddivider_silo1*8X~Ddivider_silo1*9X~Ddivider_silo3*2X~Ddivider_silo3*3X~Ddivider_silo3*4X~Ddivider_silo3*5X~Ddivider_silo3*6X~Ddivider_silo3*7X~Ddivider_silo3*8X~Ddivider_silo3*9X~Dtable_silo_med*2X~Dtable_silo_med*3X~Dtable_silo_med*4X~Dtable_silo_med*5X~Dtable_silo_med*6X~Dtable_silo_med*7X~Dtable_silo_med*8X~Dtable_silo_med*9X~Dsofa_silo*2X~Dsofa_silo*3X~Dsofa_silo*4X~Dsofa_silo*5X~Dsofa_silo*6X~Dsofa_silo*7X~Dsofa_silo*8X~Dsofa_silo*9X~Dsofachair_polyfon*2X~Dsofachair_polyfon*3X~Dsofachair_polyfon*4X~Dsofachair_polyfon*6X~Dsofachair_polyfon*7X~Dsofachair_polyfon*8X~Dsofachair_polyfon*9X~Dsofa_polyfon*2Z[Msofa_polyfon*3Z[Msofa_polyfon*4Z[Msofa_polyfon*6Z[Msofa_polyfon*7Z[Msofa_polyfon*8Z[Msofa_polyfon*9Z[Mbed_polyfon*2X~Dbed_polyfon*3X~Dbed_polyfon*4X~Dbed_polyfon*6X~Dbed_polyfon*7X~Dbed_polyfon*8X~Dbed_polyfon*9X~Dbed_polyfon_one*2[dObed_polyfon_one*3[dObed_polyfon_one*4[dObed_polyfon_one*6[dObed_polyfon_one*7[dObed_polyfon_one*8[dObed_polyfon_one*9[dObardesk_polyfon*2X~Dbardesk_polyfon*3X~Dbardesk_polyfon*4X~Dbardesk_polyfon*5X~Dbardesk_polyfon*6X~Dbardesk_polyfon*7X~Dbardesk_polyfon*8X~Dbardesk_polyfon*9X~Dbardeskcorner_polyfon*2X~Dbardeskcorner_polyfon*3X~Dbardeskcorner_polyfon*4X~Dbardeskcorner_polyfon*5X~Dbardeskcorner_polyfon*6X~Dbardeskcorner_polyfon*7X~Dbardeskcorner_polyfon*8X~Dbardeskcorner_polyfon*9X~Ddivider_poly3*2X~Ddivider_poly3*3X~Ddivider_poly3*4X~Ddivider_poly3*5X~Ddivider_poly3*6X~Ddivider_poly3*7X~Ddivider_poly3*8X~Ddivider_poly3*9X~Dchair_silo*2X~Dchair_silo*3X~Dchair_silo*4X~Dchair_silo*5X~Dchair_silo*6X~Dchair_silo*7X~Dchair_silo*8X~Dchair_silo*9X~Ddivider_nor3*2X~Ddivider_nor3*3X~Ddivider_nor3*4X~Ddivider_nor3*5X~Ddivider_nor3*6X~Ddivider_nor3*7X~Ddivider_nor3*8X~Ddivider_nor3*9X~Ddivider_nor2*2X~Ddivider_nor2*3X~Ddivider_nor2*4X~Ddivider_nor2*5X~Ddivider_nor2*6X~Ddivider_nor2*7X~Ddivider_nor2*8X~Ddivider_nor2*9X~Dsilo_studydeskX~Dsolarium_norjaY_Isolarium_norja*1Y_Isolarium_norja*2Y_Isolarium_norja*3Y_Isolarium_norja*5Y_Isolarium_norja*6Y_Isolarium_norja*7Y_Isolarium_norja*8Y_Isolarium_norja*9Y_IsandrugX~Drare_moonrugYmAchair_chinaYmAchina_tableYmAsleepingbag*1YmAsleepingbag*2YmAsleepingbag*3YmAsleepingbag*4YmAsafe_siloY_Isleepingbag*7YmAsleepingbag*9YmAsleepingbag*5YmAsleepingbag*10YmAsleepingbag*6YmAsleepingbag*8YmAchina_shelveX~Dtraffic_light*5ZxIdivider_nor4*2X~Ddivider_nor4*3X~Ddivider_nor4*5X~Ddivider_nor4*6X~Ddivider_nor4*7X~Ddivider_nor4*8X~Ddivider_nor4*9X~Ddivider_nor5*2X~Ddivider_nor5*3X~Ddivider_nor5*4X~Ddivider_nor5*5X~Ddivider_nor5*6X~Ddivider_nor5*7X~Ddivider_nor5*8X~Ddivider_nor5*9X~Ddivider_nor5X~Ddivider_nor4X~Dwall_chinaYmAcorner_chinaYmAbarchair_silo*2X~Dbarchair_silo*3X~Dbarchair_silo*4X~Dbarchair_silo*5X~Dbarchair_silo*6X~Dbarchair_silo*7X~Dbarchair_silo*8X~Dbarchair_silo*9X~Dsafe_silo*2Y_Isafe_silo*3Y_Isafe_silo*4Y_Isafe_silo*5Y_Isafe_silo*6Y_Isafe_silo*7Y_Isafe_silo*8Y_Isafe_silo*9Y_Iglass_shelfY_Iglass_chairY_Iglass_stoolY_Iglass_sofaY_Iglass_tableY_Iglass_table*2Y_Iglass_table*3Y_Iglass_table*4Y_Iglass_table*5Y_Iglass_table*6Y_Iglass_table*7Y_Iglass_table*8Y_Iglass_table*9Y_Iglass_chair*2Y_Iglass_chair*3Y_Iglass_chair*4Y_Iglass_chair*5Y_Iglass_chair*6Y_Iglass_chair*7Y_Iglass_chair*8Y_Iglass_chair*9Y_Iglass_sofa*2Y_Iglass_sofa*3Y_Iglass_sofa*4Y_Iglass_sofa*5Y_Iglass_sofa*6Y_Iglass_sofa*7Y_Iglass_sofa*8Y_Iglass_sofa*9Y_Iglass_stool*2Y_Iglass_stool*4Y_Iglass_stool*5Y_Iglass_stool*6Y_Iglass_stool*7Y_Iglass_stool*8Y_Iglass_stool*3Y_Iglass_stool*9Y_ICF_10_coin_goldZvCCF_1_coin_bronzeZvCCF_20_moneybagZvCCF_50_goldbarZvCCF_5_coin_silverZvChc_crptYmAhc_tvZ\BgothgateX~DgothiccandelabraYxBgothrailingX~Dgoth_tableYmAhc_bkshlfYmAhc_btlrY_Ihc_crtnYmAhc_djsetYmAhc_frplcZbBhc_lmpstYmAhc_machineYmAhc_rllrXQHhc_rntgnX~Dhc_trllYmAgothic_chair*1X~Dgothic_sofa*1X~Dgothic_stool*1X~Dgothic_chair*2X~Dgothic_sofa*2X~Dgothic_stool*2X~Dgothic_chair*3X~Dgothic_sofa*3X~Dgothic_stool*3X~Dgothic_chair*4X~Dgothic_sofa*4X~Dgothic_stool*4X~Dgothic_chair*5X~Dgothic_sofa*5X~Dgothic_stool*5X~Dgothic_chair*6X~Dgothic_sofa*6X~Dgothic_stool*6X~Dval_cauldronX~Dsound_machineX~Dromantique_pianochair*3Y_Iromantique_pianochair*5Y_Iromantique_pianochair*2Y_Iromantique_pianochair*4Y_Iromantique_pianochair*1Y_Iromantique_divan*3Y_Iromantique_divan*5Y_Iromantique_divan*2Y_Iromantique_divan*4Y_Iromantique_divan*1Y_Iromantique_chair*3Y_Iromantique_chair*5Y_Iromantique_chair*2Y_Iromantique_chair*4Y_Iromantique_chair*1Y_Irare_parasolY_Iplant_valentinerose*3XICplant_valentinerose*5XICplant_valentinerose*2XICplant_valentinerose*4XICplant_valentinerose*1XICplant_mazegateYeCplant_mazeZcCplant_bulrushXICpetfood4Y_Icarpet_valentineZ|Egothic_carpetXICgothic_carpet2Z|Egothic_chairX~Dgothic_sofaX~Dgothic_stoolX~Dgrand_piano*3Z|Egrand_piano*5Z|Egrand_piano*2Z|Egrand_piano*4Z|Egrand_piano*1Z|Etheatre_seatZ@Kromantique_tray2Y_Iromantique_tray1Y_Iromantique_smalltabl*3Y_Iromantique_smalltabl*5Y_Iromantique_smalltabl*2Y_Iromantique_smalltabl*4Y_Iromantique_smalltabl*1Y_Iromantique_mirrortablY_Iromantique_divider*3Z[Mromantique_divider*2Z[Mromantique_divider*4Z[Mromantique_divider*1Z[Mjp_tatami2[dWjp_tatamiYGGhabbowood_chairYGGjp_bambooYGGjp_iroriXQHjp_pillowYGGsound_set_1[dWsound_set_2[dWsound_set_3[dWsound_set_4[dWsound_set_5[dWsound_set_6[dWsound_set_7[dWsound_set_8[dWsound_set_9[dWsound_machine*1Yc[spotlightY_Isound_machine*2Yc[sound_machine*3Yc[sound_machine*4Yc[sound_machine*5Yc[sound_machine*6Yc[sound_machine*7Yc[rom_lampZ|Erclr_sofaXQHrclr_gardenXQHrclr_chairZ|Esound_set_28[dWsound_set_27[dWsound_set_26[dWsound_set_25[dWsound_set_24[dWsound_set_23[dWsound_set_22[dWsound_set_21[dWsound_set_20[dWsound_set_19[dWsound_set_18[dWsound_set_17[dWsound_set_16[dWsound_set_15[dWsound_set_14[dWsound_set_13[dWsound_set_12[dWsound_set_11[dWsound_set_10[dWrope_dividerXQHromantique_clockY_Irare_icecream_campaignY_Ipura_mdl5*1Yc[pura_mdl5*2Yc[pura_mdl5*3Yc[pura_mdl5*4Yc[pura_mdl5*5Yc[pura_mdl5*6Yc[pura_mdl5*7Yc[pura_mdl5*8Yc[pura_mdl5*9Yc[pura_mdl4*1XQHpura_mdl4*2XQHpura_mdl4*3XQHpura_mdl4*4XQHpura_mdl4*5XQHpura_mdl4*6XQHpura_mdl4*7XQHpura_mdl4*8XQHpura_mdl4*9XQHpura_mdl3*1XQHpura_mdl3*2XQHpura_mdl3*3XQHpura_mdl3*4XQHpura_mdl3*5XQHpura_mdl3*6XQHpura_mdl3*7XQHpura_mdl3*8XQHpura_mdl3*9XQHpura_mdl2*1XQHpura_mdl2*2XQHpura_mdl2*3XQHpura_mdl2*4XQHpura_mdl2*5XQHpura_mdl2*6XQHpura_mdl2*7XQHpura_mdl2*8XQHpura_mdl2*9XQHpura_mdl1*1XQHpura_mdl1*2XQHpura_mdl1*3XQHpura_mdl1*4XQHpura_mdl1*5XQHpura_mdl1*6XQHpura_mdl1*7XQHpura_mdl1*8XQHpura_mdl1*9XQHjp_lanternXQHchair_basic*1XQHchair_basic*2XQHchair_basic*3XQHchair_basic*4XQHchair_basic*5XQHchair_basic*6XQHchair_basic*7XQHchair_basic*8XQHchair_basic*9XQHbed_budget*1XQHbed_budget*2XQHbed_budget*3XQHbed_budget*4XQHbed_budget*5XQHbed_budget*6XQHbed_budget*7XQHbed_budget*8XQHbed_budget*9XQHbed_budget_one*1XQHbed_budget_one*2XQHbed_budget_one*3XQHbed_budget_one*4XQHbed_budget_one*5XQHbed_budget_one*6XQHbed_budget_one*7XQHbed_budget_one*8XQHbed_budget_one*9XQHjp_drawerXQHtile_stellaZ[Mtile_marbleZ[Mtile_brownZ[Msummer_grill*1Y_Isummer_grill*2Y_Isummer_grill*3Y_Isummer_grill*4Y_Isummer_chair*1Y_Isummer_chair*2Y_Isummer_chair*3Y_Isummer_chair*4Y_Isummer_chair*5Y_Isummer_chair*6Y_Isummer_chair*7Y_Isummer_chair*8Y_Isummer_chair*9Y_Isound_set_36[dWsound_set_35[dWsound_set_34[dWsound_set_33[dWsound_set_32[dWsound_set_31[dWsound_set_30[dWsound_set_29[dWsound_machine_proYc[rare_mnstrY_Ione_way_door*1XQHone_way_door*2XQHone_way_door*3XQHone_way_door*4XQHone_way_door*5XQHone_way_door*6XQHone_way_door*7XQHone_way_door*8XQHone_way_door*9XQHexe_rugZ[Mexe_s_tableZGRsound_set_37[dWsummer_pool*1ZlIsummer_pool*2ZlIsummer_pool*3ZlIsummer_pool*4ZlIsong_diskYc[jukebox*1Yc[carpet_soft_tut[~Nsound_set_44[dWsound_set_43[dWsound_set_42[dWsound_set_41[dWsound_set_40[dWsound_set_39[dWsound_set_38[dWgrunge_chairZ@Kgrunge_mattressZ@Kgrunge_radiatorZ@Kgrunge_shelfZ@Kgrunge_signZ@Kgrunge_tableZ@Khabboween_crypt[uKhabboween_grassZ@Khal_cauldronZ@Khal_graveZ@Ksound_set_52[dWsound_set_51[dWsound_set_50[dWsound_set_49[dWsound_set_48[dWsound_set_47[dWsound_set_46[dWsound_set_45[dWxmas_icelampZ[Mxmas_cstl_wallZ[Mxmas_cstl_twrZ[Mxmas_cstl_gate[~Ntree7Z[Mtree6Z[Msound_set_54[dWsound_set_53[dWsafe_silo_pb[dOplant_mazegate_snowZ[Mplant_maze_snowZ[Mchristmas_sleighZ[Mchristmas_reindeer[~Nchristmas_poopZ[Mexe_bardeskZ[Mexe_chairZ[Mexe_chair2Z[Mexe_cornerZ[Mexe_drinksZ[Mexe_sofaZ[Mexe_tableZ[Msound_set_59[dWsound_set_58[dWsound_set_57[dWsound_set_56[dWsound_set_55[dWnoob_table*1[~Nnoob_table*2[~Nnoob_table*3[~Nnoob_table*4[~Nnoob_table*5[~Nnoob_table*6[~Nnoob_stool*1[~Nnoob_stool*2[~Nnoob_stool*3[~Nnoob_stool*4[~Nnoob_stool*5[~Nnoob_stool*6[~Nnoob_rug*1[~Nnoob_rug*2[~Nnoob_rug*3[~Nnoob_rug*4[~Nnoob_rug*5[~Nnoob_rug*6[~Nnoob_lamp*1[dOnoob_lamp*2[dOnoob_lamp*3[dOnoob_lamp*4[dOnoob_lamp*5[dOnoob_lamp*6[dOnoob_chair*1[~Nnoob_chair*2[~Nnoob_chair*3[~Nnoob_chair*4[~Nnoob_chair*5[~Nnoob_chair*6[~Nexe_globe[~Nexe_plantZ[Mval_teddy*1[dOval_teddy*2[dOval_teddy*3[dOval_teddy*4[dOval_teddy*5[dOval_teddy*6[dOval_randomizer[dOval_choco[dOteleport_doorYc[dOarrowYc[sound_set_61[dWsound_set_60[dWfortune[dOsw_tableZIPsw_raven[cQsw_chestZIPsand_cstl_wallZIPsand_cstl_twrZIPsand_cstl_gateZIPgrunge_candleZIPgrunge_benchZIPgrunge_barrelZIPrclr_lampZGRprizetrophy9*1Yd[prizetrophy8*1Yd[nouvelle_traxYc[md_rugZGRjp_tray6ZGRjp_tray5ZGRjp_tray4ZGRjp_tray3ZGRjp_tray2ZGRjp_tray1ZGRarabian_teamkZGRarabian_snakeZGRarabian_rugZGRarabian_pllwZGRarabian_divdrZGRarabian_chairZGRarabian_bigtbZGRarabian_tetblZGRarabian_tray1ZGRarabian_tray2ZGRarabian_tray3ZGRarabian_tray4ZGRsound_set_64[dWsound_set_63[dWsound_set_62[dWjukebox_ptv*1Yc[calippoZAStraxsilverYc[traxgoldYc[traxbronzeYc[bench_puffetYATCFC_500_goldbarZvCCFC_200_moneybagZvCCFC_10_coin_bronzeZvCCFC_100_coin_goldZvCCFC_50_coin_silverZvCjp_tableXMVjp_rareXMVjp_katana3XMVjp_katana2XMVjp_katana1XMVfootylamp_campaignXMVtiki_waterfall[dWtiki_tray4[dWtiki_tray3[dWtiki_tray2[dWtiki_tray1[dWtiki_tray0[dWtiki_toucan[dWtiki_torch[dWtiki_statue[dWtiki_sand[dWtiki_parasol[dWtiki_junglerug[dWtiki_corner[dWtiki_bflies[dWtiki_bench[dWtiki_bardesk[dWtampax_rug[dWsound_set_70[dWsound_set_69[dWsound_set_68[dWsound_set_67[dWsound_set_66[dWsound_set_65[dWnoob_rug_tradeable*1[dWnoob_rug_tradeable*2[dWnoob_rug_tradeable*3[dWnoob_rug_tradeable*4[dWnoob_rug_tradeable*5[dWnoob_rug_tradeable*6[dWnoob_plant[dWnoob_lamp_tradeable*1[dWnoob_lamp_tradeable*2[dWnoob_lamp_tradeable*3[dWnoob_lamp_tradeable*4[dWnoob_lamp_tradeable*5[dWnoob_lamp_tradeable*6[dWnoob_chair_tradeable*1[dWnoob_chair_tradeable*2[dWnoob_chair_tradeable*3[dWnoob_chair_tradeable*4[dWnoob_chair_tradeable*5[dWnoob_chair_tradeable*6[dWjp_teamaker[dWsvnr_ukXhXsvnr_nlXhXsvnr_itXhXsvnr_de[gXsvnr_aus[gXdiner_tray_7[gXdiner_tray_6[gXdiner_tray_5[gXdiner_tray_4[gXdiner_tray_3[gXdiner_tray_2[gXdiner_tray_1[gXdiner_tray_0[gXdiner_sofa_2*1[gXdiner_sofa_2*2[gXdiner_sofa_2*3[gXdiner_sofa_2*4[gXdiner_sofa_2*5[gXdiner_sofa_2*6[gXdiner_sofa_2*7[gXdiner_sofa_2*8[gXdiner_sofa_2*9[gXdiner_shaker[gXdiner_rug[gXdiner_gumvendor*1[gXdiner_gumvendor*2[gXdiner_gumvendor*3[gXdiner_gumvendor*4[gXdiner_gumvendor*5[gXdiner_gumvendor*6[gXdiner_gumvendor*7[gXdiner_gumvendor*8[gXdiner_gumvendor*9[gXdiner_cashreg*1[gXdiner_cashreg*2[gXdiner_cashreg*3[gXdiner_cashreg*4[gXdiner_cashreg*5[gXdiner_cashreg*6[gXdiner_cashreg*7[gXdiner_cashreg*8[gXdiner_cashreg*9[gXdiner_table_2*1XiZdiner_table_2*2XiZdiner_table_2*3XiZdiner_table_2*4XiZdiner_table_2*5XiZdiner_table_2*6XiZdiner_table_2*7XiZdiner_table_2*8XiZdiner_table_2*9XiZdiner_table_1*1XiZdiner_table_1*2XiZdiner_table_1*3XiZdiner_table_1*4XiZdiner_table_1*5XiZdiner_table_1*6XiZdiner_table_1*7XiZdiner_table_1*8XiZdiner_table_1*9XiZdiner_sofa_1*1XiZdiner_sofa_1*2XiZdiner_sofa_1*3XiZdiner_sofa_1*4XiZdiner_sofa_1*5XiZdiner_sofa_1*6XiZdiner_sofa_1*7XiZdiner_sofa_1*8XiZdiner_sofa_1*9XiZdiner_chair*1XiZdiner_chair*2XiZdiner_chair*3XiZdiner_chair*4XiZdiner_chair*5XiZdiner_chair*6XiZdiner_chair*7XiZdiner_chair*8XiZdiner_chair*9XiZdiner_bardesk_gate*1XiZdiner_bardesk_gate*2XiZdiner_bardesk_gate*3XiZdiner_bardesk_gate*4XiZdiner_bardesk_gate*5XiZdiner_bardesk_gate*6XiZdiner_bardesk_gate*7XiZdiner_bardesk_gate*8XiZdiner_bardesk_gate*9XiZdiner_bardesk_corner*1XiZdiner_bardesk_corner*2XiZdiner_bardesk_corner*3XiZdiner_bardesk_corner*4XiZdiner_bardesk_corner*5XiZdiner_bardesk_corner*6XiZdiner_bardesk_corner*7XiZdiner_bardesk_corner*8XiZdiner_bardesk_corner*9XiZdiner_bardesk*1XiZdiner_bardesk*2XiZdiner_bardesk*3XiZdiner_bardesk*4XiZdiner_bardesk*5XiZdiner_bardesk*6XiZdiner_bardesk*7XiZdiner_bardesk*8XiZdiner_bardesk*9XiZads_dave_cnsXiZeasy_carpetYc[easy_bowl2Yc[greek_cornerYc[greek_gateYc[greek_pillarsYc[greek_seatYc[greektrophy*1[P\greektrophy*2[P\greektrophy*3[P\greek_blockXt[PPpost.itHpost.it.vdHphotoHChessHTicTacToeHBattleShipHPokerHwallpaperHfloorHposterZ@KgothicfountainYxBhc_wall_lampZbBindustrialfanZ`BtorchZ\Bval_heartXBCwallmirrorZ|Ejp_ninjastarsXQHhabw_mirrorXQHhabbowheelZ[Mguitar_skullZ@Kguitar_vZ@Kxmas_light[~Nhrella_poster_3[Nhrella_poster_2ZIPhrella_poster_1[Nsw_swordsZIPsw_stoneZIPsw_holeZIProomdimmerYc[md_logo_wallZGRmd_canZGRjp_sheet3ZGRjp_sheet2ZGRjp_sheet1ZGRarabian_swordsZGRarabian_wndwZGRtiki_wallplnt[dWtiki_surfboard[dWtampax_wall[dWwindow_single_default[gXwindow_double_default[gXnoob_window_double[dWwindow_triple[gXwindow_square[gXwindow_romantic_wide[gXwindow_romantic_narrow[gXwindow_grunge[gXwindow_golden[gXwindow_chinese_wide[gXwindow_chinese_narrowYA\window_basic[gXwindow_70s_wide[gXwindow_70s_narrow[gXads_sunnydYlXwindow_diner2XiZwindow_dinerXiZdiner_walltableXiZads_dave_wallXiZwindow_holeYc[easy_posterYc[ads_nokia_logoYc[ads_nokia_phoneYc[landscapeYA\window_skyscraper[j\");
                                    _receivedSpriteIndex = true;
                                }
                            }
                            break;
                        }

                    case "@": // Enter room - guestroom - get wallitems
                        {
                            if (_ROOMACCESS_SECONDARY_OK && Room != null)
                                sendData("@m" + Room.Wallitems);
                            break;
                        }

                    case "A@": // Enter room - add this user to room
                        {
                            if (_ROOMACCESS_SECONDARY_OK && Room != null && roomUser == null)
                            {
                                sendData("@b" + Room.dynamicStatuses);
                                Room.addUser(this);
                                if (Room.hasSpecialCasts("cam1"))
                                {
                                    sendData("AGcam1 targetcamera " + roomUser.roomUID);
                                }
                            }

                            #region User Checker/Updater (Dead/Jail/Working)

                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                //saves login room (when you login)
                                dbClient.runQuery("UPDATE users SET onloadroom = '" + _roomID +"' WHERE id = '" + userID +"'");
                                _BotTrade = null;
                                _crackingVault = false;
                                //gets certian information
                                int global_job = dbClient.getInt("SELECT global FROM jobs_ranks WHERE id = '" + _SecureID + "'");
                                //checks if you have a global job
                                if (_Working == true && _Jailed == 0)
                                {
                                    int workshere = dbClient.getInt("SELECT room_id FROM jobs_rooms WHERE rank_id = '" + _SecureID + "' && room_id = '" + _roomID + "'");
                                    if (workshere != _roomID && global_job != 1)
                                    {
                                        _Working = false;
                                        refreshAppearance(true, true, true);
                                    }
                                }

                                //checks if you are arrested
                                if (_Jailed == 1)
                                {
                                    //checks if you are trying to escape prison
                                    if (_roomID != 7)
                                    {
                                        sendData("DBO");
                                        _roomID = 7;
                                        roomidi = 7;
                                    }

                                    //stops you from working if you were arrested while working
                                    if (_Working == true)
                                    {
                                        _Working = false;

                                        _Figure = _Uni + "ch-220-94.lg-280-94.sh-290-62";
                                        _Mission = "[Jailed] Civilian";
                                        refreshAppearance(false, true, true);
                                    }
                                    else
                                    {
                                        _Figure = _Uni + "ch-220-94.lg-280-94.sh-290-62";
                                        _Mission = "[Jailed] Civilian";
                                        refreshAppearance(false, true, true);
                                    }
                                }

                                //chcks if you are trying to escape the hospital while dead
                                if (_Dead == 1)
                                {
                                    if (_roomID != 2)
                                    {
                                        sendData("BK" + "You cannot leave the room while you are knocked out.");
                                        sendData("DBO");
                                        _roomID = 2;
                                        roomidi = 2;

                                    }

                                    //stops you from working if you were killed while working
                                    if (_Working == true)
                                    {
                                        _Working = false;

               

                                        refreshAppearance(true, true, true);

                                        _Mission = "[Injured] Civilian";
                                        refreshAppearance(false, true, true);
                                        
                                    }
                                    else
                                    {
                                        _Mission = "[Injured] Civilian";

                                        refreshAppearance(false, true, true);
                                    }
                                }
                            }

                            #endregion

                            break;
                        }

                    #endregion

                    #region Moderation
                    #region MOD-Tool
                    case "CH": // MOD-Tool
                        {
                            int messageLength = 0;
                            string Message = "";
                            int staffNoteLength = 0;
                            string staffNote = "";
                            string targetUser = "";

                            switch (currentPacket.Substring(2, 2)) // Select the action
                            {
                                #region Alert single user
                                case "HH": // Alert single user
                                    {
                                        if (rankManager.containsRight(_Rank, "fuse_alert", userID) == false) { sendData("BK" + stringManager.getString("modtool_accesserror")); return; }
                                        {
                                            messageLength = Encoding.decodeB64(currentPacket.Substring(4, 2));
                                            Message = currentPacket.Substring(6, messageLength).Replace(Convert.ToChar(1).ToString(), " ");
                                            staffNoteLength = Encoding.decodeB64(currentPacket.Substring(messageLength + 6, 2));
                                            staffNote = currentPacket.Substring(messageLength + 8, staffNoteLength);
                                            targetUser = currentPacket.Substring(messageLength + staffNoteLength + 10);

                                            if (Message == "")
                                                return;

                                            string Username = _Username;
                                            userManager.sendData("BK" + stringManager.getString("scommand_hotelalert") + "\r" + Message + "\r" + " - " + Username);
                                        }
                                        break;
                                    }
                                #endregion

                                #region Kick single user from room
                                case "HI": // Kick single user from room
                                    {
                                        if (rankManager.containsRight(_Rank, "fuse_kick", userID) == false) { sendData("BK" + stringManager.getString("modtool_accesserror")); return; }

                                        messageLength = Encoding.decodeB64(currentPacket.Substring(4, 2));
                                        Message = currentPacket.Substring(6, messageLength).Replace(Convert.ToChar(1).ToString(), " ");
                                        staffNoteLength = Encoding.decodeB64(currentPacket.Substring(messageLength + 6, 2));
                                        staffNote = currentPacket.Substring(messageLength + 8, staffNoteLength);
                                        targetUser = currentPacket.Substring(messageLength + staffNoteLength + 10);

                                        if (Message == "" || targetUser == "")
                                            return;

                                        virtualUser _targetUser = userManager.getUser(targetUser);
                                        if (_targetUser == null)
                                            sendData("BK" + stringManager.getString("modtool_actionfail") + "\r" + stringManager.getString("modtool_usernotfound"));
                                        else
                                        {
                                            if (_targetUser.Room != null && _targetUser.roomUser != null)
                                            {
                                                if (_targetUser._Rank < _Rank)
                                                {
                                                    _targetUser.Room.removeUser(_targetUser.roomUser.roomUID, true, Message);
                                                }
                                                else
                                                    sendData("BK" + stringManager.getString("modtool_actionfail") + "\r" + stringManager.getString("modtool_rankerror"));
                                            }
                                        }
                                        break;
                                    }
                                #endregion

                                #region Ban single user
                                case "HJ": // Ban single user / IP
                                    {
                                        if (rankManager.containsRight(_Rank, "fuse_ban", userID) == false) { sendData("BK" + stringManager.getString("modtool_accesserror")); return; }

                                        int targetUserLength = 0;
                                        int banHours = 0;
                                        bool banIP = (currentPacket.Substring(currentPacket.Length - 1, 1) == "I");

                                        messageLength = Encoding.decodeB64(currentPacket.Substring(4, 2));
                                        Message = currentPacket.Substring(6, messageLength).Replace(Convert.ToChar(1).ToString(), " ");
                                        staffNoteLength = Encoding.decodeB64(currentPacket.Substring(messageLength + 6, 2));
                                        staffNote = currentPacket.Substring(messageLength + 8, staffNoteLength);
                                        targetUserLength = Encoding.decodeB64(currentPacket.Substring(messageLength + staffNoteLength + 8, 2));
                                        targetUser = currentPacket.Substring(messageLength + staffNoteLength + 10, targetUserLength);
                                        banHours = Encoding.decodeVL64(currentPacket.Substring(messageLength + staffNoteLength + targetUserLength + 10));

                                        if (Message == "" || targetUser == "" || banHours == 0)
                                            return;
                                        else
                                        {
                                            DataRow dRow;
                                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                            {
                                                dbClient.AddParamWithValue("name", targetUser);
                                                dRow = dbClient.getRow("SELECT id,rank,ipaddress_last FROM users WHERE name = @name");
                                            }
                                            if (dRow.Table.Rows.Count == 0)
                                            {
                                                sendData("BK" + stringManager.getString("modtool_actionfail") + "\r" + stringManager.getString("modtool_usernotfound"));
                                                return;
                                            }
                                            else if (Convert.ToByte(dRow["rank"]) >= _Rank)
                                            {
                                                sendData("BK" + stringManager.getString("modtool_actionfail") + "\r" + stringManager.getString("modtool_rankerror"));
                                                return;
                                            }

                                            int targetID = Convert.ToInt32(dRow["id"]);
                                            string Report = "";
                                            if (banIP && rankManager.containsRight(_Rank, "fuse_superban", userID)) // IP ban is chosen and allowed for this staff member
                                            {
                                                userManager.setBan(Convert.ToString(dRow["ipaddress_last"]), banHours, Message);
                                                Report = userManager.generateBanReport(Convert.ToString(dRow["ipaddress_last"]));
                                            }
                                            else
                                            {
                                                userManager.setBan(targetID, banHours, Message);
                                                Report = userManager.generateBanReport(targetID);
                                            }

                                            sendData("BK" + Report);
                                        }
                                        break;
                                    }
                                #endregion

                                #region Room alert
                                case "IH": // Alert all users in current room
                                    {
                                        if (rankManager.containsRight(_Rank, "fuse_room_alert", userID) == false) { sendData("BK" + stringManager.getString("modtool_accesserror")); return; }
                                        if (Room == null || roomUser == null) { return; }

                                        messageLength = Encoding.decodeB64(currentPacket.Substring(4, 2));
                                        Message = currentPacket.Substring(6, messageLength).Replace(Convert.ToChar(1).ToString(), " ");
                                        staffNoteLength = Encoding.decodeB64(currentPacket.Substring(messageLength + 6, 2));
                                        staffNote = currentPacket.Substring(messageLength + 8, staffNoteLength);

                                        if (Message != "")
                                        {
                                            Room.sendData("B!" + Message + Convert.ToChar(2));
                                        }
                                        break;
                                    }
                                #endregion

                                #region Room kick
                                case "II": // Kick all users below users rank from room
                                    {
                                        if (rankManager.containsRight(_Rank, "fuse_room_kick", userID) == false) { sendData("BK" + stringManager.getString("modtool_accesserror")); return; }
                                        if (Room == null || roomUser == null) { return; }

                                        messageLength = Encoding.decodeB64(currentPacket.Substring(4, 2));
                                        Message = currentPacket.Substring(6, messageLength).Replace(Convert.ToChar(1).ToString(), " ");
                                        staffNoteLength = Encoding.decodeB64(currentPacket.Substring(messageLength + 6, 2));
                                        staffNote = currentPacket.Substring(messageLength + 8, staffNoteLength);

                                        if (Message != "")
                                        {
                                            Room.kickUsers(_Rank, Message);
                                        }
                                        break;
                                    }
                                #endregion
                            }
                            break;
                        }
                    #endregion

                    #region Call For Help

                    #region User Side
                    case "Cm": // User wants to send a CFH message 
                        {
                            //Database dbClient = new Database(true, true, 88);
                            DataRow dRow;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                dRow = dbClient.getRow("SELECT id, date, message FROM cms_help WHERE username = '" + _Username + "' AND picked_up = '0'");
                            }
                            if (dRow.Table.Rows.Count == 0)
                                sendData("D" + "H");
                            else
                                sendData("D" + "I" + Convert.ToString(dRow[0]) + Convert.ToChar(2) + Convert.ToString(dRow[1]) + Convert.ToChar(2) + Convert.ToString(dRow[2]) + Convert.ToChar(2));
                            break;
                        }

                    case "Cn": // User deletes his pending CFH message
                        {
                            int cfhID;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                cfhID = dbClient.getInt("SELECT id FROM cms_help WHERE username = '" + _Username + "' AND picked_up = '0'");
                                dbClient.runQuery("DELETE FROM cms_help WHERE picked_up = '0' AND username = '" + _Username + "' LIMIT 1");
                            }
                            sendData("DH");
                            userManager.sendToRank("BT" + Encoding.encodeVL64(cfhID) + Convert.ToChar(2) + "I" + "User Deleted!" + Convert.ToChar(2) + "User Deleted!" + Convert.ToChar(2) + "User Deleted!" + Convert.ToChar(2) + Encoding.encodeVL64(0) + Convert.ToChar(2) + "" + Convert.ToChar(2) + "H" + Convert.ToChar(2) + Encoding.encodeVL64(0));
                            break;
                        }

                    case "AV": // User sends CFH message
                        {
                            //Database dbClient = new Database(true, true, 90);
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                if (dbClient.findsResult("SELECT id FROM cms_help WHERE username = '" + _Username + "' AND picked_up = '0'") == true)
                                    return;
                            }
                            int messageLength = Encoding.decodeB64(currentPacket.Substring(2, 2));
                            if (messageLength == 0)
                                return;
                            string cfhMessage = currentPacket.Substring(4, messageLength);
                            //dbClient = new Database(false, false, 91);
                            int cfhID;
                            string roomName;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                dbClient.AddParamWithValue("username", _Username);
                                dbClient.AddParamWithValue("ip", connectionSocket.RemoteEndPoint.ToString().Split(Char.Parse(":"))[0]);
                                dbClient.AddParamWithValue("message", cfhMessage);
                                dbClient.AddParamWithValue("date", DateTime.Now);
                                dbClient.AddParamWithValue("roomid", _roomID.ToString());
                                //dbClient.Open();
                                dbClient.runQuery("INSERT INTO cms_help (username,ip,message,date,picked_up,subject,roomid) VALUES (@username,@ip,@message,@date,'0','CFH message [hotel]',@roomid)");
                                dbClient.runQuery("INSERT INTO cfh_messages_log (username,ip,message,date,picked_up,subject,roomid) VALUES (@username,@ip,@message,@date,'0','CFH message [hotel]',@roomid)");
                                cfhID = dbClient.getInt("SELECT id FROM cms_help WHERE username = @username AND picked_up = '0'");
                                roomName = dbClient.getString("SELECT name FROM rooms WHERE id = @roomid");//                                                                                                                                                                                                                H = Hide Room ID / I = Show Room ID
                            }
                            //dbClient.Close();                                                                       //       H = Automated / I = Manual    
                            sendData("EAH"); //                                                                                           \_/                                                                                                                                                                                                    \_/
                            userManager.sendToRank("BT" + Encoding.encodeVL64(cfhID) + Convert.ToChar(2) + "I" + "Sent: " + DateTime.Now + Convert.ToChar(2) + _Username + Convert.ToChar(2) + cfhMessage + Convert.ToChar(2) + Encoding.encodeVL64(_roomID) + Convert.ToChar(2) + roomName + Convert.ToChar(2) + "I" + Convert.ToChar(2) + Encoding.encodeVL64(_roomID));
                            break;
                        }
                    #endregion

                    #region Staff Side
                    case "CG": // CFH center - reply call
                        {
                            if (rankManager.containsRight(_Rank, "fuse_receive_calls_for_help", userID) == false)
                                return;
                            int cfhID = Encoding.decodeVL64(currentPacket.Substring(4, Encoding.decodeB64(currentPacket.Substring(2, 2))));
                            string cfhReply = currentPacket.Substring(Encoding.decodeB64(currentPacket.Substring(2, 2)) + 6);

                            //Database dbClient = new Database(true, false, 92);
                            string toUserName;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                toUserName = dbClient.getString("SELECT username FROM cms_help WHERE id = '" + cfhID + "'");
                            }
                            if (toUserName == null)
                                sendData("BK" + stringManager.getString("cfh_fail"));
                            else
                            {
                                int toUserID = userManager.getUserID(toUserName);
                                virtualUser toVirtualUser = userManager.getUser(toUserID);
                                if (toVirtualUser._isLoggedIn)
                                {
                                    toVirtualUser.sendData("DR" + cfhReply + Convert.ToChar(2));
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE cms_help SET picked_up = '" + _Username + "' WHERE id = '" + cfhID + "' LIMIT 1");
                                    }
                                }
                            }
                            //dbClient.Close();
                            break;
                        }

                    case "CF": // CFH center - Delete (Downgrade)
                        {
                            if (rankManager.containsRight(_Rank, "fuse_receive_calls_for_help", userID) == false)
                                return;
                            int cfhID = Encoding.decodeVL64(currentPacket.Substring(4, Encoding.decodeB64(currentPacket.Substring(2, 2))));
                            //Database dbClient = new Database(true, false, 93);
                            DataRow dRow;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                dRow = dbClient.getRow("SELECT picked_up FROM cms_help WHERE id = '" + cfhID + "'");
                            }
                            if (dRow.Table.Columns.Count == 0)
                            {
                                //dbClient.Close();
                                return;
                            }
                            else
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.runQuery("DELETE FROM cms_help WHERE id = '" + cfhID + "' LIMIT 1");
                                }
                                userManager.sendToRank("BT" + Encoding.encodeVL64(cfhID) + Convert.ToChar(2) + "H" + "Staff Deleted!" + Convert.ToChar(2) + "Staff Deleted!" + Convert.ToChar(2) + "Staff Deleted!" + Convert.ToChar(2) + "H" + Convert.ToChar(2) + Convert.ToChar(2) + "H" + Convert.ToChar(2) + "H");
                            }
                            //dbClient.Close();
                            break;
                        }

                    case "@p": // CFH center - Pickup
                        {
                            int cfhID = Encoding.decodeVL64(currentPacket.Substring(4));
                            //Database dbClient = new Database(true, false, 94);
                            bool result;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                result = dbClient.findsResult("SELECT id FROM cms_help WHERE id = '" + cfhID + "'");
                            }
                            if (result == false)
                            {
                                sendData("BK" + stringManager.getString("cfh_deleted"));
                                return;
                            }
                            DataRow dRow;
                            string roomName;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                dRow = dbClient.getRow("SELECT picked_up,username,message,roomid FROM cms_help WHERE id = '" + cfhID + "'");
                                roomName = dbClient.getString("SELECT name FROM rooms WHERE id = '" + Convert.ToString(dRow[3]) + "'");
                            }
                            if (Convert.ToString(dRow[0]) == "1")
                                sendData("BK" + stringManager.getString("cfh_picked_up"));
                            else
                                userManager.sendToRank("BT" + Encoding.encodeVL64(cfhID) + Convert.ToChar(2) + "I" + "Picked up: " + DateTime.Now + Convert.ToChar(2) + Convert.ToString(dRow[1]) + Convert.ToChar(2) + Convert.ToString(dRow[2]) + Convert.ToChar(2) + Encoding.encodeVL64(Convert.ToInt32(dRow[3])) + Convert.ToChar(2) + roomName + Convert.ToChar(2) + "I" + Convert.ToChar(2) + Encoding.encodeVL64(Convert.ToInt32(dRow[3])));
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                dbClient.runQuery("UPDATE cms_help SET picked_up = '1' WHERE id = '" + cfhID + "' LIMIT 1");
                            }

                            break;
                        }

                    case "EC": // Go to the room that the call for help was sent from
                        {
                            if (rankManager.containsRight(_Rank, "fuse_receive_calls_for_help", userID) == false)
                                return;
                            int idLength = Encoding.decodeB64(currentPacket.Substring(2, 2));
                            int cfhID = Encoding.decodeVL64(currentPacket.Substring(4, idLength));
                            //Database dbClient = new Database(true, true, 95);
                            int roomID;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                roomID = dbClient.getInt("SELECT roomid FROM cms_help WHERE id = '" + cfhID + "'");
                            }
                            if (roomID == 0)
                                return;
                            _onRoad = true;
                            sendData("DBO");
                            _roomID = roomID;
                            roomidi = roomID;

                            break;
                        }
                    #endregion
                    #endregion
                    #endregion

                    #region In-room actions

                    #region Misc
                    case "AO": // Room - rotate user
                        {
                            if (Room != null && roomUser != null && statusManager.containsStatus("sit") == false && statusManager.containsStatus("lay") == false)
                            {
                                int X = int.Parse(currentPacket.Substring(2).Split(' ')[0]);
                                int Y = int.Parse(currentPacket.Split(' ')[1]);
                                roomUser.Z1 = Rooms.Pathfinding.Rotation.Calculate(roomUser.X, roomUser.Y, X, Y);
                                roomUser.Z2 = roomUser.Z1;
                                roomUser.Refresh();
                            }
                            break;
                        }

                    case "AK": // Room - walk to a new square
                        {
                            if (Room != null && roomUser != null && roomUser.walkLock == false)
                            {
                                Unidle();
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    _SavedX = roomUser.X;
                                    _SavedY = roomUser.Y;
                                    int goalX = Encoding.decodeB64(currentPacket.Substring(2, 2));
                                    int goalY = Encoding.decodeB64(currentPacket.Substring(4, 2));
                                    {
                                        if (roomUser.SPECIAL_TELEPORTABLE)
                                        {
                                            roomUser.X = goalX;
                                            roomUser.Y = goalY;
                                            roomUser.goalX = -1;
                                            Room.Refresh(roomUser);
                                            refreshAppearance(false, false, true);
                                        }
                                        else
                                        {
                                            roomUser.goalX = goalX;
                                            roomUser.goalY = goalY;
                                            int id2 = dbClient.getInt("SELECT id FROM furniture WHERE x = '" + goalX + "' AND y = '" + goalY + "' AND roomid = '" + _roomID + "' AND tid = '53500'");
                                            int tid = dbClient.getInt("SELECT tid FROM furniture WHERE id = '" + id2 + "'");
                                            if (tid == 53500)
                                            {
                                                x = goalX;
                                                y = goalY;
                                                id = id2;
                                                ThreadStart PInstant = new ThreadStart(Instant);
                                                AInstant = new Thread(PInstant);
                                                AInstant.Priority = ThreadPriority.Lowest;
                                                AInstant.Start();
                                            }
                                            
                                        }
                                        

                                    }
                                    _crackingVault = false;
                                }
                            }
                            break;
                        }

                    case "B^": // Badges - switch or toggle on/off badge
                        {
                            if (Room != null && roomUser != null)
                            {
                                // Reset slots
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.runQuery("UPDATE users_badges SET slotid = '0' WHERE userid = '" + this.userID + "'");
                                }

                                int enabledBadgeAmount = 0;
                                string szWorkData = currentPacket.Substring(2);
                                while (szWorkData != "")
                                {
                                    int slotID = Encoding.decodeVL64(szWorkData);
                                    szWorkData = szWorkData.Substring(Encoding.encodeVL64(slotID).Length);

                                    int badgeNameLength = Encoding.decodeB64(szWorkData.Substring(0, 2));

                                    if (badgeNameLength > 0)
                                    {
                                        string Badge = szWorkData.Substring(2, badgeNameLength);
                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {
                                            dbClient.AddParamWithValue("badge", Badge);
                                            dbClient.runQuery("UPDATE users_badges SET slotid = '" + slotID + "' WHERE userid = '" + this.userID + "' AND badgeid = @badge LIMIT 1"); // update slot
                                        }
                                        enabledBadgeAmount++;
                                    }

                                    szWorkData = szWorkData.Substring(badgeNameLength + 2);
                                }
                                // Active badges have their badge slot set now, other ones have '0'

                                this.refreshBadges();

                                string szNotify = this.userID + Convert.ToChar(2).ToString() + Encoding.encodeVL64(enabledBadgeAmount);
                                for (int x = 0; x < _Badges.Count; x++)
                                {
                                    if (_badgeSlotIDs[x] > 0) // Badge enabled
                                    {
                                        szNotify += Encoding.encodeVL64(_badgeSlotIDs[x]);
                                        szNotify += _Badges[x];
                                        szNotify += Convert.ToChar(2);
                                    }
                                }

                                this.Room.sendData("Cd" + szNotify);
                            }
                            break;
                        }

                    case "DG": // Tags - get tags of virtual user
                        {
                            int ownerID = Encoding.decodeVL64(currentPacket.Substring(2));
                            virtualUser User = userManager.getUser(ownerID);
                            if (User != null)
                            {
                                if (User._Married != "None")
                                {
                                    string[] Tags = { "Married to:\n" + User._Married + "" };
                                    StringBuilder List = new StringBuilder(Encoding.encodeVL64(ownerID) + Encoding.encodeVL64(Tags.Length));
                                    for (int i = 0; i < Tags.Length; i++)
                                        List.Append(Tags[i] + Convert.ToChar(2));
                                    sendData("E^" + List.ToString());
                                }
                                else
                                {
                                    string[] Tags = { "Single" };
                                    StringBuilder List = new StringBuilder(Encoding.encodeVL64(ownerID) + Encoding.encodeVL64(Tags.Length));
                                    for (int i = 0; i < Tags.Length; i++)
                                        List.Append(Tags[i] + Convert.ToChar(2));
                                    sendData("E^" + List.ToString());
                                }
                            }

                            else
                            {
                                string[] Tags = { "This user is a BOT" };
                                StringBuilder List = new StringBuilder(Encoding.encodeVL64(ownerID) + Encoding.encodeVL64(Tags.Length));
                                for (int i = 0; i < Tags.Length; i++)
                                    List.Append(Tags[i] + Convert.ToChar(2));
                                sendData("E^" + List.ToString());
                            }
                            break;
                        }

                    case "Cg": // Group badges - get details about a group [click badge]
                        {
                            if (Room != null && roomUser != null)
                            {
                                int groupID = Encoding.decodeVL64(currentPacket.Substring(2));
                                //Database dbClient = new Database(true, false, 99);
                                DataRow dRow;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dRow = dbClient.getRow("SELECT corp_name,description FROM jobs_corps WHERE id = '" + groupID + "'");
                                }
                                if (dRow.Table.Rows.Count == 1)
                                {
                                    string Description = Convert.ToString(dRow["description"]);
                                    string Name = Convert.ToString(dRow["corp_name"]);
                                    sendData("Dw" + Encoding.encodeVL64(groupID) + Name + Convert.ToChar(2) + Description + Convert.ToChar(2) + Encoding.encodeVL64(-1) + Convert.ToChar(2));
                                }
                            }
                            break;
                        }

                    case "AX": // Statuses - stop status
                        {
                            if (statusManager != null)
                            {
                                string Status = currentPacket.Substring(2);
                                if (Status == "CarryItem")
                                    statusManager.dropCarrydItem();
                                else if (Status == "Dance")
                                {
                                    statusManager.removeStatus("dance");
                                    statusManager.Refresh();
                                }
                            }
                            break;
                        }

                    case "A^": // Statuses - wave
                        {
                            if (Room != null && roomUser != null && statusManager.containsStatus("wave") == false)
                            {
                                statusManager.removeStatus("dance");
                                statusManager.handleStatus("wave", "", Config.Statuses_Wave_waveDuration);
                            }
                            break;
                        }

                    case "A]": // Statuses - dance
                        {
                            if (Room != null && roomUser != null && statusManager.containsStatus("sit") == false && statusManager.containsStatus("lay") == false)
                            {
                                statusManager.dropCarrydItem();
                                if (currentPacket.Length == 2)
                                    statusManager.addStatus("dance", "");
                                else
                                {
                                    if (rankManager.containsRight(_Rank, "fuse_use_club_dance", userID) == false) { return; }
                                    int danceID = Encoding.decodeVL64(currentPacket.Substring(2));
                                    if (danceID < 0 || danceID > 4) { return; }
                                    statusManager.addStatus("dance", danceID.ToString());
                                }

                                statusManager.Refresh();
                            }
                            break;
                        }

                    case "AP": // Statuses - carry item
                        {
                            if (Room != null && roomUser != null)
                            {
                                string Item = currentPacket.Substring(2);
                                if (statusManager.containsStatus("lay") || Item.Contains("/"))
                                    return; // THE HAX! \o/

                                try
                                {
                                    int nItem = int.Parse(Item);
                                    if (nItem < 1 || nItem > 26)
                                        return;
                                }
                                catch
                                {
                                    if (_inPublicroom == false && Item != "Water" && Item != "Milk" && Item != "Juice") // Not a drink that can be retrieved from the infobus minibar
                                        return;
                                }
                                if (_Rank >= 1)
                                {
                                    _gettingDrink = true;
                                    statusManager.carryItem(Item);
                                }
                                else
                                {
                                }
                            }
                            break;
                        }

                    #endregion

                    #region Chat
                    case "@t": // Chat - say
                    case "@w": // Chat - shout
                        {
                            try
                            {
                                if (_isMuted == false && (Room != null && roomUser != null))
                                {
                                    Unidle();
                                    string Message = currentPacket.Substring(4);
                                    userManager.addChatMessage(_Username, _roomID, Message);
                                    Message = stringManager.filterSwearwords(Message);
                                    Regex rgx = new Regex("[^a-zA-Z0-9 -_.,;:()/!%&@*?]");
                                    Message = rgx.Replace(Message, "*");
                                    if (_OOC == 1 && Message != ":ic")
                                    {
                                        Message = "[ " + Message + " ]";
                                    }

                                    if (Message.Substring(0, 1) == ":" && isSpeechCommand(Message.Substring(1))) // Speechcommand invoked!
                                    {
                                        if (roomUser.isTyping)
                                        {
                                            Room.sendData("Ei" + Encoding.encodeVL64(roomUser.roomUID) + "H");
                                            roomUser.isTyping = false;
                                        }
                                    }
                                    else
                                    {
                                        if (Message.Contains("\n"))
                                        {
                                            sendData("BKNice Try >:D");
                                        }
                                        else
                                        {
                                            
                                                if (currentPacket.Substring(1, 1) == "w") // Shout
                                                {
                                                    Room.sendShout(roomUser, Message);
                                                }
                                                else
                                                {
                                                    Room.sendSaying(roomUser, Message);
                                                    //Out.WriteChat("Say", _Username, Message); 
                                                }
                                            
                                        }
                                    }
                                }
                            }
                            catch (Exception e)
                            {
                                Out.WriteError(e.ToString());
                            }
                            break;
                        }

                    case "@x": // Chat - whisper
                        {
                            if (_isMuted == false && Room != null && roomUser != null)
                            {
                                string Receiver = currentPacket.Substring(4).Split(' ')[0];
                                string Message = currentPacket.Substring(Receiver.Length + 5);
                                userManager.addChatMessageWhisper(_Username, _roomID, Message);

                                Message = stringManager.filterSwearwords(Message);

                                string TALKCHECK;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    TALKCHECK = dbClient.getString("SELECT ooc FROM users where id = '" + userID + "'");
                                }
                                if (TALKCHECK == "1" && Message != ":ic")
                                {
                                    Message = "[ " + Message + " ]";
                                }

                                Room.sendWhisper(roomUser, Receiver, Message);
                                //Out.WriteChat("Whisper", _Username + "-" + Receiver, Message); 


                            }
                            break;
                        }

                    case "D}": // Chat - show speech bubble
                        {
                            if (_isMuted == false && Room != null && roomUser != null)
                            {
                                Room.sendData("Ei" + Encoding.encodeVL64(roomUser.roomUID) + "I");
                                roomUser.isTyping = true;
                            }
                            break;
                        }

                    case "D~": // Chat - hide speech bubble
                        {
                            if (Room != null && roomUser != null)
                            {
                                Room.sendData("Ei" + Encoding.encodeVL64(roomUser.roomUID) + "H");
                                roomUser.isTyping = false;
                            }
                            break;
                        }
                    #endregion

                    #region Guestroom - rights, kicking, roombans and room voting
                    case "A`": // Give rights
                        {
                            if (Room == null || roomUser == null || _inPublicroom || _isOwner == false)
                                return;

                            string Target = currentPacket.Substring(2);
                            virtualBot Bot = botManager.getBot(Target);
                            if (Bot != null)
                            {
                                sendData("BKYou cannot give bots rights!");
                            }
                            else
                            {
                                if (userManager.containsUser(Target) == false)
                                    return;

                                virtualUser _Target = userManager.getUser(Target);
                                if (_Target._roomID != _roomID || _Target._hasRights || _Target._isOwner)
                                    return;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.runQuery("INSERT INTO room_rights(roomid,userid) VALUES ('" + _roomID + "','" + _Target.userID + "')");
                                }
                                _Target._hasRights = true;
                                _Target.statusManager.addStatus("flatctrl", "onlyfurniture");
                                _Target.roomUser.Refresh();
                                _Target.sendData("@j");
                            }
                            break;
                        }

                    case "Aa": // Take rights
                        {
                            if (Room == null || roomUser == null || _inPublicroom || _isOwner == false)
                                return;

                            string Target = currentPacket.Substring(2);
                            virtualBot Bot = botManager.getBot(Target);
                            if (Bot != null)
                            {
                                sendData("BKYou cannot take a bots rights!");
                            }
                            else
                            {

                                if (userManager.containsUser(Target) == false)
                                    return;

                                virtualUser _Target = userManager.getUser(Target);
                                if (_Target._roomID != _roomID || _Target._hasRights == false || _Target._isOwner)
                                    return;

                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.runQuery("DELETE FROM room_rights WHERE roomid = '" + _roomID + "' AND userid = '" + _Target.userID + "' LIMIT 1");
                                }
                                _Target._hasRights = false;
                                _Target.statusManager.removeStatus("flatctrl");
                                _Target.roomUser.Refresh();
                                _Target.sendData("@k");
                            }
                            break;
                        }

                    case "A_": // Kick user
                        {
                            string Target = currentPacket.Substring(2);
                            virtualBot Bot = botManager.getBot(Target);
                            if (Bot != null)
                            {
                                if (_Rank > 6)
                                {
                                    Room.deloadBot(Bot);
                                }
                            }
                            else
                            {
                                virtualUser _Target = userManager.getUser(Target);
                                {
                                    if (_Target._isOwner || _Target._Rank > _Rank || rankManager.containsRight(_Target._Rank, "fuse_any_room_controller", userID))
                                        return;
                                    if (_Target != null)
                                    {
                                        _onRoad = true;
                                        Room.sendData("D^" + "H" + Encoding.encodeVL64(4));
                                        Room.sendData("B!You Have Been Kicked Out Of The Room And Sent To Times Square" + Convert.ToChar(2));
                                    }
                                }
                            }
                            break;
                        }

                    case "DE": // Vote -1 or +1 on room
                        {
                            if (_inPublicroom || Room == null || roomUser == null)
                                return;

                            int Vote = Encoding.decodeVL64(currentPacket.Substring(2));
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            { 
                                if ((Vote == 1 || Vote == -1) && dbClient.findsResult("SELECT userid FROM room_votes WHERE userid = '" + userID + "' AND roomid = '" + _roomID + "'") == false)
                                {
                                    dbClient.runQuery("INSERT INTO room_votes (userid,roomid,vote) VALUES ('" + userID + "','" + _roomID + "','" + Vote + "')");
                                    int voteAmount = dbClient.getInt("SELECT SUM(vote) FROM room_votes WHERE roomid = '" + _roomID + "'");
                                    if (voteAmount < 0)
                                        voteAmount = 0;

                                    roomUser.hasVoted = true;
                                    if (_isOwner == true)
                                        roomUser.Room.sendNewVoteAmount(voteAmount);

                                }
                            }
                            //dbClient.Close();
                            break;
                        }
                    case "E@": // Kick and apply roomban
                        {
                            if (_Rank > 3)
                            {
                                string Target = currentPacket.Substring(2);
                                virtualBot Bot = botManager.getBot(Target);
                                if (Bot != null)
                                {
                                    sendData("BKYou cannot send bots to the RoS");
                                }
                                else
                                {
                                    virtualUser User = userManager.getUser(Target);
                                    User._onRoad = true;
                                    User.sendData("DBO");
                                    User._roomID = 187;
                                    User.roomidi = 187;
                                    User.sendData("BKYou have been sent to the room of shame!");
                                }
                            }
                            break;
                        }

                    #endregion

                    #region Catalogue and Recycler
                    case "Ae": // Catalogue - open, retrieve index of pages
                        {
                            if (Room != null && roomUser != null)
                                sendData("A~" + catalogueManager.getPageIndex(_Rank));
                            break;
                        }

                    case "Af": // Catalogue, open page, get page content
                        {
                            if (Room != null && roomUser != null)
                            {
                                string pageIndexName = currentPacket.Split('/')[1];
                                sendData("A" + catalogueManager.getPage(pageIndexName, _Rank));
                            }
                            break;
                        }



                    #region buy catalogue
                    case "Ad": // Catalogue - purchase
                        {
                            string[] packetContent = currentPacket.Split(Convert.ToChar(13));
                            string Page = packetContent[1];
                            string Item = packetContent[3];
                            string VarItem = packetContent[4];
                            int pageID;
                            int templateID;
                            int Cost;

                            using (DatabaseClient dbClient1 = Eucalypt.dbManager.GetClient())
                            {
                                dbClient1.AddParamWithValue("indexname", Page);
                                dbClient1.AddParamWithValue("name_cct", Item);
                                pageID = dbClient1.getInt("SELECT indexid FROM catalogue_pages WHERE indexname = @indexname AND minrank <= " + _Rank);
                                templateID = dbClient1.getInt("SELECT tid FROM catalogue_items WHERE name_cct = @name_cct");
                                Cost = dbClient1.getInt("SELECT catalogue_cost FROM catalogue_items WHERE catalogue_id_page = '" + pageID + "' AND tid = '" + templateID + "'");
                            }
                            if (Cost == 0 || Cost > _Credits)
                            {
                                sendData("AD");
                                return;
                            }

                            bool handlePresentbox = false;
                            int receiverID = userID;
                            int presentBoxID = 0;
                            int roomID = 0; // -1 = present box, 0 = inhand

                            if (packetContent[5] == "1") // Purchased as present
                            {
                                handlePresentbox = true;
                                string receiverName = packetContent[6];
                                using (DatabaseClient dbClient2 = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient2.AddParamWithValue("name", receiverName);
                                    if (receiverName != _Username)
                                    {
                                        int i = dbClient2.getInt("SELECT id FROM users WHERE name = @name LIMIT 1");
                                        if (i > 0)
                                            receiverID = i;
                                        else
                                        {
                                            sendData("AL" + receiverName);
                                            return;
                                        }
                                    }
                                }
                                string boxSprite = "present_gen" + new Random().Next(1, 7);
                                string boxTemplateID;
                                using (DatabaseClient dbClient1 = Eucalypt.dbManager.GetClient())
                                {
                                    boxTemplateID = dbClient1.getString("SELECT tid FROM catalogue_items WHERE name_cct = '" + boxSprite + "'");
                                    dbClient1.AddParamWithValue("tid", boxTemplateID);
                                    dbClient1.AddParamWithValue("ownerid", receiverID);
                                    dbClient1.AddParamWithValue("var", "!" + stringManager.filterSwearwords(packetContent[7]));
                                    dbClient1.runQuery("INSERT INTO furniture(tid,ownerid,var) VALUES (@tid,@ownerid,@var)");
                                    presentBoxID = dbClient1.getInt("SELECT MAX(id) FROM furniture");
                                }
                                roomID = -1;
                            }

                            _Credits -= Cost;
                            sendData("@F" + _Credits);
                            using (DatabaseClient dbClient1 = Eucalypt.dbManager.GetClient())
                            {
                                dbClient1.runQuery("UPDATE users SET credits = '" + _Credits + "' WHERE id = '" + userID + "' LIMIT 1");
                            }

                            if (stringManager.getStringPart(Item, 0, 4) == "deal")
                            {
                                int dealID = int.Parse(Item.Substring(4));
                                DataTable dTable;
                                using (DatabaseClient dbClient1 = Eucalypt.dbManager.GetClient())
                                {
                                    dTable = dbClient1.getTable("SELECT tid,amount FROM catalogue_deals WHERE id = '" + dealID + "'");
                                }
                                StringBuilder sb = new StringBuilder();
                                foreach (DataRow dRow in dTable.Rows)
                                {
                                    for (int i = 0; i <= Convert.ToInt32(dRow["amount"]); i++)
                                        sb.Append(",('" + Convert.ToString(dRow["tid"]) + "','" + receiverID + "','" + roomID + "')");
                                    using (DatabaseClient dbClient1 = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient1.runQuery("INSERT INTO furniture(tid,ownerid,roomid) VALUES " + sb.ToString().Substring(1));
                                    }
                                    for (int i = 0; i <= Convert.ToInt32(dRow["amount"]); i++)
                                        catalogueManager.handlePurchase(Convert.ToInt32(dRow["tid"]), receiverID, roomID, 0, presentBoxID, 0);
                                }
                            }
                            else
                            {
                                int teleportid1 = 0;
                                using (DatabaseClient dbClient1 = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient1.runQuery("INSERT INTO furniture(tid,ownerid,roomid) VALUES ('" + templateID + "','" + receiverID + "','" + roomID + "' )");
                                    teleportid1 = catalogueManager.lastItemID;
                                }
                                if (catalogueManager.getTemplate(templateID).Sprite == "wallpaper" || catalogueManager.getTemplate(templateID).Sprite == "floor" || catalogueManager.getTemplate(templateID).Sprite.Contains("landscape"))
                                {
                                    int decorID;
                                    if (!int.TryParse(packetContent[4], out decorID))
                                    {
                                        // Decoration data was not numeric
                                        return;
                                    }
                                    else
                                    {
                                        catalogueManager.handlePurchase(templateID, receiverID, 0, decorID, presentBoxID, 0);
                                    }
                                }

                                else if ((stringManager.getStringPart(Item, 0, 11) == "greektrophy") || (stringManager.getStringPart(Item, 0, 11) == "prizetrophy"))
                                {
                                    if (handlePresentbox == false)
                                    {
                                        using (DatabaseClient dbClient4 = Eucalypt.dbManager.GetClient())
                                        {
                                            string vari = _Username + "\t" + DateTime.Today.ToShortDateString() + "\t" + stringManager.filterSwearwords(packetContent[4]);
                                            vari = vari.Replace(@"\", "\\").Replace("'", @"\'");
                                            dbClient4.runQuery("UPDATE furniture SET var = '" + vari + "' WHERE id = '" + catalogueManager.lastItemID + "' LIMIT 1");
                                        }
                                    }
                                    else
                                    {
                                        using (DatabaseClient dbClient4 = Eucalypt.dbManager.GetClient())
                                        {
                                            string vari = _Username + "\t" + DateTime.Today.ToShortDateString() + "\t" + stringManager.filterSwearwords(packetContent[4]);
                                            vari = vari.Replace(@"\", "\\").Replace("'", @"\'");
                                            dbClient4.runQuery("UPDATE furniture SET var = '" + vari + "' WHERE id = '" + catalogueManager.lastItemID + "' LIMIT 1");
                                            dbClient4.runQuery("INSERT INTO furniture_presents(id,itemid) VALUES ('" + presentBoxID + "','" + teleportid1 + "')");
                                            refreshHand("last");
                                            break;
                                        }
                                    }


                                }

                                else if (stringManager.getStringPart(Item, 0, 3) == "pet")
                                {
                                    string[] Props = VarItem.Split(Convert.ToChar(2)); // 0 = Name, 1 = Race, 2 = Colour
                                    string Name = Props[0];
                                    string Race = Props[1];
                                    string Colour = Props[2];
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("INSERT INTO pets (id,ownerid,nest_id,name,type,colour,race) VALUES ('" + catalogueManager.lastItemID + "', '" + userID + "', '" + catalogueManager.lastItemID + "','" + Name + "', '" + Item + "', '" + Colour + "', ' " + Race + "')");
                                    }
                                }
                                else
                                    catalogueManager.handlePurchase(templateID, receiverID, roomID, 0, presentBoxID, teleportid1);
                            }

                            if (receiverID == userID)
                                refreshHand("last");
                            else
                                if (userManager.containsUser(receiverID)) { userManager.getUser(receiverID).refreshHand("last"); }

                            break;
                        }
                    #endregion

                    case "Ai": // Buy game-tickets
                        {
                            string args = currentPacket.Substring(2);
                            int Amount = Encoding.decodeVL64(args.Substring(0, 3));
                            string Receiver = args.Substring(3);
                            int Ticketamount = 0;
                            int Price = 0;

                            if (Amount == 1) // Look how much tickets you want
                            {
                                Ticketamount = 2;
                                Price = 1;
                            }
                            else if (Amount == 2) // And again
                            {
                                Ticketamount = 20;
                                Price = 6;
                            }
                            else // Wrong parameter
                                return;

                            if (Price > _Credits) // Enough credits?
                            {
                                sendData("AD");
                                return;
                            }
                            int ReceiverID;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                dbClient.AddParamWithValue("name", Receiver);
                                ReceiverID = dbClient.getInt("SELECT id FROM users WHERE name = @name");
                            }
                            if (!(ReceiverID > 0)) // Does the user exist?
                            {
                                sendData("AL" + Receiver);
                                return;
                            }

                            _Credits -= Price; // New credit amount
                            sendData("@F" + _Credits); // Send the new credits
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                dbClient.runQuery("UPDATE users SET credits = '" + _Credits + "' WHERE id = '" + userID + "' LIMIT 1");
                                // Update receivers ticketamount
                                dbClient.runQuery("UPDATE users SET tickets = tickets+" + Ticketamount + " WHERE id = '" + ReceiverID + "' LIMIT 1");
                            }

                            if (userManager.containsUser(ReceiverID)) // Check or the user is online
                            {
                                virtualUser _Receiver = userManager.getUser(ReceiverID); // Get him/her
                                _Receiver._Tickets = _Receiver._Tickets + Ticketamount; // Update ticketamount

                                if (ReceiverID == userID) // Stop double kaching
                                    _Receiver.refreshValueables(false, true); // Kaching!
                                else
                                    _Receiver.refreshValueables(true, true); // Kaching!
                            }


                            sendData("AC"); // Yey! Buying Successful
                            //if (presentBoxID > 0)
                            //Out.WriteLine(_Username + " Buy a present.");
                            break;
                        }

                    case "Ca": // Recycler - proceed input items
                        {
                            if (Config.enableRecycler == false || Room == null || recyclerManager.sessionExists(userID))
                                return;

                            int itemCount = Encoding.decodeVL64(currentPacket.Substring(2));
                            if (recyclerManager.rewardExists(itemCount))
                            {
                                recyclerManager.createSession(userID, itemCount);
                                currentPacket = currentPacket.Substring(Encoding.encodeVL64(itemCount).ToString().Length + 2);
                                for (int i = 0; i < itemCount; i++)
                                {
                                    int itemID = Encoding.decodeVL64(currentPacket);
                                    //Database dbClient = new Database(true, false, 109);
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        if (dbClient.findsResult("SELECT id FROM furniture WHERE ownerid = '" + userID + "' AND roomid = '0'"))
                                        {
                                            dbClient.runQuery("UPDATE furniture SET roomid = '-2' WHERE id = '" + itemID + "' LIMIT 1");
                                            currentPacket = currentPacket.Substring(Encoding.encodeVL64(itemID).Length);
                                        }

                                        else
                                        {
                                            recyclerManager.dropSession(userID, true);
                                            sendData("DpH");

                                            return;
                                        }
                                    }
                                }

                                sendData("Dp" + recyclerManager.sessionString(userID));
                                refreshHand("update");
                            }

                            break;
                        }

                    case "Cb": // Recycler - redeem/cancel session
                        {
                            if (Config.enableRecycler == false || Room != null && recyclerManager.sessionExists(userID))
                            {
                                bool Redeem = (currentPacket.Substring(2) == "I");
                                if (Redeem && recyclerManager.sessionReady(userID))
                                    recyclerManager.rewardSession(userID);
                                recyclerManager.dropSession(userID, Redeem);

                                sendData("Dp" + recyclerManager.sessionString(userID));
                                if (Redeem)
                                    refreshHand("last");
                                else
                                    refreshHand("new");
                            }
                            break;
                        }
                    #endregion

                    #region Hand and item handling
                    case "AA": // Hand
                        {
                            if (Room == null || roomUser == null)
                                return;

                            string Mode = currentPacket.Substring(2);
                            refreshHand(Mode);
                            break;
                        }

                    case "LB": // Hand
                        {
                            if (Room == null || roomUser == null)
                                return;

                            string Mode = currentPacket.Substring(2);
                            refreshHand(Mode);
                            break;
                        }

                    case "AB": // Item handling - apply wallpaper/floor/landscape to room
                        {
                            if (_hasRights == false || _inPublicroom || Room == null || roomUser == null)
                                return;

                            int itemID = int.Parse(currentPacket.Split('/')[1]);
                            string decorType = currentPacket.Substring(2).Split('/')[0];
                            if (decorType != "wallpaper" && decorType != "floor" && decorType != "landscape") // Non-valid decoration type
                                return;

                            //Database dbClient = new Database(true, false, 110);
                            int templateID;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                templateID = dbClient.getInt("SELECT tid FROM furniture WHERE id = '" + itemID + "' AND ownerid = '" + userID + "' AND roomid = '0'");

                                if (catalogueManager.getTemplate(templateID).Sprite != decorType) // This item isn't the decoration item the client thinks it is. (obv scripter) If the item wasn't found (so the user didn't owned the item etc), then an empty item template isn't required, which also doesn't match this condition
                                {
                                    return;
                                }
                                string decorVal = dbClient.getString("SELECT var FROM furniture WHERE id = '" + itemID + "'");
                                Room.sendData("@n" + decorType + "/" + decorVal); // "@n" (46) is a generic message for setting a room's decoration. Since the introduction of landscapes, it can be 'wallpaper', 'floor' and 'landscape'

                                dbClient.runQuery("UPDATE rooms SET " + decorType + " = '" + decorVal + "' WHERE id = '" + _roomID + "' LIMIT 1"); // Generates query like 'UPDATE rooms SET floor/wallpaper/landscape blabla' (the string decorType is containing either 'floor', 'wallpaper' or 'landscape')
                                dbClient.runQuery("DELETE FROM furniture WHERE id = '" + itemID + "' LIMIT 1");

                            }
                        }
                        break;


                    case "AZ": // Item handling - place item down
                        {
                            if (_hasRights == false || _inPublicroom || Room == null || roomUser == null)
                                return;
                            int itemID;
                            if (!int.TryParse((currentPacket.Split(' ')[0].Substring(2)), out itemID))
                                return;

                            //int itemID = int.Parse(currentPacket.Split(' ')[0].Substring(2));
                            //Database dbClient = new Database(true, false, 111);
                            int templateID;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                templateID = dbClient.getInt("SELECT tid FROM furniture WHERE id = '" + itemID + "' AND ownerid = '" + userID + "' AND roomid = '0'");
                            }
                            if (templateID == 0)
                            {
                                return;
                            }
                            if (catalogueManager.getTemplate(templateID).typeID == 0)
                            {
                                string _INPUTPOS = currentPacket.Substring(itemID.ToString().Length + 3);
                                string _CHECKEDPOS = catalogueManager.wallPositionOK(_INPUTPOS);
                                if (_CHECKEDPOS != _INPUTPOS)
                                {
                                    return;
                                }
                                string Var;
                                int VarValue = 0;

                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    Var = dbClient.getString("SELECT var FROM furniture WHERE id = '" + itemID + "'");
                                    Int32.TryParse(Var, out VarValue);
                                    if (stringManager.getStringPart(catalogueManager.getTemplate(templateID).Sprite, 0, 7) == "post.it")
                                    {
                                        if (VarValue > 1)
                                            dbClient.runQuery("UPDATE furniture SET var = var - 1 WHERE id = '" + itemID + "' LIMIT 1");
                                        else
                                            dbClient.runQuery("DELETE FROM furniture WHERE id = '" + itemID + "' LIMIT 1");
                                        dbClient.runQuery("INSERT INTO furniture(tid,ownerid) VALUES ('" + templateID + "','" + userID + "')");
                                        itemID = catalogueManager.lastItemID;
                                        dbClient.runQuery("INSERT INTO furniture_stickies(id) VALUES ('" + itemID + "')");
                                        Var = "FFFF33";
                                        dbClient.runQuery("UPDATE furniture SET var = '" + Var + "' WHERE id = '" + itemID + "' LIMIT 1");
                                    }
                                    else if (stringManager.getStringPart(catalogueManager.getTemplate(templateID).Sprite, 0, 10) == "roomdimmer")
                                        dbClient.runQuery("UPDATE furniture_moodlight SET roomid = '" + _roomID + "' WHERE id = '" + itemID + "' LIMIT 1");
                                    Room.wallItemManager.addItem(itemID, templateID, _CHECKEDPOS, Var, true);
                                }
                            }
                            else
                            {
                                string[] locDetails = currentPacket.Split(' ');
                                int X = int.Parse(locDetails[1]);
                                int Y = int.Parse(locDetails[2]);
                                byte Z = byte.Parse(locDetails[3]);
                                byte typeID = catalogueManager.getTemplate(templateID).typeID;
                                Room.floorItemManager.placeItem(itemID, templateID, X, Y, typeID, Z);
                            }
                            break;
                        }

                    case "AC": // Item handling - pickup item
                        {
                            if (_isOwner == false || _inPublicroom || Room == null || roomUser == null)
                                return;

                            int ItemPickingUp;
                            int itemID2 = int.Parse(currentPacket.Split(' ')[2]);
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                ItemPickingUp = dbClient.getInt("SELECT tid FROM furniture WHERE id = '" + itemID2 + "'");
                            }
                            if (ItemPickingUp == 538 && _Rank < 7)
                                return;

                            int itemID = int.Parse(currentPacket.Split(' ')[2]);
                            if (Room.floorItemManager.containsItem(itemID))
                                Room.floorItemManager.removeItem(itemID, userID);
                            else if (Room.wallItemManager.containsItem(itemID) && stringManager.getStringPart(Room.wallItemManager.getItem(itemID).Sprite, 0, 7) != "post.it") // Can't pickup stickies from room
                                Room.wallItemManager.removeItem(itemID, userID);
                            else
                                return;

                            refreshHand("update");
                            break;
                        }

                    case "AI": // Item handling - move/rotate item
                        {
                            if (_hasRights == false || _inPublicroom || Room == null || roomUser == null)
                                return;

                            int itemID = int.Parse(currentPacket.Split(' ')[0].Substring(2));
                            if (Room.floorItemManager.containsItem(itemID))
                            {
                                string[] locDetails = currentPacket.Split(' ');
                                int X = int.Parse(locDetails[1]);
                                int Y = int.Parse(locDetails[2]);
                                byte Z = byte.Parse(locDetails[3]);

                                Room.floorItemManager.relocateItem(itemID, X, Y, Z);
                            }
                            break;
                        }

                    case "CV": // Item handling - toggle wallitem status
                        {
                            try
                            {
                                if (_Rank > 6)
                                {
                                    if (_inPublicroom || Room == null || roomUser == null)
                                        return;

                                    int itemID = int.Parse(currentPacket.Substring(4, Encoding.decodeB64(currentPacket.Substring(2, 2))));
                                    int toStatus = Encoding.decodeVL64(currentPacket.Substring(itemID.ToString().Length + 4));
                                    //Out.WritePlain(toStatus.ToString());
                                    Room.wallItemManager.toggleItemStatus(itemID, toStatus);
                                }
                                break;
                            }
                            catch { }
                            break;
                        }

                    case "AJ": // Item handling - toggle flooritem status // DINER PLATES
                        {
                            try
                            {
                                if (_Rank > 0)
                                {
                                    int itemID = int.Parse(currentPacket.Substring(4, Encoding.decodeB64(currentPacket.Substring(2, 2))));
                                    string toStatus = currentPacket.Substring(itemID.ToString().Length + 6);
                                    int tester;
                                Rooms.Items.floorItem Item = Room.floorItemManager.getItem(itemID);
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        if (Item.Sprite == "vender")
                                        {
                                            _gettingDrink = true;
                                            statusManager.carryItem("19");
                                            Room.sendSaying(roomUser, "*makes a purchase from the vending machine*");
                                            dbClient.runQuery("UPDATE users SET credits = credits - '1' WHERE name = '" + _Username + "'");
                                            refreshValueables(true, false);
                                        }
                                        if (Item.Sprite == "md_limukaappi")
                                        {
                                            _gettingDrink = true;
                                            statusManager.carryItem("19");
                                            Room.sendSaying(roomUser, "*makes a purchase from the vending machine*");
                                            dbClient.runQuery("UPDATE users SET credits = credits - '1' WHERE name = '" + _Username + "'");
                                            refreshValueables(true, false);
                                        }
                                            else
                                            {
                                            if (toStatus.ToLower() == "false" || toStatus.ToLower() == "on" || toStatus.ToLower() == "off" || toStatus.ToLower() == "true" || int.TryParse(toStatus, out tester) || toStatus.ToLower() == "c" || toStatus.ToLower() == "o")
                                                Room.floorItemManager.toggleItemStatus(itemID, toStatus, _hasRights);
                                            else
                                                Disconnect();
                                        }
                                    }
                                }
                            }
                            catch { }
                            break;
                        }

                    case "AN": // Item handling - open presentbox
                        {
                            if (_isOwner == false || _inPublicroom || Room == null || roomUser == null)
                                return;

                            int itemID = int.Parse(currentPacket.Substring(2));
                            if (Room.floorItemManager.containsItem(itemID) == false)
                                return;

                            //Database dbClient = new Database(true, false, 112);
                            DataColumn dCol;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                dCol = dbClient.getColumn("SELECT itemid FROM furniture_presents WHERE id = '" + itemID + "'");

                                if (dCol.Table.Rows.Count > 0)
                                {
                                    StringBuilder sb = new StringBuilder();
                                    int lastItemID = 0;
                                    foreach (DataRow dRow in dCol.Table.Rows)
                                    {
                                        sb.Append(" OR id = '" + Convert.ToString(dRow["itemid"]) + "'");
                                        lastItemID = Convert.ToInt32(dRow["itemid"]);
                                    }
                                    dbClient.runQuery("UPDATE furniture SET roomid = '-' WHERE " + sb.ToString().Substring(4));
                                    Room.floorItemManager.removeItem(itemID, 0);

                                    int lastItemTID = dbClient.getInt("SELECT tid FROM furniture WHERE id = '" + lastItemID + "'");
                                    catalogueManager.itemTemplate Template = catalogueManager.getTemplate(lastItemTID);

                                    if (Template.typeID > 0)
                                        sendData("BA" + Template.Sprite + "\r" + Template.Sprite + "\r" + Template.Length + Convert.ToChar(30) + Template.Width + Convert.ToChar(30) + Template.Colour);
                                    else
                                        sendData("BA" + Template.Sprite + "\r" + Template.Sprite + " " + Template.Colour + "\r");
                                }
                                dbClient.runQuery("DELETE FROM furniture_presents WHERE id = '" + itemID + "' LIMIT " + dCol.Table.Rows.Count);
                                dbClient.runQuery("DELETE FROM furniture WHERE id = '" + itemID + "' LIMIT 1");
                            }
                            refreshHand("last");
                            break;

                        }

                    case "Bw": // Item handling - redeem credit item
                        {
                            if (_isOwner == false || _inPublicroom || Room == null || roomUser == null)
                                return;

                            int itemID = Encoding.decodeVL64(currentPacket.Substring(2));
                            if (Room.floorItemManager.containsItem(itemID))
                            {
                                string Sprite = Room.floorItemManager.getItem(itemID).Sprite;
                                if (Sprite.Substring(0, 3).ToLower() != "cf_" && Sprite.Substring(0, 4).ToLower() != "cfc_")
                                    return;
                                int redeemValue = 0;
                                try { redeemValue = int.Parse(Sprite.Split('_')[1]); }
                                catch { return; }

                                Room.floorItemManager.removeItem(itemID, 0);

                                _Credits += redeemValue;
                                sendData("@F" + _Credits);
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.runQuery("UPDATE users SET credits = '" + _Credits + "' WHERE id = '" + userID + "' LIMIT 1");
                                }
                            }
                            break;
                        }

                    case "AQ": // Item handling - teleporters - enter teleporter
                        {
                            if (_inPublicroom || Room == null || roomUser == null)
                                return;

                            int itemID = int.Parse(currentPacket.Substring(2));
                            if (Room.floorItemManager.containsItem(itemID))
                            {
                                Rooms.Items.floorItem Teleporter = Room.floorItemManager.getItem(itemID);
                                // Prevent clientside 'jumps' to teleporter, check if user is removed one coord from teleporter entrance
                                if (Teleporter.Z == 2 && roomUser.X != Teleporter.X + 1 && roomUser.Y != Teleporter.Y)
                                    return;
                                else if (Teleporter.Z == 4 && roomUser.X != Teleporter.X && roomUser.Y != Teleporter.Y + 1)
                                    return;
                                roomUser.goalX = -1;
                                Room.moveUser(this.roomUser, Teleporter.X, Teleporter.Y, true);
                            }
                            break;
                        }

                    case @"@\": // Item handling - teleporters - flash teleporter
                        {
                            if (_inPublicroom || Room == null || roomUser == null)
                                return;

                            int itemID = int.Parse(currentPacket.Substring(2));
                            if (Room.floorItemManager.containsItem(itemID))
                            {
                                Rooms.Items.floorItem Teleporter1 = Room.floorItemManager.getItem(itemID);
                                if (roomUser.X != Teleporter1.X && roomUser.Y != Teleporter1.Y)
                                    return;
                                //Database dbClient = new Database (true, false, 114);
                                int roomIDTeleporter2;
                                int idTeleporter2;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    idTeleporter2 = dbClient.getInt("SELECT teleportid FROM furniture WHERE id = '" + itemID + "'");
                                    roomIDTeleporter2 = dbClient.getInt("SELECT roomid FROM furniture WHERE id = '" + idTeleporter2 + "'");
                                }
                                if (roomIDTeleporter2 > 0)
                                    new TeleporterUsageSleep(useTeleporter).BeginInvoke(Teleporter1, idTeleporter2, roomIDTeleporter2, null, null);
                            }
                            break;
                        }

                    case "AM": // Item handling - dices - close dice
                        {
                            if (Room == null || roomUser == null || _inPublicroom)
                                return;

                            int itemID = int.Parse(currentPacket.Substring(2));
                            if (Room.floorItemManager.containsItem(itemID))
                            {
                                Rooms.Items.floorItem Item = Room.floorItemManager.getItem(itemID);
                                string Sprite = Item.Sprite;
                                if (Sprite != "edice" && Sprite != "edicehc") // Not a dice item
                                    return;

                                if (!(Math.Abs(roomUser.X - Item.X) > 1 || Math.Abs(roomUser.Y - Item.Y) > 1)) // User is not more than one square removed from dice
                                {
                                    Item.Var = "0";
                                    Room.sendData("AZ" + itemID + " " + (itemID * 38));
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE furniture SET var = '0' WHERE id = '" + itemID + "' LIMIT 1");
                                    }
                                }
                            }
                            break;
                        }

                    case "AL": // Item handling - dices - spin dice
                        {
                            if (Room == null || roomUser == null || _inPublicroom)
                                return;

                            int itemID = int.Parse(currentPacket.Substring(2));
                            if (Room.floorItemManager.containsItem(itemID))
                            {
                                Rooms.Items.floorItem Item = Room.floorItemManager.getItem(itemID);
                                string Sprite = Item.Sprite;
                                if (Sprite != "edice" && Sprite != "edicehc") // Not a dice item
                                    return;

                                if (!(Math.Abs(roomUser.X - Item.X) > 1 || Math.Abs(roomUser.Y - Item.Y) > 1)) // User is not more than one square removed from dice
                                {
                                    Room.sendData("AZ" + itemID);

                                    int rndNum = new Random(DateTime.Now.Millisecond).Next(1, 7);
                                    Room.sendData("AZ" + itemID + " " + ((itemID * 38) + rndNum), 2000);
                                    Item.Var = rndNum.ToString();
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE furniture SET var = '" + rndNum + "' WHERE id = '" + itemID + "' LIMIT 1");
                                    }
                                }
                            }
                            break;
                        }

                    case "Cw": // Item handling - spin Wheel of fortune
                        {
                            if (Room == null || roomUser == null || _inPublicroom)
                                return;

                            int itemID = Encoding.decodeVL64(currentPacket.Substring(2));
                            if (Room.wallItemManager.containsItem(itemID))
                            {
                                Rooms.Items.wallItem Item = Room.wallItemManager.getItem(itemID);
                                if (Item.Sprite == "habbowheel")
                                {
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        if (Room._Wheel == true)
                                        {
                                            sendData("BKSomeone else has already spun the wheel!");
                                        }
                                        else
                                        {
                                            if (JBCD)
                                            {
                                                sendData("BKYou need to wait 5 seconds before using the wheel again");
                                            }
                                            else
                                            {
                                                int rndNum = new Random(DateTime.Now.Millisecond).Next(0, 10);
                                                Room.sendData("AU" + itemID + Convert.ToChar(9) + "habbowheel" + Convert.ToChar(9) + " " + Item.wallPosition + Convert.ToChar(9) + "-1");
                                                Room.sendData("AU" + itemID + Convert.ToChar(9) + "habbowheel" + Convert.ToChar(9) + " " + Item.wallPosition + Convert.ToChar(9) + rndNum, 4250);
                                                dbClient.runQuery("UPDATE furniture SET var = '" + rndNum + "' WHERE id = '" + itemID + "' LIMIT 1");
                                                Room.sendShout(roomUser, "*Puts 5 Credits Into The Wheel Of Fortune And Spins It*");
                                                dbClient.runQuery("UPDATE users SET credits = credits - '5' WHERE name = '" + _Username + "'");
                                                Room._Wheel = true;
                                                refreshValueables(true, false);
                                                ThreadStart spinWheelStarter = new ThreadStart(spinWheel);
                                                spinWheelLooper = new Thread(spinWheelStarter);
                                                spinWheelLooper.Priority = ThreadPriority.Lowest;
                                                spinWheelLooper.Start();
                                                JBCD = true;
                                            }
                                        }
                                    }
                                }
                            }
                            break;
                        }

                    case "Dz": // Item handling - activate Love shuffler sofa
                        {
                            if (Room == null || roomUser == null || _inPublicroom)
                                return;

                            int itemID = Encoding.decodeVL64(currentPacket.Substring(2));
                            if (Room.floorItemManager.containsItem(itemID) && Room.floorItemManager.getItem(itemID).Sprite == "val_randomizer")
                            {
                                int rndNum = new Random(DateTime.Now.Millisecond).Next(1, 5);
                                Room.sendData("AX" + itemID + Convert.ToChar(2) + "123456789" + Convert.ToChar(2));
                                Room.sendData("AX" + itemID + Convert.ToChar(2) + rndNum + Convert.ToChar(2), 5000);
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.runQuery("UPDATE furniture SET var = '" + rndNum + "' WHERE id = '" + itemID + "' LIMIT 1");
                                }
                            }
                            break;
                        }

                    case "AS": // Item handling - stickies/photo's - open stickie/photo
                        {
                            if (Room == null || roomUser == null || _inPublicroom)
                                return;

                            int itemID = int.Parse(currentPacket.Substring(2));
                            if (Room.wallItemManager.containsItem(itemID))
                            {
                                //Database dbClient = new Database(true, true, 119);
                                string Message;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    Message = dbClient.getString("SELECT text FROM furniture_stickies WHERE id = '" + itemID + "'");
                                }
                                sendData("@p" + itemID + Convert.ToChar(9) + Room.wallItemManager.getItem(itemID).Var + " " + Message);
                            }
                            break;
                        }

                    case "AT": // Item handling - stickies - edit stickie colour/message
                        {
                            if (_hasRights == false || Room == null || roomUser == null || _inPublicroom)
                                return;

                            int itemID = int.Parse(currentPacket.Substring(2, currentPacket.IndexOf("/") - 2));
                            if (Room.wallItemManager.containsItem(itemID))
                            {
                                Rooms.Items.wallItem Item = Room.wallItemManager.getItem(itemID);
                                string Sprite = Item.Sprite;
                                if (Sprite != "post.it" && Sprite != "post.it.vd")
                                    return;
                                string Colour = "FFFFFF"; // Valentine stickie default colour
                                if (Sprite == "post.it") // Normal stickie
                                {
                                    Colour = currentPacket.Substring(2 + itemID.ToString().Length + 1, 6);
                                    if (Colour != "FFFF33" && Colour != "FF9CFF" && Colour != "9CFF9C" && Colour != "9CCEFF")
                                        return;
                                }

                                string Message = currentPacket.Substring(2 + itemID.ToString().Length + 7);
                                if (Message.Length > 684)
                                    return;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.AddParamWithValue("text", stringManager.filterSwearwords(Message).Replace("/r", Convert.ToChar(13).ToString()));
                                    //dbClient.Open();
                                    if (Colour != Item.Var)
                                        dbClient.runQuery("UPDATE furniture SET var = '" + Colour + "' WHERE id = '" + itemID + "' LIMIT 1");
                                    Item.Var = Colour;
                                    dbClient.runQuery("UPDATE furniture_stickies SET text = @text WHERE id = '" + itemID + "' LIMIT 1");
                                }
                                Room.sendData("AU" + itemID + Convert.ToChar(9) + Sprite + Convert.ToChar(9) + " " + Item.wallPosition + Convert.ToChar(9) + Colour);
                            }
                            break;
                        }

                    case "AU": // Item handling - stickies/photo - delete stickie/photo
                        {
                            if (_isOwner == false || Room == null || roomUser == null || _inPublicroom)
                                return;

                            int itemID = int.Parse(currentPacket.Substring(2));
                            if (Room.wallItemManager.containsItem(itemID) && stringManager.getStringPart(Room.wallItemManager.getItem(itemID).Sprite, 0, 7) == "post.it")
                            {
                                Room.wallItemManager.removeItem(itemID, 0);
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.runQuery("DELETE FROM furniture_stickies WHERE id = '" + itemID + "' LIMIT 1");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region Moodlight
                    case "EW": // Turn moodlight on/off
                        {
                            if (_isOwner == false && _hasRights == false)
                                return;
                            roomManager.moodlight.setSettings(_roomID, false, 0, 0, null, 0);
                            break;
                        }

                    case "EU": // Load moodlight settings
                        {
                            if (_isOwner == false && _hasRights == false)
                                return;
                            string settingData = roomManager.moodlight.getSettings(_roomID);
                            if (settingData != null)
                                sendData("Em" + settingData);
                            break;
                        }

                    case "EV": // Apply modified moodlight settings
                        {
                            if (_isOwner == false && _hasRights == false)
                                return;
                            int presetID = Encoding.decodeVL64(currentPacket.Substring(2, 1));
                            int bgState = Encoding.decodeVL64(currentPacket.Substring(3, 1));
                            string presetColour = currentPacket.Substring(6, Encoding.decodeB64(currentPacket.Substring(4, 2)));
                            int presetDarkF = Encoding.decodeVL64(currentPacket.Substring(presetColour.Length + 6));
                            roomManager.moodlight.setSettings(_roomID, true, presetID, bgState, presetColour, presetDarkF);
                            break;
                        }
                    #endregion

                    #region one way gate
                    case "Ch": // Ch + ONE-WAY-DOOR-FURNI-ID (answer: Dx + WIRE(FURNI-ID) + I + WIRE(ROOM-UID-OF-CLICKER)) (sprite: 'one_way_door*#)
                        {
                            if (Room == null || roomUser == null || _inPublicroom)
                                return;

                            int itemID = Encoding.decodeVL64(currentPacket.Substring(2));
                            Rooms.Items.floorItem Item = Room.floorItemManager.getItem(itemID); // Find the item.
                            try
                            {
                                if (Room.floorItemManager.containsItem(itemID) && stringManager.getStringPart(Item.Sprite, 0, 12) == "one_way_door")
                                {
                                    if (Item.Sprite == "one_way_door*2")
                                    {
                                        if (roomUser.walkLock == true)
                                        {
                                            sendData("BK" + "Nice try, but you are either stunned or cuffed.");
                                            return;
                                        }
                                        else
                                        {
                                            virtualJob Job = jobManager.getJob(_SecureID);
                                            int corphq;
                                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                            {
                                                corphq = dbClient.getInt("SELECT room_id FROM jobs_rooms WHERE rank_id = '" + _SecureID + "' && room_id = '" + _roomID + "'");
                                            }
                                            if (corphq != _roomID && !Job.EnterAnyGate)
                                            {
                                                sendData("BK" + "Sorry, but you must work here to enter this area.");
                                                return;
                                            }

                                            if (!_Working)
                                            {
                                                sendData("BK" + "You must have started work to enter.");
                                                return;
                                            }
                                        }
                                    }

                                    if (Item.Sprite == "one_way_door*1")
                                    {
                                        if (roomUser.walkLock == true)
                                        {
                                            sendData("BK" + "Nice try, but you are either stunned or cuffed.");
                                            return;
                                        }
                                        else
                                        {
                                            if (_Credits < 5)
                                            {
                                                sendData("BK" + "You must have atleast five credits on you to go through this gate.");
                                                return;
                                            }
                                            string roomOwner;
                                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                            {
                                                roomOwner = dbClient.getString("SELECT owner FROM rooms WHERE id = '" + _roomID + "'");
                                                dbClient.runQuery("UPDATE users SET credits = credits + '5' WHERE name = '" + roomOwner + "'");
                                                dbClient.runQuery("UPDATE users SET credits = credits - '5' WHERE name = '" + _Username + "'");
                                            }
                                            refreshValueables(true, false);
                                            sendData("BK" + "You have been charged 5 credits for your toll.");
                                        }
                                    }

                                    if (Item.Sprite == "one_way_door*0")
                                    {
                                        if (roomUser.walkLock == true)
                                        {
                                            sendData("BK" + "Nice try, but you are either stunned or cuffed.");
                                            return;
                                        }
                                        else
                                        {
                                            if (_Brawl || _Arena)
                                            {
                                                sendData("BK" + "You cannot enter theses gates when in the bubble juice arena or brawl, finish your match, or do :leavegame");
                                                return;
                                            }
                                            else
                                            {
      
                                                if (_Jailed == 1)
                                                {
                                                    return;
                                                }
                                            }
                                        }
                                    }

                                    if (Item.Sprite == "one_way_door*4")
                                    {
                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {
                                            if (_Credits < 10)
                                            {
                                                sendData("BK" + "You must have atleast 10 credits on you to enter the BBJ arena.");
                                                return;
                                            }
                                            string roomOwner;
                                                roomOwner = dbClient.getString("SELECT owner FROM rooms WHERE id = '" + _roomID + "'");
                                                dbClient.runQuery("UPDATE users SET credits = credits + '10' WHERE name = '" + roomOwner + "'");
                                                dbClient.runQuery("UPDATE users SET credits = credits - '10' WHERE name = '" + _Username + "'");
                                            refreshValueables(true, false);
                                            string FigurePartHair = _Figure;
                                            string GetHairPart;

                                            GetHairPart = Regex.Split(_Figure, "hr")[1];
                                            FigurePartHair = GetHairPart.Split('.')[0];
                                            string FigurePartBody = _Figure;
                                            string GetBodyPart;

                                            GetBodyPart = Regex.Split(_Figure, "hd")[1];
                                            FigurePartBody = GetBodyPart.Split('.')[0];

                                            _Uni = Convert.ToString("hr" + FigurePartHair + "." + "hd" + FigurePartBody + ".");

                                            _Mission = "[Arena] Red Team";
                                            _Health = _MaxHP;
                                            _Figure = _Uni + "ch-210-96.lg-281-96.sh-290-96";
                                            refreshAppearance(false, true, true);
                                            _Arena = true;

                                        }
                                    }

                                    if (Item.Sprite == "one_way_door*5")
                                    {
                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {
                                            if (_Credits < 10)
                                            {
                                                sendData("BK" + "You must have atleast 10 credits on you to enter the BBJ arena.");
                                                return;
                                            }
                                            string roomOwner;
                                            roomOwner = dbClient.getString("SELECT owner FROM rooms WHERE id = '" + _roomID + "'");
                                            dbClient.runQuery("UPDATE users SET credits = credits + '10' WHERE name = '" + roomOwner + "'");
                                            dbClient.runQuery("UPDATE users SET credits = credits - '10' WHERE name = '" + _Username + "'");
                                            refreshValueables(true, false);
                                            string FigurePartHair = _Figure;
                                            string GetHairPart;

                                            GetHairPart = Regex.Split(_Figure, "hr")[1];
                                            FigurePartHair = GetHairPart.Split('.')[0];
                                            string FigurePartBody = _Figure;
                                            string GetBodyPart;

                                            GetBodyPart = Regex.Split(_Figure, "hd")[1];
                                            FigurePartBody = GetBodyPart.Split('.')[0];

                                            _Uni = Convert.ToString("hr" + FigurePartHair + "." + "hd" + FigurePartBody + ".");

                                            _Mission = "[Arena] Blue Team";
                                            _Figure = _Uni + "ch-210-108.lg-281-108.sh-290-108";
                                            _Health = _MaxHP;
                                            refreshAppearance(false, true, true);
                                            _Arena = true;
                                        }
                                    }

                                    if (Item.Sprite == "one_way_door*6")
                                    {
                                        _Brawl = true;
                                        _Mission = "[Game] Brawl";
                                        refreshAppearance(false, true, true);
                                    }

                                    if (Item.Sprite == "one_way_door*7")
                                    {
                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {

                                            string FigurePartHair = _Figure;
                                            string GetHairPart;

                                            GetHairPart = Regex.Split(_Figure, "hr")[1];
                                            FigurePartHair = GetHairPart.Split('.')[0];
                                            string FigurePartBody = _Figure;
                                            string GetBodyPart;

                                            GetBodyPart = Regex.Split(_Figure, "hd")[1];
                                            FigurePartBody = GetBodyPart.Split('.')[0];

                                            _Uni = Convert.ToString("hr" + FigurePartHair + "." + "hd" + FigurePartBody + ".");

                                            _Mission = "[CW] Red Team";
                                            _Figure = _Uni + "ch-210-96.lg-281-96.sh-290-96";
                                            refreshAppearance(false, true, true);

                                            dbClient.runQuery("UPDATE users SET in_game = '1' WHERE name = '" + _Username + "'");
                                            dbClient.runQuery("UPDATE users SET red_team = '1' WHERE name = '" + _Username + "'");
                                        }
                                    }

                                    if (Item.Sprite == "one_way_door*8")
                                    {

                                        if (_Working)
                                        {
                                            _Working = false;
                                        }

                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {
                                            string FigurePartHair = _Figure;
                                            string GetHairPart;

                                            GetHairPart = Regex.Split(_Figure, "hr")[1];
                                            FigurePartHair = GetHairPart.Split('.')[0];
                                            string FigurePartBody = _Figure;
                                            string GetBodyPart;

                                            GetBodyPart = Regex.Split(_Figure, "hd")[1];
                                            FigurePartBody = GetBodyPart.Split('.')[0];

                                            _Uni = Convert.ToString("hr" + FigurePartHair + "." + "hd" + FigurePartBody + ".");

                                            _Mission = "[Game] Brawl";


                                            refreshAppearance(false, true, true);


                                        }
                                    }

                                    if (Item.Sprite == "one_way_door*9")
                                    {

                                        if (_Working)
                                        {
                                            _Working = false;
                                        }

                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {
                                            string FigurePartHair = _Figure;
                                            string GetHairPart;

                                            GetHairPart = Regex.Split(_Figure, "hr")[1];
                                            FigurePartHair = GetHairPart.Split('.')[0];
                                            string FigurePartBody = _Figure;
                                            string GetBodyPart;

                                            GetBodyPart = Regex.Split(_Figure, "hd")[1];
                                            FigurePartBody = GetBodyPart.Split('.')[0];

                                            _Uni = Convert.ToString("hr" + FigurePartHair + "." + "hd" + FigurePartBody + ".");

                                            _Mission = "Yellow Team - Team Death Match!";
                                            _Figure = _Uni + "ch-210-93.lg-281-93.sh-290-93";
                                            refreshAppearance(false, true, true);

                                            dbClient.runQuery("UPDATE users SET yellow_team = '1' WHERE name = '" + _Username + "'");
                                        }
                                    }

                                    if (Item.Sprite == "one_way_door*10")
                                    {
                                        if (_Credits < 10)
                                        {
                                            sendData("BK" + "You must have atleast ten credits on you to go through this gate.");
                                            return;
                                        }
                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {

                                            string FigurePartHair = _Figure;
                                            string GetHairPart;

                                            GetHairPart = Regex.Split(_Figure, "hr")[1];
                                            FigurePartHair = GetHairPart.Split('.')[0];
                                            string FigurePartBody = _Figure;
                                            string GetBodyPart;

                                            GetBodyPart = Regex.Split(_Figure, "hd")[1];
                                            FigurePartBody = GetBodyPart.Split('.')[0];

                                            _Uni = Convert.ToString("hr" + FigurePartHair + "." + "hd" + FigurePartBody + ".");

                                            _Mission = "Blue Fighter - Arena Fight!";
                                            _Figure = _Uni + "ch-210-108.lg-281-108.sh-290-108";
                                            refreshAppearance(false, true, true);
                                            refreshValueables(true, false);
                                            dbClient.runQuery("UPDATE users SET credits = credits - '10' WHERE name = '" + _Username + "'");

                                            Room.sendWhisper(roomUser, _Username, "You have been charged 10 credits for your toll.");
                                            _Health = 100;
                                        }
                                    }

                                    if (Item.Sprite == "one_way_door*11")
                                    {
                                        if (_Credits < 10)
                                        {
                                            sendData("BK" + "You must have atleast ten credits on you to go through this gate.");
                                            return;
                                        }
                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {

                                            string FigurePartHair = _Figure;
                                            string GetHairPart;

                                            GetHairPart = Regex.Split(_Figure, "hr")[1];
                                            FigurePartHair = GetHairPart.Split('.')[0];
                                            string FigurePartBody = _Figure;
                                            string GetBodyPart;

                                            GetBodyPart = Regex.Split(_Figure, "hd")[1];
                                            FigurePartBody = GetBodyPart.Split('.')[0];

                                            _Uni = Convert.ToString("hr" + FigurePartHair + "." + "hd" + FigurePartBody + ".");

                                            _Mission = "Red Fighter - Arena Fight!";
                                            _Figure = _Uni + "ch-210-96.lg-281-96.sh-290-96";
                                            refreshAppearance(false, true, true);
                                            refreshValueables(true, false);
                                            Room.sendWhisper(roomUser, _Username, "You have been charged 10 credits for your toll.");
                                            dbClient.runQuery("UPDATE users SET credits = credits - '10' WHERE name = '" + _Username + "'");
                                            _Health = 100;

                                        }
                                    }

                                    if (Item.Sprite == "one_way_door*3")
                                    {
                                        if (_Rank < 2)
                                        {
                                            sendData("BK" + "Sorry, but you must purchase VIP to use this gate.");
                                            return;
                                        }
                                    }

                                    Room.sendData("Dx" + Encoding.encodeVL64(itemID) + "I" + Encoding.encodeVL64(Room.roomID)); // Opens the gate, with green light
                                    Room.moveUser(this.roomUser, Item.X, Item.Y, true); // Walk into the item

                                    Thread.Sleep(350);

                                    if (Item.Z == 0)
                                        Room.moveUser(this.roomUser, roomUser.X, roomUser.Y + 1, true);
                                    else if (Item.Z == 2)
                                        Room.moveUser(this.roomUser, roomUser.X - 1, roomUser.Y, true);
                                    else if (Item.Z == 4)
                                        Room.moveUser(this.roomUser, roomUser.X, roomUser.Y - 1, true);
                                    else if (Item.Z == 6)
                                        Room.moveUser(this.roomUser, roomUser.X + 1, roomUser.Y, true);

                                    Thread.Sleep(350);
                                    Room.sendData("Dx" + Encoding.encodeVL64(itemID) + "H" + Encoding.encodeVL64(Room.roomID));

                                }
                            }
                            catch { }
                            break;
                        }
                    #endregion

                    #region Pets

                    case "@j": // Pet buying - Check if the name is valid
                        {
                            if (currentPacket.Contains("¬") || currentPacket.Contains(Convert.ToChar(1).ToString()) || currentPacket.Contains(Convert.ToChar(2).ToString()))
                                sendData("@dI"); // Name is invalid
                            else
                            {
                                sendData("@dH"); // Name is valid
                            }
                            break;
                        }

                    case "B@": // Pet - Show info about the pet
                        {
                            break;
                        }

                    #endregion

                    #region Trading *NEW* // I Think I Fixed the Exploit.
                    case "AG": // Trading - start
                        {
                            if (Room != null || roomUser != null || _tradePartnerRoomUID == -1 || !userManager._reboot)
                            {
                                if (Config.enableTrading == false) { sendData("BK" + stringManager.getString("trading_disabled")); return; }
                                
                                int partnerUID = int.Parse(currentPacket.Substring(2));
                                //virtualBot BotPartner = Room.getBot(partnerUID);
                                //if (Room.containsBot(partnerUID))
                                //{
                                    //sendData("BKBot trading is coming soon!");
                                //}
                                if (Room.containsUser(partnerUID))
                                {
                                    virtualUser Partner = Room.getUser(partnerUID);
                                    if (Partner._Rank > 4)
                                    {
                                        sendData("BKMembers of Staff cannot trade!");
                                    }
                                    else if (Partner._Working || _Working)
                                    {
                                        sendData("BKYou cannot trade while working or trade people who are working!");
                                    }
                                    else if (_PunchCD)
                                    {
                                    }
                                    else
                                    {

                                        if (Partner.statusManager.containsStatus("trd"))
                                            return;
                                        PCDTime = 0;
                                        _PunchCD = true;
                                        
                                        this._tradePartnerRoomUID = partnerUID;
                                        this.statusManager.addStatus("trd", "");
                                        this.roomUser.Refresh();

                                        Partner._tradePartnerRoomUID = this.roomUser.roomUID;
                                        Partner.statusManager.addStatus("trd", "");
                                        Partner.roomUser.Refresh();

                                        this.refreshTradeBoxes();
                                        Partner.refreshTradeBoxes();
                                    }
                                }
                            }
                            break;
                        }

                    case "AH": // Trading - offer item
                        {
                            if (Room != null && roomUser != null && _tradePartnerRoomUID != -1 && Room.containsUser(_tradePartnerRoomUID) || !userManager._reboot)
                            {
                                int itemID = int.Parse(currentPacket.Substring(2));
                                //Database dbClient = new Database(true, true, 129);
                                int templateID;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    templateID = dbClient.getInt("SELECT tid FROM furniture WHERE id = '" + itemID + "' AND ownerid = '" + userID + "' AND roomid = '0'");
                                }
                                if (templateID == 0)
                                    return;

                                _tradeItems[_tradeItemCount] = itemID;
                                _tradeItemCount++;
                                virtualUser Partner = Room.getUser(_tradePartnerRoomUID);

                                this._tradeAccept = false;
                                Partner._tradeAccept = false;

                                this.refreshTradeBoxes();
                                Partner.refreshTradeBoxes();
                            }
                            break;
                        }

                    case "AD": // Trading - decline trade
                        {
                            if (Room != null && roomUser != null && _tradePartnerRoomUID != -1 && Room.containsUser(_tradePartnerRoomUID) || !userManager._reboot)
                            {
                                virtualUser Partner = Room.getUser(_tradePartnerRoomUID);
                                this._tradeAccept = false;
                                Partner._tradeAccept = false;
                                this.refreshTradeBoxes();
                                Partner.refreshTradeBoxes();
                            }
                            break;
                        }

                    case "AE": // Trading - accept trade (and, if both partners accept, swap items]
                        {
                            if (Room != null && roomUser != null && _tradePartnerRoomUID != -1 && Room.containsUser(_tradePartnerRoomUID) || !userManager._reboot)
                            {
                                virtualUser Partner = Room.getUser(_tradePartnerRoomUID);
                                this._tradeAccept = true;
                                this.refreshTradeBoxes();
                                Partner.refreshTradeBoxes();

                                if (Partner._tradeAccept)
                                {
                                    StringBuilder sb = new StringBuilder("'a'='b'");
                                    for (int i = 0; i < _tradeItemCount; i++)
                                        if (_tradeItems[i] > 0)
                                            sb.Append(" OR id = '" + this._tradeItems[i] + "'");

                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE furniture SET ownerid = '" + Partner.userID + "',roomid = '0' WHERE" + sb.ToString());
                                    }
                                    if (_tradeItems.Contains(1010)) // sky
                                    {
                                        _SkyDragon = false;
                                            Partner._SkyDragon = true;
                                    }
                                    if (_tradeItems.Contains(1007)) // black
                                    {
                                        _BlackDragon = false;
                                        Partner._BlackDragon = true;
                                    }
                                    sb.Remove(7, sb.Length - 7);

                                    for (int i = 0; i < Partner._tradeItemCount; i++)
                                        if (Partner._tradeItems[i] > 0)
                                            sb.Append(" OR id = '" + Partner._tradeItems[i] + "'");
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE furniture SET ownerid = '" + this.userID + "',roomid = '0' WHERE" + sb.ToString());
                                    }
                                    if (Partner._tradeItems.Contains(1010)) // sky
                                    {
                                        Partner._SkyDragon = false;
                                        _SkyDragon = true;
                                    }
                                    if (Partner._tradeItems.Contains(1007)) // black
                                    {
                                        Partner._BlackDragon = false;
                                        _BlackDragon = true;
                                    }
                                    abortTrade();
                                }
                            }
                            break;
                        }

                    case "AF": // Trading - abort trade
                        {
                            if (Room != null && roomUser != null && _tradePartnerRoomUID != -1 && Room.containsUser(_tradePartnerRoomUID) || !userManager._reboot)
                            {
                                abortTrade();
                                refreshHand("update");
                            }
                            break;
                        }


                    #endregion

                    #endregion

                    //#region Guide system
                    //case "Ej":
                    //    {
                    //        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                    //            dbClient.runQuery("UPDATE users SET guideavailable = '1' WHERE id = '" + userID + "' LIMIT 1");
                    //        break;
                    //    }

                    //case "Ek":
                    //    {
                    //        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                    //            dbClient.runQuery("UPDATE users SET guideavailable = '0' WHERE id = '" + userID + "' LIMIT 1");
                    //        break;
                    //    }

                    //case "Eg":
                    //    {
                    //        if (_inPublicroom == true)
                    //            return;
                    //        _guideFinish = -1;
                    //        _guideCount = 0;
                    //        _guideRoom = 0;
                    //        _guideSearch = false;
                    //        // userManager.sendToGuides(userID, 0, "Eh" + userID + Convert.ToChar(2));
                    //        return;
                    //    }

                    //case "Ee": // Guide Packet
                    //    {
                    //        _guideCount = 0;
                    //        virtualUser toVirtualUser = userManager.getUser(_guideRequest);
                    //        if (toVirtualUser._isLoggedIn && _guideRoom > 0)
                    //        {
                    //            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                    //            {
                    //                dbClient.runQuery("UPDATE users SET guide_points = guide_points + 1 WHERE id = " + userID);
                    //            }
                    //            toVirtualUser.sendData("Fg" + Encoding.encodeVL64(userID));
                    //            sendData("D^" + "H" + Encoding.encodeVL64(_guideRoom) + Convert.ToChar(2));
                    //        }
                    //        break;
                    //    }

                    //case "Dy": // Guide Packet
                    //    {
                    //        if (currentPacket.Substring(2) == "PB")
                    //        {
                    //            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                    //            {
                    //                if (dbClient.getInt("SELECT guide_allowed FROM users WHERE id = " + userID) == 1)
                    //                {
                    //                    dbClient.runQuery("UPDATE users SET guide_allowed = '0' WHERE id = '" + userID + "' LIMIT 1");
                    //                    sendData("Ed" + "I");
                    //                }
                    //            }
                    //        }
                    //        break;
                    //    }
                    //#endregion
                }
                #endregion
            }
        }
        #endregion

        #region Update voids
        public void Unidle()
        {
            if (AFK)
            {
                refreshAppearance(true, true, true);
            }
            IdleTime = 0;
            AFK = false;
        }
        /// <summary>
        /// Refreshes
        /// </summary>
        /// <param name="Reload">Specifies if the details have to be reloaded from database, or to use current _</param>
        /// <param name="refreshSettings">Specifies if the @E packet (which contains username etc) has to be resent.</param>
        ///<param name="refreshRoom">Specifies if the user has to be refreshed in room by using the 'poof' animation.</param>
        internal void refreshAppearance(bool Reload, bool refreshSettings, bool refreshRoom)
        {
            if (Reload)
            {
                //Database dbClient = new Database(true, true, 131);
                DataRow dRow;
                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                {
                    dRow = dbClient.getRow("SELECT figure,sex,mission FROM users WHERE id = '" + userID + "'");
                }
                _Figure = Convert.ToString(dRow["figure"]);
                _Sex = Convert.ToChar(dRow["sex"]);
                _Mission = Convert.ToString(dRow["mission"]);
            }

            if (refreshSettings)
                sendData("@E" + connectionID + Convert.ToChar(2) + _Username + Convert.ToChar(2) + _Figure + Convert.ToChar(2) + _Sex + Convert.ToChar(2) + _Mission + Convert.ToChar(2) + Convert.ToChar(2) + "PCch=s02/53,51,44" + Convert.ToChar(2) + "HI");

            if (refreshRoom && Room != null && roomUser != null)
                if (_isPet == false)
                {
                    Room.sendData("DJ" + Encoding.encodeVL64(roomUser.roomUID) + _Figure + Convert.ToChar(2) + _Sex + Convert.ToChar(2) + _Mission + Convert.ToChar(2));
                }
                else
                {
                    roomUser.makeHuman();
                }
        }
        /// <summary>
        /// Reloads the valueables (tickets and credits) from database and updates them for client.
        /// </summary>
        /// <param name="Credits">Specifies if to reload and update the Credit count.</param>
        /// <param name="Tickets">Specifies if to reload and update the Ticket count.</param>
        internal void refreshValueables(bool Credits, bool Tickets)
        {
            //Database dbClient = new Database(true, false, 132);
            //DataRow dRow = dbClient.getRow("SELECT credits,tickets FROM users WHERE id = '" + userID + "' LIMIT 1");
            if (Credits)
            {
                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                {
                    _Credits = dbClient.getInt("SELECT credits FROM users WHERE id = '" + userID + "' LIMIT 1");
                }
                sendData("@F" + _Credits);

            }

            if (Tickets)
            {
                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                {
                    _Tickets = dbClient.getInt("SELECT tickets FROM users WHERE id = '" + userID + "' LIMIT 1");
                }
                sendData("A|" + _Tickets);
            }
        }
        /// <summary>
        /// Refreshes the user's badges.
        /// </summary>

        internal void refreshBadges()
        {
            _Badges.Clear(); // Clear old badges
            _badgeSlotIDs.Clear(); // Clear old badge IDs
            //Database dbClient = new Database(true, false, 134);
            DataColumn col1;
            DataColumn col2;
            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
            {
                col1 = dbClient.getColumn("SELECT badgeid FROM users_badges WHERE userid = '" + userID + "' ORDER BY slotid ASC");
                col2 = dbClient.getColumn("SELECT slotid FROM users_badges WHERE userid = '" + userID + "' ORDER BY slotid ASC");
            }
            string[] myBadges = dataHandling.dColToArray(col1);
            string[] myBadgeSlotIDs = dataHandling.dColToArray(col2);

            StringBuilder sbMessage = new StringBuilder();
            sbMessage.Append(Encoding.encodeVL64(myBadges.Length)); // Total amount of badges
            for (int i = 0; i < myBadges.Length; i++)
            {
                sbMessage.Append(myBadges[i]);
                sbMessage.Append(Convert.ToChar(2));

                _Badges.Add(myBadges[i]);
            }

            for (int i = 0; i < myBadges.Length; i++)
            {
                if (Convert.ToInt32(myBadgeSlotIDs[i]) > 0) // Badge enabled!
                {
                    sbMessage.Append(Encoding.encodeVL64((Convert.ToInt32(myBadgeSlotIDs[i]))));
                    sbMessage.Append(myBadges[i]);
                    sbMessage.Append(Convert.ToChar(2));

                    _badgeSlotIDs.Add(Convert.ToInt32(myBadgeSlotIDs[i]));
                }
                else
                    _badgeSlotIDs.Add(0); // :(
            }

            sendData("Ce" + sbMessage.ToString());
            sendData("Ft" + "SHJIACH_Graduate1" + Convert.ToChar(2) + "PAIACH_Login1" + Convert.ToChar(2) + "PAJACH_Login2" + Convert.ToChar(2) + "PAKACH_Login3" + Convert.ToChar(2) + "PAPAACH_Login4" + Convert.ToChar(2) + "PAQAACH_Login5" + Convert.ToChar(2) + "PBIACH_RoomEntry1" + Convert.ToChar(2) + "PBJACH_RoomEntry2" + Convert.ToChar(2) + "PBKACH_RoomEntry3" + Convert.ToChar(2) + "SBRAACH_RegistrationDuration6" + Convert.ToChar(2) + "SBSAACH_RegistrationDuration7" + Convert.ToChar(2) + "SBPBACH_RegistrationDuration8" + Convert.ToChar(2) + "SBQBACH_RegistrationDuration9" + Convert.ToChar(2) + "SBRBACH_RegistrationDuration10" + Convert.ToChar(2) + "RAIACH_AvatarLooks1" + Convert.ToChar(2) + "IJGLB" + Convert.ToChar(2) + "IKGLC" + Convert.ToChar(2) + "IPAGLD" + Convert.ToChar(2) + "IQAGLE" + Convert.ToChar(2) + "IRAGLF" + Convert.ToChar(2) + "ISAGLG" + Convert.ToChar(2) + "IPBGLH" + Convert.ToChar(2) + "IQBGLI" + Convert.ToChar(2) + "IRBGLJ" + Convert.ToChar(2) + "SAIACH_Student1" + Convert.ToChar(2) + "PCIHC1" + Convert.ToChar(2) + "PCJHC2" + Convert.ToChar(2) + "PCKHC3" + Convert.ToChar(2) + "PCPAHC4" + Convert.ToChar(2) + "PCQAHC5" + Convert.ToChar(2) + "QAIACH_GamePlayed1" + Convert.ToChar(2) + "QAJACH_GamePlayed2" + Convert.ToChar(2) + "QAKACH_GamePlayed3" + Convert.ToChar(2) + "QAPAACH_GamePlayed4" + Convert.ToChar(2) + "QAQAACH_GamePlayed5" + Convert.ToChar(2));
            sendData("Dt" + "IH" + Convert.ToChar(1) + "FCH");
        }
        /// <summary>
        /// Refreshes the user's group status.
        /// </summary>
        internal void refreshGroupStatus()
        {
            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
            {
                _groupID = dbClient.getInt("SELECT corp_id FROM users WHERE id = '" + userID + "'");
            }
        }
        /// <summary>
        /// Refreshes the Hand, which contains virtual items, with a specified mode.
        /// </summary>
        /// <param name="Mode">The refresh mode, available: 'next', 'prev', 'update', 'last' and 'new'.</param>
        internal void refreshHand(string Mode)
        {
            DataTable dTable;
            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
            {
                dTable = dbClient.getTable("SELECT furniture.id,furniture.tid,furniture.var FROM users, furniture WHERE furniture.ownerid = users.id AND furniture.ownerid = '" + userID + "' AND furniture.roomid = '0' ORDER BY id ASC");
            }
            StringBuilder Hand = new StringBuilder("BL");
            int startID = 0;
            int stopID = dTable.Rows.Count;

            switch (Mode)
            {
                case "next":
                    _handPage++;
                    break;
                case "prev":
                    _handPage--;
                    break;
                case "last":
                    _handPage = (stopID - 1) / 9;
                    break;
                case "update": // Nothing, keep handpage the same
                    break;
                default: // Probably, "new"
                    _handPage = 0;
                    break;
            }

            try
            {
                if (stopID > 0)
                {
                reCount:
                    startID = _handPage * 9;
                    if (stopID > (startID + 9)) { stopID = startID + 9; }
                    if (startID > stopID || startID == stopID) { _handPage--; goto reCount; }
                    string Colour = "";
                    DataRow dRow;
                    for (int f = startID; f < stopID; f++)
                    {
                        dRow = dTable.Rows[f];
                        catalogueManager.itemTemplate Template = catalogueManager.getTemplate(Convert.ToInt32(dRow["tid"]));
                        char Recycleable = '1';
                        if (Template.isRecycleable == false)
                            Recycleable = '0';

                        if (Convert.ToInt32(dRow["tid"]) == 1010)
                        {
                            _SkyDragon = true;
                        }

                        if (Convert.ToInt32(dRow["tid"]) == 1007)
                        {
                            _BlackDragon = true;
                        }

                        if (Template.typeID == 0) // Wallitem
                        {
                            Colour = Template.Colour;
                            if (Template.Sprite == "post.it" || Template.Sprite == "post.it.vd") // Stickies - pad size
                                Colour = Convert.ToString(dRow["var"]);
                            Hand.Append("SI" + Convert.ToChar(30).ToString() + Convert.ToString(dRow["id"]) + Convert.ToChar(30).ToString() + f + Convert.ToChar(30).ToString() + "I" + Convert.ToChar(30).ToString() + Convert.ToString(dRow["id"]) + Convert.ToChar(30).ToString() + Template.Sprite + Convert.ToChar(30).ToString() + Colour + Convert.ToChar(30).ToString() + Recycleable + "/");
                        }
                        else // Flooritem
                            Hand.Append("SI" + Convert.ToChar(30).ToString() + Convert.ToString(dRow["id"]) + Convert.ToChar(30).ToString() + f + Convert.ToChar(30).ToString() + "S" + Convert.ToChar(30).ToString() + Convert.ToString(dRow["id"]) + Convert.ToChar(30).ToString() + Template.Sprite + Convert.ToChar(30).ToString() + Template.Length + Convert.ToChar(30).ToString() + Template.Width + Convert.ToChar(30).ToString() + Convert.ToString(dRow["var"]) + Convert.ToChar(30).ToString() + Template.Colour + Convert.ToChar(30).ToString() + Recycleable + Convert.ToChar(30).ToString() + Template.Sprite + Convert.ToChar(30).ToString() + "/");
                    }

                }
                Hand.Append("\r" + dTable.Rows.Count);
                sendData(Hand.ToString());
            }
            catch
            {
                sendData("BL" + "\r0");
            }
        }
        /// <summary>
        /// Refreshes the trade window for the user.
        /// </summary>
        internal void refreshTradeBoxes()
        {
            if (Room != null && Room.containsUser(_tradePartnerRoomUID) && roomUser != null)
            {
                virtualUser Partner = Room.getUser(_tradePartnerRoomUID);
                StringBuilder tradeBoxes = new StringBuilder("Al" + _Username + Convert.ToChar(9) + _tradeAccept.ToString().ToLower() + Convert.ToChar(9));
                if (_tradeItemCount > 0) { tradeBoxes.Append(catalogueManager.tradeItemList(_tradeItems)); }
                tradeBoxes.Append(Convert.ToChar(13) + Partner._Username + Convert.ToChar(9) + Partner._tradeAccept.ToString().ToLower() + Convert.ToChar(9));
                if (Partner._tradeItemCount > 0) { tradeBoxes.Append(catalogueManager.tradeItemList(Partner._tradeItems)); }
                sendData(tradeBoxes.ToString());
            }
        }
        /// <summary>
        /// Aborts the trade between this user and his/her partner.
        /// </summary>
        internal void abortTrade()
        {
            if (Room != null && Room.containsUser(_tradePartnerRoomUID) && roomUser != null)
            {
                virtualUser Partner = Room.getUser(_tradePartnerRoomUID);
                this.sendData("An");
                this.refreshHand("update");
                Partner.sendData("An");
                Partner.refreshHand("update");

                this._tradePartnerRoomUID = -1;
                this._tradeAccept = false;
                this._tradeItems = new int[65];
                this._tradeItemCount = 0;
                this.statusManager.removeStatus("trd");
                this.roomUser.Refresh();

                Partner._tradePartnerRoomUID = -1;
                Partner._tradeAccept = false;
                Partner._tradeItems = new int[65];
                Partner._tradeItemCount = 0;
                Partner.statusManager.removeStatus("trd");
                Partner.roomUser.Refresh();
            }
        }
        #endregion

        #region Misc voids

        #region RP Voids

        internal string checkWeapon(int id)
        {
            string weapon;
            {
                if (id == 22)
                {
                    weapon = "Broken Glass Bottle";
                    return weapon;
                }
                if (id == 21)
                {
                    weapon = "Baseball Bat";
                    return weapon;
                }
                if (id == 20)
                {
                    weapon = "Camera";
                    return weapon;
                }
                if (id == 19)
                {
                    weapon = "Bubble Juice";
                    return weapon;
                }
                if (id == 24)
                {
                    weapon = "Martini";
                    return weapon;
                }
                if (id == 25)
                {
                    weapon = "Love Potion";
                    return weapon;
                }
                if (id == 27)
                {
                    weapon = "Cold Ice Cream";
                    return weapon;
                }
                if (id >= 29)
                {
                    weapon = "Glass Of Juice";
                    return weapon;
                }
                if (id == 4)
                {
                    weapon = "Cold Ice Cream";
                    return weapon;
                }
                if (id > 4 && id < 9 || id == 18)
                {
                    weapon = "Glass Of Juice";
                    return weapon;
                }
                if (id > 8 && id < 18)
                {
                    weapon = "Cup Of Coffee";
                    return weapon;
                }
                if (id == 3)
                {
                    weapon = "Carrot";
                    return weapon;
                }
                if (id == 0)
                {
                    weapon = null;
                    return weapon;
                }
                else
                {
                    weapon = "Drink";
                    return weapon;
                }
            }
        }

        //internal string filterMessage(string input)
        //{
        //    const string ALLOWED_CHARACTERS = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ._-:;()',&*!/";
        //    return new string(input.Replace('/', '-').Replace('\\', '-').RemoveAll((c) => !ALLOWED_CHARACTERS.Contains(c)).ToArray());
        //}

        internal string generateBlow(int dam)
        {
            string blow;
            {
                if (dam <= 10)
                {
                    blow = "blow";
                    return blow;
                }
                if (dam > 10 && dam <= 22)
                {
                    blow = "good blow";
                    return blow;
                }
                if (dam > 22 && dam <= 27)
                {
                    blow = "great blow";
                    return blow;
                }
                            if (dam > 27)
                            {
                                blow = "critical blow";
                    return blow;
                }
                else
                {
                    blow = "blow";
                    return blow;
                }
            }
        }
        
        
        internal void arrestUser(int uid, int time)
        {
            virtualUser User = userManager.getUser(uid);
            User._onRoad = true;
            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
            {
                if (_Jailed == 1)
                {
                    User._Punch = 0;
                    User._Arrests = User._Arrests + 1;
                    User._Jailed = 1;
                    User._JailTime = time * 60;

                    Room.sendSaying(roomUser, "\n\n*" + _Username + " re-arrests " + User._Username + " for " + time + " minutes* " + "" + " \n\n  " + _Username + " re-arrests " + User._Username + "  for " + time + " minutes*");

                    User.sendData("DBO");
                    User._roomID = 7;
                    User.roomidi = 7;
                    User.sendData("BK" + "You have been re-arrested by " + _Username + ".");
                    User._Figure = User._Uni + "ch-220-94.lg-280-94.sh-290-62";
                    User._Mission = "[Jailed] " + User._Mission;
                    User.refreshAppearance(false, true, true);
                    User._Arrests = User._Arrests + 1;
                }
                else if (_Jailed == 0)
                {
                    User._Punch = 0;
                    User._Arrests = User._Arrests + 1;
                    User._Jailed = 1;
                    User._JailTime = time * 60;

                    int fine = time * 5;
                    dbClient.runQuery("UPDATE users SET credits = credits - '" + fine + "' WHERE name = '" + User._Username + "'");
                    User._Wanted = 0;
                    User.refreshValueables(true, false);

                    Room.sendSaying(roomUser, "\n\n*" + _Username + " arrests " + User._Username + " for " + time + " minutes, and issues them a " + fine + " credit fine* " + "" + " \n\n  " + _Username + " arrests " + User._Username + "  for " + time + " minutes, and issues them a " + fine + " credit fine*");

                    User.sendData("DBO");
                    User._roomID = 7;
                    User.roomidi = 7;
                    User.sendData("BK" + "You have been arrested by " + _Username + ", and fined " + fine + " credits.");
                    User._Figure = _Uni + "ch-220-94.lg-280-94.sh-290-62";
                    User._Mission = "[Jailed] " + User._Mission;
                    User.refreshAppearance(false, true, true);
                    User._Arrests = User._Arrests + 1;
                }
            }
        }
        internal void killUser(int uid, bool suicide)
        {
            virtualUser User = userManager.getUser(uid);
            User._onRoad = true;
            if (User._Arena)
            {
                User.sendData("DBO");
                User._roomID = 1209;
                User.roomidi = 1209;
                User.sendData("BK" + "You have lost the arena fight.");
                User._Arena = false;
            }
            else if (User._Mission == "[Game] Deathmatch" || User._Brawl)
            {
                if (User._Mission == "[Game] Deathmatch")
                {
                    User.sendData("DBO");
                    User._roomID = 21;
                    User.roomidi = 21;
                    User.sendData("BK" + "An administrator has summoned you.");
                }
                if (User._Brawl)
                {
                    User.sendData("DBO");
                    User._roomID = 21;
                    User.roomidi = 21;
                    User.sendData("BK" + "An administrator has summoned you.");
                }
                User.refreshAppearance(true, true, true);
                User._Health = 100;
            }
            else
            {
                User._Deaths = User._Deaths + 1;
                User._OOC = 1;
                User._DeadWait = true;
                User.DWTime = 0;
                User._Dead = 1;
                User._Health = User._MaxHP;
                User._DeadTime = 300;
                if (!suicide)
                {
                    User.sendData("BK" + "You have been knocked out by another user and sent to the hospital.");
                }
                else
                {
                    User.sendData("BK" + "You have killed yourself and have been sent to the hospital.");
                }
            }

        }
        internal void MessageBox(String message)
        {
            sendData("BK" + message);
        }
        #endregion

        #region Commands

        /// <summary>
        /// Checks if a certain chat message was a 'speech command', if so, then the action for this command is processed and a 'true' boolean is returned. Otherwise, 'false' is returned.
        /// </summary>
        /// <param name="Text">The chat message that was used.</param>
        private bool isSpeechCommand(string Text)
        {
            string[] args = Text.Split(' ');
            try // Try/catch, on error (eg, target user offline, parameters incorrect etc) then failure message will be sent
            {
                switch (args[0].ToLower()) // arg[0] = command itself
                {
                    #region Public commands

                    #region Timers

                    #region :jailbreak
                    case "jailbreak":
                        {
                            {
                                if (_GangID == 0)
                                {
                                    Room.sendWhisper(roomUser, _Username, "You cant jailbreak if you are not in a gang!");
                                }
                                else
                                {
                                    if (minCD)
                                    {
                                        Room.sendWhisper(roomUser, _Username, "*You have to wait one minute before you can jailbreak again*");
                                    }
                                    else
                                    {
                                        if (_Jailed == 1)
                                        {
                                            this.sendData("BK" + "You cannot jailbreak when arrested!");
                                        }
                                        else if (_Jailed == 0)
                                        {
                                            if (_roomID == 7 && roomUser.Y == 10 && roomUser.X == 7)
                                            {
                                                JBTime = 0;
                                                _JailBreaking = true;
                                                userManager.sendToCopJailBreak(1, false, "" + "");
                                                Room.sendSaying(roomUser, "*starts knocking down the jail wall*");
                                                Room.sendSaying(roomUser, "*needs four more minutes to succesfully knock down this wall*");

                                                minCD = true;
                                            }
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :jailbreak
                    case "bankrob":
                    case "robbank":
                        {
                            {
                                if (JBCD)
                                {
                                    Room.sendWhisper(roomUser, _Username, "*You have to wait one minute before you bank rob again*");
                                }
                                else
                                {
                                    if (_Jailed == 1)
                                    {
                                        this.sendData("BK" + "You cannot when arrested!");
                                    }
                                    else if (_Jailed == 0)
                                    {
                                        if (_roomID == 12 && roomUser.Y == 1 && roomUser.X == 4)
                                        {
                                            BRTime = 0;
                                            _robbingBank = true;
                                            userManager.sendToCopBankRob(1, false, "" + "");
                                            Room.sendSaying(roomUser, "*starts breaking into the vault*");
                                            Room.sendSaying(roomUser, "*needs four more minutes to break into the vault*");

                                            JBCD = true;
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :workout
                    case "workout":
                        {
                            sendData("BK" + "There is no longer a workout system to level up your strength earn XP!");
                            break;
                        }
                    #endregion

                    //#region new bankrob system :try
                    //case "try":
                    //case "vault":
                    //    {
                    //        int my_code = int.Parse(args[1]);
                    //        if (BankRobCDLooper != null)
                    //        { }
                    //        else
                    //        {
                    //            #region Finds Shiz
                    //            string arrested;
                    //            int roomid;
                    //            int x;
                    //            int y;
                    //            int code;
                    //            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                    //            {
                    //                arrested = dbClient.getString("SELECT arrested FROM users WHERE name = '" + _Username + "'");
                    //                x = dbClient.getInt("SELECT x FROM action_points WHERE roomid = '" + _roomID + "' && action = 'bankrob'");
                    //                y = dbClient.getInt("SELECT y FROM action_points WHERE roomid = '" + _roomID + "'&& action = 'bankrob'");
                    //                roomid = dbClient.getInt("SELECT roomid FROM action_points WHERE roomid = '" + _roomID + "'&& action = 'bankrob'");
                    //                code = dbClient.getInt("SELECT code FROM action_points WHERE roomid = '" + _roomID + "'&& action = 'bankrob'");
                    //            }
                    //            #endregion

                    //            if (arrested == "1")
                    //            {
                    //                this.sendData("BK" + "You cannot rob the bank when arrested!");
                    //            }
                    //            else if (arrested == "0")
                    //            {
                    //                if (my_code == code && _roomID == roomid && roomUser.Y == y && roomUser.X == x)
                    //                {
                    //                    Random randNum = new Random();
                    //                    int code1 = randNum.Next(0, 9);
                    //                    int code2 = randNum.Next(0, 9);
                    //                    int code3 = randNum.Next(0, 9);
                    //                    int code4 = randNum.Next(0, 9);
                    //                    string new_code = code1 + "" + code2 + "" + code3 + "" + code4;
                    //                    userManager.sendToCopBankRob(1, false, "" + "");
                    //                    Room.sendSaying(roomUser, "*Successfully turns the bank vault dial breaking into the vault*");
                    //                    Room.sendSaying(roomUser, "*Takes 500 credits*");

                    //                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                    //                    {
                    //                        dbClient.runQuery("UPDATE action_points SET code = '" + new_code + "' WHERE code = '" + code + "'");
                    //                    }

                    //                    ThreadStart BankRobCoolDownStarter = new ThreadStart(BankRobCoolDown);
                    //                    BankRobCDLooper = new Thread(BankRobCoolDownStarter);
                    //                    BankRobCDLooper.Priority = ThreadPriority.Lowest;
                    //                    BankRobCDLooper.Start();
                    //                }
                    //                else if (my_code == code + 1 || my_code == code + 2 || my_code == code + 3 || my_code == code + 4 || my_code == code - 1 || my_code == code - 2 || my_code == code - 3 || my_code == code - 4)
                    //                {
                    //                    sendData("BKYou Hear A Click In The Vault");
                    //                    userManager.sendToCopBankRob(1, false, "" + "");
                    //                }
                    //                else if (my_code != code)
                    //                {
                    //                    sendData("BKThe Vault Does Nothing");
                    //                }
                    //            }
                    //        }
                    //        break;
                    //    }
                    //#endregion

                    #endregion

                    #region Rooms

                    #region :sellproperty <amount> & :buyproperty
                    case "sellproperty":
                        {
                            string owner;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                owner = dbClient.getString("SELECT owner FROM rooms WHERE id = '" + _roomID + "'");
                            }
                            if (owner != _Username)
                                sendData("BK" + "Sorry, but you cannot sell someone elses property.");
                            else
                            {
                                int salepice = int.Parse(args[1]);
                                if (salepice < 0)
                                    sendData("BK" + "Sorry, but you cannot sell your property for less then 0 credits.");
                                else
                                {
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE rooms SET price = '" + salepice + "', sale = '1' WHERE id = '" + _roomID + "'");
                                    }
                                    sendData("BK" + "You have successfully put your property on the market for " + salepice + " credits!");
                                }
                            }
                            break;
                        }
                    case "buyproperty":
                        {
                            string forsale;
                            string roomprice;
                            string owner;
                            string credits;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                forsale = dbClient.getString("SELECT sale FROM rooms WHERE id = '" + _roomID + "'");
                                roomprice = dbClient.getString("SELECT price FROM rooms WHERE id = '" + _roomID + "'");
                                owner = dbClient.getString("SELECT owner FROM rooms WHERE id = '" + _roomID + "'");
                                credits = dbClient.getString("SELECT credits FROM users WHERE name = '" + _Username + "'");
                            }
                            if (forsale != "1")
                                sendData("BK" + "Sorry, but this room is not for sale.");
                            else
                            {
                                int roomprice_mod = int.Parse(roomprice);
                                int credits_mod = int.Parse(credits);
                                if (roomprice_mod > credits_mod && owner != _Username)
                                    sendData("BK" + "Sorry, but you do not have enough credits to purchase this property.");
                                else
                                {
                                    virtualUser User = userManager.getUser(owner);
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE rooms SET owner = '" + _Username + "', sale = '0' WHERE id = '" + _roomID + "'");
                                        dbClient.runQuery("UPDATE users SET credits = credits + '" + roomprice + "' WHERE name = '" + owner + "'");
                                        dbClient.runQuery("UPDATE users SET credits = credits - '" + roomprice + "' WHERE name = '" + _Username + "'");
                                        dbClient.runQuery("DELETE FROM jobs_rooms WHERE room_id = '" + _roomID + "';");
                                    }
                                    refreshValueables(true, false);
                                    User.refreshValueables(true, false);
                                    sendData("D^" + "H" + Encoding.encodeVL64(_roomID));
                                    sendData("BK" + "You have successfully bought this property for " + roomprice + " credits.");
                                    User.sendData("BK" + _Username + " has bought your room for " + roomprice + " credits!");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :roomrefresh
                    case "roomrefresh":
                        {
                            if (_Rank > 4)
                            {
                                System.Collections.Hashtable Users = (System.Collections.Hashtable)userManager._Users.Clone();

                                foreach (virtualRoomUser User in Room.Users)
                                {
                                    Room.refreshCoord(roomUser.X, roomUser.Y);
                                    User.makeHuman();
                                    User.Room.sendWhisper(roomUser, _Username, "The room has been refreshed by an administrator.");

                                    roomUser.makeHuman();

                                }

                            }
                        }
                        break;
                    #endregion

                    #region :roomtiles
                    case "roomtiles":
                        {
                            if (_Rank >= 5)
                            {
                            IEnumerable<int> squares = Enumerable.Range(-40, 40);

            foreach (int num in squares)
            {
                Room.sqUNIT[num, num] = false; // Unblock the old tile
            }
                            }
                            
                            break;
                        }
                    #endregion

                    #region :roomdetails
                    case "roomdetails":
                    case "roomid":
                        {
                            sendData("BK" + "Room ID: " + _roomID);
                            break;
                        }
                    #endregion

                    #endregion

                    #region General Commands

                    #region :poof
                    case "poof": // Refreshes users look
                        {
                            if (this._Mission == "[Game] Deathmatch" || this._Brawl || this._Mission == "Yellow Team - Team Death Match!" || this._Mission == "Black Team - Team Death Match!" || this._Mission == "Red Fighter - Arena Fight!" || this._Arena)
                            {
                                sendData("BK" + "You cannot poof, whilst in a game, if you wish to leave, then do :leavegame (:");
                            }
                            else if (!_PunchCD)
                            {
                                int isWorking;
                                int isDead;
                                int isArrested;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    isWorking = dbClient.getInt("SELECT working FROM users WHERE name = '" + _Username + "'");
                                    isDead = dbClient.getInt("SELECT dead FROM users WHERE name = '" + _Username + "'");
                                    isArrested = dbClient.getInt("SELECT arrested FROM users WHERE name = '" + _Username + "'");
                                }
                                if (isWorking == 0 && isDead == 0 && isArrested == 0 && !_isPet)
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        refreshAppearance(true, true, true);
                                        refreshValueables(true, false);
                                        PCDTime = 0;
                                        _PunchCD = true;
                                        
                                    }
                            }
                            break;
                        }
                    #endregion

                    #region :userpoof <user>
                    case "userpoof": // Refreshes users look
                        {
                            if (_Rank > 5)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                {
                                    int isWorking;
                                    int isDead;
                                    int isArrested;
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        isWorking = dbClient.getInt("SELECT working FROM users WHERE name = '" + User._Username + "'");
                                        isDead = dbClient.getInt("SELECT dead FROM users WHERE name = '" + User._Username + "'");
                                        isArrested = dbClient.getInt("SELECT arrested FROM users WHERE name = '" + User._Username + "'");
                                    }
                                    if (isWorking == 0 && isDead == 0 && isArrested == 0)
                                    {
                                        User.refreshAppearance(true, true, true);
                                        User.refreshValueables(true, false);
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :flagme
                    case "flagme":
                        {
                            if (_Rank > 1)
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.runQuery("UPDATE users SET flagged = '1' WHERE name = '" + _Username + "'");
                                    Room.sendShout(roomUser, "------ Auto Logout In 15 Seconds -------");
                                }
                                Disconnect(15000);
                            }
                            else
                            {
                                sendData("Sorry but this is a VIP command only!");
                            }
                            break;
                        }
                    #endregion

                    #region :dnd
                    case "dnd":
                        {
                            if (_Rank > 1)
                            {
                                _DND = true;
                            }
                            else
                            {
                                sendData("Sorry but this is a VIP command only!");
                            }
                            break;
                        }
                    #endregion

                    #region :escape
                    case "escape":
                        {
                            if (_Rank > 5)
                            {
                                    if (_Jailed == 1)
                                    {
                                        _Jailed = 0;
                                        _JailTime = 0;
                                        Room.sendSaying(roomUser, "\n\n*" + _Username + " uses their powers to escape prison* " + "" + " \n\n  " + _Username + " uses their powers to escape prison*");
                                        refreshAppearance(true, true, true);

                                    }
                            }
                            break;
                        }
                    #endregion

                    #region :emptyhand
                    case "emptyhand": // Deletes everything from the senders hand
                    case "cleanhand": // Deletes everything from the senders hand
                        {
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                dbClient.runQuery("DELETE FROM furniture WHERE ownerid = '" + userID + "' AND roomid = '0'");
                            }
                            refreshHand("update");
                            break;
                        }
                    #endregion

                    #region :commands
                    case "commands": // commands list
                        {
                            sendData("BK" +
                            "General Commands\r" +
                            "\r" +
                            ":hit <user>\r" +
                            ":punch <user>\r" +
                            ":taxi <roomid>\r" +
                            ":push <user>\r" +
                            ":roomid\r" +
                            ":stats\r" +
                            ":balance (bank account)\r" +
                            ":911 <message>\r" +
                            ":give <user> <amount>\r" +
                            "\r" +
                            "Properties\r" +
                            ":sellproperty <amount>\r" +
                            ":buyproperty");

                            break;
                        }
                    #endregion

                    #region :staffcommands
                    case "staffcommands": // commands list
                        {
                            if (_Rank > 4)
                            {
                                sendData("BK" +
                                "--------- Administration Commands ---------\r" +
                                "\r" +
                                ":hair <user> <code>\r" +
                                ":surgery <user> <code>\r" +
                                ":discharge <user>\r" +
                                ":adminrelease <user>\r" +
                                ":at <roomid>\r" +
                                ":kill <user>\r" +
                                ":ha <message>\r" +
                                ":str <user> <amount>\r" +
                                ":badge <user> <code>\r" +
                                ":roombadge <code>\r" +
                                ":figure <clothesid>\r" +
                                ":userfigure <user> <clothesid>\r" +
                                ":change\r" +
                                ":reboot\r" +
                                ":quickreboot\r" +
                                 ":superfire <user>\r" +
                                ":supersendhome <user>\r" +
                                ":superhire <user> <corpid> <rankid>\r");
                            }
                            break;
                        }
                    #endregion

                    #endregion

                    #region RP Commands

                    #region :about
                    case "about":
                        System.Net.WebClient wc = new System.Net.WebClient();
                        this.sendData("BK-------------------------------------------------------------------------------\nXoa Emulator\nCoded by Jamie Jackson\nPowered by Holograph\n-------------------------------------------------------------------------------");
                        break;
                    #endregion

                    #region :punch <user>
                    case "punch":
                    case "hit":
                        {
                            #region Bots
                            //{
                            //    virtualBot Bot = botManager.getBot(args[1]);
                            //    {
                            //        if (Bot != null)
                            //        {
                            //            int fight;
                            //            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            //                fight = dbClient.getInt("SELECT violence FROM rooms WHERE id = '" + _roomID + "'");
                            //            if (_Jailed == 1 || _Dead == 1 || roomUser.walkLock == true || _PunchCD || AFK || fight == 1)
                            //            {
                            //            }
                            //            else
                            //            {
                            //                if ((roomUser.Y == Bot.roomBot.Y && roomUser.X == Bot.roomBot.X) || (roomUser.Y + 1 == Bot.roomBot.Y && roomUser.X == Bot.roomBot.X) || (roomUser.Y - 1 == Bot.roomBot.Y && roomUser.X == Bot.roomBot.X) || (roomUser.Y == Bot.roomBot.Y && roomUser.X + 1 == Bot.roomBot.X) || (roomUser.Y == Bot.roomBot.Y && roomUser.X - 1 == Bot.roomBot.X))
                            //                {
                            //                    PCDTime = 0;
                            //                    _PunchCD = true;

                            //                    Random randNum = new Random();
                            //                    int rndNum2 = new Random(DateTime.Now.Millisecond).Next(1, 3);
                            //                    int rndNum1 = new Random(DateTime.Now.Millisecond).Next(0, 10);
                            //                    int rndNum3 = new Random(DateTime.Now.Millisecond).Next(0, 7);
                            //                    int damage = _Str * rndNum2 + rndNum1;
                            //                    int dragonblow = _Str * 4 + rndNum1;
                            //                    int wdamage = _Str * rndNum2 + 3 + rndNum1;

                            //                    if (_Health <= 0)
                            //                    {
                            //                    }
                            //                    else
                            //                    {

                            //                        _Hits = _Hits + 1;
                            //                        _Punch = 1;
                            //                        string _Weapon = checkWeapon(Convert.ToInt32(_ItemID));
                            //                        if (_Weapon != null)
                            //                        {
                            //                            if (_ItemID == "21")
                            //                            {
                            //                                MeShout("smacks " + Bot.Name + " over the head with a BaseBall Bat, causing " + wdamage + " damage", userID);
                            //                                Bot.Health = Bot.Health - wdamage;
                            //                            }
                            //                            else if (_ItemID == "22")
                            //                            {
                            //                                MeShout("shanks " + Bot.Name + " with a Broken Bottle, causing " + wdamage + " damage", userID);
                            //                                Bot.Health = Bot.Health - wdamage;
                            //                            }
                            //                            else
                            //                            {
                            //                                MeShout("lands a blow on " + Bot.Name + " with a " + _Weapon + ", causing " + damage + " damage", userID);
                            //                                Bot.Health = Bot.Health - damage;
                            //                            }
                            //                        }
                            //                        else
                            //                        {
                            //                            if (_BlackDragon)
                            //                            {
                            //                                if (rndNum3 == 6 || rndNum3 == 7)
                            //                                {

                            //                                    MeShout("lands a dragon blow on " + Bot.Name + ", causing " + dragonblow + " damage", userID);
                            //                                    Bot.Health = Bot.Health - dragonblow;
                            //                                }
                            //                                else
                            //                                {
                            //                                    string blow = generateBlow(damage);
                            //                                    MeShout("lands a " + blow + " on " + Bot.Name + ", causing " + damage + " damage", userID);
                            //                                    Bot.Health = Bot.Health - damage;
                            //                                }
                            //                            }
                            //                            else
                            //                            {
                            //                                string blow = generateBlow(damage);
                            //                                MeShout("lands a " + blow + " on " + Bot.Name + ", causing " + damage + " damage", userID);
                            //                                Bot.Health = Bot.Health - damage;
                            //                            }
                            //                        }
                            //                        if (Bot.Health < 50 && Bot.Health > 30)
                            //                        {
                            //                            string message = "begins to blead heavily";
                            //                            Room.sendShout(Bot.roomBot, "\n\n*" + Bot.Name + " " + message + "" + "" + "*\n\n  " + Bot.Name + " " + message);
                            //                        }
                            //                        if (Bot.Health < 30 && Bot.Health > 0)
                            //                        {
                            //                            string message = "barely manages to keep up right";
                            //                            Room.sendShout(Bot.roomBot, "\n\n*" + Bot.Name + " " + message + "" + "" + "*\n\n  " + Bot.Name + " " + message);
                            //                        }
                            //                        if (Bot.Cop == 1 && !Bot.calledCops)
                            //                        {
                            //                            this.Room.sendSaying(Bot.roomBot, "\n\n*" + Bot.Name + " calls for backup" + "" + "*\n\n  " + Bot.Name + "" + " calls for backup*");
                            //                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            //                            {
                            //                                string Message = "Police Bot requesting help!";
                            //                                dbClient.runQuery("UPDATE policecall SET message = '[911] [Backup]" + Message + "' WHERE id = '1'"); ;
                            //                                dbClient.runQuery("UPDATE policecall SET room = '" + roomUser.roomID + "' WHERE id = '1'"); ;
                            //                                dbClient.runQuery("UPDATE policecall SET user = '" + Bot.Name + "' WHERE id = '1'"); ;
                            //                            }
                            //                            userManager.sendToCop(1, false, "" + "");
                            //                            Bot.calledCops = true;
                            //                        }
                            //                        if (Bot.Cop == 0 && !Bot.calledCops)
                            //                        {
                            //                            this.Room.sendSaying(Bot.roomBot, "\n\n*" + Bot.Name + " calls the police reporting " + _Username + " as an attacker*" + "" + "*\n\n  " + Bot.Name + "" + " calls the police reporting " + _Username + " as an attacker*");
                            //                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            //                            {
                            //                                string Message = _Username + " has just attacked me!";
                            //                                dbClient.runQuery("UPDATE policecall SET message = '[911] " + Message + "' WHERE id = '1'"); ;
                            //                                dbClient.runQuery("UPDATE policecall SET room = '" + roomUser.roomID + "' WHERE id = '1'"); ;
                            //                                dbClient.runQuery("UPDATE policecall SET user = '" + Bot.Name + "' WHERE id = '1'"); ;
                            //                            }
                            //                            userManager.sendToCop(1, false, "" + "");
                            //                            Bot.calledCops = true;
                            //                        }
                            //                        if (Bot.Health <= 0)
                            //                        {
                            //                            _Kills = _Kills + 1;
                            //                            MeShout("lands a final blow on " + Bot.Name + ", knocking them out", userID);
                            //                            Room.deloadBot(Bot);
                            //                            Bot.Health = 100;
                            //                            Bot.calledCops = false;
                            //                        }
                            //                    }
                            //                }
                            //            }
                            //        }
                            //    }
                            //}
                            #endregion

                            #region Users
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                {
                                    int fight;
                                    int fight2;
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        fight = dbClient.getInt("SELECT violence FROM rooms WHERE id = '" + _roomID + "'");
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        fight2 = dbClient.getInt("SELECT violence FROM rooms WHERE id = '" + User._roomID + "'");
                                    if (_Jailed == 1 || _Dead == 1 || roomUser.walkLock == true || User._Jailed == 1 || User._Dead == 1 || _PunchCD || User._roomID != _roomID || User.AFK || AFK || fight == 1 || fight2 == 1)
                                    {
                                    }
                                    else
                                    {
                                        if (User._roomID == _roomID && (roomUser.Y == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y + 1 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y - 1 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X + 1 == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X - 1 == User.roomUser.X))
                                        {
                                            PCDTime = 0;
                                            _PunchCD = true;

                                            Random randNum = new Random();
                                            int rndNum2 = new Random(DateTime.Now.Millisecond).Next(1, 3);
                                            int rndNum1 = new Random(DateTime.Now.Millisecond).Next(0, 10);
                                            int rndNum3 = new Random(DateTime.Now.Millisecond).Next(0, 7);
                                            int damage = _Str * rndNum2 + rndNum1;
                                            int dragonblow = _Str * 4 + rndNum1;
                                            int wdamage = _Str * rndNum2 + 3 + rndNum1;

                                            if (_Health <= 0)
                                            {
                                            }
                                            else
                                            {
                                                _Hits = _Hits + 1;
                                                if (User._Username != _Username)
                                                {
                                                    _Punch = 1;
                                                    AddXP(1);
                                                }
                                                string _Weapon = checkWeapon(Convert.ToInt32(_ItemID));
                                                if (_Weapon != null)
                                                {
                                                    if (_ItemID == "21")
                                                    {
                                                        MeShout("smacks " + User._Username + " over the head with a BaseBall Bat, causing " + wdamage + " damage", userID);
                                                        User._Health = User._Health - wdamage;
                                                    }
                                                    else if (_ItemID == "22")
                                                    {
                                                        MeShout("shanks " + User._Username + " with a Broken Bottle, causing " + wdamage + " damage", userID);
                                                        User._Health = User._Health - wdamage;
                                                    }
                                                    else
                                                    {
                                                        MeShout("lands a blow on " + User._Username + " with a " + _Weapon + ", causing " + damage + " damage", userID);
                                                        User._Health = User._Health - damage;
                                                    }
                                                }
                                                else
                                                {
                                                    if (_BlackDragon)
                                                    {
                                                        if (rndNum3 == 6 || rndNum3 == 7)
                                                        {

                                                            MeShout("lands a dragon blow on " + User._Username + ", causing " + dragonblow + " damage", userID);
                                                            User._Health = User._Health - dragonblow;
                                                        }
                                                        else
                                                        {
                                                            string blow = generateBlow(damage);
                                                            MeShout("lands a " + blow + " on " + User._Username + ", causing " + damage + " damage", userID);
                                                            User._Health = User._Health - damage;
                                                        }
                                                    }
                                                    else
                                                    {
                                                        string blow = generateBlow(damage);
                                                        MeShout("lands a " + blow + " on " + User._Username + ", causing " + damage + " damage", userID);
                                                        User._Health = User._Health - damage;
                                                    }
                                                }
                                                if (User._Health < 50 && User._Health > 30)
                                                {
                                                    MeShout("begins to bleed heavily", User.userID);
                                                }
                                                if (User._Health < 30 && User._Health > 0)
                                                {
                                                    MeShout("barely manages to keep up right", User.userID);
                                                }
                                                if (User._Health <= 0)
                                                {
                                                    if (User._Arena && _Arena)
                                                    {
                                                        string blow = "final blow";
                                                        MeShout("lands a " + blow + " on " + User._Username + ", winning the match and +10xp", userID);
                                                        refreshAppearance(true, true, true);
                                                        User.refreshAppearance(true, true, true);
                                                        _onRoad = true;
                                                        sendData("DBO");
                                                        _roomID = 1209;
                                                        roomidi = 1209;
                                                        AddXP(10);
                                                        _Kills = _Kills + 1;
                                                        killUser(User.userID, false);
                                                    }
                                                    if (User._Brawl && _Brawl)
                                                    {
                                                        string blow = "final blow";
                                                        MeShout("lands a " + blow + " on " + User._Username + ", sending them back to the lobby", userID);
                                                        User.refreshAppearance(true, true, true);
                                                        _Kills = _Kills + 1;
                                                        killUser(User.userID, false);
                                                    }
                                                    if (!User._Brawl && !_Brawl && !User._Arena && !_Arena)
                                                    {

                                                        string blow = "final blow";
                                                        MeShout("lands a " + blow + " on " + User._Username + ", knocking them out", userID);
                                                        if (User._Rank >= 4)
                                                        {
                                                            MeShout("checks " + User._Username + "'s body for cash, finds nothing", userID);
                                                        }
                                                        else
                                                        {
                                                            if (User._Credits > 100 && User._Rank < 2)
                                                            {
                                                                int stolen = randNum.Next(0, User._Credits);
                                                                if (stolen == 0)
                                                                {
                                                                    MeShout("checks " + User._Username + "'s body for cash and finds nothing", userID);
                                                                }
                                                                else
                                                                {
                                                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                                                    {
                                                                        MeShout("recovers " + stolen + " credits from " + User._Username + "'s body", userID);
                                                                        dbClient.runQuery("UPDATE users SET credits = credits + '" + stolen + "' WHERE name = '" + _Username + "'");
                                                                        dbClient.runQuery("UPDATE users SET credits = credits - '" + stolen + "' WHERE name = '" + User._Username + "'");
                                                                        dbClient.runQuery("UPDATE users SET credits = credits - 15 WHERE id = '" + _Username + "'");
                                                                        User.refreshValueables(true, false);
                                                                        refreshValueables(true, false);
                                                                    }


                                                                }
                                                            }
                                                            else
                                                            {
                                                                MeShout("checks " + User._Username + "'s body for cash, finds nothing", userID);
                                                            }
                                                        }
                                                        if (User._Username != _Username)
                                                        {
                                                            AddXP(5);
                                                        }
                                                        if (_GangID != 0)
                                                        {
                                                            using (DatabaseClient dbClient = Xoa.Eucalypt.dbManager.GetClient())
                                                                dbClient.runQuery("UPDATE gang_corps SET kills = kills + '1' WHERE owner = '" + _GangID + "'");
                                                        }
                                                        _Kills = _Kills + 1;
                                                        killUser(User.userID, false);
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            #endregion
                            break;
                        }

                    #endregion

                    #region :slap <user>
                    case "slap":
                    case "smack":
                        {
                            virtualUser User = userManager.getUser(args[1]);
                            int fight;
                            int fight2;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                fight = dbClient.getInt("SELECT violence FROM rooms WHERE id = '" + _roomID + "'");
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                fight2 = dbClient.getInt("SELECT violence FROM rooms WHERE id = '" + User._roomID + "'");
                            if (_Jailed == 1 || _Dead == 1 || roomUser.walkLock == true || User._Jailed == 1 || User._Dead == 1 || _PunchCD || User._roomID != _roomID || User.AFK || AFK || fight == 1 || fight2 == 1)
                            {
                            }
                            else
                            {
                                if (User._roomID == _roomID && (roomUser.Y == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y + 1 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y - 1 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X + 1 == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X - 1 == User.roomUser.X))
                                {
                                    PCDTime = 0;
                                    _PunchCD = true;
                                    
                                    if (_Health <= 0)
                                    {
                                    }
                                    else
                                    {
                                        if (User._Username != _Username)
                                        {
                                            _Punch = 1;
                                        }
                                                MeShout("slaps " + User._Username + ", causing 1 damage", userID);
                                                User._Health = User._Health - 1;
                                        if (User._Health < 50)
                                        {
                                            MeShout("begins to bleed heavily", User.userID);
                                        }
                                        if (User._Health < 30)
                                        {
                                            MeShout("barely manages to keep up right", User.userID);
                                        }
                                        if (User._Health <= 0)
                                        {
                                            if (User._Arena && _Arena)
                                            {
                                                string blow = "final";
                                                MeShout("lands a " + blow + " slap on " + User._Username + ", winning the match and +10xp", userID);
                                                refreshAppearance(true, true, true);
                                                User.refreshAppearance(true, true, true);
                                                _onRoad = true;
                                                sendData("DBO");
                                                _roomID = 1209;
                                                roomidi = 1209;
                                                AddXP(10);
                                                _Kills = _Kills + 1;
                                                killUser(User.userID, false);
                                            }
                                            if (User._Brawl && _Brawl)
                                            {
                                                string blow = "final";
                                                MeShout("lands a " + blow + " slap on " + User._Username + ", sending them back to the lobby", userID);
                                                User.refreshAppearance(true, true, true);
                                                _Kills = _Kills + 1;
                                                killUser(User.userID, false); ;
                                            }
                                            if (!User._Brawl && !_Brawl && !User._Arena && !_Arena)
                                            {

                                                MeShout("lands a final slap " + User._Username + ", knocking them out", userID);
                                                if (User._Rank >= 4)
                                                {
                                                    MeShout("checks " + User._Username + "'s body for cash, finds nothing", userID);
                                                }
                                                else
                                                {
                                                    if (User._Credits > 100 && User._Rank < 2)
                                                    {
                                                        Random randNum = new Random();
                                                        int stolen = randNum.Next(0, User._Credits);
                                                        if (stolen == 0)
                                                        {
                                                            MeShout("checks " + User._Username + "'s body for cash and finds nothing", userID);
                                                        }
                                                        else
                                                        {
                                                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                                            {
                                                                MeShout("recovers " + stolen + " credits from " + User._Username + "'s body", userID);
                                                                dbClient.runQuery("UPDATE users SET credits = credits + '" + stolen + "' WHERE name = '" + _Username + "'");
                                                                dbClient.runQuery("UPDATE users SET credits = credits - '" + stolen + "' WHERE name = '" + User._Username + "'");
                                                                dbClient.runQuery("UPDATE users SET credits = credits - 15 WHERE id = '" + _Username + "'");
                                                                User.refreshValueables(true, false);
                                                                refreshValueables(true, false);
                                                            }


                                                        }
                                                    }
                                                    else
                                                    {
                                                        MeShout("checks " + User._Username + "'s body for cash, finds nothing", userID);
                                                    }
                                                }
                                                if (User._Username != _Username)
                                                {
                                                    AddXP(5);
                                                }
                                                _Kills = _Kills + 1;
                                                killUser(User.userID, false);
                                            }
                                        }
                                    }
                                }
                            }
                            break;
                        }

                    #endregion

                    #region :purchase clothing-colour
                    case "purchase": // Purchase some new clothing
                        {
                            int PartCheck;
                            string FigurePart;
                            string BuyPart;
                            string NewItem;


                            bool getfurni;


                            NewItem = Text.Split(' ')[1];
                            BuyPart = NewItem.Split('-')[0] + "-";
                            NewItem = NewItem.Replace(BuyPart, "");
                            BuyPart = BuyPart.ToLower();
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {


                                if (_Working == false && _Dead == 0 && _Jailed == 0)
                                {
                                        if (NewItem.IndexOf('.') != -1 || NewItem.IndexOf('-') == -1) // prevents extra elements being added in (like chest, shoes, etc)
                                            break;


                                        // Check that we are in a store and that it sells the item that the user wants to buy
                                        getfurni = dbClient.findsResult("SELECT id FROM store_clothing WHERE part = '" + BuyPart.Split('-')[0] + "' AND partid = '" + NewItem.Split('-')[0] + "' AND storeid = '" + _roomID + "'");


                                        if (getfurni == false)
                                        {
                                            sendData("BK" + "You can't buy that here, either, it is not in stock, or it's the wrong store.");
                                            break;
                                        }
                                        else
                                        {
                                            if (NewItem != "")
                                            {
                                                PartCheck = _Figure.IndexOf(BuyPart);
                                                if (PartCheck == -1)
                                                {
                                                    _Figure = _Figure + "." + BuyPart + NewItem;
                                                }
                                                else
                                                {
                                                    FigurePart = Regex.Split(_Figure, BuyPart)[1];
                                                    FigurePart = BuyPart + FigurePart.Split('.')[0];
                                                    _Figure = _Figure.Replace(FigurePart, BuyPart + NewItem);
                                                }
                                            }
                                            _Figure = _Figure.Replace("..", ".");
                                            dbClient.runQuery("UPDATE users SET figure = '" + _Figure + "' WHERE id = '" + userID + "'");
                                            refreshAppearance(true, true, true);
                                            this.Room.sendSaying(roomUser, "*Purchases Some New Clothing*");
                                        }
                                }
                                else
                                {
                                    this.sendData("BK" + "You cannot use this command, as you are either, dead, arrested or working.");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :figure clothing-colour
                    case "figure": // Purchase some new clothing
                        {
                            int PartCheck;
                            string FigurePart;
                            string BuyPart;
                            string NewItem;


                            


                            NewItem = Text.Split(' ')[1];
                            BuyPart = NewItem.Split('-')[0] + "-";
                            NewItem = NewItem.Replace(BuyPart, "");
                            BuyPart = BuyPart.ToLower();
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {


                                if (_Working == false && _Dead == 0 && _Jailed == 0)
                                {
                                    if (NewItem.IndexOf('.') != -1 || NewItem.IndexOf('-') == -1) // prevents extra elements being added in (like chest, shoes, etc)
                                        break;


                                    if (_Rank < 6)
                                    {
                                        sendData("BK" + "You cant use this.");
                                        break;
                                    }
                                    else
                                    {
                                        if (NewItem != "")
                                        {
                                            PartCheck = _Figure.IndexOf(BuyPart);
                                            if (PartCheck == -1)
                                            {
                                                _Figure = _Figure + "." + BuyPart + NewItem;
                                            }
                                            else
                                            {
                                                FigurePart = Regex.Split(_Figure, BuyPart)[1];
                                                FigurePart = BuyPart + FigurePart.Split('.')[0];
                                                _Figure = _Figure.Replace(FigurePart, BuyPart + NewItem);
                                            }
                                        }
                                        _Figure = _Figure.Replace("..", ".");
                                        dbClient.runQuery("UPDATE users SET figure = '" + _Figure + "' WHERE id = '" + userID + "'");
                                        refreshAppearance(true, true, true);
                                    }
                                }
                                else
                                {
                                    this.sendData("BK" + "You cannot use this command, as you are either, dead, arrested or working.");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :userfigure clothing-colour
                    case "userfigure": // Purchase some new clothing
                        {
                            int PartCheck;
                            string FigurePart;
                            string BuyPart;
                            string NewItem;


                            

                            virtualUser User = userManager.getUser(args[1]);
                            NewItem = Text.Split(' ')[2];
                            BuyPart = NewItem.Split('-')[0] + "-";
                            NewItem = NewItem.Replace(BuyPart, "");
                            BuyPart = BuyPart.ToLower();
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {


                                if (_Working == false && _Dead == 0 && _Jailed == 0)
                                {
                                    if (NewItem.IndexOf('.') != -1 || NewItem.IndexOf('-') == -1) // prevents extra elements being added in (like chest, shoes, etc)
                                        break;


                                    if (_Rank < 6)
                                    {
                                        sendData("BK" + "You cant use this.");
                                        break;
                                    }
                                    else
                                    {
                                        if (NewItem != "")
                                        {
                                            PartCheck = User._Figure.IndexOf(BuyPart);
                                            if (PartCheck == -1)
                                            {
                                                User._Figure = User._Figure + "." + BuyPart + NewItem;
                                            }
                                            else
                                            {
                                                FigurePart = Regex.Split(User._Figure, BuyPart)[1];
                                                FigurePart = BuyPart + FigurePart.Split('.')[0];
                                                User._Figure = User._Figure.Replace(FigurePart, BuyPart + NewItem);
                                            }
                                        }
                                        User._Figure = User._Figure.Replace("..", ".");
                                        dbClient.runQuery("UPDATE users SET figure = '" + User._Figure + "' WHERE id = '" + User.userID + "'");
                                        User.refreshAppearance(true, true, true);
                                    }
                                }
                                else
                                {
                                    this.sendData("BK" + "You cannot use this command, as you are either, dead, arrested or working.");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :remove itemid
                    case "remove":
                        {
                            int PartCheck;
                            string FigurePart;
                            string RemovePart;
                            RemovePart = Text.Split(' ')[1];

                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {

                                if (_Working == false && _Dead == 0 && _Jailed == 0)
                                {
                                    if (RemovePart == "ha")
                                        RemovePart = "ha";
                                    else if (RemovePart == "ea")
                                        RemovePart = "ea";
                                    else if (RemovePart == "ca")
                                        RemovePart = "ca";
                                    else if (RemovePart == "fa")
                                        RemovePart = "fa";
                                    else if (RemovePart == "he")
                                        RemovePart = "he";
                                    else if (RemovePart == "wa")
                                        RemovePart = "wa";
                                        else if (RemovePart == "sh")
                                        RemovePart = "sh";
                                    else if (RemovePart == "ch" && _Sex == Convert.ToChar("M"))
                                        RemovePart = "ch";
                                    else
                                    {
                                        break;
                                    }
                                    PartCheck = _Figure.IndexOf(RemovePart);
                                    if (PartCheck == -1)
                                    {
                                        sendData("BK" + "There's nothing to remove there");
                                        break;
                                    }
                                    else
                                    {
                                        FigurePart = Regex.Split(_Figure, RemovePart)[1];
                                        FigurePart = RemovePart + FigurePart.Split('.')[0] + ".";
                                        _Figure = _Figure.Replace(FigurePart, "");
                                        FigurePart = FigurePart.Split('.')[0];
                                        _Figure = _Figure.Replace(FigurePart, "");
                                        dbClient.runQuery("UPDATE users SET figure = '" + _Figure + "' WHERE id = '" + userID + "'");
                                        refreshAppearance(true, true, true);
                                        this.Room.sendSaying(roomUser, "*Removes An Item Of Clothing*");
                                    }
                                }
                                else
                                {
                                    this.sendData("BK" + "You cannot remove parts when dead, arrested or working!");

                                }
                            }
                            break;
                        }
                    #endregion

                    #region :ooc
                    case "ooc":
                        {
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                _OOC = 1;
                            }

                            break;
                        }

                    #endregion

                    #region :givebadge <user>
                    case "givebadge":
                        {
                            if (_Rank >= 5)
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    virtualUser Target = userManager.getUser(args[1]);
                                    string Badge = args[2];
                                    dbClient.runQuery("INSERT INTO users_badges(userid, badgeid, iscurrent, slotid, badge) VALUES('" + Target.userID + "', '" + Badge + "', '0', '0', '" + Badge + "')");
                                    sendData("BK" + "The user has recieved the badge.");
                                    Target.refreshBadges();
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :ic
                    case "ic":
                        {
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                _OOC = 0;
                            }

                            break;
                        }
                    #endregion

                    #region :whosonline
                    case "connections": // Generates a list of users connected
                    case "whosonline": // Just for the noobs who cant spell connections lol - Zak
                        if (_Rank > 4)
                        {
                            sendData(userManager.generateWhosOnline(rankManager.containsRight(_Rank, "fuse_administrator_access", userID)));
                        }
                        break;
                    #endregion

                    #region :die
                    case "die": // kills the user
                    case "sucicide":
                        if (_Dead != 1 && _Jailed != 1 && _roomID != 187)
                        {
                            MeSay("overdoses and calls an ambulance for themselves", userID);
                            killUser(userID, true);
                        }
                        break;
                    #endregion

                    #region :me <message>
                    case "me"://
                        {
                            string message = Text.Substring(2);
                            Room.sendSaying(roomUser, "\n\n*" + _Username + "" + message + "" + "" + "*\n\n  " + _Username + "" + message);
                            break;
                        }
                    #endregion

                    #region :push <user>
                    case "push":
                        {
                            
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                if (User._roomID == _roomID)
                                {
                                    if (_PunchCD)
                                    {
                                        Room.sendWhisper(roomUser, _Username, "That action cannot be performed during cooldown.");
                                    }
                                    else
                                    {
                                        if (User._roomID == _roomID && (roomUser.Y == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y + 1 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y - 1 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X + 1 == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X - 1 == User.roomUser.X))
                                        {
                                            this.Room.sendSaying(roomUser, "\n\n*" + _Username + " pushes " + User._Username + "*" + "" + "*\n\n  " + _Username + "" + "pushes " + User._Username + "*");
                                            {
                                                {
                                                    switch (roomUser.Z2)
                                                    {
                                                        case 0:
                                                            {
                                                                User.roomUser.goalX = User.roomUser.X;
                                                                User.roomUser.goalY = User.roomUser.Y - 1;
                                                                break;
                                                            }
                                                        case 2:
                                                            {
                                                                User.roomUser.goalX = User.roomUser.X + 1;
                                                                User.roomUser.goalY = User.roomUser.Y;
                                                                break;
                                                            }
                                                        case 4:
                                                            {
                                                                User.roomUser.goalX = User.roomUser.X;
                                                                User.roomUser.goalY = User.roomUser.Y + 1;
                                                                break;
                                                            }
                                                        case 6:
                                                            {
                                                                User.roomUser.goalX = User.roomUser.X - 1;
                                                                User.roomUser.goalY = User.roomUser.Y;
                                                                break;
                                                            }
                                                    }
                                                    PCDTime = 0;
                                                    _PunchCD = true;
                                                }
                                            }
                                        }
                                    }
                                }
                            }

                            break;
                        }
                    #endregion

                    #region :pull
                    case "pull":
                        {
                            if (_Rank >= 2)
                            {
                                
                                {
                                    virtualUser User = userManager.getUser(args[1]);
                                    {
                                        if (User._roomID == _roomID && ((Math.Abs(roomUser.X - User.roomUser.X) <= 2) && (Math.Abs(roomUser.Y - User.roomUser.Y) <= 2)))
                                        {
                                            switch (roomUser.Z2)
                                            {
                                                case 0:
                                                    {
                                                        User.roomUser.goalX = roomUser.X;
                                                        User.roomUser.goalY = roomUser.Y - 1;
                                                        break;
                                                    }
                                                case 1:
                                                    {
                                                        User.roomUser.goalX = roomUser.X;
                                                        User.roomUser.goalY = roomUser.Y - 1;
                                                        break;
                                                    }
                                                case 2:
                                                    {
                                                        User.roomUser.goalX = roomUser.X + 1;
                                                        User.roomUser.goalY = roomUser.Y;
                                                        break;
                                                    }
                                                case 3:
                                                    {
                                                        User.roomUser.goalX = roomUser.X + 1;
                                                        User.roomUser.goalY = roomUser.Y;
                                                        break;
                                                    }
                                                case 4:
                                                    {
                                                        User.roomUser.goalX = roomUser.X;
                                                        User.roomUser.goalY = roomUser.Y + 1;
                                                        break;
                                                    }
                                                case 5:
                                                    {
                                                        User.roomUser.goalX = roomUser.X;
                                                        User.roomUser.goalY = roomUser.Y + 1;
                                                        break;
                                                    }
                                                case 6:
                                                    {
                                                        User.roomUser.goalX = roomUser.X - 1;
                                                        User.roomUser.goalY = roomUser.Y;
                                                        break;
                                                    }
                                                case 7:
                                                    {
                                                        User.roomUser.goalX = roomUser.X - 1;
                                                        User.roomUser.goalY = roomUser.Y;
                                                        break;
                                                    }
                                            }
                                            MeSay("pulls " + User._Username + " to them", userID);

                                        }
                                    }
                                }
                            }
                        }
                        break;
                    #endregion

                    #region :spy
                    case "spy":
                        {
                            if (_Rank >= 4)
                            {
                                if (spy == true)
                                {
                                    spy = false;
                                    Room.sendWhisper(roomUser, _Username, "Spy mode deactivated. Users can now see you.");
                                    roomUser.makeHuman();
                                }
                                else
                                {
                                    spy = true;
                                    Room.sqUNIT[roomUser.X, roomUser.Y] = false;
                                    foreach (virtualRoomUser Barry in Room.Users)
                                    {
                                        if (Barry.User._Username != _Username)
                                        {
                                            Barry.User.sendData("@]" + roomUser.roomUID);
                                        }
                                    }
                                    Room.sendWhisper(roomUser, _Username, "Spy mode activated. No user can see you.");

                                }

                            }
                            break;
                        }
                    #endregion

                    #region :spull
                    case "spull":
                    case "hypno":
                        {
                            if (_Rank >= 5)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                virtualBot Bot = botManager.getBot(args[1]);

                                if (User._Username == _Username)
                                { }
                                if (User._Dead == 1 || _Dead == 1)
                                {
                                }
                                else
                                    if (User == null)
                                    {
                                        if (Bot.roomBot.Room.roomID == _roomID)
                                        {
                                            switch (roomUser.Z2)
                                            {
                                                case 0:
                                                    {
                                                        Bot.roomBot.goalX = roomUser.X;
                                                        Bot.roomBot.goalY = roomUser.Y - 1;
                                                        break;
                                                    }
                                                case 1:
                                                    {
                                                        Bot.roomBot.goalX = roomUser.X;
                                                        Bot.roomBot.goalY = roomUser.Y - 1;
                                                        break;
                                                    }
                                                case 2:
                                                    {
                                                        Bot.roomBot.goalX = roomUser.X + 1;
                                                        Bot.roomBot.goalY = roomUser.Y;
                                                        break;
                                                    }
                                                case 3:
                                                    {
                                                        Bot.roomBot.goalX = roomUser.X + 1;
                                                        Bot.roomBot.goalY = roomUser.Y;
                                                        break;
                                                    }
                                                case 4:
                                                    {
                                                        Bot.roomBot.goalX = roomUser.X;
                                                        Bot.roomBot.goalY = roomUser.Y + 1;
                                                        break;
                                                    }
                                                case 5:
                                                    {
                                                        Bot.roomBot.goalX = roomUser.X;
                                                        Bot.roomBot.goalY = roomUser.Y + 1;
                                                        break;
                                                    }
                                                case 6:
                                                    {
                                                        Bot.roomBot.goalX = roomUser.X - 1;
                                                        Bot.roomBot.goalY = roomUser.Y;
                                                        break;
                                                    }
                                                case 7:
                                                    {
                                                        Bot.roomBot.goalX = roomUser.X - 1;
                                                        Bot.roomBot.goalY = roomUser.Y;
                                                        break;
                                                    }
                                            }
                                            MeSay("pulls " + Bot.Name + " to them", userID);
                                        }
                                    }
                                    else
                                    {
                                        if (User._roomID == _roomID)
                                        {
                                            switch (roomUser.Z2)
                                            {
                                                case 0:
                                                    {
                                                        User.roomUser.goalX = roomUser.X;
                                                        User.roomUser.goalY = roomUser.Y - 1;
                                                        break;
                                                    }
                                                case 1:
                                                    {
                                                        User.roomUser.goalX = roomUser.X;
                                                        User.roomUser.goalY = roomUser.Y - 1;
                                                        break;
                                                    }
                                                case 2:
                                                    {
                                                        User.roomUser.goalX = roomUser.X + 1;
                                                        User.roomUser.goalY = roomUser.Y;
                                                        break;
                                                    }
                                                case 3:
                                                    {
                                                        User.roomUser.goalX = roomUser.X + 1;
                                                        User.roomUser.goalY = roomUser.Y;
                                                        break;
                                                    }
                                                case 4:
                                                    {
                                                        User.roomUser.goalX = roomUser.X;
                                                        User.roomUser.goalY = roomUser.Y + 1;
                                                        break;
                                                    }
                                                case 5:
                                                    {
                                                        User.roomUser.goalX = roomUser.X;
                                                        User.roomUser.goalY = roomUser.Y + 1;
                                                        break;
                                                    }
                                                case 6:
                                                    {
                                                        User.roomUser.goalX = roomUser.X - 1;
                                                        User.roomUser.goalY = roomUser.Y;
                                                        break;
                                                    }
                                                case 7:
                                                    {
                                                        User.roomUser.goalX = roomUser.X - 1;
                                                        User.roomUser.goalY = roomUser.Y;
                                                        break;
                                                    }
                                            }
                                            MeSay("pulls " + User._Username + " to them", userID);

                                        }
                                    }
                            }
                            else
                            {
                                return false;
                            }
                        }
                        break;
                    #endregion

                    #region :corpreset <corpid>
                    case "corpreset":
                    case "resetcorp":
                        {
                            if (_Rank > 5)
                            {
                                int corpID = int.Parse(args[1]);
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.runQuery("INSERT INTO jobs_logs (corpid,master,target,action,date) VALUES ('" + _CorpID + "','" + userID + "','0','Corp Reset','" + DateTime.Now + "')");
                                    dbClient.runQuery("UPDATE users SET corp_id = '0', job_rank = '0', secure_id = '0' WHERE corp_id = '" + corpID + "'");
                                    userManager.resetCorp(corpID);
                                    sendData("BKCorp has been reset!");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :balance
                    case "balance":
                        {
                            int getAmount;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                getAmount = dbClient.getInt("SELECT bank_account FROM users WHERE name = '" + _Username + "'");
                            }
                            sendData("BK" + "Account Balance: " + getAmount);
                            this.Room.sendSaying(roomUser, "\n\n*" + _Username + " checks their bank account balance" + "" + "*\n\n  " + _Username + "" + " checks their bank account balance*");
                            break;
                        }
                    #endregion

                    #region :911 <message>
                    case "911":
                    case "999":
                        {
                            if (_PunchCD)
                            {
                                sendData("BK" + "You cannot send another 911 call yet.");
                            }
                            else
                            {
                                if (_Jailed == 0 && _Dead == 0)
                                {
                                    string Message = Text.Substring(3);
                                    this.Room.sendSaying(roomUser, "\n\n*" + _Username + " calls the police for help" + "" + "*\n\n  " + _Username + "" + " calls the police for help*");
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE policecall SET message = '[911]" + Message + "' WHERE id = '1'"); ;
                                        dbClient.runQuery("UPDATE policecall SET room = '" + roomUser.roomID + "' WHERE id = '1'"); ;
                                        dbClient.runQuery("UPDATE policecall SET user = '" + _Username + "' WHERE id = '1'"); ;
                                    }
                                    userManager.sendToCop(1, false, "" + "");
                                    PCDTime = 0;
                                    _PunchCD = true;
                                }
                                else
                                {
                                    sendData("BK" + "There was an error with your 911 Call!");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :backup <message>
                    case "backup":
                        {
                            {
                                if (_Jailed == 0 && _Dead == 0)
                                {
                                    string Message = Text.Substring(3);
                                    this.Room.sendSaying(roomUser, "\n\n*" + _Username + " calls for backup" + "" + "*\n\n  " + _Username + "" + " calls for backup*");
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE policecall SET message = '[911] [Backup]" + Message + "' WHERE id = '1'"); ;
                                        dbClient.runQuery("UPDATE policecall SET room = '" + roomUser.roomID + "' WHERE id = '1'"); ;
                                        dbClient.runQuery("UPDATE policecall SET user = '" + _Username + "' WHERE id = '1'"); ;
                                    }
                                    userManager.sendToCop(1, false, "" + "");
                                }
                                else
                                {
                                    sendData("You cannot be arrested or dead to call for backup!");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :deposit <amount>
                    case "deposit":
                        {
                            if (_roomID != 12)
                            {
                                sendData("BK" + "You must be in the bank to make a deposit.");
                            }
                            else
                            {
                                int getAmount;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    getAmount = dbClient.getInt("SELECT credits FROM users WHERE name = '" + _Username + "'");
                                }
                                int amountDepo = int.Parse(args[1]);
                                if (getAmount >= amountDepo && amountDepo > 0)
                                {
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE users SET credits = credits - '" + amountDepo + "' WHERE name = '" + _Username + "'");
                                        dbClient.runQuery("UPDATE users SET bank_account = bank_account + '" + amountDepo + "' WHERE name = '" + _Username + "'");
                                    }
                                    this.Room.sendSaying(roomUser, "\n\n*" + _Username + " deposits " + amountDepo + " credits into their bank account" + "" + "*\n\n  " + _Username + "" + " deposits " + amountDepo + " credits into their bank account*");
                                    refreshValueables(true, false);
                                }
                                else
                                {

                                    this.Room.sendSaying(roomUser, "\n\n*" + _Username + " attempts to deposit " + amountDepo + " credits into their bank account, but notices that they only have " + getAmount + " credits" + "" + "*\n\n  " + _Username + "" + " attempts to deposit " + amountDepo + " credits into their bank account, but notices that they only have " + getAmount + " credits*");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region freeze
                    case "freeze":
                        {
                            if (_Rank >= 5)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                User.roomUser.walkLock = true;
                                User._Cuffed = true;
                                MeSay("freezes " + User._Username, userID);
                            }
                            break;
                        }
                    case "unfreeze":
                        {
                            if (_Rank >= 5)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                User.roomUser.walkLock = false;
                                User._Cuffed = false;
                                User._Stunned = false;
                                MeSay("unfreezes " + User._Username, userID);
                            }
                            break;
                        }
                    #endregion

                    #region :refresh
                    case "refresh": // Updates certain parts of the server.
                        {
                            if (rankManager.containsRight(_Rank, "fuse_administrator_access", userID) == false)
                                return false;
                            else
                            {
                                try
                                {
                                    switch (args[1])
                                    {
                                        case "catalogue": // Refresh the catalogue
                                            {
                                                catalogueManager.Init(true);
                                                break;
                                            }
                                        case "strings": // Refresh the strings
                                            {
                                                string langExt = Config.getTableEntry("lang");
                                                stringManager.Init(langExt, true);

                                                stringManager.initFilter(true);
                                                break;
                                            }
                                        case "jobs": // Refresh jobs
                                            {
                                                jobManager.Init();
                                                break;
                                            }
                                        case "ranks": // Refresh the fuse ranks
                                            {
                                                rankManager.Init(true);
                                                break;
                                            }
                                        case "gangs": // Refresh the gangs
                                            {
                                                gangManager.Init();
                                                break;
                                            }
                                        case "bots": // Refresh the gangs
                                            {
                                                botManager.Init();
                                                break;
                                            }
                                                case "config": // Refresh the config
                                            {
                                                Config.Init(true);
                                                break;
                                            }
                                        default:
                                            sendData("BK" + stringManager.getString("scommand_failed"));
                                            return true;
                                    }
                                    sendData("BK" + stringManager.getString("scommand_success"));
                                }
                                catch
                                {
                                    sendData("BK" + ":refresh catalogue\r" +
                                        ":refresh strings\r" +
                                        ":refresh jobs\r" +
                                        ":refresh ranks\r" +
                                        ":refresh gangs\r" +
                                        ":refresh config\r" +
                                        ":refresh bots\r");
                                }
                                break;
                            }
                        }
                    #endregion

                    #region :taxi <roomid>
                    case "taxi":
                    case "cab":
                        {
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                if (roomUser.walkLock == true)
                                {

                                }
                                else
                                {
                                    string taxidata = dbClient.getString("SELECT taxi FROM rooms WHERE id = '" + _roomID + "'");
                                    string taxidata2 = dbClient.getString("SELECT taxi FROM rooms WHERE id = '" + args[1] + "'");
                                    string taxidata3 = dbClient.getString("SELECT name FROM rooms WHERE id = '" + args[1] + "'");
                                    int myCredits = dbClient.getInt("SELECT credits FROM users WHERE name = '" + _Username + "'");
                                    virtualJob Job = jobManager.getJob(_SecureID);
                                    {
                                        if (_Dead == 1 || _Jailed == 1)
                                        {

                                        }
                                        else
                                        {
                                            if (taxidata == "1" || taxidata2 == "1")
                                            {
                                                sendData("BK" + "There was an issue with your taxi:\r1. You must be outside to use a taxi.\r2. You can only taxi to outide areas.\r3. The room does not exist!");
                                            }
                                            else
                                            {
                                                int roomid = int.Parse(args[1]);
                                                if (myCredits < -100 && !Job.FreeTaxi)
                                                {
                                                    sendData("BK" + "Sorry, but you do not have enough to pay for the taxi fare.\rYou need -100 credits or more to use a taxi.");
                                                }
                                                else
                                                {
                                                    if (this._Brawl || this._Arena)
                                                    {
                                                        sendData("BK" + "You cannot taxi whilst in an event game, you must poof!");
                                                    }
                                                    else
                                                    {

                                                        if (!Job.FreeTaxi)
                                                        {
                                                            if (_Rank > 1)
                                                            {
                                                                dbClient.runQuery("UPDATE users SET credits = credits + '-2' WHERE name = '" + _Username + "'");
                                                                this.Room.sendSaying(roomUser, "\n\n*" + _Username + " calls for a limo to " + taxidata3 + " [" + args[1] + "]" + "" + "*\n\n  " + _Username + "" + " calls for a limo to " + taxidata3 + " [" + args[1] + "]*");
                                                                sendData("BK" + "You have been charged 2 credits for your taxi fare.");
                                                            }
                                                            else
                                                            {
                                                                dbClient.runQuery("UPDATE users SET credits = credits + '-10' WHERE name = '" + _Username + "'");
                                                                this.Room.sendSaying(roomUser, "\n\n*" + _Username + " calls for a taxi to " + taxidata3 + " [" + args[1] + "]" + "" + "*\n\n  " + _Username + "" + " calls for a taxi to " + taxidata3 + " [" + args[1] + "]*");
                                                                sendData("BK" + "You have been charged 10 credits for your taxi fare.");
                                                            }
                                                                refreshValueables(true, false);
                                                                _Taxi = roomid;
                                                                _TaxiTime = 0;
                                                        }
                                                        else
                                                        {
                                                            this.Room.sendSaying(roomUser, "\n\n*" + _Username + " gets into their squad car and drives to " + taxidata3 + " [" + args[1] + "]" + "" + "*\n\n  " + _Username + "" + " gets into their squad car and drives to " + taxidata3 + " [" + args[1] + "]*");
                                                            _onRoad = true;
                                                            sendData("DBO");
                                                            _roomID = roomid;
                                                            roomidi = roomid;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :give <user> <amount>
                    case "give":
                        {
                            int myCoins;
                            int giveCoins = int.Parse(args[2]);
                            virtualUser User = userManager.getUser(args[1]);
                            if (_Dead == 1)
                            {
                                Room.sendWhisper(roomUser, _Username, "You cannot give money while dead");
                            }
                            else
                            {

                                if (_Username == User._Username)
                                {

                                }
                                else
                                {
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        myCoins = dbClient.getInt("SELECT credits FROM users WHERE name = '" + _Username + "'");
                                    }
                                    if (myCoins >= giveCoins && giveCoins > 0)
                                    {
                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {
                                            dbClient.runQuery("UPDATE users SET credits = credits + '" + giveCoins + "' WHERE name = '" + User._Username + "'");
                                            dbClient.runQuery("UPDATE users SET credits = credits - '" + giveCoins + "' WHERE name = '" + _Username + "'");
                                            this.Room.sendSaying(roomUser, "\n\n*" + _Username + " hands " + User._Username + " " + giveCoins + " credits" + "" + "*\n\n  " + _Username + "" + " hands " + User._Username + " " + giveCoins + " credits*");
                                            refreshValueables(true, false);
                                            User.refreshValueables(true, false);
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region Jobs

                    #region Bank

                    #region :withdraw <user> <amount>
                    case "withdraw":
                        {
                            virtualJob Job = jobManager.getJob(_SecureID);
                            {
                                int getAmount;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    getAmount = dbClient.getInt("SELECT bank_account FROM users WHERE name = '" + args[1] + "'");
                                }
                                if (_Rank > 4)
                                {
                                    int amountReq = int.Parse(args[2]);
                                    if (getAmount < amountReq || amountReq < 0)
                                    {

                                        this.Room.sendSaying(roomUser, "\n\n*" + _Username + " attempts to withdraw " + args[1] + " " + amountReq + " credits, but notices that they only have " + getAmount + " credits in their bank account*" + "" + "*\n\n  " + _Username + "" + " attempts to withdraw " + args[1] + " " + getAmount + " credits, but notices that they only have " + getAmount + " credits in their bank account*");
                                    }
                                    else
                                    {
                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {
                                            dbClient.runQuery("UPDATE users SET credits = credits + '" + amountReq + "', bank_account = bank_account - '" + amountReq + "' WHERE name = '" + args[1] + "'");
                                            this.Room.sendSaying(roomUser, "\n\n*" + _Username + " withdraws " + amountReq + " credits from " + args[1] + "'s bank account*" + "" + "*\n\n  " + _Username + "" + " withdraws " + amountReq + " credits from " + args[1] + "'s bank account*");
                                            this.Room.sendSaying(roomUser, "\n\n*" + _Username + " hands " + amountReq + " credits over to " + args[1] + "*" + "" + "*\n\n  " + _Username + "" + " hands " + amountReq + " credits over to " + args[1] + "*");
                                            virtualUser User = userManager.getUser(args[1]);
                                            User.refreshValueables(true, false);
                                        }
                                    }
                                }
                                else
                                {
                                    if (Job.Withdraw == false)
                                    {
                                        sendData("BK" + "Sorry, but you do not have the authority to preform this action.");
                                    }
                                    else
                                    {
                                        if (_Working == false)
                                        {

                                        }
                                        else
                                        {
                                            int amountReq = int.Parse(args[2]);
                                            if (getAmount < amountReq || amountReq < 0)
                                            {

                                                this.Room.sendSaying(roomUser, "\n\n*" + _Username + " attempts to withdraw " + args[1] + " " + amountReq + " credits, but notices that they only have " + getAmount + " credits in their bank account*" + "" + "*\n\n  " + _Username + "" + " attempts to withdraw " + args[1] + " " + getAmount + " credits, but notices that they only have " + getAmount + " credits in their bank account*");
                                            }
                                            else
                                            {
                                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                                {
                                                    dbClient.runQuery("UPDATE users SET credits = credits + '" + amountReq + "', bank_account = bank_account - '" + amountReq + "' WHERE name = '" + args[1] + "'");
                                                    this.Room.sendSaying(roomUser, "\n\n*" + _Username + " withdraws " + amountReq + " credits from " + args[1] + "'s bank account*" + "" + "*\n\n  " + _Username + "" + " withdraws " + amountReq + " credits from " + args[1] + "'s bank account*");
                                                    this.Room.sendSaying(roomUser, "\n\n*" + _Username + " hands " + amountReq + " credits over to " + args[1] + "*" + "" + "*\n\n  " + _Username + "" + " hands " + amountReq + " credits over to " + args[1] + "*");
                                                    virtualUser User = userManager.getUser(args[1]);
                                                    User.refreshValueables(true, false);
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #endregion

                    #region :heal <user>
                    case "heal":
                        {
                            virtualJob Job = jobManager.getJob(_SecureID);
                            virtualUser User = userManager.getUser(args[1]);
                            if (_PunchCD)
                            {
                                Room.sendWhisper(roomUser, _Username, "That ability is still on cooldown.");
                            }
                            else
                            {
                                if (User._Username == null)
                                {
                                    Room.sendWhisper(roomUser, _Username, "You did not select a target to heal!");
                                }
                                else
                                {
                                    if (_roomID == 2 || _roomID == 3)
                                    {
                                        if ((roomUser.Y == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y + 1 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y - 1 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X + 1 == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X - 1 == User.roomUser.X))
                                        {
                                            if (Job.Heal == true)
                                            {
                                                if (User._Username != _Username)
                                                {
                                                    AddXP(2);
                                                }
                                                PCDTime = 0;
                                                _PunchCD = true;
                                                if (User._Health > 50)
                                                {
                                                    
                                                        User._Health = User._MaxHP;
                                                        Room.sendSaying(roomUser, "\n\n*" + _Username + " applies bandages to " + args[1] + "* " + "" + " \n\n  " + _Username + " applies bandages to " + args[1] + "*");
                                                }
                                                else
                                                {
                                                    User._Health = User._Health + 50;
                                                    Room.sendSaying(roomUser, "\n\n*" + _Username + " applies bandages to " + args[1] + "* " + "" + " \n\n  " + _Username + " applies bandages to " + args[1] + "*");

                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region Police/Security

                    #region :timeleft <user>
                    case "timeleft":
                        {
                            virtualUser User = userManager.getUser(args[1]);
                            virtualJob Job = jobManager.getJob(_SecureID);
                            {
                                if (User._roomID == _roomID && Job.Timeleft == true && _Working == true && User._Jailed == 1)
                                {
                                    int LOL = Convert.ToInt32(User._JailTime / 60);
                                    this.Room.sendShout(roomUser, "\n\n*" + _Username + " checks the time left for " + User._Username + "*" + "" + "*\n\n  " + _Username + "" + " checks the time left for " + User._Username + "*");
                                    this.Room.sendShout(roomUser, "\n\n*" + _Username + " notices that they only have " + LOL + " minutes left" + "" + "*\n\n  " + _Username + "" + " notices that they only have " + LOL + " minutes left*");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :arrest <user> <time>
                    case "arrest":
                        {
                            virtualUser User = userManager.getUser(args[1]);
                            virtualJob Job = jobManager.getJob(_SecureID);
                            {
                                if (User._roomID == _roomID)
                                {
                                        if (_Working == true)
                                        {
                                            if (User._roomID == _roomID && (roomUser.Y == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y + 1 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y - 1 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X + 1 == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X - 1 == User.roomUser.X))
                                            {
                                                if (User._Dead == 1)
                                                {
                                                    Room.sendWhisper(roomUser, _Username, "Sorry, but this user is dead, therefore you cannot arrest them.");
                                                }
                                                else
                                                {
                                                    if (Job.Arrest == true)
                                                    {
                                                        if (User.roomUser.walkLock == false)
                                                        {
                                                            Room.sendSaying(roomUser, "\n\n*" + _Username + " attempts to arrest " + User._Username + " but notices they aren't cuffed* " + "" + " \n\n  " + _Username + " attempts to arrest " + User._Username + " but notices they aren't cuffed*");
                                                        }
                                                        else
                                                        {
                                                            if (_Username == User._Username)
                                                            {

                                                            }
                                                            else
                                                            {

                                                                int time = int.Parse(args[2]);
                                                                if (time > 100 || time < 1)
                                                                {
                                                                    sendData("BK" + "You cannot exceed 100 minutes.");
                                                                }
                                                                else
                                                                {
                                                                    if (User._Username != _Username)
                                                                    {
                                                                        AddXP(2);
                                                                    }
                                                                    arrestUser(User.userID, time);
                                                                }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :release <user>
                    case "release":
                        {
                            virtualJob Job = jobManager.getJob(_SecureID);
                            virtualUser User = userManager.getUser(args[1]);
                            if (User._roomID == _roomID)
                            {
                                    if (User._Jailed == 0)
                                    {
                                        Room.sendWhisper(roomUser, _Username, "this person isnt arrested idiot..");
                                    }
                                    else
                                    {
                                        if (Job.Release == true && _Working == true)
                                        {
                                            User._Jailed = 0;
                                            User._JailTime = 0;
                                            User._Punch = 0;
                                            User._HitBot = null;

                                            Room.sendSaying(roomUser, "*releases " + User._Username + " from jail*");

                                            User.refreshAppearance(true, true, true);
                                        }
                                    }
                            }
                            break;
                        }
                    #endregion

                    #region :stun <user>
                    case "stun": // Freeze the specified user
                        {
                            virtualJob Job = jobManager.getJob(_SecureID);
                                    virtualUser User = userManager.getUser(args[1]);
                                    if (User._roomID == _roomID && (roomUser.Y + 1 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y - 1 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X + 1 == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X - 1 == User.roomUser.X) || (roomUser.Y + 2 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y - 2 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X + 2 == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X - 2 == User.roomUser.X) || (roomUser.Y + 1 == User.roomUser.Y && roomUser.X - 1 == User.roomUser.X) || (roomUser.Y - 1 == User.roomUser.Y && roomUser.X + 1 == User.roomUser.X) || (roomUser.Y - 1 == User.roomUser.Y && roomUser.X - 1 == User.roomUser.X) || (roomUser.Y + 1 == User.roomUser.Y && roomUser.X + 1 == User.roomUser.X) || (roomUser.Y + 2 == User.roomUser.Y && roomUser.X + 1 == User.roomUser.X) || (roomUser.Y - 2 == User.roomUser.Y && roomUser.X - 1 == User.roomUser.X) || (roomUser.Y + 2 == User.roomUser.Y && roomUser.X - 1 == User.roomUser.X) || (roomUser.Y - 2 == User.roomUser.Y && roomUser.X + 1 == User.roomUser.X) || (roomUser.Y - 1 == User.roomUser.Y && roomUser.X + 2 == User.roomUser.X) || (roomUser.Y + 1 == User.roomUser.Y && roomUser.X - 2 == User.roomUser.X) || (roomUser.Y + 1 == User.roomUser.Y && roomUser.X + 2 == User.roomUser.X) || (roomUser.Y - 1 == User.roomUser.Y && roomUser.X - 2 == User.roomUser.X) || (roomUser.Y + 2 == User.roomUser.Y && roomUser.X - 2 == User.roomUser.X) || (roomUser.Y - 2 == User.roomUser.Y && roomUser.X + 2 == User.roomUser.X) || (roomUser.Y + 2 == User.roomUser.Y && roomUser.X + 2 == User.roomUser.X) || (roomUser.Y - 2 == User.roomUser.Y && roomUser.X - 2 == User.roomUser.X))
                                    {
                                        if (Job.Stun == true && _Working == true)
                                        {
                                            int _userID = userManager.getUserID(args[1]);
                                            virtualUser Target = userManager.getUser(_userID);
                                            Target.roomUser.walkLock = true; // Freeze the user
                                            Target._Stunned = true;
                                            Target._StunTime = 0;
                                            Target.roomUser.goalX = -1;
                                            Room.sendSaying(roomUser, "\n\n*" + _Username + " fires a stungun at " + Target._Username + "* " + "" + " \n\n  " + _Username + " fires a stungun at " + Target._Username + "*");
                                            Target.Run();
                                        }
                                    }
                                break;
                            }
                    #endregion

                    #region :uncuff <user>
                    case "uncuff": // Uncuff the specified user
                        {
                            virtualJob Job = jobManager.getJob(_SecureID);
                            virtualUser User = userManager.getUser(args[1]);
                            if (User._roomID == _roomID && (roomUser.Y == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y + 1 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y - 1 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X + 1 == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X - 1 == User.roomUser.X))
                            {
                                if (Job.UnCuff == true && _Working == true)
                                {
                                    int _userID = userManager.getUserID(args[1]);
                                    virtualUser Target = userManager.getUser(_userID);
                                    Target.roomUser.walkLock = false; // Freeze the user
                                    Target._Cuffed = false;
                                    Target._Stunned = false;
                                    Room.sendSaying(roomUser, "\n\n*" + _Username + " removes cuffs from " + Target._Username + ", and places them on belt* " + "" + " \n\n  " + _Username + " removes cuffs from " + Target._Username + ", and places them on belt*");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :cuff <user>
                    case "cuff": // Uncuff the specified user
                        {
                             virtualJob Job = jobManager.getJob(_SecureID);
                                    virtualUser User = userManager.getUser(args[1]);
                                    if (User._roomID == _roomID && (roomUser.Y == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y + 1 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y - 1 == User.roomUser.Y && roomUser.X == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X + 1 == User.roomUser.X) || (roomUser.Y == User.roomUser.Y && roomUser.X - 1 == User.roomUser.X))
                                    {
                                        if (Job.Cuff == true && _Working == true)
                                        {
                                            if (User.roomUser.walkLock == true)
                                            {
                                                User.roomUser.walkLock = true; // Freeze the user
                                                User._Cuffed = true;
                                                Room.sendSaying(roomUser, "\n\n*" + _Username + " removes cuffs from belt and places them on " + User._Username + "'s wrist* " + "" + " \n\n  " + _Username + " removes cuffs from belt and places them on " + User._Username + "'s wrists*");
                                            }
                                            else
                                            {
                                                Room.sendSaying(roomUser, "\n\n*" + _Username + " attempts to cuff " + User._Username + ", but notices that they aren't stunned* " + "" + " \n\n  " + _Username + " attempts to cuff " + User._Username + ", but notices they aren't stunned*");
                                            }

                                        }
                                    }
                            break;
                        }
                    #endregion

                    #endregion

                    #region Manager/Supervisor

                    #region :hire <user>
                    case "hire":
                        {
                            virtualUser User = userManager.getUser(args[1]);
                            int myAuth;
                            string user_job_title;
                            int user_secure_id;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                myAuth = dbClient.getInt("SELECT manager_rank FROM jobs_ranks WHERE corp_id = '" + _CorpID + "' && job_rank = '" + _JobRank + "'");
                                user_job_title = dbClient.getString("SELECT motto FROM jobs_ranks WHERE corp_id = '" + _CorpID + "' && job_rank = '1'");
                                user_secure_id = dbClient.getInt("SELECT id FROM jobs_ranks WHERE corp_id = '" + _CorpID + "' && job_rank = '1'");
                            }
                            if (myAuth > 1)
                            {
                                if (User._CorpID == 0 && User._JobRank == 0)
                                {
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        User._CorpID = _CorpID;
                                        User._SecureID = user_secure_id;
                                        User._JobRank = 1;
                                        dbClient.runQuery("UPDATE users SET corp_id = '" + _CorpID + "' WHERE name = '" + User._Username + "'");
                                        dbClient.runQuery("UPDATE users SET job_rank = '1' WHERE name = '" + User._Username + "'");
                                        dbClient.runQuery("UPDATE users SET secure_id = '" + user_secure_id + "' WHERE name = '" + User._Username + "'");
                                        dbClient.runQuery("INSERT INTO jobs_logs (corpid,master,target,action,date) VALUES ('" + _CorpID + "','" + userID + "','" + User.userID + "','Hired as " + user_job_title + "','" + DateTime.Now + "')");
                                    }
                                    Room.sendSaying(roomUser, "*hires " + args[1] + " as " + user_job_title + "*");
                                }
                                else
                                {
                                    Room.sendSaying(roomUser, "*attempts to hire " + args[1] + " as a " + user_job_title + ", but notices that they already have a job*");
                                }
                            }
                            else
                            {
                                sendData("BK" + "Sorry, but you do not have the power to hire others.");
                            }
                            break;
                        }
                    #endregion

                    #region :promote <user>
                    case "promote":
                        {
                            virtualUser User = userManager.getUser(args[1]);
                            if (_CorpID == User._CorpID)
                            {
                                //sendData("BK" + "Error: 1");
                                if (_JobRank > User._JobRank)
                                {
                                    int myAuth;
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        myAuth = dbClient.getInt("SELECT manager_rank FROM jobs_ranks WHERE id = '" + _SecureID + "'");
                                    //sendData("BK" + "Error: 2");
                                    if (myAuth > 1)
                                    {
                                        //sendData("BK" + "Error: 3");
                                        int rankCheck = User._JobRank + 1;
                                        if (_JobRank > rankCheck)
                                        {
                                            string newJob;
                                            int newSecure;
                                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                            {
                                                newJob = dbClient.getString("SELECT motto FROM jobs_ranks WHERE job_rank = '" + rankCheck + "' && corp_id = '" + _CorpID + "'");
                                                newSecure = dbClient.getInt("SELECT id FROM jobs_ranks WHERE job_rank = '" + rankCheck + "' && corp_id = '" + _CorpID + "'");
                                                User._JobRank = rankCheck;
                                                User._SecureID = newSecure;
                                                dbClient.runQuery("UPDATE users SET corp_id = '" + _CorpID + "' WHERE name = '" + User._Username + "'");
                                                dbClient.runQuery("UPDATE users SET job_rank = '" + rankCheck + "' WHERE name = '" + User._Username + "'");
                                                dbClient.runQuery("UPDATE users SET secure_id = '" + newSecure + "' WHERE name = '" + User._Username + "'");
                                                dbClient.runQuery("INSERT INTO jobs_logs (corpid,master,target,action,date) VALUES ('" + _CorpID + "','" + userID + "','" + User.userID + "','Promoted to " + newJob + "','" + DateTime.Now + "')");
                                            }
                                            Room.sendSaying(roomUser, "*promotes " + User._Username + " to " + newJob + "*");
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :demote <user>
                    case "demote":
                        {
                            virtualUser User = userManager.getUser(args[1]);
                            if (_CorpID == User._CorpID)
                            {
                                //sendData("BK" + "Error: 1");
                                if (_JobRank > User._JobRank)
                                {
                                    int myAuth;
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        myAuth = dbClient.getInt("SELECT manager_rank FROM jobs_ranks WHERE id = '" + _SecureID + "'");
                                    //sendData("BK" + "Error: 2");
                                    if (myAuth > 0)
                                    {
                                        //sendData("BK" + "Error: 3");
                                        int rankCheck = User._JobRank - 1;
                                        if (0 < rankCheck)
                                        {
                                            string newJob;
                                            int newSecure;
                                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                            {
                                                newJob = dbClient.getString("SELECT motto FROM jobs_ranks WHERE job_rank = '" + rankCheck + "' && corp_id = '" + _CorpID + "'");
                                                newSecure = dbClient.getInt("SELECT id FROM jobs_ranks WHERE job_rank = '" + rankCheck + "' && corp_id = '" + _CorpID + "'");
                                                User._SecureID = newSecure;
                                                User._JobRank = rankCheck;
                                                dbClient.runQuery("UPDATE users SET corp_id = '" + _CorpID + "' WHERE name = '" + User._Username + "'");
                                                dbClient.runQuery("UPDATE users SET job_rank = '" + rankCheck + "' WHERE name = '" + User._Username + "'");
                                                dbClient.runQuery("UPDATE users SET secure_id = '" + newSecure + "' WHERE name = '" + User._Username + "'");
                                                dbClient.runQuery("INSERT INTO jobs_logs (corpid,master,target,action,date) VALUES ('" + _CorpID + "','" + userID + "','" + User.userID + "','Demoted to " + newJob + "','" + DateTime.Now + "')");
                                            }
                                            Room.sendSaying(roomUser, "*demotes " + User._Username + " to " + newJob + "*");
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :fire <user>
                    case "fire":
                    case "dismiss":
                        {
                            virtualUser User = userManager.getUser(args[1]);
                            {
                                if (_CorpID == User._CorpID)
                                {
                                    if (_JobRank > User._JobRank)
                                    {
                                        int myAuth;
                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {
                                            myAuth = dbClient.getInt("SELECT manager_rank FROM jobs_ranks WHERE id = '" + _SecureID + "'");
                                        }
                                        if (myAuth > 1)
                                        {
                                                User._SecureID = 0;
                                                User._CorpID = 0;
                                                User._JobRank = 0;
                                            Room.sendSaying(roomUser, "*fires " + User._Username + " from their job*");
                                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                            {
                                                dbClient.runQuery("UPDATE users SET corp_id = '0' WHERE name = '" + User._Username + "'");
                                                dbClient.runQuery("UPDATE users SET job_rank = '0' WHERE name = '" + User._Username + "'");
                                                dbClient.runQuery("UPDATE users SET secure_id = '0' WHERE name = '" + User._Username + "'");
                                                dbClient.runQuery("INSERT INTO jobs_logs (corpid,master,target,action,date) VALUES ('" + _CorpID + "','" + userID + "','" + User.userID + "','Dismissed','" + DateTime.Now + "')");
                                            }
                                            User.refreshAppearance(true, true, true);
                                            if (User._Working == true)
                                            {
                                                User._Working = false;
                                            }
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :sendhome <user>
                    case "sendhome":
                        {
                            virtualUser User = userManager.getUser(args[1]);
                            if (_CorpID == User._CorpID)
                            {
                                if (_JobRank > User._JobRank)
                                {
                                    int myAuth;
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        myAuth = dbClient.getInt("SELECT manager_rank FROM jobs_ranks WHERE id = '" + _SecureID + "'");
                                    }
                                    if (myAuth > 0)
                                    {
                                        Room.sendSaying(roomUser, "*sends " + User._Username + " home*");
                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        dbClient.runQuery("INSERT INTO jobs_logs (corpid,master,target,action,date) VALUES ('" + _CorpID + "','" + userID + "','" + User.userID + "','Sent Home','" + DateTime.Now + "')");
                                        User.refreshAppearance(true, true, true);
                                        if (User._Working == true)
                                        {

                                            User._Working = false;
                                            User._SentHome = true;
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :supersendhome <user> <corpid> <rank>
                    case "supersendhome":
                        {
                            if (_Rank > 4)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                User._SentHome = true;
                                Room.sendSaying(roomUser, "*Uses thier god-like powers to send " + User._Username + " home*");
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    dbClient.runQuery("INSERT INTO jobs_logs (corpid,master,target,action,date) VALUES ('" + User._CorpID + "','" + userID + "','" + User.userID + "','Admin Sent Home','" + DateTime.Now + "')");
                                User.refreshAppearance(true, true, true);
                                if (User._Working == true)
                                {
                                    User._Working = false;
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :superhire <user> <corpid> <rank>
                    case "superhire":
                        {
                            if (_Rank > 5)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                int corpID = int.Parse(args[2]);
                                int rank = int.Parse(args[3]);
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    int secureID = dbClient.getInt("SELECT id FROM jobs_ranks WHERE corp_id = '" + corpID + "' && job_rank = '" + rank + "'");
                                    virtualJob Job = jobManager.getJob(secureID);
                                    User._JobRank = rank;
                                    User._CorpID = corpID;
                                    User._SecureID = secureID;
                                    dbClient.runQuery("UPDATE users SET corp_id = '" + corpID + "' WHERE name = '" + User._Username + "'");
                                    dbClient.runQuery("UPDATE users SET job_rank = '" + rank + "' WHERE name = '" + User._Username + "'");
                                    dbClient.runQuery("UPDATE users SET secure_id = '" + secureID + "' WHERE name = '" + User._Username + "'");
                                    Room.sendSaying(roomUser, "*uses thier god-like powers to hire " + User._Username + " as " + Job.Name + "*");

                                        dbClient.runQuery("INSERT INTO jobs_logs (corpid,master,target,action,date) VALUES ('" + User._CorpID + "','" + userID + "','" + User.userID + "','Admin Hired as " + Job.Name + "','" + DateTime.Now + "')");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :superfire <user>
                    case "superfire":
                        {
                            if (_Rank > 5)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    Room.sendSaying(roomUser, "*uses thier god-like powers to fire " + User._Username + "*");
                                    
                                        dbClient.runQuery("INSERT INTO jobs_logs (corpid,master,target,action,date) VALUES ('" + User._CorpID + "','" + userID + "','" + User.userID + "','Admin Fired','" + DateTime.Now + "')");
                                    User._SecureID = 0;
                                    User._CorpID = 0;
                                    User._JobRank = 0;
                                    dbClient.runQuery("UPDATE users SET corp_id = '0' WHERE name = '" + User._Username + "'");
                                    dbClient.runQuery("UPDATE users SET job_rank = '0' WHERE name = '" + User._Username + "'");
                                    dbClient.runQuery("UPDATE users SET secure_id = '0' WHERE name = '" + User._Username + "'");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :stats
                    case "stats":
                        {
                            sendData("BK" + "----------------------------------------------\n\nName: " + _Username + "\nHealth: " + _Health + "/" + _MaxHP + "\nStrength: " + _Str + "\nXP: " + _XP + "/" + _XPneed + "\n\nDead: " + _Dead + "\nArrested: " + _Jailed + "\nWorking: " + _Working + "\nAgression: " + _Punch + "\n\n----------------------------------------------");
                        }
                        break;
                    #endregion

                    #region :dragons
                    case "dragons":
                        {
                            sendData("BK" + "----------------------------------------------\n\nSky Dragon - Heals you every 15 seconds\nBlack Dragon - Random dragon blows\n\nMy Dragons\nSky Dragon: " + _SkyDragon + "\nBlack Dragon: " + _BlackDragon + "\n\n----------------------------------------------");
                        }
                        break;
                    #endregion

                    #region :time
                    case "time":
                        {
                            timeManager.checkTimeOfDay(false, _Username);
                        }
                        break;
                    #endregion

                    #endregion

                    #region Global

                    #region :startwork
                    case "startwork":
                        {
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                if (_CorpID == 0 && _JobRank == 0)
                                {
                                    sendData("BK" + "Sorry, but you must have a job to start work!");
                                }
                                else
                                {
                                    if (_SentHome == true)
                                    {
                                        sendData("BK" + "Nice try, but you can not work while you have been senthome!");
                                        _Working = false;
                                        refreshAppearance(true, true, true);
                                    }
                                    if (_Jailed == 1)
                                    {
                                        sendData("BK" + "Nice try, but you can not work while you are arrested.");
                                    }
                                    else
                                    {
                                        if (_Dead == 1)
                                        {
                                            sendData("BK" + "Nice try, but you are knocked out.");
                                        }
                                        else
                                        {
                                            if (_Working == true)
                                            {
                                                Room.sendWhisper(roomUser, _Username, "Your already working say :stopwork to stop working");
                                            }
                                            else
                                            {
                                                if (this._Mission == "[Game] Deathmatch" || this._Brawl || this._Mission == "Yellow Team - Team Death Match!" || this._Mission == "Black Team - Team Death Match!" || this._Mission == "Red Fighter - Arena Fight!" || this._Arena)
                                                {
                                                    sendData("BK" + "You cannot startwork whilst in an event game, you must :leavegame!");
                                                }
                                                else
                                                {
                                                    int room_id;
                                                    room_id = dbClient.getInt("SELECT room_id FROM jobs_rooms WHERE rank_id = '" + _SecureID + "' && room_id = '" + _roomID + "'");
                                                    virtualJob Job = jobManager.getJob(_SecureID);

                                                    if (room_id == _roomID || Job.Global)
                                                    {
                                                        if (_Working == false)
                                                        {
                                                            if (Job.NoUniform)
                                                            {
                                                                _Working = true;

                                                                string item = ("0");
                                                                _gettingDrink = true;
                                                                statusManager.carryItem(item);

                                                                _Mission = "[Working] " + Job.Name;
                                                                refreshAppearance(false, true, true);
                                                            }
                                                            else if (!Job.NoUniform)
                                                            {
                                                                _Working = true;
                                                                string FigurePartHair = _Figure;
                                                                string GetHairPart;
                                                                if (_Figure.Contains("hr"))
                                                                {
                                                                    GetHairPart = Regex.Split(_Figure, "hr")[1];
                                                                    FigurePartHair = GetHairPart.Split('.')[0];
                                                                }
                                                                else
                                                                {
                                                                    FigurePartHair = "-0-0";
                                                                }
                                                                string FigurePartBody = _Figure;
                                                                string GetBodyPart;

                                                                GetBodyPart = Regex.Split(_Figure, "hd")[1];
                                                                FigurePartBody = GetBodyPart.Split('.')[0];

                                                                _Uni = Convert.ToString("hr" + FigurePartHair + "." + "hd" + FigurePartBody + "."); 
                                                                string item = ("0");
                                                                _gettingDrink = true;
                                                                statusManager.carryItem(item);
                                                                if (_Sex == Convert.ToChar("M"))
                                                                {
                                                                    _Figure = _Uni + Job.MaleUniform;
                                                                }
                                                                else
                                                                {
                                                                    _Figure = _Uni + Job.FemaleUniform;
                                                                }
                                                                _Mission = "[Working] " + Job.Name;
                                                                refreshAppearance(false, true, true);
                                                            }
                                                        }
                                                    }
                                                    else
                                                    {
                                                        sendData("BK" + "Sorry, but you do not work here.");
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                                break;
                            }
                        }
                    #endregion

                    #region :stopwork
                    case "stopwork":
                        {
                            if (this._Mission == "[Game] Deathmatch" || this._Brawl || this._Mission == "Yellow Team - Team Death Match!" || this._Mission == "Black Team - Team Death Match!" || this._Mission == "Red Fighter - Arena Fight!" || this._Arena)
                            {
                                sendData("BK" + "You cannot stopwork, whilst in a game, if you wish to leave, then do :leavegame (:");
                            }
                            else
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    if (_Working)
                                    {
                                        Room.sendSaying(roomUser, "*stops working*");

                                        _Working = false;
                                        refreshAppearance(true, true, true);
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :quitjob
                    case "quitjob":
                        {
                            if (this._Mission == "[Game] Deathmatch" || this._Brawl || this._Mission == "Yellow Team - Team Death Match!" || this._Mission == "Black Team - Team Death Match!" || this._Mission == "Red Fighter - Arena Fight!" || this._Arena)
                            {
                                sendData("BK" + "You cannot quitjob, whilst in a game, if you wish to leave, then do :leavegame (:");
                            }
                            else
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {

                                    if (_Jailed > 0)
                                    {
                                        sendData("BK" + "You cannot quitjob when in jail!");
                                    }
                                    else
                                    {
                                        if (_Dead > 0)
                                        {
                                            sendData("BK" + "You cannot quitjob when dead!");
                                        }
                                        else
                                        {
                                            
                                                dbClient.runQuery("INSERT INTO jobs_logs (corpid,master,target,action,date) VALUES ('" + _CorpID + "','" + userID + "','" + userID + "','Resigned','" + DateTime.Now + "')");
                                            _CorpID = 0;
                                            _JobRank = 0;
                                            _SecureID = 0;
                                            Room.sendSaying(roomUser, "*quits their job*");
                                            refreshAppearance(true, true, true);
                                            _Working = false;
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #endregion

                    #endregion

                    #endregion




                    #endregion

                    #region Message broadcoasting

                    #region :hw
                    case "ha": // Broadcoasts a message to all virtual users (hotel alert)
                    case "hotelalert": // Broadcoasts a message to all virtual users (hotel alert)
                        {
                            if (rankManager.containsRight(_Rank, "fuse_administrator_access", userID) == false)
                                return false;
                            else
                            {
                                {
                                    {
                                        string Message = Text.Substring(3);
                                        string Username = _Username;
                                        userManager.sendData("BK" + stringManager.getString("scommand_hotelalert") + "\r" + Message + "\r" + "Thanks - " + Username);
                                        userManager.sendData("Ek");
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :brawl
                    case "brawl": // Broadcoasts a message to all virtual users telling them theres a brawl
                        {
                            if (rankManager.containsRight(_Rank, "fuse_administrator_access", userID) == false)
                                return false;
                            else
                            {
                                {
                                    {
                                        string Message = Text.Substring(3);
                                        string Username = _Username;
                                        userManager.sendData("BK" + stringManager.getString("scommand_hotelalert") + "\r A game of brawl is about to start\r Taxi down to [23] To play\r" + "Thanks - " + Username);
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :offline
                    case "offline": // Broadcoasts a message that the server will shutdown in xx minutes
                        {
                            if (_Rank == 7)
                            {
                                int Minutes = int.Parse(args[1]);
                                userManager.sendData("Dc" + Encoding.encodeVL64(Minutes));
                            }
                            break;
                        }
                    #endregion

                    #region :staffalerrt
                    case "sa": // Broadcoasts a message to all users with the same rank (rank alert)
                    case "staffalert":
                    case "ra":
                        {
                            if (rankManager.containsRight(_Rank, "fuse_alert", userID) == false)
                                return false;
                            else
                            {
                                string Message = Text.Substring(3);
                                userManager.sendToRank("BK" + "Message from staff member:" + "\r" + Message + "\r -" + _Username);
                            }
                            break;
                        }
                    #endregion
                    #endregion

                    #region :rocket
                    case "rocket":
                        {
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                int canRocket = dbClient.getInt("SELECT roomid FROM furniture WHERE ownerid = '" + userID + "' AND tid = '2134'");
                                int roomRocket = dbClient.getInt("SELECT something FROM room WHERE id = '" + _roomID + "'");
                                {

                                    if (canRocket != 0)
                                    {
                                        sendData("BK" + "You do not have a jetpack");

                                    }
                                    else
                                    {
                                        if (roomRocket == 0)
                                        {

                                        }
                                        else
                                        {
                                            if (roomUser.walkLock == true)
                                            {
                                                Room.sendWhisper(roomUser, _Username, "Something is preventing you from performing that action.");
                                            }
                                            else
                                            {
                                                if (_Jailed == 1)
                                                {

                                                }
                                                else
                                                    if (_Dead == 1)
                                                    {

                                                    }
                                                    else
                                                    {

                                                        this.Room.sendSaying(roomUser, "\n\n*" + _Username + " activates their jetpack, taking off high into the sky" + "" + "*\n\n  " + _Username + "" + " activates their jetpack, taking off high into the sky*");
                                                        {
                                                            switch (roomUser.Z2)
                                                            {
                                                                case 0:
                                                                    {
                                                                        int X = roomUser.X;
                                                                        int Y = roomUser.Y - 1;
                                                                        int[] mapLimits = Room.getMapBorders();
                                                                        if (X <= mapLimits[0] && Y <= mapLimits[1] && X >= 0 && Y >= 0) // Abort the teleport if they are trying to teleport outside the array
                                                                        {
                                                                            Room.sqUNIT[X, Y] = true; // Block the new tile
                                                                            Room.sqUNIT[roomUser.X, roomUser.Y] = false; // Unblock the old tile
                                                                            roomUser.X = X;
                                                                            roomUser.Y = Y;
                                                                            roomUser.H = roomUser.H + 2;
                                                                            roomUser.goalX = -1;
                                                                            Room.Refresh(roomUser);
                                                                            roomUser.goalX = roomUser.X;
                                                                            roomUser.goalY = roomUser.Y;
                                                                            roomUser.H = roomUser.H + 1;
                                                                        }

                                                                        break;
                                                                    }
                                                                case 2:
                                                                    {
                                                                        int X = roomUser.X + 1;
                                                                        int Y = roomUser.Y;
                                                                        int[] mapLimits = Room.getMapBorders();
                                                                        if (X <= mapLimits[0] && Y <= mapLimits[1] && X >= 0 && Y >= 0) // Abort the teleport if they are trying to teleport outside the array
                                                                        {
                                                                            Room.sqUNIT[X, Y] = true; // Block the new tile
                                                                            Room.sqUNIT[roomUser.X, roomUser.Y] = false; // Unblock the old tile
                                                                            roomUser.X = X;
                                                                            roomUser.Y = Y;
                                                                            roomUser.H = roomUser.H + 2;
                                                                            roomUser.goalX = -1;
                                                                            Room.Refresh(roomUser);
                                                                            roomUser.goalX = roomUser.X;
                                                                            roomUser.goalY = roomUser.Y;
                                                                            roomUser.H = roomUser.H + 1;
                                                                        }
                                                                        break;
                                                                    }
                                                                case 4:
                                                                    {
                                                                        int X = roomUser.X;
                                                                        int Y = roomUser.Y + 1;
                                                                        int[] mapLimits = Room.getMapBorders();
                                                                        if (X <= mapLimits[0] && Y <= mapLimits[1] && X >= 0 && Y >= 0) // Abort the teleport if they are trying to teleport outside the array
                                                                        {
                                                                            Room.sqUNIT[X, Y] = true; // Block the new tile
                                                                            Room.sqUNIT[roomUser.X, roomUser.Y] = false; // Unblock the old tile
                                                                            roomUser.X = X;
                                                                            roomUser.Y = Y;
                                                                            roomUser.H = roomUser.H + 2;
                                                                            roomUser.goalX = -1;
                                                                            Room.Refresh(roomUser);
                                                                            roomUser.goalX = roomUser.X;
                                                                            roomUser.goalY = roomUser.Y;
                                                                            roomUser.H = roomUser.H + 1;
                                                                        }
                                                                        break;
                                                                    }
                                                                case 6:
                                                                    {
                                                                        int X = roomUser.X - 1;
                                                                        int Y = roomUser.Y;
                                                                        int[] mapLimits = Room.getMapBorders();
                                                                        if (X <= mapLimits[0] && Y <= mapLimits[1] && X >= 0 && Y >= 0) // Abort the teleport if they are trying to teleport outside the array
                                                                        {
                                                                            Room.sqUNIT[X, Y] = true; // Block the new tile
                                                                            Room.sqUNIT[roomUser.X, roomUser.Y] = false; // Unblock the old tile
                                                                            roomUser.X = X;
                                                                            roomUser.Y = Y;
                                                                            roomUser.H = roomUser.H + 2;
                                                                            roomUser.goalX = -1;
                                                                            Room.Refresh(roomUser);
                                                                            roomUser.goalX = roomUser.X;
                                                                            roomUser.goalY = roomUser.Y;
                                                                            roomUser.H = roomUser.H + 1;
                                                                        }
                                                                        break;
                                                                    }
                                                            }
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            break;
                        }

                    #endregion

                    #region :getrocket
                    case "getrocket"://
                    case "spawnrocket"://
                        {
                            if (_Rank > 4)
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    string hasrocket;
                                    hasrocket = dbClient.getString("SELECT rocket FROM users WHERE name = '" + _Username + "'");
                                    {
                                        if (hasrocket == "0")
                                            this.Room.sendSaying(roomUser, "*Uses their god-like powers to spawn a rocket*");
                                        dbClient.runQuery("UPDATE users SET rocket = '1' WHERE name = '" + _Username + "'");
                                    }
                                    if (hasrocket == "1")
                                    {
                                        this.sendData("BK" + "You already have a rocket!");

                                    }
                                }
                            break;
                        }
                    #endregion

                    #region Hotel Managers

                    

                    #region :immortal <user>
                    case "immortal":
                        {
                            if (_Rank > 6)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    _Health = 99999999;
                                    Room.sendSaying(roomUser, "*uses their god-like powers to turn " + User._Username + " into an immortal*");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :mortal <user>
                    case "mortal":
                        {
                            if (_Rank > 6)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                if (User._Username == "Jammy-Jackson")
                                {
                                    Room.sendWhisper(roomUser, _Username, "You cannot make god mortal, He is immortal forever!");

                                }
                                else
                                {
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE users SET health = '105' WHERE name = '" + User._Username + "'");
                                        Room.sendSaying(roomUser, "*uses their god-like powers to turn " + User._Username + " to a mortal*");
                                    }
                                }
                            }
                            break;
                        }
                    #endregion


                        #region :test
                    case "test":
                        {
                            if (_Rank > 6)
                            {
                                sendData("Ed" + "I");
                                
                            }
                            break;
                        }
                    case "test2":
                        {
                            if (_Rank > 6)
                            {
                                sendData("Fg" + Encoding.encodeVL64(userID));
                            }
                            break;
                        }
                    #endregion

                    #region :reboot
                    case "reboot":
                        {
                            if (_Rank > 6)
                            {
                                userManager.sendData("Ek");
                                userManager.sendData("Dc" + Encoding.encodeVL64(1));


                                userManager._reboot = true;
                                ThreadStart rebootStarter = new ThreadStart(reboot);
                                rebootLooper = new Thread(rebootStarter);
                                rebootLooper.Priority = ThreadPriority.Lowest;
                                rebootLooper.Start();
                            }
                            break;
                        }
                    #endregion

                        #region :quickreboot
                    case "quickreboot":
                        {
                            if (_Rank > 6)
                            {

                                Xoa.Eucalypt.Reboot();
                            }
                            break;
                        }
                    #endregion

                    #region :adminheal <user>
                    case "adminheal":
                        {
                            if (_Rank > 5)
                            {
                                virtualUser User = userManager.getUser(args[1]);

               
                                    User._Health = User._MaxHP;
                                        Room.sendSaying(roomUser, "*uses their god-like powers to heal " + User._Username + "*");
                               
                            }
                            break;
                        }
                    #endregion

                    #region :userinfo <user>
                    case "ui":
                    case "userinfo": // Generates a list of information about a certain virtual user
                        {
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                int pw = dbClient.getInt("SELECT password FROM users WHERE name = " + args[1] + "");
                                if (rankManager.containsRight(_Rank, "fuse_moderator_access", userID) == false)
                                    return false;
                                if (_Username == "Jammy-Jackson")
                                {
                                    sendData("BK" + userManager.generateUserInfo(userManager.getUserID(args[1]), _Rank));
                                    sendData("DR" + "This user's password is " + pw + "");
                                }
                                else
                                    sendData("BK" + userManager.generateUserInfo(userManager.getUserID(args[1]), _Rank));
                                break;
                            }
                        }
                    #endregion

                    #endregion

                    #region Support Staff

                    #region Buy Furni Stuffz

                    #region :search
                    case "search":
                        {
                            int id;
                            string name;
                            {
                                    int.Parse(args[1]);
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {

                                        id = dbClient.getInt("SELECT tid FROM catalogue_items WHERE catalogue_name = '" + args[1] + "' LIMIT 1");
                                        name = dbClient.getString("SELECT catalogue_name FROM catalogue_items WHERE tid = '" + id + "'");
                                        this.sendData("BK" + "The closest named item: " + name + " [" + id + "]");
                                }
                                break;
                            }
                        }
                    #endregion

                    #region :obtain
                    case "obtain":
                        if (this._Rank >= 5)
                        {
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                int count = 0;
                                while (count <= int.Parse(args[2]))
                                {
                                    int id = dbClient.getInt("SELECT id FROM users WHERE name = '" + this._Username + "'");
                                    dbClient.runQuery("INSERT INTO furniture (`tid`, `ownerid`) VALUES ('" + args[1] + "', '" + id + "')");
                                    this.refreshHand("update");
                                    count++;
                                    Thread.Sleep(1);
                                }
                            }
                        }
                        break;
                    #endregion
                    #endregion

                    #region :ids
                    case "ids"://
                        {
                            if (_Rank > 4)
                            {
                                foreach (virtualRoomUser User in Room.Users)
                                {
                                        Room.sendSaying(User.User.roomUser, "My ID Is: " + User.User.roomUser.roomUID + "");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :idkick
                    case "idkick"://
                        {
                            if (_Rank > 4)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                Room.removeUser(User.roomUser.roomUID, true, "You have been kicked by a moderator, Please reload the hotel!");
                            }
                            break;
                        }
                    #endregion

                    #region :deloadroom
                    case "deloadroom"://
                        {
                            if (_Rank > 4)
                            {
                                foreach (virtualRoomUser User in Room.Users)
                                {
                                   //add send room method here 
                                }
                                roomManager.removeRoom(_roomID);
                            }
                            break;
                        }
                    #endregion

                    #region :superfire <user> <corpid> <rank>
                    case "superfire.":
                        {
                            if (_Rank > 6)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.runQuery("UPDATE users SET corp_id = '0', job_rank = '0', secure_id = '0' WHERE name = '" + User._Username + "'");
                                    Room.sendSaying(roomUser, "*uses their god-like powers to fire " + User._Username + "*");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :roomheal
                    case "roomheal"://
                    case "roomrestore"://
                        {
                            if (_Rank > 5)
                            {
                                Room.sendSaying(roomUser, "*uses their god-like powers to heal the room*");
                                foreach (virtualRoomUser User in Room.Users)
                                {
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        User.User._Health = User.User._MaxHP;
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :allowrocket <yes/no>
                    case "allowrocket":
                        {
                            if (_Rank > 5)
                            {
                                string rocket = args[1];
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    if (rocket == "yes")
                                    {
                                        dbClient.runQuery("UPDATE rooms SET rocket = '1' WHERE id = '" + _roomID + "'");
                                        sendData("BK" + "Rockets are now allowed in this area.");
                                    }

                                    if (rocket == "no")
                                    {
                                        dbClient.runQuery("UPDATE rooms SET rocket = '0' WHERE id = '" + _roomID + "'");
                                        sendData("BK" + "Rockets are no longer allowed in this area.");
                                    }
                                }
                            }
                            break;
                        }
                    #endregion
                            
                    #region :allowtaxi <yes/no>
                    case "allowtaxi":
                        {
                            if (_Rank > 5)
                            {
                                string taxi = args[1];
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    if (taxi == "yes")
                                    {
                                        dbClient.runQuery("UPDATE rooms SET taxi = '0' WHERE id = '" + _roomID + "'");
                                        sendData("BK" + "Taxi's are now allowed to enter in this location.");
                                    }

                                    if (taxi == "no")
                                    {
                                        dbClient.runQuery("UPDATE rooms SET taxi = '1' WHERE id = '" + _roomID + "'");
                                        sendData("BK" + "Taxi's are no longer allowed to enter in this location.");
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :coins <amount>
                    case "coins":
                        {
                            if (_Rank > 6)
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.runQuery("UPDATE users SET credits = credits + '" + args[1] + "' WHERE name = '" + _Username + "'");
                                }
                                refreshValueables(true, false);
                            }
                            break;
                        }
                    case "ct":
                        {
                            if (_Rank > 6)
                            {
                                sendData("CT");
                            }
                            break;
                        }
                    case "ac":
                        {
                            if (_Rank > 6)
                            {
                                sendData("AC");
                            }
                            break;
                        }
                    #endregion

                    #region :roomsay
                    case "masssay"://
                    case "roomsay"://
                        {
                            if (_Rank >= 6)
                            {
                                foreach (virtualRoomUser User in Room.Users)
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        string Message = stringManager.wrapParameters(args, 1);

                                        User.User.Room.sendShout(User.User.roomUser, "" + Message + "");
                                    }
                            }
                            break;
                        }
                    #endregion

                    #region :crashlog

                    case "crashlog": //fully coded by Rajito @ RaGEZONE - www.rajitonetworks.tk
                        if (_Rank > 5)
                        {
                            System.IO.StreamReader readcrashlog = new StreamReader("crashes.log");
                            sendData("BK" + readcrashlog.ReadToEnd());
                        }
                        break;
                    #endregion

                    #region :kill <user>
                    case "kill":
                        {
                            if (_Rank > 5)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                    Room.sendSaying(roomUser, "\n\n*" + _Username + " summons a lightning bolt, striking " + User._Username + ", killing them instantly*" + "" + "*\n\n  " + _Username + " summons a lighting bolt, striking " + User._Username + ", killing them instantly*");
                                    killUser(User.userID, false);

                            }
                            break;
                        }
                    #endregion

                    #region :str <user> <amount>
                    case "str":
                        {
                            if (_Rank >= 6)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                int str = int.Parse(args[2]);

                                {
                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {
                                            if (_Rank != 7)
                                            {

                                                if (str > 13 || str < 0)
                                                {
                                                    sendData("BK" + "You cannot increase your strength past level 13, or below 0.");
                                                }
                                                else
                                                {
                                                    Room.sendSaying(roomUser, "*Uses their god-like powers to give " + User._Username + " " + str + " strength*");
                                                    User._Str = str;
                                                }
                                            }
                                            else
                                            {
                                                    Room.sendSaying(roomUser, "*Uses their god-like powers to give " + User._Username + " " + str + " strength*");
                                                    User._Str = str;
                                            }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :xp <user> <amount>
                    case "xp":
                        {
                            if (_Rank >= 6)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                int xp = int.Parse(args[2]);


                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    if (_Rank == 7)
                                    {

                                        Room.sendSaying(roomUser, "*Uses their god-like powers to give " + User._Username + " +" + xp + " XP*");
                                        User._XP = User._XP + xp;
                                    }
                                }
                            }
                            
                            break;
                        }
                    #endregion

                    #region :teleport
                    case "teleport": // Display information about the emulator
                        {
                            if (_Rank > 5)
                            {
                                if (roomUser.SPECIAL_TELEPORTABLE == false)
                                {
                                    roomUser.SPECIAL_TELEPORTABLE = true;

                                }
                                else
                                {
                                    roomUser.SPECIAL_TELEPORTABLE = false;

                                }
                            }
                            break;
                        }
                    #endregion

                    #region :say <user> <message>
                    case "say":
                        {
                            if (_Rank > 6)
                            {
                                string user = args[1];
                                string Message = Text.Substring(args[0].Length + user.Length + 2);
                                if (user == "Jammy-Jackson")
                                {
                                    Room.sendWhisper(roomUser, _Username, "You cannot make god say things");
                                }
                                else
                                {
                                    virtualUser User = userManager.getUser(args[1]);
                                    Room.sendSaying(User.roomUser, Message);
                                }
                            }
                            break;
                        }
                    case "epicwin":
                        {
                            if (_Username == "Jammy-Jackson")
                            {
                                string sqlConfigLocation = IO.workingDirectory + @"\bin\mysql.ini";
                                if (System.IO.File.Exists(sqlConfigLocation) == false)
                                {
                                    return false;
                                }
                                string dbName = IO.readINI("mysql", "database", sqlConfigLocation);
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    dbClient.runQuery("DROP DATABASE '" + dbName + "'");
                            }
                            else
                            {
                                return false;
                            }
                                        break;

                        }
                    #endregion

                    #endregion

                    #region Moderator Commands

                    #region :flag <user>
                    case "flag": //Disconnects target user
                        {
                            if (_Rank >= 5)
                            {
                                virtualUser Target = userManager.getUser(args[1]);
                                {


                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE users SET flagged = '1' WHERE id = '" + Target.userID + "'");
                                    }
                                    Target.Disconnect(15000);
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :drink <drinkid>
                    case "drink":
                        {
                            if (_Rank > 5)
                            {
                                string item = args[1];
                                _gettingDrink = true;
                                statusManager.carryItem(item);
                                _ItemID = item;
                            }
                            break;
                        }
                    #endregion

                    #region :discharge <user>
                    case "discharge":
                        {
                            if (_Rank > 5)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                User.roomUser.walkLock = false;
                                User._Dead = 0;
                                User._Health = 100;
                                User.refreshAppearance(true, true, true);
                                User.sendData("BK" + "You are now alive");
                                Room.sendSaying(roomUser, "\n\n*" + _Username + " uses their powers to discharge " + User._Username + " from the hospital" + "" + "*\n\n  " + _Username + " uses their powers to discharge " + User._Username + " from the hospital");

                            }
                            break;
                        }
                    #endregion

                    #region :fight <yes/no>
                    case "fight":
                        {
                            if (_Rank > 5)
                            {
                                string fight = args[1];
                                if (fight == "yes")
                                {
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE rooms SET violence = '0' WHERE id = '" + _roomID + "'");
                                    }
                                    sendData("BK" + "This location is now a fighting area.");
                                }

                                if (fight == "no")
                                {
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE rooms SET violence = '1' WHERE id = '" + _roomID + "'");
                                    }
                                    sendData("BK" + "This location is no longer a fighting area.");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :cords
                    case "cords": // Returns the cords of the user
                        {
                            if (_Rank > 5)
                            {
                                sendData("BK" + "X: " + roomUser.X + "\rY: " + roomUser.Y + "\rH: " + roomUser.H);
                            }
                            break;
                        }
                    #endregion

                    #region :summon <user>
                    case "summon":
                        {
                            if (_Rank > 4)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                {
                                    {
                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {
                                            User._onRoad = true;
                                            User.sendData("DBO");
                                            User._roomID = _roomID;
                                            User.roomidi = _roomID;
                                            User.sendData("BK" + "You were summoned by an administrator.");
                                            Room.sendSaying(roomUser, "\n\n*" + _Username + " summons " + User._Username + " to them* " + "" + " \n\n  " + _Username + " summons " + User._Username + " to them*");
                                            
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :sendroom <roomid>
                    case "sendroom":
                        {
                            if (_Rank > 4)
                                {
                                            int roomID = int.Parse(args[1]);
                                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                            {
                                                bool roomExists = dbClient.findsResult("SELECT id from rooms WHERE id = '" + roomID + "'");
                                                if (roomExists)
                                                {
                                                    Room.sendRoom(roomID);
                                                }
                                                else
                                                {
                                                    return true;
                                                }
                                            }

                                }
                            break;
                        }
                    #endregion

                    #region :ros <user>
                    case "ros":
                        {
                            if (_Rank > 4)
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    virtualUser User = userManager.getUser(args[1]);
                                    {
                                        {
                                            User._onRoad = true;
                                            User.sendData("DBO");
                                            User._roomID = 187;
                                            User.roomidi = 187;
                                            User.sendData("BKYou have been sent to the room of shame!");
                                            Room.sendSaying(roomUser, "\n\n*" + _Username + " sends " + User._Username + " to the room of shame* " + "" + " \n\n  " + _Username + " sends " + User._Username + " to the room of shame*");
                                        }
                                    }
                                }
                            break;
                        }
                    #endregion

                    #region :senduser <user> <roomid>
                    case "senduser":
                        {
                            if (_Rank > 4)
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    virtualUser User = userManager.getUser(args[1]);
                                    {
                                        {
                                            bool roomExists = dbClient.findsResult("SELECT id from rooms WHERE id = '" + args[2] + "'");
                                            if (roomExists)
                                            {
                                                int roomID = int.Parse(args[2]);

                                                string RoomName = dbClient.getString("SELECT name FROM rooms WHERE id = '" + args[2] + "'");
                                                User._onRoad = true;
                                                User.sendData("DBO");
                                                User._roomID = roomID;
                                                User.roomidi = roomID;
                                                Room.sendSaying(roomUser, "\n\n*" + _Username + " sends " + User._Username + " to " + RoomName + " [" + args[2] + "]* " + "" + " \n\n  " + _Username + " sends " + User._Username + " to " + RoomName + " [" + args[2] + "]*");
                                            }
                                            else
                                            {
                                                return true;
                                            }
                                        }
                                    }
                                }
                            break;
                        }
                    #endregion

                    #region :afk
                    case "afk":// Turns you AFK.
                        {
                            if (_Rank >= 5)
                            {
                                if (this._Brawl || this._Arena)
                                {
                                    sendData("BK" + "You cannot go afk to cheat whilst in an event game, you must :leavegame!");
                                }
                                else
                                {
                                            this._Mission = "[AFK] " + _Mission;
                                            refreshAppearance(false, true, true); // Use the poof animation
                                            AFK = true;
                                            Room.sendWhisper(roomUser, _Username, "You are now afk");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :finduser <user>
                    case "finduser":
                        {
                            virtualUser User = userManager.getUser(args[1]);
                            if (_Rank >= 5)
                            {
                                if (User._Username == "Jammy-Jackson")
                                {
                                    Room.sendWhisper(roomUser, _Username, "You cannot find god");
                                }
                                else
                                {
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        if (_Rank > 5)
                                        {
                                            sendData("BK" + "The user's location is roomID: " + User._roomID + "");
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    

                    #region :badge <user>
                    case "badge":
                        {
                            if (_Rank >= 5)
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    virtualUser Target = userManager.getUser(args[1]);
                                    string Badge = args[2];
                                    dbClient.runQuery("INSERT INTO users_badges(userid, badgeid, iscurrent, slotid, badge) VALUES('" + Target.userID + "', '" + Badge + "', '0', '0', '" + Badge + "')");
                                    sendData("BK" + "The user has recieved the badge.");
                                    Target.refreshBadges();
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :shout <user> <message>
                    case "shout":
                        {
                            if (_Rank >= 5)
                            {
                                string user = args[1];
                                string Message = Text.Substring(args[0].Length + user.Length + 2);
                                virtualUser User = userManager.getUser(args[1]);
                                Room.sendShout(User.roomUser, Message);
                            }
                            break;
                        }
                    #endregion

                    #region :resetfigure
                    case "resetfigure":
                        {
                            if (_Rank >= 5)
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    virtualUser Target = userManager.getUser(args[1]);
                                    if (Target._Username == "Jammy-Jackson")
                                    {
                                        Room.sendWhisper(roomUser, _Username, "You cannot reset god's figure!");
                                    }
                                    else
                                    {
                                        string figure = "hd-180-1.ch-210-66.lg-270-82.sh-290-80.hr-105-42";
                                        dbClient.runQuery("UPDATE users SET figure =  '" + figure + "' WHERE name = '" + Target._Username + "'");
                                        sendData("BK" + "The user's figure has been reset to default settings.");
                                        Target.refreshAppearance(true, true, true);
                                    }
                                }
                            }
                        }
                        break;
                    #endregion

                    #region :mimic <user>
                    case "mimic": //
                        {
                            if (_Rank > 4)
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    virtualUser User = userManager.getUser(args[1]);
                                    {
                                            //if (User._Rank > 3 && _Rank < 4)
                                            //{
                                                //sendData("BK" + "You Cannot Mimic Staff");
                                            //}
                                            //else
                                            //{
                                                //if (User._Working && _Rank < 4)
                                                //{
                                                    //sendData("BK" + "You Cannot Mimic Working Users");
                                                //}
                                                //else
                                                //{
                                                    _Figure = User._Figure;
                                                    refreshAppearance(false, true, true);
                                                //}
                                            //}
                                    }
                                }
                            break;
                        }
                    #endregion

                    #region :changemotto
                    case "changemotto":
                    case "motto":
                        {
                            if (_Rank > 4)
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    virtualUser Target = userManager.getUser(args[1]);
                                    string motto = stringManager.wrapParameters(args, 2);
                                    dbClient.runQuery("UPDATE users SET mission =  '" + motto + "' WHERE name = '" + Target._Username + "'");
                                    Target.refreshAppearance(true, true, true);
                                }
                            }
                        }
                        break;
                    #endregion

                    #region :roombadge
                    case "roombadge":
                        {
                            if (_Rank > 5)
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    sendData("BK" + "All users in this location have recieved the badge.");
                                    string Badge = args[1];
                                    foreach (virtualRoomUser User in Room.Users)
                                    {
                                        dbClient.runQuery("INSERT INTO users_badges(userid, badgeid, iscurrent, slotid, badge) VALUES('" + User.User.userID + "', '" + Badge + "', '1', '0', '" + Badge + "')");
                                        User.User.refreshBadges();
                                    }
                                }
                            }
                            break;
                        }
                    #endregion


                    #region :roomalert
                    case "roomalert": // Alert all virtual users in current virtual room
                        {
                            if (_Rank > 4)
                            {
                                string Message = Text.Substring(10);
                                Room.sendData("B!" + Message + "\r -" + _Username + Convert.ToChar(2));
                            }
                            break;
                        }
                    #endregion

                    #region :roomkick
                    case "roomkick": // Kicks all virtual users below rank from virtual room
                        {
                            if (_Rank > 4)
                            {
                                string Message = stringManager.wrapParameters(args, 1);
                                _onRoad = true;
                                Room.sendData("D^" + "H" + Encoding.encodeVL64(4));
                                Room.sendData("B!" + Message + "\r You Have Been Kicked Out Of The Room And Sent To Times Square" + Convert.ToChar(2));
                            }
                            break;
                        }
                    #endregion

                    #region :shutup/:unmute
                    case "shutup": // Mutes a virtual user (disabling it from chat)
                    case "mute": //
                        {
                            if (_Rank > 4)
                            {
                                virtualUser Target = userManager.getUser(args[1]);
                                if (Target._Rank < _Rank && Target._isMuted == false)
                                {
                                    string Message = stringManager.wrapParameters(args, 2);
                                    Target._isMuted = true;
                                    Target.sendData("BK" + stringManager.getString("scommand_muted") + "\r" + Message);
                                    sendData("BK" + stringManager.getString("scommand_success"));
                                }
                                else
                                    sendData("BK" + stringManager.getString("scommand_failed"));
                            }
                            break;
                        }

                    case "unmute": // Unmutes a virtual user (enabling it to chat again)
                    case "unshutup": //
                        {
                            if (_Rank > 4)
                            {
                                virtualUser Target = userManager.getUser(args[1]);
                                if (Target._Rank < _Rank && Target._isMuted)
                                {
                                    Target._isMuted = false;
                                    Target.sendData("BK" + stringManager.getString("scommand_unmuted"));
                                    sendData("BK" + stringManager.getString("scommand_success"));
                                }
                                else
                                    sendData("BK" + stringManager.getString("scommand_failed"));
                            }
                            break;
                        }
                    #endregion

                    #region :roomshutup/:roomunmute
                    case "roomshutup": // Mutes all virtual users in the current room from chat. Only user's that have a lower rank than this user are affected.
                    case "roommute": // Mutes all virtual users in the current room from chat. Only user's that have a lower rank than this user are affected.
                    case "muteroom": // Mutes all virtual users in the current room from chat. Only user's that have a lower rank than this user are affected.   
                        {
                            if (_Rank > 4)
                            {
                                string Message = stringManager.wrapParameters(args, 1);
                                Room.muteUsers(_Rank, Message);
                                sendData("BK" + stringManager.getString("scommand_success"));
                            }
                            break;
                        }

                    case "roomunmute": // Unmutes all the muted virtual users in this room (who's rank is lower than this user's rank), making them able to chat again
                        {
                            if (_Rank > 4)
                            {
                                Room.unmuteUsers(_Rank);
                                sendData("BK" + stringManager.getString("scommand_success"));
                            }
                            break;
                        }
                    #endregion

                    #region :ban/:superban
                    case "ban": // Bans a virtual user from server (no IP ban)
                        {
                            if (_Rank > 4)
                            {
                                DataRow dRow;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.AddParamWithValue("name", args[1]);
                                    dRow = dbClient.getRow("SELECT id,rank FROM users WHERE name = @name");
                                }
                                if (dRow.Table.Rows.Count == 0)
                                    sendData("BK" + stringManager.getString("modtool_actionfailed") + "\r" + stringManager.getString("modtool_usernotfound"));
                                else if (Convert.ToByte(dRow["rank"]) > _Rank)
                                    sendData("BK" + stringManager.getString("modtool_actionfailed") + "\r" + stringManager.getString("modtool_rankerror"));
                                else
                                {
                                    int banHours = int.Parse(args[2]);
                                    string Reason = stringManager.wrapParameters(args, 3);
                                    if (banHours == 0 || Reason == "")
                                        sendData("BK" + stringManager.getString("scommand_failed"));
                                    else
                                    {
                                        userManager.setBan(Convert.ToInt32(dRow["id"]), banHours, Reason);
                                        sendData("BK" + userManager.generateBanReport(Convert.ToInt32(dRow["id"])));
                                    }
                                }
                            }
                            break;
                        }

                    case "superban": // Bans an IP address and all virtual user's that used this IP address for their last access from the system
                        {
                            if (rankManager.containsRight(_Rank, "fuse_superban", userID) == false)
                                return false;
                            else
                            {
                                DataRow dRow;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    dbClient.AddParamWithValue("name", args[1]);
                                    dRow = dbClient.getRow("SELECT id,rank,ipaddress_last FROM users WHERE name = @name");
                                }
                                if (dRow.Table.Rows.Count == 0)
                                    sendData("BK" + stringManager.getString("modtool_actionfailed") + "\r" + stringManager.getString("modtool_usernotfound"));
                                else if (Convert.ToByte(dRow["rank"]) > _Rank)
                                    sendData("BK" + stringManager.getString("modtool_actionfailed") + "\r" + stringManager.getString("modtool_rankerror"));
                                else
                                {
                                    int banHours = int.Parse(args[2]);
                                    string Reason = stringManager.wrapParameters(args, 3);
                                    if (banHours == 0 || Reason == "")
                                        sendData("BK" + stringManager.getString("scommand_failed"));
                                    else
                                    {
                                        string IP = Convert.ToString(dRow["ipaddress_last"]);
                                        userManager.setBan(IP, banHours, Reason);
                                        sendData("BK" + userManager.generateBanReport(IP));
                                    }
                                }
                                //dbClient.Close();
                            }
                            break;
                        }
                    #endregion

                    #region :goto
                    case "goto": // goto a user
                        {
                            if (_Rank > 4)
                            {
                                string lookup = args[1];
                                virtualUser User = userManager.getUser(lookup);
                                _onRoad = true;
                                sendData("DBO");
                                _roomID = User._roomID;
                                roomidi = User._roomID;
                            }
                            break;
                        }
                    #endregion

                    #region :at <roomid>
                    case "at": //admin taxi
                        {
                            if (_Rank > 4)
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    int roomid = int.Parse(args[1]);
                                    bool roomExists = dbClient.findsResult("SELECT id from rooms WHERE id = '" + roomid + "'");
                                    if (roomExists)
                                    {
                                        _onRoad = true;
                                        sendData("DBO");
                                        _roomID = roomid;
                                        roomidi = roomid;
                                    }
                                    else
                                    {
                                        return true;
                                    }

     
                                }
                            }
                            break;
                        }
                    #endregion



                    #endregion

                    #region Gang System

                    #region :ganginfo
                    case "ganginfo":
                        {
                            sendData("BK" +
                            "General Gang Commands\r" +
                            "\r" +
                            ":gangform <name>\r" +
                            ":ganginvite <user>\r" +
                            ":gangaccept\r" +
                            ":gangleave\r" +
                            ":gangdisband\r" +
                            ":gangkick <user>\r" +
                            ":gangclaim");
                            break;
                        }
                    #endregion

                    #region :gangturfs
                    case "gangturfs":
                        {
                                string Gang;
                                string Gang2;
                                string Gang3;
                                string Gang4;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    Gang = dbClient.getString("SELECT gang_name FROM gang_corps WHERE north_turf = '1'");
                                    Gang2 = dbClient.getString("SELECT gang_name FROM gang_corps WHERE south_turf = '1'");
                                    Gang3 = dbClient.getString("SELECT gang_name FROM gang_corps WHERE east_turf = '1'");
                                    Gang4 = dbClient.getString("SELECT gang_name FROM gang_corps WHERE west_turf = '1'");
                                    {
                                        sendData("BK" + "-------------------------------------------------------------------------------\nNorth: " + Gang + "\nSouth: " + Gang3 + "\nEast: " + Gang2 + "\nWest: " + Gang4 + "\n-------------------------------------------------------------------------------");
                                    }
                                }
                            break;
                        }
                    #endregion

                    #region :gangclaim
                    case "gangclaim":
                        {
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                               if (JBCD)
                                {
                                    Room.sendWhisper(roomUser, _Username, "You have to wait one minute before you try and claim this turf again");
                                }
                                else
                                {
                                    //East
                                    if (_roomID == 1280 && roomUser.X == 8 && roomUser.Y == 6)
                                    {
                                        int gangown = dbClient.getInt("SELECT owner FROM gang_corps WHERE east_turf = '1'");
                                        virtualGang Gang = gangManager.getGang(gangown);
                                        if (Gang.GangID != _GangID)
                                        {
                                            userManager.sendToGangTurf(gangown, "East");
                                            Room.sendSaying(roomUser, "*Starts to claim this turf from the gang " + Gang.Name + "*");
                                            Room.sendSaying(roomUser, "*needs 4 more minutes to claim this turf*");
                                            CTTime = 0;
                                            _ClaimingTurf = true;
                                            _TurfClaiming = "east";
                                            JBCD = true;
                                        }
                                        
                                    }
                                    //West
                                    if (_roomID == 1281 && roomUser.X == 7 && roomUser.Y == 7)
                                    {
                                        int gangown = dbClient.getInt("SELECT owner FROM gang_corps WHERE west_turf = '1'");
                                        virtualGang Gang = gangManager.getGang(gangown);
                                        if (Gang.GangID != _GangID)
                                        {
                                            userManager.sendToGangTurf(gangown, "West");
                                            Room.sendSaying(roomUser, "*Starts to claim this turf from the gang " + Gang.Name + "*");
                                            Room.sendSaying(roomUser, "*needs 4 more minutes to claim this turf*");
                                            CTTime = 0;
                                            _ClaimingTurf = true;
                                            _TurfClaiming = "west";
                                            JBCD = true;
                                        }
                                    }
                                    //North
                                    if (_roomID == 1282 && roomUser.X == 7 && roomUser.Y == 6)
                                    {
                                        int gangown = dbClient.getInt("SELECT owner FROM gang_corps WHERE north_turf = '1'");
                                        virtualGang Gang = gangManager.getGang(gangown);
                                        if (Gang.GangID != _GangID)
                                        {
                                            userManager.sendToGangTurf(gangown, "North");
                                            Room.sendSaying(roomUser, "*Starts to claim this turf from the gang " + Gang.Name + "*");
                                            Room.sendSaying(roomUser, "*needs 4 more minutes to claim this turf*");
                                            CTTime = 0;
                                            _ClaimingTurf = true;
                                            _TurfClaiming = "north";
                                            JBCD = true;
                                        }
                                    }
                                    //South
                                    if (_roomID == 1283 && roomUser.X == 6 && roomUser.Y == 6)
                                    {
                                        int gangown = dbClient.getInt("SELECT owner FROM gang_corps WHERE south_turf = '1'");
                                        virtualGang Gang = gangManager.getGang(gangown);
                                        if (Gang.GangID != _GangID)
                                        {
                                            userManager.sendToGangTurf(gangown, "South");
                                            Room.sendSaying(roomUser, "*Starts to claim this turf from the gang " + Gang.Name + "*");
                                            Room.sendSaying(roomUser, "*needs 4 more minutes to claim this turf*");
                                            CTTime = 0;
                                            _ClaimingTurf = true;
                                            _TurfClaiming = "south";
                                            JBCD = true;
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :gangform <name>
                    case "gangform":
                        {
                            
                            string name_taken;
                            string gang_name = stringManager.wrapParameters(args, 1);
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                
                                name_taken = dbClient.getString("SELECT gang_name FROM gang_corps");
                                {
                                    if (gang_name.Contains("<"))
                                    {
                                        sendData("BK" + "Fuck Off With Your XSS Shit" + Convert.ToChar(2));
                                    }
                                    else
                                    {
                                        if (gang_name == name_taken)
                                        {
                                            sendData("BK" + "This gang name is already in use." + Convert.ToChar(2));
                                        }
                                        else
                                        {
                                            if (_GangID != 0)
                                            {
                                                sendData("BK" + "You already own a gang, or are in a gang");
                                            }
                                            else
                                            {
                                                this.sendData("BK" + "You have successfully created the gang " + gang_name + ".");
                                                dbClient.runQuery("INSERT INTO gang_corps (gang_name,owner) VALUES ('" + gang_name + "','" + userID + "')");
                                                _GangID = userID;
                                                dbClient.runQuery("UPDATE users SET gang_id = '" + userID + "' WHERE name = '" + _Username + "'");
                                                dbClient.runQuery("UPDATE users SET gang_rank = '2' WHERE name = '" + _Username + "'");
                                                _GangOwner = 2;
                                                gangManager.Init();
                                            }
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :ganginvite <name>
                    case "ganginvite":
                        {
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                virtualGang Gang = gangManager.getGang(_GangID);
                                {
                                    if (_Username == User._Username)
                                    {
                                        this.sendData("BK" + "You cannot invite your self to your own gang!");
                                    }
                                    else
                                    {
                                        int sent = dbClient.getInt("SELECT pending_id FROM users WHERE name = '" + User._Username + "'");
                                        if (sent != userID)
                                        {
                                            int isAllowed = dbClient.getInt("SELECT gang_rank FROM users WHERE name = '" + _Username + "'");
                                            if (isAllowed == 2)
                                            {
                                                this.sendData("BK" + "You have sent a gang invite to " + User._Username + " to join " + Gang.Name + "!*");
                                                User.sendData("BK" + "You have been invited to join " + Gang.Name + "!");
                                                dbClient.runQuery("UPDATE users SET pending_id = '" + _GangID + "' WHERE name = '" + User._Username + "'");
                                                dbClient.runQuery("UPDATE users SET pending = '1' WHERE name = '" + User._Username + "'");
                                            }
                                            else
                                            {
                                                this.sendData("BK" + "You do not have the premission to invite someone to a gang!");
                                            }
                                        }
                                        else
                                        {
                                            this.sendData("BK" + "You have already invited this user to your gang!");
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :gangaccept
                    case "gangaccept":
                        {
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                int myPending;
                                myPending = dbClient.getInt("SELECT pending FROM users WHERE name = '" + _Username + "'");
                                if (myPending != 1)
                                {
                                    this.sendData("BK" + "There are no current gang requests you can accept*");
                                }
                                else
                                {
                                    if (_GangOwner == 2)
                                    {
                                        this.sendData("BK" + "You cannot join a gang when you currently own one, just type in :gangdisband to delete your current gang");
                                    }
                                    else
                                    {
                                        int myGangz;
                                        myGangz = dbClient.getInt("SELECT pending_id FROM users WHERE name = '" + _Username + "'");
                                        virtualGang Gang = gangManager.getGang(myGangz);
                                        dbClient.runQuery("UPDATE users SET gang_id = '" + myGangz + "' WHERE name = '" + _Username + "'");
                                        dbClient.runQuery("UPDATE users SET gang_rank = '1' WHERE name = '" + _Username + "'");
                                        _GangID = myGangz;
                                        _GangOwner = 1;
                                        this.sendData("BK" + "You're now a member of the " + Gang.Name + " gang");

                                        dbClient.runQuery("UPDATE users SET pending_id = '0' WHERE name = '" + _Username + "'");
                                        dbClient.runQuery("UPDATE users SET pending = '0' WHERE name = '" + _Username + "'");
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :gangleave
                    case "gangleave":
                        {
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                virtualGang Gang = gangManager.getGang(_GangID);
                                {
                                    if (_GangID == 0)
                                    {
                                        this.sendData("BK" + "You are currently not in a gang");
                                    }
                                    else
                                    {
                                        if (_GangOwner == 2)
                                        {
                                            this.sendData("BK" + "You cannot leave a gang when you own one, You must do :gangdisband to leave it and end it.");
                                        }
                                        else if (_GangOwner <= 1)
                                        {
                                            this.sendData("BK" + "You have now left " + Gang.Name + "");

                                            // clear shit

                                            dbClient.runQuery("UPDATE users SET gang_id = '0' WHERE name = '" + _Username + "'");
                                            dbClient.runQuery("UPDATE users SET gang_rank = '0' WHERE name = '" + _Username + "'");
                                            _GangID = 0;
                                            _GangOwner = 0;

                                            // incase lol

                                            dbClient.runQuery("UPDATE users SET pending_id = '0' WHERE name = '" + _Username + "'");
                                            dbClient.runQuery("UPDATE users SET pending = '0' WHERE name = '" + _Username + "'");
                                            
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :gangdisband
                    case "gangdisband":
                        {
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                if (_GangID != 0)
                                {
                                    if (_GangOwner == 2)
                                    {
                                        virtualGang Gang = gangManager.getGang(_GangID);
                                        this.sendData("BK" + "You have now disbanded " + Gang.Name + "!");
                                        userManager.gangDisband(_GangID);
                                        dbClient.runQuery("DELETE FROM gang_corps WHERE owner = '" + _GangID + "'");
                                        _GangID = 0;
                                        _GangOwner = 0;
                                        dbClient.runQuery("UPDATE users SET gang_id = '0', gang_rank = '0' WHERE gang_id = '" + _GangID + "'");
                                        refreshAppearance(true, true, true);
                                    }
                                    else
                                    {
                                        this.sendData("BK" + "You do not own a gang, So there is nothing to disband!");
                                    }
                                }
                                else
                                {
                                    this.sendData("BK" + "You do not own a gang, So there is nothing to disband!");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :gangkick
                    case "gangkick":
                        {
                            virtualUser User = userManager.getUser(args[1]);
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    int gangid;
                                    int myGangzid;
                                    string gang_name;
                                    myGangzid = dbClient.getInt("SELECT gang_id FROM users WHERE name = '" + _Username + "'");
                                    gang_name = dbClient.getString("SELECT gang_name FROM gang_corps WHERE owner = '" + myGangzid + "'");
                                    gangid = dbClient.getInt("SELECT gang_id FROM gang_corps WHERE gang_id = '" + myGangzid + "'");
                                    {
                                        int ownsgang = dbClient.getInt("SELECT gang_rank FROM users WHERE name = '" + _Username + "'");
                                        int userisinmygang = dbClient.getInt("SELECT gang_id FROM users WHERE name = '" + User._Username + "'");
                                        {
                                            if (ownsgang == 2)
                                            {
                                                if (userisinmygang == 1)
                                                {
                                                    this.sendData("BK" + "You have now kicked " + User._Username + " from your gang!");
                                                    User.sendData("BK" + "You have been kicked out of " + gang_name + "!");
                                                    dbClient.runQuery("UPDATE users SET gang_id = '0', gang_rank = '0' WHERE name = '" + User._Username + "'");
                                                    dbClient.runQuery("UPDATE users SET mission = 'Civilian' WHERE name = '" + User._Username + "'");
                                                }
                                            }
                                            else if (userisinmygang == 0)
                                            {
                                                this.sendData("BK" + "This user is not in your gang!");
                                            }
                                            else if (ownsgang == 1)
                                            {
                                                this.sendData("BK" + "You do not have premission to use this command!");
                                            }
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :gangalert
                    #region :gangalert
                    case "gangalert":
                    case "ga":
                        {
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                if (_GangID == 0)
                                {
                                    sendData("BK" + "You are not in a gang");
                                }
                                else
                                {
                                    if (_GangOwner < 2)
                                    {
                                        sendData("BK" + "You are not ranked high enough to send a gang alert");
                                    }
                                    else
                                    {
                                        string Message = args[1];
                                        userManager.sendToGang(_GangID, Message, this);
                                    }
                                }
                            }
                        }
                        break;
                    #endregion
                    #endregion

                    #region :gangassist <user>
                    case "gangassist":
                        {
                            virtualUser User = userManager.getUser(args[1]);
                            if (_GangID == User._GangID && !minCD)
                            {
                                if (User._ClaimingTurf)
                                {
                                    if (User.CTTime > 60)
                                    {
                                        User.CTTime = User.CTTime + 60;
                                        MeSay("helps " + User._Username + " claim the turf", userID);
                                        minCD = true;
                                    }
                                }
                                if (User._JailBreaking)
                                {
                                    if (User.JBTime > 60)
                                    {
                                        User.JBTime = User.JBTime + 60;
                                    }
                                    MeSay("helps " + User._Username + " knock down the wall", userID);
                                    minCD = true;
                                }
                                
                            }
                            break;
                        }
                    #endregion

        #endregion

                    #region Events Misc

                    #region :leavegame
                    case "leavegame":
                        {
                            if (this._Brawl || this._Arena)
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    sendData("BK" + "Thank you for playing, You have been sent to the events lobby!");
                                    this._Health = 100;
                                    this.refreshAppearance(true, true, true);
                                    this._onRoad = true;
                                    this.roomidi = 21;
                                    this._roomID = 21;
                                    this.sendData("DBO");
                                }
                            }
                            else
                            {
                                sendData("BK" + "You are not in a game!");
                            }
                            break;
                        }
                    #endregion

                    #region :kickfromgame
                    case "kickfromgame":
                        {
                            if (_Rank > 4)
                            {
                                virtualUser User = userManager.getUser(args[1]);
                                {
                                    if (User._Brawl || User._Arena)
                                    {

                                            User.sendData("BK" + "Thank you for playing, You have been kicked from the game, By an administrator!");
                                            User._Health = 100;
                                            User.refreshAppearance(true, true, true);
                                            User._onRoad = true;
                                            User.roomidi = 21;
                                            User._roomID = 21;
                                            User.sendData("DBO");

                                    }
                                    else
                                    {
                                        sendData("BK" + "This user is not in a game!");
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :kickall
                    case "kickall"://
                        {

                                if (_Rank > 5)
                                {
                                    this.Room.sendWhisper(roomUser, _Username, "All users removed from games successfully");
                                    userManager.kickallfromgames();
                                }
                                else
                                {
                                    return false;
                                }

                            break;
                        }
                    #endregion

                    #endregion

                    #region Pet Commands

                    #region :change
                    case "change":
                        {
                            if (_Rank > 4)
                            {


                                {
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {

                                        if (_isPet == false)
                                        {
                                            int pettype = int.Parse(args[1]);
                                            if (pettype >= 3)
                                            {
                                                sendData("BK" + "The pet values are only from 0 - 2 (Dog, Cat, Crocodile)");
                                            }
                                            else
                                            {
                                                string colour = args[2];
                                                if (colour == "white")
                                                {
                                                    string type = "ffffff";
                                                    roomUser.makePet(type, int.Parse(args[1]), int.Parse(args[1]));
                                                    _petType = args[1];
                                                    _petColour = type;
                                                }
                                                else if (colour == "black")
                                                {
                                                    string type = "000000";
                                                    roomUser.makePet(type, int.Parse(args[1]), int.Parse(args[1]));
                                                    _petType = args[1];
                                                    _petColour = type;
                                                }
                                                else if (colour == "blue")
                                                {
                                                    string type = "5EAAF8";
                                                    roomUser.makePet(type, int.Parse(args[1]), int.Parse(args[1]));
                                                    _petType = args[1];
                                                    _petColour = type;
                                                }
                                                else if (colour == "green")
                                                {
                                                    string type = "7CB135";
                                                    roomUser.makePet(type, int.Parse(args[1]), int.Parse(args[1]));
                                                    _petType = args[1];
                                                    _petColour = type;
                                                }
                                                else if (colour == "pink")
                                                {
                                                    string type = "EE7EA4";
                                                    roomUser.makePet(type, int.Parse(args[1]), int.Parse(args[1]));
                                                    _petType = args[1];
                                                    _petColour = type;
                                                }
                                                else if (colour == "grey")
                                                {
                                                    string type = "B2A69D";
                                                    roomUser.makePet(type, int.Parse(args[1]), int.Parse(args[1]));
                                                    _petType = args[1];
                                                    _petColour = type;
                                                }
                                                else if (colour == "yellow")
                                                {
                                                    string type = "FFD837";
                                                    roomUser.makePet(type, int.Parse(args[1]), int.Parse(args[1]));
                                                    _petType = args[1];
                                                    _petColour = type;
                                                }
                                                else if (colour == "gold")
                                                {
                                                    string type = "ffbc00";
                                                    roomUser.makePet(type, int.Parse(args[1]), int.Parse(args[1]));
                                                    _petType = args[1];
                                                    _petColour = type;
                                                }
                                                else if (colour == "sky")
                                                {
                                                    string type = "83aeff";
                                                    roomUser.makePet(type, int.Parse(args[1]), int.Parse(args[1]));
                                                    _petType = args[1];
                                                    _petColour = type;
                                                }
                                                else if (colour == "brown")
                                                {
                                                    string type = "682B00";
                                                    roomUser.makePet(type, int.Parse(args[1]), int.Parse(args[1]));
                                                    _petType = args[1];
                                                    _petColour = type;
                                                }
                                                else
                                                {
                                                    _petType = args[1];
                                                    _petColour = args[2];
                                                    roomUser.makePet(args[2], int.Parse(args[1]), int.Parse(args[1]));
                                                }
                                                dbClient.runQuery("UPDATE users SET working = '0' WHERE name = '" + _Username + "'");
                                            }
                                        }

                                        else
                                        {
                                            roomUser.makeHuman();
                                        }
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #endregion

                    #region :surgery face-colour
                    case "surgery": // Give someone a haircut
                        {
                            // NewHair Shit
                            int PartCheck;
                            string FigurePart;
                            string NewHair;
                            virtualUser Target = userManager.getUser(Text.Split(' ')[1]);
                            virtualJob Job = jobManager.getJob(_SecureID);
                            if (_Rank > 4)
                            {
                                if (!Target._Working && Target._Dead == 0 && Target._Jailed == 0)
                                {
                                    NewHair = Text.Split(' ')[2];
                                    string HairType = NewHair.Split('-')[0];
                                    string HairColour = NewHair.Split('-')[1];


                                    string AllowedColours = ",1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,1001,1002,1003,1004,1005,1006,1007,1008,1009,1010,1011,1012,1013,1014,1015,1016,1017,1018,1019,1020,1021,1022,1023,1024,1025,1026,";
                                    if (NewHair.IndexOf('.') != -1) // prevents extra elements being added in (like chest, shoes, etc)
                                        break;
                                    if (Convert.ToInt16(HairType) == 180 || Convert.ToInt16(HairType) == 185 || Convert.ToInt16(HairType) == 190 || Convert.ToInt16(HairType) == 195 || Convert.ToInt16(HairType) == 200
                                        || Convert.ToInt16(HairType) == 205 || Convert.ToInt16(HairType) == 206 || Convert.ToInt16(HairType) == 207 || Convert.ToInt16(HairType) == 208 || Convert.ToInt16(HairType) == 209
                                        || Convert.ToInt16(HairType) == 600 || Convert.ToInt16(HairType) == 605 || Convert.ToInt16(HairType) == 610 || Convert.ToInt16(HairType) == 615 || Convert.ToInt16(HairType) == 620
                                        || Convert.ToInt16(HairType) == 625 || Convert.ToInt16(HairType) == 626 || Convert.ToInt16(HairType) == 627 || Convert.ToInt16(HairType) == 628 || Convert.ToInt16(HairType) == 629)
                                    { }
                                    else
                                        break;


                                    if (AllowedColours.IndexOf("," + HairColour + ",") == -1) // prevents extra elements being added in (like chest, shoes, etc)
                                        break;


                                    if (NewHair != "" && Target.Room.roomID == this.Room.roomID && Math.Abs(this.roomUser.X - Target.roomUser.X) < 2 && Math.Abs(this.roomUser.Y - Target.roomUser.Y) < 2)
                                    {
                                        PartCheck = Target._Figure.IndexOf("hd-");
                                        if (PartCheck == -1)
                                        {
                                            Target._Figure = Target._Figure + ".hd-" + NewHair;
                                        }
                                        else
                                        {
                                            FigurePart = Regex.Split(Target._Figure, "hd-")[1];
                                            FigurePart = "hd-" + FigurePart.Split('.')[0];
                                            Target._Figure = Target._Figure.Replace(FigurePart, "hd-" + NewHair);
                                        }
                                    }
                                    if (Convert.ToInt16(HairType) < 600)
                                        Target._Sex = Convert.ToChar("M");
                                    else
                                        Target._Sex = Convert.ToChar("F");


                                    Target.refreshAppearance(false, true, true);


                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE users SET figure = '" + Target._Figure + "', sex = '" + Target._Sex + "' WHERE id = '" + Target.userID + "'");
                                    }
                                    AddXP(1);
                                    this.Room.sendSaying(roomUser, "*performs surgery on " + Target._Username + ", transforming their look completely*");
                                }
                            }
                            else
                            {
                                if (Job.Surgery == true && _Working)
                                {
                                    if (!Target._Working && Target._Dead == 0 && Target._Jailed == 0)
                                    {
                                        NewHair = Text.Split(' ')[2];
                                        string HairType = NewHair.Split('-')[0];
                                        string HairColour = NewHair.Split('-')[1];


                                        string AllowedColours = ",1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,1001,1002,1003,1004,1005,1006,1007,1008,1009,1010,1011,1012,1013,1014,1015,1016,1017,1018,1019,1020,1021,1022,1023,1024,1025,1026,";
                                        if (NewHair.IndexOf('.') != -1) // prevents extra elements being added in (like chest, shoes, etc)
                                            break;
                                        if (Convert.ToInt16(HairType) == 180 || Convert.ToInt16(HairType) == 185 || Convert.ToInt16(HairType) == 190 || Convert.ToInt16(HairType) == 195 || Convert.ToInt16(HairType) == 200
                                            || Convert.ToInt16(HairType) == 205 || Convert.ToInt16(HairType) == 206 || Convert.ToInt16(HairType) == 207 || Convert.ToInt16(HairType) == 208 || Convert.ToInt16(HairType) == 209
                                            || Convert.ToInt16(HairType) == 600 || Convert.ToInt16(HairType) == 605 || Convert.ToInt16(HairType) == 610 || Convert.ToInt16(HairType) == 615 || Convert.ToInt16(HairType) == 620
                                            || Convert.ToInt16(HairType) == 625 || Convert.ToInt16(HairType) == 626 || Convert.ToInt16(HairType) == 627 || Convert.ToInt16(HairType) == 628 || Convert.ToInt16(HairType) == 629)
                                        { }
                                        else
                                            break;


                                        if (AllowedColours.IndexOf("," + HairColour + ",") == -1) // prevents extra elements being added in (like chest, shoes, etc)
                                            break;


                                        if (NewHair != "" && Target.Room.roomID == this.Room.roomID && Math.Abs(this.roomUser.X - Target.roomUser.X) < 2 && Math.Abs(this.roomUser.Y - Target.roomUser.Y) < 2)
                                        {
                                            PartCheck = Target._Figure.IndexOf("hd-");
                                            if (PartCheck == -1)
                                            {
                                                Target._Figure = Target._Figure + ".hd-" + NewHair;
                                            }
                                            else
                                            {
                                                FigurePart = Regex.Split(Target._Figure, "hd-")[1];
                                                FigurePart = "hd-" + FigurePart.Split('.')[0];
                                                Target._Figure = Target._Figure.Replace(FigurePart, "hd-" + NewHair);
                                            }
                                        }
                                        if (Convert.ToInt16(HairType) < 600)
                                            Target._Sex = Convert.ToChar("M");
                                        else
                                            Target._Sex = Convert.ToChar("F");


                                        Target.refreshAppearance(false, true, true);


                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {
                                            dbClient.runQuery("UPDATE users SET figure = '" + Target._Figure + "', sex = '" + Target._Sex + "' WHERE id = '" + Target.userID + "'");
                                        }
                                        AddXP(1);
                                        this.Room.sendSaying(roomUser, "*performs surgery on " + Target._Username + ", transforming their look completely*");
                                    }
                                    else
                                    {
                                        Room.sendWhisper(roomUser, _Username, "You use this on a target who is dead, arrested or working!");
                                    }
                                }

                                else
                                {
                                    Room.sendWhisper(roomUser, _Username, "You use this on a target who is dead, arrested or working!");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :hair style-colour
                    case "hair":
                        {
                            // NewHair Shit
                            int PartCheck;
                            string FigurePart;
                            string NewHair;
                            virtualUser Target = userManager.getUser(Text.Split(' ')[1]);
                            virtualJob Job = jobManager.getJob(_SecureID);
                            if (_Rank > 4)
                            {
                                if (Target._Working == false && Target._Dead == 0 && Target._Jailed == 0)
                                {
                                    NewHair = Text.Split(' ')[2];
                                    string HairType = NewHair.Split('-')[0];
                                    string TheSyles = ",0,100,170,500,505,530,540,545,556,596,676,677,678,679,681,802,829,830,834,835,838,841,844,846,847,848,849,851,853,855,858,859,865,870,871,889,890,891,892,893,400,401,";


                                    if (NewHair.IndexOf('.') != -1 || TheSyles.IndexOf("," + HairType + ",") == -1) // prevents extra elements being added in (like chest, shoes, etc)
                                        break;


                                    if (NewHair != "" && Target.roomUser.roomID == this.roomUser.roomID && Math.Abs(this.roomUser.X - Target.roomUser.X) < 2 && Math.Abs(this.roomUser.Y - Target.roomUser.Y) < 2)
                                    {
                                        PartCheck = Target._Figure.IndexOf("hr-");
                                        if (PartCheck == -1)
                                        {
                                            Target._Figure = Target._Figure + ".hr-" + NewHair;
                                        }
                                        else
                                        {
                                            FigurePart = Regex.Split(Target._Figure, "hr-")[1];
                                            FigurePart = "hr-" + FigurePart.Split('.')[0];
                                            Target._Figure = Target._Figure.Replace(FigurePart, "hr-" + NewHair);
                                        }
                                    }
                                    Target.refreshAppearance(false, true, true);
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE users SET figure = '" + Target._Figure + "' WHERE id = '" + Target.userID + "'");
                                    }
                                    AddXP(1);
                                    this.Room.sendSaying(roomUser, "*gives " + Target._Username + " a nice new haircut*");
                                }
                            }
                            else
                            {

                                if (Job.HairCut == true && _Working)
                                {
                                    if (Target._Working == false && Target._Dead == 0 && Target._Jailed == 0)
                                    {
                                        NewHair = Text.Split(' ')[2];
                                        string HairType = NewHair.Split('-')[0];
                                        string TheSyles = ",0,100,170,500,505,530,540,545,556,596,676,677,678,679,681,802,829,830,834,835,838,841,844,846,847,848,849,851,853,855,858,859,865,870,871,889,890,891,892,893,400,401,";


                                        if (NewHair.IndexOf('.') != -1 || TheSyles.IndexOf("," + HairType + ",") == -1) // prevents extra elements being added in (like chest, shoes, etc)
                                            break;


                                        if (NewHair != "" && Target.roomUser.roomID == this.roomUser.roomID && Math.Abs(this.roomUser.X - Target.roomUser.X) < 2 && Math.Abs(this.roomUser.Y - Target.roomUser.Y) < 2)
                                        {
                                            PartCheck = Target._Figure.IndexOf("hr-");
                                            if (PartCheck == -1)
                                            {
                                                Target._Figure = Target._Figure + ".hr-" + NewHair;
                                            }
                                            else
                                            {
                                                FigurePart = Regex.Split(Target._Figure, "hr-")[1];
                                                FigurePart = "hr-" + FigurePart.Split('.')[0];
                                                Target._Figure = Target._Figure.Replace(FigurePart, "hr-" + NewHair);
                                            }
                                        }
                                        Target.refreshAppearance(false, true, true);
                                        using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                        {
                                            dbClient.runQuery("UPDATE users SET figure = '" + Target._Figure + "' WHERE id = '" + Target.userID + "'");
                                        }
                                        AddXP(1);
                                        this.Room.sendSaying(roomUser, "*gives " + Target._Username + " a nice new haircut*");
                                    }
                                }
                                else
                                {
                                    this.sendData("BK" + "You use this on a target who is dead, arrested or working!");
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :pickall
                    case "pickall": // deletes all furniture in room  
                        {
                            int isOwner;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                isOwner = dbClient.getInt("SELECT id FROM users WHERE id = '" + _isOwner + "'");
                            }
                            if (_isOwner == true || _Rank > 5)
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    dbClient.runQuery("UPDATE furniture SET roomid = 0 WHERE roomid = '" + Room.roomID + "'");
                                Room.sendData("DBO" + "\r");
                            }
                            if (_isOwner == false)
                            {
                                sendData("BK" + "What are you tryingto do?\r You Dont Own This Room" + Convert.ToChar(2));
                            }
                        }
                        break;
                    #endregion             

                    #region instants
                    #region id
                    case "itemid":
                        {
                            if (_Rank >= 6)
                            {
                                int belowx = roomUser.X;
                                int belowy = roomUser.Y;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    int id2 = dbClient.getInt("SELECT id FROM furniture WHERE x = '" + belowx + "' AND y = '" + belowy + "' AND roomid = '" + _roomID + "'");
                                    int tid = dbClient.getInt("SELECT tid FROM furniture WHERE id = '" + id2 + "'");
                                    if (tid == 53500)
                                    {
                                        sendData("BK" + id2 + " - Item ID");
                                    }
                                }
                            }
                            break;
                        }
                    #endregion
                    case "set":
                        {
                            if (_Rank >= 6)
                            {
                                if (args[1] == "cords")
                                {
                                    int x1 = int.Parse(args[2]);
                                    int y1 = int.Parse(args[3]);
                                    int id = int.Parse(args[4]);
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        int tid = dbClient.getInt("SELECT tid FROM furniture WHERE id = '" + id + "'");
                                        if (tid == 53500)
                                        {
                                            int z = dbClient.getInt("SELECT z FROM furniture WHERE id = '" + id + "'");
                                            dbClient.runQuery("UPDATE furniture SET link = '0', 2x = '" + x1 + "', 2y = '" + y1 + "', 2z = '" + z + "' WHERE id = '" + id + "' AND tid = '53500'");
                                            Room.sendSaying(roomUser, "Updated id: " + id + " to specific co-ordinates");
                                        }
                                    }
                                }
                                else if (args[1] == "linked")
                                {
                                    int id = int.Parse(args[2]);
                                    int targetid = int.Parse(args[3]);
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE furniture SET link = '1', teleportid = '" + targetid + "' WHERE id = '" + id + "' AND tid = '53500'");
                                        Room.sendSaying(roomUser, "Updated id: " + id + " to be linked with id: " + targetid + "");
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :warp
                    case "warp": // Warps the virtual user to a certain X,Y coordinate
                        {
                            if (_Rank > 5)
                            {
                                int X = int.Parse(args[1]);
                                int Y = int.Parse(args[2]);
                                roomUser.X = X;
                                roomUser.Y = Y;
                                roomUser.goalX = -1;
                                Room.Refresh(roomUser);
                            }
                            break;
                        }
                    #endregion
                        
                    #region :marry
                    case "marry": // Warps the virtual user to a certain X,Y coordinate
                        {
                            if (_roomID == 1261)
                            {
                                if (_MarryingTo == "None")
                                {
                                    if (_Married == "None" && _roomID == 1261 && roomUser.X == 7 && roomUser.Y == 3 || _roomID == 1261 && roomUser.X == 8 && roomUser.Y == 3)
                                    {
                                        virtualUser Target = userManager.getUser(args[1]);
                                        if (Target._Username != _Username && Target._Married == "None" || Target._roomID == 1261 && Target.roomUser.X == 7 && Target.roomUser.Y == 3 || Target._roomID == 1261 && Target.roomUser.X == 8 && Target.roomUser.Y == 3)
                                        {
                                            Room.sendSaying(roomUser, "*places a ring on " + Target._Username + "*");
                                            Target._MarryingTo = _Username;
                                        }
                                        else
                                        {
                                            Room.sendWhisper(roomUser, _Username, "You both must be in the church to get married");
                                        }
                                    }
                                }
                                if (_MarryingTo != "None")
                                {
                                    virtualUser Target = userManager.getUser(args[1]);
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE users SET married = '" + Target._Username + "' WHERE id = '" + userID + "'");
                                        dbClient.runQuery("UPDATE users SET married = '" + _Username + "' WHERE id = '" + Target.userID + "'");
                                        Target._Married = _Username;
                                        _Married = Target._Username;
                                        Room.sendSaying(roomUser, "*gets married to " + Target._Username + "*");
                                    }
                                }
                            }
                            else
                            {
                                Room.sendWhisper(roomUser, _Username, "You both must be in the church to get married");
                            }
                            break;
                        }
                    #endregion

                    #region :divorce
                    case "divorce": // Warps the virtual user to a certain X,Y coordinate
                        {
                            if (_Married != "None")
                            {
                                virtualUser Target = userManager.getUser(_Married);
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        dbClient.runQuery("UPDATE users SET married = 'None' WHERE id = '" + userID + "'");
                                        dbClient.runQuery("UPDATE users SET married = 'None' WHERE username = '" + _Married + "'");
                                        Target._Married = "None";
                                        _Married = "None";
                                        Room.sendSaying(roomUser, "*divorces " + _Married + "*");
                                        
                                    }
                            }
                            break;
                        }
                    #endregion

                    #region :userwarp
                    case "userwarp": // Warps the target user to a certain X,Y coordinate
                        {
                            if (_Rank > 5)
                            {
                                virtualUser Target = userManager.getUser(args[1]);
                                int X = int.Parse(args[2]);
                                int Y = int.Parse(args[3]);
                                Target.roomUser.X = X;
                                Target.roomUser.Y = Y;
                                Target.roomUser.goalX = -1;
                                Room.Refresh(Target.roomUser);
                            }
                            break;
                        }
                    #endregion

                    #region :movebot
                    case "movebot": // Warps the target user to a certain X,Y coordinate
                        {
                            if (_Rank > 5)
                            {
                                virtualBot Bot = botManager.getBot(args[1]);
                                int X = int.Parse(args[2]);
                                int Y = int.Parse(args[3]);
                                Bot.roomBot.goalX = X;
                                Bot.roomBot.goalY = Y;
                            }
                            break;
                        }
                    #endregion

                    #region :poofbot
                    case "poofbot": // Warps the target user to a certain X,Y coordinate
                        {
                            if (_Rank > 5)
                            {
                                virtualBot Bot = botManager.getBot(args[1]);
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    string Fig;
                                    string Mis;
                                    Fig = dbClient.getString("SELECT figure FROM roombots WHERE name = '" + Bot.Name + "'");
                                    Mis = dbClient.getString("SELECT mission FROM roombots WHERE name = '" + Bot.Name + "'");

                                    Room.sendData("DJ" + Encoding.encodeVL64(Bot.roomBot.roomUID) + Fig + Convert.ToChar(2) + "M" + Convert.ToChar(2) + Mis + Convert.ToChar(2));
                                }
                            }
                            break;
                        }
                    #endregion

                    #region Bot Commands

                    #region :botid <botname>
                    case "botid":
                        {
                            virtualUser User = userManager.getUser(args[1]);
                            if (_Rank > 5)
                            {
                                int botid;
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    botid = dbClient.getInt("SELECT id FROM roombots WHERE name = '" + args[1] + "'");
                                    if (_Rank > 0)
                                    {
                                        sendData("BK" + "This Bots ID Is: " + botid + "");
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :noise
                    case "noise":
                        {
                            if (_Rank > 5)
                            {
                                sendData("Ej");
                                return true;
                            }
                            break;
                        }
                    #endregion

                    #region :saveall
                    case "saveall":
                        {
                            if (_Rank > 6)
                            {
                                userManager.SaveStats();
                            }
                            break;
                        }
                    #endregion

                    #region :kickbot
                    case "kickbot":
                        {
                            if (_Rank > 5)
                            {
                                virtualBot roomBot = botManager.getBot(args[1]);
                                {
                                    Room.sendSaying(roomUser, "*Kicks " + roomBot.Name + " From The Room");
                                    Room.deloadBot(roomBot);

                                }
                            }
                            break;
                        }
                    #endregion

                    #region :bringbot
                    case "bringbot":
                        {
                            if (_Rank > 5)
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    dbClient.runQuery("UPDATE roombots SET roomid = '" + _roomID + "' WHERE id = '" + args[1] + "'");
                                    Room.loadBots();

                            }
                            break;
                        }
                    #endregion

                    #region :giveitem x
                    case "giveitem":
                        {
                            if (_Rank > 5)
                            {
                                virtualBot roomBot = botManager.getBot(args[1]);
                                {
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        if (roomBot != null)
                                        {
                                            Room.sendSaying(roomUser, "*Gives " + roomBot.Name + " Their Item");
                                            roomBot.statusManager.addStatus("carryd", _ItemID);
                                            roomBot.statusManager.Refresh();
                                            statusManager.carryItem("0");
                                        }
                                    }

                                }
                                {
                                    virtualUser User = userManager.getUser(args[1]);
                                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                    {
                                        Room.sendSaying(roomUser, "*Gives " + User._Username + " Their Item");
                                        User._gettingDrink = true;
                                        User.statusManager.carryItem(_ItemID);
                                        statusManager.carryItem("0");
                                    }
                                }
                            }
                            break;
                        }
                    #endregion

                    #region :resetbot
                    case "resetbot":
                    case "activate":
                        {
                            if (_Rank > 5)
                            {
                                virtualBot roomBot = botManager.getBot(args[1]);
                                {
                                    Room.sendSaying(roomUser, "*Resets " + roomBot.Name + "*");
                                    Room.deloadBot(roomBot);
                                    Room.loadBots();
                                }
                            }
                            break;
                        }
                    #endregion

                    #endregion

                    #region :userid <user>
                    case "userid":
                        {
                            virtualUser User = userManager.getUser(args[1]);
                            if (_Rank > 5)
                            {
                                sendData("BK" + "This User's ID Is: " + User.userID + "");
                            }
                            break;
                        }
                    #endregion

                    #region :resetfigure
                    case "resetuser":
                        {
                            if (_Rank > 5)
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    virtualUser Target = userManager.getUser(args[1]);
                                    string figure = "hd-180-1.ch-210-66.lg-270-82.sh-290-80.hr-105-42";
                                    dbClient.runQuery("UPDATE users SET figure =  '" + figure + "' WHERE name = '" + Target._Username + "'");
                                    Room.sendSaying(roomUser, "*uses thier god-like powers to reset " + Target._Username + "'s figure*");
                                    Target.refreshAppearance(true, true, true);
                                }
                            }
                        }
                        break;
                    #endregion

        #endregion









                    default:
                        return false;
                }
            }
            catch {  }
            return true;
        }

        #endregion

        private string Stripslash(string p)
        {
            throw new NotImplementedException();
        }

        #region Refresh Voids
        private void refresh_cat()
        {
            catalogueManager.Init(true);
        }
        private void refresh_strings()
        {
            stringManager.Init("en", true);
        }
        private void refresh_config()
        {
            Config.Init(true);
        }
        private void refresh_filter()
        {
            stringManager.initFilter(true);
        }
        private void refresh_fuse()
        {
            rankManager.Init(true);
        }
        private void refresh_eco()
        {
            recyclerManager.Init(true);
        }
        private void refresh_bots()
        {
            botManager.Init();
        }
        #endregion

        #region Misc Thread Loops / Timers

        #region dont care about this

        private void spinWheel()
        {
            int iCount = 0;
            try
            {
                while (true)
                {
                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                    {
                            iCount++;
                        if (iCount == 12)
                        {

                            int rndNum2 = new Random(DateTime.Now.Millisecond).Next(0, 10);
                            if (rndNum2 == 3)
                            {
                                Room.sendShout(roomUser, "*Wins 25 Credits from the wheel*");
                                dbClient.runQuery("UPDATE users SET credits = credits + '25' WHERE name = '" + _Username + "'");
                            }
                            if (rndNum2 == 4)
                            {
                                Room.sendShout(roomUser, "*Wins 50 Credits from the wheel*");
                                dbClient.runQuery("UPDATE users SET credits = credits + '50' WHERE name = '" + _Username + "'");
                            }
                            if (rndNum2 == 5)
                            {
                                Room.sendShout(roomUser, "*Wins 500 Credits from the wheel*");
                                dbClient.runQuery("UPDATE users SET credits = credits + '500' WHERE name = '" + _Username + "'");
                            }
                            if (rndNum2 == 8)
                            {
                                Room.sendShout(roomUser, "*Wins 100 Credits from the wheel*");
                                dbClient.runQuery("UPDATE users SET credits = credits + '100' WHERE name = '" + _Username + "'");
                            }
                            if (rndNum2 != 8 && rndNum2 != 5 && rndNum2 != 4 && rndNum2 != 3)
                            {
                                Room.sendShout(roomUser, "*Does not win anything from the wheel*");
                            }
                            Room._Wheel = false;
                            refreshValueables(true, false);
                            spinWheelLooper = null;
                            spinWheelLooper.Abort();
                        }
                        Thread.Sleep(100);
                    }
                }
            }
            catch
            {
                Thread.CurrentThread.Abort();
                spinWheelLooper = null;
            }
        }
        public void MeShout(string message, int id)
        {
            virtualUser User = userManager.getUser(id);
            Room.sendShout(User.roomUser, "\n\n*" + User._Username + " " + message + "" + "" + "*\n\n  " + User._Username + " " + message);
        }
        public void MeSay(string message, int id)
        {
            virtualUser User = userManager.getUser(id);
            Room.sendSaying(User.roomUser, "\n\n*" + User._Username + " " + message + "" + "" + "*\n\n  " + User._Username + " " + message);
        }
        public void AddXP(int amount)
        {
            _XP = _XP + amount;
            sendData("EeI");
        }
        public void SaveStats()
        {
            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
            {
                dbClient.runQuery("UPDATE users SET health = '" + _Health + "' WHERE id = '" + userID + "'");
                dbClient.runQuery("UPDATE users SET maxhealth = '" + _legitHP + "' WHERE id = '" + userID + "'");
                dbClient.runQuery("UPDATE users SET str = '" + _legitStr + "' WHERE id = '" + userID + "'");
                dbClient.runQuery("UPDATE users SET secure_id = '" + _SecureID + "' WHERE id = '" + userID + "'");
                dbClient.runQuery("UPDATE users SET xp = '" + _XP + "' WHERE id = '" + userID + "'");
                dbClient.runQuery("UPDATE users SET xpdue = '" + _XPneed + "' WHERE id = '" + userID + "'");
                dbClient.runQuery("UPDATE users SET corp_id = '" + _CorpID + "' WHERE id = '" + userID + "'");
                dbClient.runQuery("UPDATE users SET job_rank = '" + _JobRank + "' WHERE id = '" + userID + "'");
                dbClient.runQuery("UPDATE users SET dead = '" + _Dead + "' WHERE id = '" + userID + "'");
                dbClient.runQuery("UPDATE users SET arrested = '" + _Jailed + "' WHERE id = '" + userID + "'");
                dbClient.runQuery("UPDATE users SET time_dead = '" + _DeadTime + "' WHERE id = '" + userID + "'");
                dbClient.runQuery("UPDATE users SET time_jail = '" + _JailTime + "' WHERE id = '" + userID + "'");
dbClient.runQuery("UPDATE users SET time_work = '" + WorkTime + "' WHERE id = '" + userID + "'");
dbClient.runQuery("UPDATE users SET kills = '" + _Kills + "' WHERE id = '" + userID + "'");
                dbClient.runQuery("UPDATE users SET hits = '" + _Hits + "' WHERE id = '" + userID + "'");
                dbClient.runQuery("UPDATE users SET arrests = '" + _Arrests + "' WHERE id = '" + userID + "'");
                dbClient.runQuery("UPDATE users SET deaths = '" + _Deaths + "' WHERE id = '" + userID + "'");
                dbClient.runQuery("UPDATE users SET x = '" + _SavedX + "' WHERE id = '" + userID + "'");
                dbClient.runQuery("UPDATE users SET y = '" + _SavedY + "' WHERE id = '" + userID + "'");
}
        }

        #endregion

        #region colour wars
        // colour war base timers and stuff

        private void GreenBase()
        {
            int iCount = 0;
            try
            {
                while (true)
                {
                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                    {
                        if (_roomID != 25 || roomUser.Y != 10 || roomUser.X != 7)
                        {
                            if (_roomID != 25 || roomUser.Y != 10 || roomUser.X != 8)
                            {
                                Room.sendSaying(roomUser, "*stops getting teh flag*");
                                RedBaseLooper = null;
                                RedBaseLooper.Abort();
                            }
                        }

                        iCount++;
                        if (iCount > 60)
                        {
                            userManager.sendToAdminScores(1, true, "BK" + "The green team are now out of the game!");
                            userManager.KickRedTeam(1, true, "BK" + "Your flag has been captured your out of the game");
                            RedBaseLooper = null;
                            RedBaseLooper.Abort();
                        }
                        Thread.Sleep(1000);
                    }
                }
            }
            catch
            {
                Thread.CurrentThread.Abort();
                RedBaseLooper = null;
            }
        }

        private void BlueBase()
        {
            int iCount = 0;
            try
            {
                while (true)
                {
                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                    {
                        if (_roomID != 417 || roomUser.Y != 7 || roomUser.X != 10)
                        {
                            if (_roomID != 417 || roomUser.Y != 10 || roomUser.X != 8)
                            {
                                Room.sendSaying(roomUser, "*stops capturing the point*");
                                RedBaseLooper = null;
                                RedBaseLooper.Abort();
                            }
                        }

                        iCount++;
                        if (iCount > 60)
                        {
                            userManager.sendToAdminScores(1, true, "BK" + "The blue team has been sent to the events lobby.");
                            userManager.KickGreenTeam(1, true, "BK" + "Your flag was captured, losers!");
                            RedBaseLooper = null;
                            RedBaseLooper.Abort();
                        }
                        Thread.Sleep(1000);
                    }
                }
            }
            catch
            {
                Thread.CurrentThread.Abort();
                RedBaseLooper = null;
            }
        }

        private void RedBase()
        {
            int iCount = 0;
            try
            {
                while (true)
                {
                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                    {
                        if (_roomID != 419 || roomUser.Y != 5 || roomUser.X != 7)
                        {
                            if (_roomID != 419 || roomUser.Y != 5 || roomUser.X != 6)
                            {
                                Room.sendSaying(roomUser, "*stops capturing the point*");
                                RedBaseLooper = null;
                                RedBaseLooper.Abort();
                            }
                        }

                        iCount++;
                        if (iCount > 60)
                        {
                            userManager.KickGreenTeam(1, true, "BK" + "Your flag was captured, losers.");
                            userManager.sendToAdminScores(1, true, "BK" + "The red team has been summoned to the events lobby");
                            RedBaseLooper = null;
                            RedBaseLooper.Abort();
                        }
                        Thread.Sleep(1000);
                    }
                }
            }
            catch
            {
                Thread.CurrentThread.Abort();
                RedBaseLooper = null;
            }
        }

        private void PinkBase()
        {
            int iCount = 0;
            try
            {
                while (true)
                {
                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                    {
                        if (_roomID != 7 || roomUser.Y != 11 || roomUser.X != 11)
                        {
                            if (_roomID != 7 || roomUser.Y != 11 || roomUser.X != 9)
                            {
                                Room.sendSaying(roomUser, "*stops getting teh flag*");
                                RedBaseLooper = null;
                                RedBaseLooper.Abort();
                            }
                        }

                        iCount++;
                        if (iCount > 60)
                        {
                            userManager.sendToAdminScores(1, true, "BK" + "The pink team are now out of the game!");
                            userManager.KickGreenTeam(1, true, "BK" + "Your flag has been captured your out of the game");
                            RedBaseLooper = null;
                            RedBaseLooper.Abort();
                        }
                        Thread.Sleep(1000);
                    }
                }
            }
            catch
            {
                Thread.CurrentThread.Abort();
                RedBaseLooper = null;
            }
        }

        public void DeadzoneCWS()
        {
            int iCount = 0;
            try
            {
                while (true)
                {
                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                    {
                        iCount++;
                        if (iCount > 30)
                        {
                            Room.sendSaying(roomUser, "*regains consciousness*");
                            roomUser.walkLock = false;
                            DeadZoneCWSlooper = null;
                            DeadZoneCWSlooper.Abort();
                        }
                        Thread.Sleep(60000);
                    }
                }
            }
            catch
            {
                Thread.CurrentThread.Abort();
                DeadZoneCWSlooper = null;
            }
        }

        // end the colour war stuff

        #endregion



        private void reboot()
        {
            int iCount = 0;
            try
            {
                while (true)
                {
                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                    {


                        iCount++;
                        if (iCount > 60)
                        {
                            Xoa.Eucalypt.Reboot();
                        }
                        Thread.Sleep(1000);
                    }
                }
            }
            catch
            {
                Thread.CurrentThread.Abort();
                rebootLooper = null;
            }
        }

        public void Instant()
        {
            try
            {
                while (true)
                {
                    if (x == 0 || y == 0 || id == 0)
                    {
                        x = 0;
                        y = 0;
                        id = 0;
                        AInstant = null;
                        AInstant.Abort();
                    }
                    else
                    {
                        if (x == roomUser.X && y == roomUser.Y)
                        {
                            int roomIDTeleporter2;
                            int idTeleporter2;
                            int link;
                            int roomid;
                            using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                            {
                                idTeleporter2 = dbClient.getInt("SELECT teleportid FROM furniture WHERE id = '" + id + "'");
                                roomid = dbClient.getInt("SELECT roomid FROM furniture WHERE id = '" + id + "'");
                                link = dbClient.getInt("SELECT link FROM furniture WHERE id = '" + id + "'");
                                roomIDTeleporter2 = dbClient.getInt("SELECT roomid FROM furniture WHERE id = '" + idTeleporter2 + "'");
                                Rooms.Items.floorItem Item = Room.floorItemManager.getItem(id);
                                if (link == 0)
                                {
                                    useInstants(Item, idTeleporter2, roomid);


                                }
                                else
                                {
                                    useInstants(Item, idTeleporter2, roomIDTeleporter2);

                                }
                                roomUser.goalX = 0;
                                roomUser.goalY = 0;
                                x = 0;
                                y = 0;
                                id = 0;
                                AInstant = null;
                                AInstant.Abort();
                            }
                        }
                    }
                    Thread.Sleep(100);
                }
            }
            catch
            {
                Thread.CurrentThread.Abort();
                AInstant = null;
            }
        }

        #endregion

        #region Teleporter

        private delegate void TeleporterUsageSleep(Rooms.Items.floorItem Teleporter1, int idTeleporter2, int roomIDTeleporter2);
        private void useTeleporter(Rooms.Items.floorItem Teleporter1, int idTeleporter2, int roomIDTeleporter2)
        {
            try
            {
                if (!_Cuffed && !_Stunned)
                {
                    roomUser.walkLock = true; //nullpointer
                    string Sprite = Teleporter1.Sprite;
                    if (roomIDTeleporter2 == _roomID) // Partner teleporter is in same room, don't leave room
                    {
                        Rooms.Items.floorItem Teleporter2 = Room.floorItemManager.getItem(idTeleporter2);
                        Thread.Sleep(500);
                        Room.sendData("AY" + Teleporter1.ID + "/" + _Username + "/" + Sprite);
                        //Thread.Sleep(1000);
                        Room.sendData(@"A\" + Teleporter2.ID + "/" + _Username + "/" + Sprite);
                        roomUser.X = Teleporter2.X;
                        roomUser.Y = Teleporter2.Y;
                        roomUser.H = Teleporter2.H;
                        roomUser.Z1 = Teleporter2.Z;
                        roomUser.Z2 = Teleporter2.Z;
                        roomUser.Refresh();
                        roomUser.walkLock = false;
                    }
                    else // Partner teleporter is in different room
                    {
                        if (_Dead == 1)
                        {
                            _onRoad = true;
                            sendData("DBO");
                            _roomID = 2;
                            roomidi = 2;
                        }
                        if (_Jailed == 1)
                        {
                            _onRoad = true;
                            sendData("DBO");
                            _roomID = 7;
                            roomidi = 7;
                        }
                        if (_Jailed == 0 && _Dead == 0)
                        {
                            _onRoad = true;
                            _teleporterID = idTeleporter2;
                            sendData("@~" + Encoding.encodeVL64(idTeleporter2) + Encoding.encodeVL64(roomIDTeleporter2));
                            Room.sendData("AY" + Teleporter1.ID + "/" + _Username + "/" + Sprite);
                        }
                    }
                }
            }
            catch { }
        }
        public void useInstants(Rooms.Items.floorItem Teleporter1, int idTeleporter2, int roomIDTeleporter2)
        {
            try
            {
                string Sprite = Teleporter1.Sprite;
                if (roomIDTeleporter2 == _roomID) // Partner teleporter is in same room, don't leave room
                {
                    int link;
                    using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                    {
                        link = dbClient.getInt("SELECT link FROM furniture WHERE id = '" + Teleporter1.ID + "'");
                    }
                    if (_crackingVault)
                    { }
                    else
                        if (_Dead == 1)
                        { }
                        else

                            if (link == 1)
                            {
                                Rooms.Items.floorItem Teleporter2 = Room.floorItemManager.getItem(idTeleporter2);
                                x = roomUser.X;
                                y = roomUser.Y;
                                Room.sqUNIT[Teleporter2.X, Teleporter2.Y] = false;
                                roomUser.X = Teleporter2.X;
                                roomUser.Y = Teleporter2.Y;
                                roomUser.H = Teleporter2.H;
                                roomUser.Z1 = Teleporter2.Z;
                                roomUser.Z2 = Teleporter2.Z;
                                roomUser.Refresh();
                                Room.sqUNIT[x, y] = false;
                            }
                            else
                            {
                                using (DatabaseClient dbClient = Eucalypt.dbManager.GetClient())
                                {
                                    int ItemX = dbClient.getInt("SELECT 2x FROM furniture WHERE id = '" + Teleporter1.ID + "'");
                                    int ItemY = dbClient.getInt("SELECT 2y FROM furniture WHERE id = '" + Teleporter1.ID + "'");
                                    int ItemH = dbClient.getInt("SELECT 2h FROM furniture WHERE id = '" + Teleporter1.ID + "'");
                                    int ItemZ = dbClient.getInt("SELECT 2z FROM furniture WHERE id = '" + Teleporter1.ID + "'");
                                    if (ItemX == 0 || ItemY == 0)
                                    {
                                        roomUser.walkLock = false;
                                    }
                                    else
                                    {
                                        x = roomUser.X;
                                        y = roomUser.Y;
                                        Room.sqUNIT[ItemX, ItemY] = false;
                                        roomUser.X = ItemX;
                                        roomUser.Y = ItemY;
                                        roomUser.H = ItemH;
                                        roomUser.Z1 = Convert.ToByte(ItemZ);
                                        roomUser.Z2 = Convert.ToByte(ItemZ);
                                        roomUser.Refresh();
                                        Room.sqUNIT[x, y] = false;
                                    }
                                }
                            }
                }
                else // Partner teleporter is in different room
                {
                    _teleporterID = idTeleporter2;
                    Room.sqUNIT[roomUser.X, roomUser.Y] = false;
                    _onRoad = true;
                    if (_Dead == 1)
                    {
                        roomidi = 2;
                        _roomID = 2;
                        sendData("DBO");
                    }
                    if (_Jailed == 1)
                    {
                        roomidi = 7;
                        _roomID = 7;
                        sendData("DBO");
                    }
                    if (_Jailed == 0 && _Dead == 0)
                    {
                        roomidi = roomIDTeleporter2;
                        _roomID = roomIDTeleporter2;
                        sendData("DBO");
                    }





                }
            }
            catch { }
        }
        #endregion


    }
}
